/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 4.0.2
 *
 * This file is not intended to be easily readable and contains a number of
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG
 * interface file instead.
 * ----------------------------------------------------------------------------- */


#ifndef SWIGPYTHON
#define SWIGPYTHON
#endif

#define SWIG_PYTHON_THREADS
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE


#ifdef __cplusplus
/* SwigValueWrapper is described in swig.swg */
template<typename T> class SwigValueWrapper {
  struct SwigMovePointer {
    T *ptr;
    SwigMovePointer(T *p) : ptr(p) { }
    ~SwigMovePointer() { delete ptr; }
    SwigMovePointer& operator=(SwigMovePointer& rhs) { T* oldptr = ptr; ptr = 0; delete oldptr; ptr = rhs.ptr; rhs.ptr = 0; return *this; }
  } pointer;
  SwigValueWrapper& operator=(const SwigValueWrapper<T>& rhs);
  SwigValueWrapper(const SwigValueWrapper<T>& rhs);
public:
  SwigValueWrapper() : pointer(0) { }
  SwigValueWrapper& operator=(const T& t) { SwigMovePointer tmp(new T(t)); pointer = tmp; return *this; }
  operator T&() const { return *pointer.ptr; }
  T *operator&() { return pointer.ptr; }
};

template <typename T> T SwigValueInit() {
  return T();
}
#endif

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif


#if defined(__GNUC__) && defined(_WIN32) && !defined(SWIG_PYTHON_NO_HYPOT_WORKAROUND)
/* Workaround for '::hypot' has not been declared', see https://bugs.python.org/issue11566 */
# include <math.h>
#endif

#if defined(_DEBUG) && defined(SWIG_PYTHON_INTERPRETER_NO_DEBUG)
/* Use debug wrappers with the Python release dll */
# undef _DEBUG
# include <Python.h>
# define _DEBUG 1
#else
# include <Python.h>
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "4"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (strcmp(iter->type->name, c) == 0) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Identical to SWIG_TypeCheck, except strcmp is replaced with a pointer comparison
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *iter = ty->cast;
    while (iter) {
      if (iter->type == from) {
        if (iter == ty->cast)
          return iter;
        /* Move iter to the top of the linked list */
        iter->prev->next = iter->next;
        if (iter->next)
          iter->next->prev = iter->prev;
        iter->next = ty->cast;
        iter->prev = 0;
        if (ty->cast) ty->cast->prev = iter;
        ty->cast = iter;
        return iter;
      }
      iter = iter->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13



/* Compatibility macros for Python 3 */
#if PY_VERSION_HEX >= 0x03000000

#define PyClass_Check(obj) PyObject_IsInstance(obj, (PyObject *)&PyType_Type)
#define PyInt_Check(x) PyLong_Check(x)
#define PyInt_AsLong(x) PyLong_AsLong(x)
#define PyInt_FromLong(x) PyLong_FromLong(x)
#define PyInt_FromSize_t(x) PyLong_FromSize_t(x)
#define PyString_Check(name) PyBytes_Check(name)
#define PyString_FromString(x) PyUnicode_FromString(x)
#define PyString_Format(fmt, args)  PyUnicode_Format(fmt, args)
#define PyString_AsString(str) PyBytes_AsString(str)
#define PyString_Size(str) PyBytes_Size(str)	
#define PyString_InternFromString(key) PyUnicode_InternFromString(key)
#define Py_TPFLAGS_HAVE_CLASS Py_TPFLAGS_BASETYPE
#define PyString_AS_STRING(x) PyUnicode_AS_STRING(x)
#define _PyLong_FromSsize_t(x) PyLong_FromSsize_t(x)

#endif

#ifndef Py_TYPE
#  define Py_TYPE(op) ((op)->ob_type)
#endif

/* SWIG APIs for compatibility of both Python 2 & 3 */

#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_Python_str_FromFormat PyUnicode_FromFormat
#else
#  define SWIG_Python_str_FromFormat PyString_FromFormat
#endif


/* Warning: This function will allocate a new string in Python 3,
 * so please call SWIG_Python_str_DelForPy3(x) to free the space.
 */
SWIGINTERN char*
SWIG_Python_str_AsChar(PyObject *str)
{
#if PY_VERSION_HEX >= 0x03030000
  return (char *)PyUnicode_AsUTF8(str);
#elif PY_VERSION_HEX >= 0x03000000
  char *newstr = 0;
  str = PyUnicode_AsUTF8String(str);
  if (str) {
    char *cstr;
    Py_ssize_t len;
    if (PyBytes_AsStringAndSize(str, &cstr, &len) != -1) {
      newstr = (char *) malloc(len+1);
      if (newstr)
        memcpy(newstr, cstr, len+1);
    }
    Py_XDECREF(str);
  }
  return newstr;
#else
  return PyString_AsString(str);
#endif
}

#if PY_VERSION_HEX >= 0x03030000 || PY_VERSION_HEX < 0x03000000
#  define SWIG_Python_str_DelForPy3(x)
#else
#  define SWIG_Python_str_DelForPy3(x) free( (void*) (x) )
#endif


SWIGINTERN PyObject*
SWIG_Python_str_FromChar(const char *c)
{
#if PY_VERSION_HEX >= 0x03000000
  return PyUnicode_FromString(c); 
#else
  return PyString_FromString(c);
#endif
}

#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

// SWIGPY_USE_CAPSULE is no longer used within SWIG itself, but some user
// interface files check for it.
# define SWIGPY_USE_CAPSULE
# define SWIGPY_CAPSULE_NAME ("swig_runtime_data" SWIG_RUNTIME_VERSION ".type_pointer_capsule" SWIG_TYPE_TABLE_NAME)

#if PY_VERSION_HEX < 0x03020000
#define PyDescr_TYPE(x) (((PyDescrObject *)(x))->d_type)
#define PyDescr_NAME(x) (((PyDescrObject *)(x))->d_name)
#define Py_hash_t long
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred())
    PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    const char *tmp = SWIG_Python_str_AsChar(old_str);
    PyErr_Clear();
    Py_XINCREF(type);
    if (tmp)
      PyErr_Format(type, "%s %s", tmp, mesg);
    else
      PyErr_Format(type, "%s", mesg);
    SWIG_Python_str_DelForPy3(tmp);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_SetString(PyExc_RuntimeError, mesg);
  }
}

SWIGRUNTIME int
SWIG_Python_TypeErrorOccurred(PyObject *obj)
{
  PyObject *error;
  if (obj)
    return 0;
  error = PyErr_Occurred();
  return error && PyErr_GivenExceptionMatches(error, PyExc_TypeError);
}

SWIGRUNTIME void
SWIG_Python_RaiseOrModifyTypeError(const char *message)
{
  if (SWIG_Python_TypeErrorOccurred(NULL)) {
    /* Use existing TypeError to preserve stacktrace and enhance with given message */
    PyObject *newvalue;
    PyObject *type = NULL, *value = NULL, *traceback = NULL;
    PyErr_Fetch(&type, &value, &traceback);
#if PY_VERSION_HEX >= 0x03000000
    newvalue = PyUnicode_FromFormat("%S\nAdditional information:\n%s", value, message);
#else
    newvalue = PyString_FromFormat("%s\nAdditional information:\n%s", PyString_AsString(value), message);
#endif
    Py_XDECREF(value);
    PyErr_Restore(type, newvalue, traceback);
  } else {
    /* Raise TypeError using given message */
    PyErr_SetString(PyExc_TypeError, message);
  }
}

#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    define SWIG_PYTHON_USE_GIL
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  const char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

#if PY_VERSION_HEX < 0x02070000 /* 2.7.0 */
# error "This version of SWIG only supports Python >= 2.7"
#endif

#if PY_VERSION_HEX >= 0x03000000 && PY_VERSION_HEX < 0x03020000
# error "This version of SWIG only supports Python 3 >= 3.2"
#endif

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)

#ifdef SWIGPYTHON_BUILTIN
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(self, ptr, type, flags)
#else
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(NULL, ptr, type, flags)
#endif

#define SWIG_InternalNewPointerObj(ptr, type, flags)	SWIG_Python_NewPointerObj(NULL, ptr, type, flags)

#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(NULL, ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         SwigPyClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

#if defined(SWIGPYTHON_BUILTIN)

SWIGINTERN void
SwigPyBuiltin_AddPublicSymbol(PyObject *seq, const char *key) {
  PyObject *s = PyString_InternFromString(key);
  PyList_Append(seq, s);
  Py_DECREF(s);
}

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, PyObject *public_interface, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);
  if (public_interface)
    SwigPyBuiltin_AddPublicSymbol(public_interface, name);
}

#else

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, name, obj);
  Py_DECREF(obj);                            
}

#endif

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
}

/* Unpack the argument tuple */

SWIGINTERN Py_ssize_t
SWIG_Python_UnpackTuple(PyObject *args, const char *name, Py_ssize_t min, Py_ssize_t max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), (int)min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    if (min <= 1 && max >= 1) {
      Py_ssize_t i;
      objs[0] = args;
      for (i = 1; i < max; ++i) {
	objs[i] = 0;
      }
      return 2;
    }
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    Py_ssize_t l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), (int)min, (int)l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), (int)max, (int)l);
      return 0;
    } else {
      Py_ssize_t i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

SWIGINTERN int
SWIG_Python_CheckNoKeywords(PyObject *kwargs, const char *name) {
  int no_kwargs = 1;
  if (kwargs) {
    assert(PyDict_Check(kwargs));
    if (PyDict_Size(kwargs) > 0) {
      PyErr_Format(PyExc_TypeError, "%s() does not take keyword arguments", name);
      no_kwargs = 0;
    }
  }
  return no_kwargs;
}

/* A functor is a function object with one single object argument */
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#define SWIG_BUILTIN_TP_INIT	    (SWIG_POINTER_OWN << 2)
#define SWIG_BUILTIN_INIT	    (SWIG_BUILTIN_TP_INIT | SWIG_POINTER_OWN)

#ifdef __cplusplus
extern "C" {
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* SwigPyClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
  PyTypeObject *pytype;
} SwigPyClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  SwigPyClientData *data = (SwigPyClientData *)ty->clientdata;
  int fail = data ? data->implicitconv : 0;
  if (fail)
    PyErr_SetString(PyExc_TypeError, "Implicit conversion is prohibited for explicit constructors.");
  return fail;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  SwigPyClientData *data = desc ? (SwigPyClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME SwigPyClientData * 
SwigPyClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    SwigPyClientData *data = (SwigPyClientData *)malloc(sizeof(SwigPyClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
      data->newraw = PyObject_GetAttrString(data->klass, "__new__");
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, "__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
      data->delargs = !(flags & (METH_O));
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    data->pytype = 0;
    return data;
  }
}

SWIGRUNTIME void 
SwigPyClientData_Del(SwigPyClientData *data) {
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== SwigPyObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
#ifdef SWIGPYTHON_BUILTIN
  PyObject *dict;
#endif
} SwigPyObject;


#ifdef SWIGPYTHON_BUILTIN

SWIGRUNTIME PyObject *
SwigPyObject_get___dict__(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;

  if (!sobj->dict)
    sobj->dict = PyDict_New();

  Py_INCREF(sobj->dict);
  return sobj->dict;
}

#endif

SWIGRUNTIME PyObject *
SwigPyObject_long(SwigPyObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
SwigPyObject_format(const char* fmt, SwigPyObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, SwigPyObject_long(v)) == 0) {
      PyObject *ofmt = SWIG_Python_str_FromChar(fmt);
      if (ofmt) {
#if PY_VERSION_HEX >= 0x03000000
	res = PyUnicode_Format(ofmt,args);
#else
	res = PyString_Format(ofmt,args);
#endif
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
SwigPyObject_oct(SwigPyObject *v)
{
  return SwigPyObject_format("%o",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_hex(SwigPyObject *v)
{
  return SwigPyObject_format("%x",v);
}

SWIGRUNTIME PyObject *
SwigPyObject_repr(SwigPyObject *v)
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *repr = SWIG_Python_str_FromFormat("<Swig Object of type '%s' at %p>", (name ? name : "unknown"), (void *)v);
  if (v->next) {
    PyObject *nrep = SwigPyObject_repr((SwigPyObject *)v->next);
# if PY_VERSION_HEX >= 0x03000000
    PyObject *joined = PyUnicode_Concat(repr, nrep);
    Py_DecRef(repr);
    Py_DecRef(nrep);
    repr = joined;
# else
    PyString_ConcatAndDel(&repr,nrep);
# endif
  }
  return repr;  
}

/* We need a version taking two PyObject* parameters so it's a valid
 * PyCFunction to use in swigobject_methods[]. */
SWIGRUNTIME PyObject *
SwigPyObject_repr2(PyObject *v, PyObject *SWIGUNUSEDPARM(args))
{
  return SwigPyObject_repr((SwigPyObject*)v);
}

SWIGRUNTIME int
SwigPyObject_compare(SwigPyObject *v, SwigPyObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

/* Added for Python 3.x, would it also be useful for Python 2.x? */
SWIGRUNTIME PyObject*
SwigPyObject_richcompare(SwigPyObject *v, SwigPyObject *w, int op)
{
  PyObject* res;
  if( op != Py_EQ && op != Py_NE ) {
    Py_INCREF(Py_NotImplemented);
    return Py_NotImplemented;
  }
  res = PyBool_FromLong( (SwigPyObject_compare(v, w)==0) == (op == Py_EQ) ? 1 : 0);
  return res;  
}


SWIGRUNTIME PyTypeObject* SwigPyObject_TypeOnce(void);

#ifdef SWIGPYTHON_BUILTIN
static swig_type_info *SwigPyObject_stype = 0;
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
    SwigPyClientData *cd;
    assert(SwigPyObject_stype);
    cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
    assert(cd);
    assert(cd->pytype);
    return cd->pytype;
}
#else
SWIGRUNTIME PyTypeObject*
SwigPyObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyObject_TypeOnce();
  return type;
}
#endif

SWIGRUNTIMEINLINE int
SwigPyObject_Check(PyObject *op) {
#ifdef SWIGPYTHON_BUILTIN
  PyTypeObject *target_tp = SwigPyObject_type();
  if (PyType_IsSubtype(op->ob_type, target_tp))
    return 1;
  return (strcmp(op->ob_type->tp_name, "SwigPyObject") == 0);
#else
  return (Py_TYPE(op) == SwigPyObject_type())
    || (strcmp(Py_TYPE(op)->tp_name,"SwigPyObject") == 0);
#endif
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
SwigPyObject_dealloc(PyObject *v)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own == SWIG_POINTER_OWN) {
    swig_type_info *ty = sobj->ty;
    SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;

      /* PyObject_CallFunction() has the potential to silently drop
         the active exception.  In cases of unnamed temporary
         variable or where we just finished iterating over a generator
         StopIteration will be active right now, and this needs to
         remain true upon return from SwigPyObject_dealloc.  So save
         and restore. */
      
      PyObject *type = NULL, *value = NULL, *traceback = NULL;
      PyErr_Fetch(&type, &value, &traceback);

      if (data->delargs) {
        /* we need to create a temporary object to carry the destroy operation */
        PyObject *tmp = SwigPyObject_New(sobj->ptr, ty, 0);
        res = SWIG_Python_CallFunctor(destroy, tmp);
        Py_DECREF(tmp);
      } else {
        PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
        PyObject *mself = PyCFunction_GET_SELF(destroy);
        res = ((*meth)(mself, v));
      }
      if (!res)
        PyErr_WriteUnraisable(destroy);

      PyErr_Restore(type, value, traceback);

      Py_XDECREF(res);
    } 
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
    else {
      const char *name = SWIG_TypePrettyName(ty);
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", (name ? name : "unknown"));
    }
#endif
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
SwigPyObject_append(PyObject* v, PyObject* next)
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (!SwigPyObject_Check(next)) {
    PyErr_SetString(PyExc_TypeError, "Attempt to append a non SwigPyObject");
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
SwigPyObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
SwigPyObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
{
  SwigPyObject *sobj = (SwigPyObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
SwigPyObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
  if (!PyArg_UnpackTuple(args, "own", 0, 1, &val)) {
    return NULL;
  } else {
    SwigPyObject *sobj = (SwigPyObject *)v;
    PyObject *obj = PyBool_FromLong(sobj->own);
    if (val) {
      if (PyObject_IsTrue(val)) {
        SwigPyObject_acquire(v,args);
      } else {
        SwigPyObject_disown(v,args);
      }
    } 
    return obj;
  }
}

static PyMethodDef
swigobject_methods[] = {
  {"disown",  SwigPyObject_disown,  METH_NOARGS,  "releases ownership of the pointer"},
  {"acquire", SwigPyObject_acquire, METH_NOARGS,  "acquires ownership of the pointer"},
  {"own",     SwigPyObject_own,     METH_VARARGS, "returns/sets ownership of the pointer"},
  {"append",  SwigPyObject_append,  METH_O,       "appends another 'this' object"},
  {"next",    SwigPyObject_next,    METH_NOARGS,  "returns the next 'this' object"},
  {"__repr__",SwigPyObject_repr2,   METH_NOARGS,  "returns object representation"},
  {0, 0, 0, 0}  
};

SWIGRUNTIME PyTypeObject*
SwigPyObject_TypeOnce(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";

  static PyNumberMethods SwigPyObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    /* nb_divide removed in Python 3 */
#if PY_VERSION_HEX < 0x03000000
    (binaryfunc)0, /*nb_divide*/
#endif
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
#if PY_VERSION_HEX < 0x03000000
    0,   /*nb_coerce*/
#endif
    (unaryfunc)SwigPyObject_long, /*nb_int*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_long, /*nb_long*/
#else
    0, /*nb_reserved*/
#endif
    (unaryfunc)0,                 /*nb_float*/
#if PY_VERSION_HEX < 0x03000000
    (unaryfunc)SwigPyObject_oct,  /*nb_oct*/
    (unaryfunc)SwigPyObject_hex,  /*nb_hex*/
#endif
#if PY_VERSION_HEX >= 0x03050000 /* 3.5 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_matrix_multiply */
#elif PY_VERSION_HEX >= 0x03000000 /* 3.0 */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index, nb_inplace_divide removed */
#else
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_index */
#endif
  };

  static PyTypeObject swigpyobject_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyObject",                       /* tp_name */
      sizeof(SwigPyObject),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyObject_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX >= 0x03000000
      0, /* tp_reserved in 3.0.1, tp_compare in 3.0.0 but not used */
#else
      (cmpfunc)SwigPyObject_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyObject_repr,          /* tp_repr */
      &SwigPyObject_as_number,              /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      0,                                    /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigobject_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      (richcmpfunc)SwigPyObject_richcompare,/* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      swigobject_methods,                   /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpyobject_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpyobject_type) < 0)
      return NULL;
  }
  return &swigpyobject_type;
}

SWIGRUNTIME PyObject *
SwigPyObject_New(void *ptr, swig_type_info *ty, int own)
{
  SwigPyObject *sobj = PyObject_NEW(SwigPyObject, SwigPyObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} SwigPyPacked;

SWIGRUNTIME PyObject *
SwigPyPacked_repr(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return SWIG_Python_str_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
SwigPyPacked_str(SwigPyPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return SWIG_Python_str_FromFormat("%s%s", result, v->ty->name);
  } else {
    return SWIG_Python_str_FromChar(v->ty->name);
  }  
}

SWIGRUNTIME int
SwigPyPacked_compare(SwigPyPacked *v, SwigPyPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((const char *)v->pack, (const char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* SwigPyPacked_TypeOnce(void);

SWIGRUNTIME PyTypeObject*
SwigPyPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = SwigPyPacked_TypeOnce();
  return type;
}

SWIGRUNTIMEINLINE int
SwigPyPacked_Check(PyObject *op) {
  return ((op)->ob_type == SwigPyPacked_TypeOnce()) 
    || (strcmp((op)->ob_type->tp_name,"SwigPyPacked") == 0);
}

SWIGRUNTIME void
SwigPyPacked_dealloc(PyObject *v)
{
  if (SwigPyPacked_Check(v)) {
    SwigPyPacked *sobj = (SwigPyPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
SwigPyPacked_TypeOnce(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject swigpypacked_type;
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp = {
#if PY_VERSION_HEX>=0x03000000
      PyVarObject_HEAD_INIT(NULL, 0)
#else
      PyObject_HEAD_INIT(NULL)
      0,                                    /* ob_size */
#endif
      "SwigPyPacked",                       /* tp_name */
      sizeof(SwigPyPacked),                 /* tp_basicsize */
      0,                                    /* tp_itemsize */
      (destructor)SwigPyPacked_dealloc,     /* tp_dealloc */
      0,                                    /* tp_print */
      (getattrfunc)0,                       /* tp_getattr */
      (setattrfunc)0,                       /* tp_setattr */
#if PY_VERSION_HEX>=0x03000000
      0, /* tp_reserved in 3.0.1 */
#else
      (cmpfunc)SwigPyPacked_compare,        /* tp_compare */
#endif
      (reprfunc)SwigPyPacked_repr,          /* tp_repr */
      0,                                    /* tp_as_number */
      0,                                    /* tp_as_sequence */
      0,                                    /* tp_as_mapping */
      (hashfunc)0,                          /* tp_hash */
      (ternaryfunc)0,                       /* tp_call */
      (reprfunc)SwigPyPacked_str,           /* tp_str */
      PyObject_GenericGetAttr,              /* tp_getattro */
      0,                                    /* tp_setattro */
      0,                                    /* tp_as_buffer */
      Py_TPFLAGS_DEFAULT,                   /* tp_flags */
      swigpacked_doc,                       /* tp_doc */
      0,                                    /* tp_traverse */
      0,                                    /* tp_clear */
      0,                                    /* tp_richcompare */
      0,                                    /* tp_weaklistoffset */
      0,                                    /* tp_iter */
      0,                                    /* tp_iternext */
      0,                                    /* tp_methods */
      0,                                    /* tp_members */
      0,                                    /* tp_getset */
      0,                                    /* tp_base */
      0,                                    /* tp_dict */
      0,                                    /* tp_descr_get */
      0,                                    /* tp_descr_set */
      0,                                    /* tp_dictoffset */
      0,                                    /* tp_init */
      0,                                    /* tp_alloc */
      0,                                    /* tp_new */
      0,                                    /* tp_free */
      0,                                    /* tp_is_gc */
      0,                                    /* tp_bases */
      0,                                    /* tp_mro */
      0,                                    /* tp_cache */
      0,                                    /* tp_subclasses */
      0,                                    /* tp_weaklist */
      0,                                    /* tp_del */
      0,                                    /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
      0,                                    /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
      0,                                    /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
      0,                                    /* tp_print */
#endif
#ifdef COUNT_ALLOCS
      0,                                    /* tp_allocs */
      0,                                    /* tp_frees */
      0,                                    /* tp_maxalloc */
      0,                                    /* tp_prev */
      0                                     /* tp_next */
#endif
    };
    swigpypacked_type = tmp;
    type_init = 1;
    if (PyType_Ready(&swigpypacked_type) < 0)
      return NULL;
  }
  return &swigpypacked_type;
}

SWIGRUNTIME PyObject *
SwigPyPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  SwigPyPacked *sobj = PyObject_NEW(SwigPyPacked, SwigPyPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
SwigPyPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (SwigPyPacked_Check(obj)) {
    SwigPyPacked *sobj = (SwigPyPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

static PyObject *Swig_This_global = NULL;

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  if (Swig_This_global == NULL)
    Swig_This_global = SWIG_Python_str_FromChar("this");
  return Swig_This_global;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

/* TODO: I don't know how to implement the fast getset in Python 3 right now */
#if PY_VERSION_HEX>=0x03000000
#define SWIG_PYTHON_SLOW_GETSET_THIS 
#endif

SWIGRUNTIME SwigPyObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  PyObject *obj;

  if (SwigPyObject_Check(pyobj))
    return (SwigPyObject *) pyobj;

#ifdef SWIGPYTHON_BUILTIN
  (void)obj;
# ifdef PyWeakref_CheckProxy
  if (PyWeakref_CheckProxy(pyobj)) {
    pyobj = PyWeakref_GET_OBJECT(pyobj);
    if (pyobj && SwigPyObject_Check(pyobj))
      return (SwigPyObject*) pyobj;
  }
# endif
  return NULL;
#else

  obj = 0;

#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  if (PyInstance_Check(pyobj)) {
    obj = _PyInstance_Lookup(pyobj, SWIG_This());      
  } else {
    PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
    if (dictptr != NULL) {
      PyObject *dict = *dictptr;
      obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
    } else {
#ifdef PyWeakref_CheckProxy
      if (PyWeakref_CheckProxy(pyobj)) {
	PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
      }
#endif
      obj = PyObject_GetAttr(pyobj,SWIG_This());
      if (obj) {
	Py_DECREF(obj);
      } else {
	if (PyErr_Occurred()) PyErr_Clear();
	return 0;
      }
    }
  }
#else
  obj = PyObject_GetAttr(pyobj,SWIG_This());
  if (obj) {
    Py_DECREF(obj);
  } else {
    if (PyErr_Occurred()) PyErr_Clear();
    return 0;
  }
#endif
  if (obj && !SwigPyObject_Check(obj)) {
    /* a PyObject is called 'this', try to get the 'real this'
       SwigPyObject from it */ 
    return SWIG_Python_GetSwigThis(obj);
  }
  return (SwigPyObject *)obj;
#endif
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own == SWIG_POINTER_OWN) {
    SwigPyObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  int res;
  SwigPyObject *sobj;
  int implicit_conv = (flags & SWIG_POINTER_IMPLICIT_CONV) != 0;

  if (!obj)
    return SWIG_ERROR;
  if (obj == Py_None && !implicit_conv) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  }

  res = SWIG_ERROR;

  sobj = SWIG_Python_GetSwigThis(obj);
  if (own)
    *own = 0;
  while (sobj) {
    void *vptr = sobj->ptr;
    if (ty) {
      swig_type_info *to = sobj->ty;
      if (to == ty) {
        /* no type cast needed */
        if (ptr) *ptr = vptr;
        break;
      } else {
        swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
        if (!tc) {
          sobj = (SwigPyObject *)sobj->next;
        } else {
          if (ptr) {
            int newmemory = 0;
            *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
            if (newmemory == SWIG_CAST_NEW_MEMORY) {
              assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
              if (own)
                *own = *own | SWIG_CAST_NEW_MEMORY;
            }
          }
          break;
        }
      }
    } else {
      if (ptr) *ptr = vptr;
      break;
    }
  }
  if (sobj) {
    if (own)
      *own = *own | sobj->own;
    if (flags & SWIG_POINTER_DISOWN) {
      sobj->own = 0;
    }
    res = SWIG_OK;
  } else {
    if (implicit_conv) {
      SwigPyClientData *data = ty ? (SwigPyClientData *) ty->clientdata : 0;
      if (data && !data->implicitconv) {
        PyObject *klass = data->klass;
        if (klass) {
          PyObject *impconv;
          data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
          impconv = SWIG_Python_CallFunctor(klass, obj);
          data->implicitconv = 0;
          if (PyErr_Occurred()) {
            PyErr_Clear();
            impconv = 0;
          }
          if (impconv) {
            SwigPyObject *iobj = SWIG_Python_GetSwigThis(impconv);
            if (iobj) {
              void *vptr;
              res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
              if (SWIG_IsOK(res)) {
                if (ptr) {
                  *ptr = vptr;
                  /* transfer the ownership to 'ptr' */
                  iobj->own = 0;
                  res = SWIG_AddCast(res);
                  res = SWIG_AddNewMask(res);
                } else {
                  res = SWIG_AddCast(res);		    
                }
              }
            }
            Py_DECREF(impconv);
          }
        }
      }
      if (!SWIG_IsOK(res) && obj == Py_None) {
        if (ptr)
          *ptr = 0;
        if (PyErr_Occurred())
          PyErr_Clear();
        res = SWIG_OK;
      }
    }
  }
  return res;
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    swig_cast_info *tc;

    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc)
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
    if (!desc)
      return SWIG_ERROR;
    tc = SWIG_TypeCheck(desc,ty);
    if (tc) {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,vptr,&newmemory);
      assert(!newmemory); /* newmemory handling not yet implemented */
    } else {
      return SWIG_ERROR;
    }
    return SWIG_OK;
  }
}

/* Convert a packed pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = SwigPyPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, without calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(SwigPyClientData *data, PyObject *swig_this)
{
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
        Py_DECREF(inst);
        inst = 0;
      }
#endif
    }
  } else {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *empty_args = PyTuple_New(0);
    if (empty_args) {
      PyObject *empty_kwargs = PyDict_New();
      if (empty_kwargs) {
        inst = ((PyTypeObject *)data->newargs)->tp_new((PyTypeObject *)data->newargs, empty_args, empty_kwargs);
        Py_DECREF(empty_kwargs);
        if (inst) {
          if (PyObject_SetAttr(inst, SWIG_This(), swig_this) == -1) {
            Py_DECREF(inst);
            inst = 0;
          } else {
            Py_TYPE(inst)->tp_flags &= ~Py_TPFLAGS_VALID_VERSION_TAG;
          }
        }
      }
      Py_DECREF(empty_args);
    }
#else
    PyObject *dict = PyDict_New();
    if (dict) {
      PyDict_SetItem(dict, SWIG_This(), swig_this);
      inst = PyInstance_NewRaw(data->newargs, dict);
      Py_DECREF(dict);
    }
#endif
  }
  return inst;
}

SWIGRUNTIME int
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
  PyObject **dictptr = _PyObject_GetDictPtr(inst);
  if (dictptr != NULL) {
    PyObject *dict = *dictptr;
    if (dict == NULL) {
      dict = PyDict_New();
      *dictptr = dict;
    }
    return PyDict_SetItem(dict, SWIG_This(), swig_this);
  }
#endif
  return PyObject_SetAttr(inst, SWIG_This(), swig_this);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args, "swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    SwigPyObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      SwigPyObject_append((PyObject*) sthis, obj[1]);
    } else {
      if (SWIG_Python_SetSwigThis(obj[0], obj[1]) != 0)
        return NULL;
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(PyObject *self, void *ptr, swig_type_info *type, int flags) {
  SwigPyClientData *clientdata;
  PyObject * robj;
  int own;

  if (!ptr)
    return SWIG_Py_Void();

  clientdata = type ? (SwigPyClientData *)(type->clientdata) : 0;
  own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
  if (clientdata && clientdata->pytype) {
    SwigPyObject *newobj;
    if (flags & SWIG_BUILTIN_TP_INIT) {
      newobj = (SwigPyObject*) self;
      if (newobj->ptr) {
        PyObject *next_self = clientdata->pytype->tp_alloc(clientdata->pytype, 0);
        while (newobj->next)
	  newobj = (SwigPyObject *) newobj->next;
        newobj->next = next_self;
        newobj = (SwigPyObject *)next_self;
#ifdef SWIGPYTHON_BUILTIN
        newobj->dict = 0;
#endif
      }
    } else {
      newobj = PyObject_New(SwigPyObject, clientdata->pytype);
#ifdef SWIGPYTHON_BUILTIN
      newobj->dict = 0;
#endif
    }
    if (newobj) {
      newobj->ptr = ptr;
      newobj->ty = type;
      newobj->own = own;
      newobj->next = 0;
      return (PyObject*) newobj;
    }
    return SWIG_Py_Void();
  }

  assert(!(flags & SWIG_BUILTIN_TP_INIT));

  robj = SwigPyObject_New(ptr, type, own);
  if (robj && clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
    PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
    Py_DECREF(robj);
    robj = inst;
  }
  return robj;
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? SwigPyPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
    type_pointer = PyCapsule_Import(SWIGPY_CAPSULE_NAME, 0);
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

SWIGRUNTIME void
SWIG_Python_DestroyModule(PyObject *obj)
{
  swig_module_info *swig_module = (swig_module_info *) PyCapsule_GetPointer(obj, SWIGPY_CAPSULE_NAME);
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      SwigPyClientData *data = (SwigPyClientData *) ty->clientdata;
      if (data) SwigPyClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
  Swig_This_global = NULL;
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
#if PY_VERSION_HEX >= 0x03000000
 /* Add a dummy module object into sys.modules */
  PyObject *module = PyImport_AddModule("swig_runtime_data" SWIG_RUNTIME_VERSION);
#else
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} }; /* Sentinel */
  PyObject *module = Py_InitModule("swig_runtime_data" SWIG_RUNTIME_VERSION, swig_empty_runtime_method_table);
#endif
  PyObject *pointer = PyCapsule_New((void *) swig_module, SWIGPY_CAPSULE_NAME, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, "type_pointer_capsule" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = SWIG_Python_str_FromChar(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
    descriptor = (swig_type_info *) PyCapsule_GetPointer(obj, NULL);
  } else {
    swig_module_info *swig_module = SWIG_GetModule(0);
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCapsule_New((void*) descriptor, NULL, NULL);
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{  
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      const char *tmp = SWIG_Python_str_AsChar(old_str);
      const char *errmesg = tmp ? tmp : "Invalid error message";
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, errmesg);
      } else {
	PyErr_Format(type, "%s %s", errmesg, mesg);
      }
      SWIG_Python_str_DelForPy3(tmp);
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
SwigPyObject_GetDesc(PyObject *self)
{
  SwigPyObject *v = (SwigPyObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : "";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && SwigPyObject_Check(obj)) {
      const char *otype = (const char *) SwigPyObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'SwigPyObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? SWIG_Python_str_AsChar(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
          SWIG_Python_str_DelForPy3(cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int SWIGUNUSEDPARM(argnum), int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
#if SWIG_POINTER_EXCEPTION
    if (flags) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
#endif
  }
  return result;
}

#ifdef SWIGPYTHON_BUILTIN
SWIGRUNTIME int
SWIG_Python_NonDynamicSetAttr(PyObject *obj, PyObject *name, PyObject *value) {
  PyTypeObject *tp = obj->ob_type;
  PyObject *descr;
  PyObject *encoded_name;
  descrsetfunc f;
  int res = -1;

# ifdef Py_USING_UNICODE
  if (PyString_Check(name)) {
    name = PyUnicode_Decode(PyString_AsString(name), PyString_Size(name), NULL, NULL);
    if (!name)
      return -1;
  } else if (!PyUnicode_Check(name))
# else
  if (!PyString_Check(name))
# endif
  {
    PyErr_Format(PyExc_TypeError, "attribute name must be string, not '%.200s'", name->ob_type->tp_name);
    return -1;
  } else {
    Py_INCREF(name);
  }

  if (!tp->tp_dict) {
    if (PyType_Ready(tp) < 0)
      goto done;
  }

  descr = _PyType_Lookup(tp, name);
  f = NULL;
  if (descr != NULL)
    f = descr->ob_type->tp_descr_set;
  if (!f) {
    if (PyString_Check(name)) {
      encoded_name = name;
      Py_INCREF(name);
    } else {
      encoded_name = PyUnicode_AsUTF8String(name);
      if (!encoded_name)
        return -1;
    }
    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '%.200s'", tp->tp_name, PyString_AsString(encoded_name));
    Py_DECREF(encoded_name);
  } else {
    res = f(descr, obj, value);
  }
  
  done:
  Py_DECREF(name);
  return res;
}
#endif


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



#ifdef __cplusplus
extern "C" {
#endif

/* Method creation and docstring support functions */

SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name);
SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);
SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func);

#ifdef __cplusplus
}
#endif


  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_Axis swig_types[0]
#define SWIGTYPE_p_Base swig_types[1]
#define SWIGTYPE_p_Event swig_types[2]
#define SWIGTYPE_p_Hist1DBase swig_types[3]
#define SWIGTYPE_p_NetworkSimplex swig_types[4]
#define SWIGTYPE_p_PairwiseDistance swig_types[5]
#define SWIGTYPE_p_ParticleCollection swig_types[6]
#define SWIGTYPE_p_Self swig_types[7]
#define SWIGTYPE_p_Transform swig_types[8]
#define SWIGTYPE_p_WeightCollection swig_types[9]
#define SWIGTYPE_p_allocator_type swig_types[10]
#define SWIGTYPE_p_char swig_types[11]
#define SWIGTYPE_p_difference_type swig_types[12]
#define SWIGTYPE_p_double swig_types[13]
#define SWIGTYPE_p_first_type swig_types[14]
#define SWIGTYPE_p_p_PyObject swig_types[15]
#define SWIGTYPE_p_p_double swig_types[16]
#define SWIGTYPE_p_p_float swig_types[17]
#define SWIGTYPE_p_second_type swig_types[18]
#define SWIGTYPE_p_size_type swig_types[19]
#define SWIGTYPE_p_std__allocatorT_double_t swig_types[20]
#define SWIGTYPE_p_std__allocatorT_float_t swig_types[21]
#define SWIGTYPE_p_std__allocatorT_std__string_t swig_types[22]
#define SWIGTYPE_p_std__invalid_argument swig_types[23]
#define SWIGTYPE_p_std__ostream swig_types[24]
#define SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t swig_types[25]
#define SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t swig_types[26]
#define SWIGTYPE_p_std__ptrdiff_t swig_types[27]
#define SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t swig_types[28]
#define SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t swig_types[29]
#define SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t swig_types[30]
#define SWIGTYPE_p_swig__SwigPyIterator swig_types[31]
#define SWIGTYPE_p_value_type swig_types[32]
#define SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t swig_types[33]
#define SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t swig_types[34]
#define SWIGTYPE_p_wasserstein__EMDBaseT_double_t swig_types[35]
#define SWIGTYPE_p_wasserstein__EMDBaseT_float_t swig_types[36]
#define SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t swig_types[37]
#define SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t swig_types[38]
#define SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t swig_types[39]
#define SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t swig_types[40]
#define SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t swig_types[41]
#define SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t swig_types[42]
#define SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t swig_types[43]
#define SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t swig_types[44]
#define SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t swig_types[45]
#define SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t swig_types[46]
#define SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t swig_types[47]
#define SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t swig_types[48]
#define SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t swig_types[49]
#define SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t swig_types[50]
#define SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t swig_types[51]
#define SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t swig_types[52]
static swig_type_info *swig_types[54];
static swig_module_info swig_module = {swig_types, 53, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#ifdef SWIG_TypeQuery
# undef SWIG_TypeQuery
#endif
#define SWIG_TypeQuery SWIG_Python_TypeQuery

/*-----------------------------------------------
              @(target):= _wasserstein.so
  ------------------------------------------------*/
#if PY_VERSION_HEX >= 0x03000000
#  define SWIG_init    PyInit__wasserstein

#else
#  define SWIG_init    init_wasserstein

#endif
#define SWIG_name    "_wasserstein"

#define SWIGVERSION 0x040002 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) const_cast< void * >(static_cast< const void * >(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),reinterpret_cast< void** >(a)) 


#include <stdexcept>


namespace swig {
  class SwigPtr_PyObject {
  protected:
    PyObject *_obj;

  public:
    SwigPtr_PyObject() :_obj(0)
    {
    }

    SwigPtr_PyObject(const SwigPtr_PyObject& item) : _obj(item._obj)
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XINCREF(_obj);      
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    SwigPtr_PyObject(PyObject *obj, bool initial_ref = true) :_obj(obj)
    {
      if (initial_ref) {
        SWIG_PYTHON_THREAD_BEGIN_BLOCK;
        Py_XINCREF(_obj);
        SWIG_PYTHON_THREAD_END_BLOCK;
      }
    }
    
    SwigPtr_PyObject & operator=(const SwigPtr_PyObject& item) 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XINCREF(item._obj);
      Py_XDECREF(_obj);
      _obj = item._obj;
      SWIG_PYTHON_THREAD_END_BLOCK;
      return *this;      
    }
    
    ~SwigPtr_PyObject() 
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      Py_XDECREF(_obj);
      SWIG_PYTHON_THREAD_END_BLOCK;
    }
    
    operator PyObject *() const
    {
      return _obj;
    }

    PyObject *operator->() const
    {
      return _obj;
    }
  };
}


namespace swig {
  struct SwigVar_PyObject : SwigPtr_PyObject {
    SwigVar_PyObject(PyObject* obj = 0) : SwigPtr_PyObject(obj, false) { }
    
    SwigVar_PyObject & operator = (PyObject* obj)
    {
      Py_XDECREF(_obj);
      _obj = obj;
      return *this;      
    }
  };
}


// this controls some class access modifiers
#ifndef SWIG_WASSERSTEIN
# define SWIG_WASSERSTEIN
#endif

namespace wasserstein {}
using namespace wasserstein;


#include <typeinfo>
#include <stdexcept>


#if defined(__GNUC__)
#  if __GNUC__ == 2 && __GNUC_MINOR <= 96
#     define SWIG_STD_NOMODERN_STL
#  endif
#endif


#include <string>


#include <stddef.h>


#include <utility>


#include <iostream>

#if PY_VERSION_HEX >= 0x03020000
# define SWIGPY_SLICE_ARG(obj) ((PyObject*) (obj))
#else
# define SWIGPY_SLICE_ARG(obj) ((PySliceObject*) (obj))
#endif


namespace swig {
  struct stop_iteration {
  };

  struct SwigPyIterator {
  private:
    SwigPtr_PyObject _seq;

  protected:
    SwigPyIterator(PyObject *seq) : _seq(seq)
    {
    }
      
  public:
    virtual ~SwigPyIterator() {}

    // Access iterator method, required by Python
    virtual PyObject *value() const = 0;

    // Forward iterator method, required by Python
    virtual SwigPyIterator *incr(size_t n = 1) = 0;
    
    // Backward iterator method, very common in C++, but not required in Python
    virtual SwigPyIterator *decr(size_t /*n*/ = 1)
    {
      throw stop_iteration();
    }

    // Random access iterator methods, but not required in Python
    virtual ptrdiff_t distance(const SwigPyIterator &/*x*/) const
    {
      throw std::invalid_argument("operation not supported");
    }

    virtual bool equal (const SwigPyIterator &/*x*/) const
    {
      throw std::invalid_argument("operation not supported");
    }
    
    // C++ common/needed methods
    virtual SwigPyIterator *copy() const = 0;

    PyObject *next()     
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK; // disable threads       
      PyObject *obj = value();
      incr();       
      SWIG_PYTHON_THREAD_END_BLOCK; // re-enable threads
      return obj;     
    }

    /* Make an alias for Python 3.x */
    PyObject *__next__()
    {
      return next();
    }

    PyObject *previous()
    {
      SWIG_PYTHON_THREAD_BEGIN_BLOCK; // disable threads       
      decr();
      PyObject *obj = value();
      SWIG_PYTHON_THREAD_END_BLOCK; // re-enable threads       
      return obj;
    }

    SwigPyIterator *advance(ptrdiff_t n)
    {
      return  (n > 0) ?  incr(n) : decr(-n);
    }
      
    bool operator == (const SwigPyIterator& x)  const
    {
      return equal(x);
    }
      
    bool operator != (const SwigPyIterator& x) const
    {
      return ! operator==(x);
    }
      
    SwigPyIterator& operator += (ptrdiff_t n)
    {
      return *advance(n);
    }

    SwigPyIterator& operator -= (ptrdiff_t n)
    {
      return *advance(-n);
    }
      
    SwigPyIterator* operator + (ptrdiff_t n) const
    {
      return copy()->advance(n);
    }

    SwigPyIterator* operator - (ptrdiff_t n) const
    {
      return copy()->advance(-n);
    }
      
    ptrdiff_t operator - (const SwigPyIterator& x) const
    {
      return x.distance(*this);
    }
      
    static swig_type_info* descriptor() {
      static int init = 0;
      static swig_type_info* desc = 0;
      if (!init) {
	desc = SWIG_TypeQuery("swig::SwigPyIterator *");
	init = 1;
      }	
      return desc;
    }    
  };

#if defined(SWIGPYTHON_BUILTIN)
  inline PyObject* make_output_iterator_builtin (PyObject *pyself)
  {
    Py_INCREF(pyself);
    return pyself;
  }
#endif
}


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
#if PY_VERSION_HEX < 0x03000000
  } else if (PyInt_Check(obj)) {
    if (val) *val = (double) PyInt_AsLong(obj);
    return SWIG_OK;
#endif
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else
#endif
  if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (PyObject *obj, unsigned long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    unsigned long long v = PyLong_AsUnsignedLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, 0, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
      if (val) *val = (unsigned long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
  }
#endif
  return res;
}


  #define SWIG_From_long   PyInt_FromLong 


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_long_SS_long  (long long value)
{
  return ((value < LONG_MIN) || (value > LONG_MAX)) ?
    PyLong_FromLongLong(value) : PyInt_FromLong(static_cast< long >(value));
}
#endif


SWIGINTERNINLINE PyObject *
SWIG_From_ptrdiff_t  (ptrdiff_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(ptrdiff_t) <= sizeof(long)) {
#endif
    return SWIG_From_long  (static_cast< long >(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(ptrdiff_t) <= sizeof(long long) */
    return SWIG_From_long_SS_long  (static_cast< long long >(value));
  }
#endif
}


SWIGINTERNINLINE PyObject*
  SWIG_From_bool  (bool value)
{
  return PyBool_FromLong(value ? 1 : 0);
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else
#endif
  if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      return SWIG_OverflowError;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_long_SS_long (PyObject *obj, long long *val)
{
  int res = SWIG_TypeError;
  if (PyLong_Check(obj)) {
    long long v = PyLong_AsLongLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
      res = SWIG_OverflowError;
    }
  } else {
    long v;
    res = SWIG_AsVal_long (obj,&v);
    if (SWIG_IsOK(res)) {
      if (val) *val = v;
      return res;
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    const double mant_max = 1LL << DBL_MANT_DIG;
    const double mant_min = -mant_max;
    double d;
    res = SWIG_AsVal_double (obj,&d);
    if (SWIG_IsOK(res) && !SWIG_CanCastAsInteger(&d, mant_min, mant_max))
      return SWIG_OverflowError;
    if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, mant_min, mant_max)) {
      if (val) *val = (long long)(d);
      return SWIG_AddCast(res);
    }
    res = SWIG_TypeError;
  }
#endif
  return res;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_ptrdiff_t (PyObject * obj, ptrdiff_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(ptrdiff_t) <= sizeof(long)) {
#endif
    long v;
    res = SWIG_AsVal_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(ptrdiff_t) <= sizeof(long long)) {
    long long v;
    res = SWIG_AsVal_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
  }
#endif
  return res;
}


#include <algorithm>


#include <vector>


namespace swig {
  template <class Type>
  struct noconst_traits {
    typedef Type noconst_type;
  };

  template <class Type>
  struct noconst_traits<const Type> {
    typedef Type noconst_type;
  };

  /*
    type categories
  */
  struct pointer_category { };
  struct value_category { };

  /*
    General traits that provides type_name and type_info
  */
  template <class Type> struct traits { };

  template <class Type>
  inline const char* type_name() {
    return traits<typename noconst_traits<Type >::noconst_type >::type_name();
  }

  template <class Type> struct traits_info {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  /*
    Partial specialization for pointers (traits_info)
  */
  template <class Type> struct traits_info<Type *> {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  template <class Type>
  inline swig_type_info *type_info() {
    return traits_info<Type>::type_info();
  }

  /*
    Partial specialization for pointers (traits)
  */
  template <class Type> struct traits <Type *> {
    typedef pointer_category category;
    static std::string make_ptr_name(const char* name) {
      std::string ptrname = name;
      ptrname += " *";
      return ptrname;
    }
    static const char* type_name() {
      static std::string name = make_ptr_name(swig::type_name<Type>());
      return name.c_str();
    }
  };

  template <class Type, class Category>
  struct traits_as { };

  template <class Type, class Category>
  struct traits_check { };

}


namespace swig {  
  /*
    Traits that provides the from method
  */
  template <class Type> struct traits_from_ptr {
    static PyObject *from(Type *val, int owner = 0) {
      return SWIG_InternalNewPointerObj(val, type_info<Type>(), owner);
    }
  };

  template <class Type> struct traits_from {
    static PyObject *from(const Type& val) {
      return traits_from_ptr<Type>::from(new Type(val), 1);
    }
  };

  template <class Type> struct traits_from<Type *> {
    static PyObject *from(Type* val) {
      return traits_from_ptr<Type>::from(val, 0);
    }
  };

  template <class Type> struct traits_from<const Type *> {
    static PyObject *from(const Type* val) {
      return traits_from_ptr<Type>::from(const_cast<Type*>(val), 0);
    }
  };


  template <class Type>
  inline PyObject *from(const Type& val) {
    return traits_from<Type>::from(val);
  }

  template <class Type>
  inline PyObject *from_ptr(Type* val, int owner) {
    return traits_from_ptr<Type>::from(val, owner);
  }

  /*
    Traits that provides the asval/as/check method
  */
  template <class Type>
  struct traits_asptr {   
    static int asptr(PyObject *obj, Type **val) {
      int res = SWIG_ERROR;
      swig_type_info *descriptor = type_info<Type>();
      if (val) {
        Type *p = 0;
        int newmem = 0;
        res = descriptor ? SWIG_ConvertPtrAndOwn(obj, (void **)&p, descriptor, 0, &newmem) : SWIG_ERROR;
        if (SWIG_IsOK(res)) {
          if (newmem & SWIG_CAST_NEW_MEMORY) {
            res |= SWIG_NEWOBJMASK;
          }
          *val = p;
        }
      } else {
        res = descriptor ? SWIG_ConvertPtr(obj, 0, descriptor, 0) : SWIG_ERROR;
      }
      return res;
    }
  }; 

  template <class Type>
  inline int asptr(PyObject *obj, Type **vptr) {
    return traits_asptr<Type>::asptr(obj, vptr);
  }

  template <class Type> 
  struct traits_asval {
    static int asval(PyObject *obj, Type *val) {
      if (val) {
	Type *p = 0;
	int res = traits_asptr<Type>::asptr(obj, &p);
	if (!SWIG_IsOK(res)) return res;	
	if (p) {
	  typedef typename noconst_traits<Type>::noconst_type noconst_type;
	  *(const_cast<noconst_type*>(val)) = *p;
	  if (SWIG_IsNewObj(res)){
	    delete p;
	    res = SWIG_DelNewMask(res);
	  }
	  return res;
	} else {
	  return SWIG_ERROR;
	}
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };

  template <class Type> struct traits_asval<Type*> {
    static int asval(PyObject *obj, Type **val) {
      if (val) {
        typedef typename noconst_traits<Type>::noconst_type noconst_type;
        noconst_type *p = 0;
        int res = traits_asptr<noconst_type>::asptr(obj,  &p);
        if (SWIG_IsOK(res)) {
          *(const_cast<noconst_type**>(val)) = p;
	}
	return res;
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };
  
  template <class Type>
  inline int asval(PyObject *obj, Type *val) {
    return traits_asval<Type>::asval(obj, val);
  }

  template <class Type> 
  struct traits_as<Type, value_category> {
    static Type as(PyObject *obj) {
      Type v;
      int res = asval(obj, &v);
      if (!obj || !SWIG_IsOK(res)) {
	if (!PyErr_Occurred()) {
	  ::SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
      return v;
    }
  };

  template <class Type> 
  struct traits_as<Type, pointer_category> {
    static Type as(PyObject *obj) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res) && v) {
	if (SWIG_IsNewObj(res)) {
	  Type r(*v);
	  delete v;
	  return r;
	} else {
	  return *v;
	}
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
    }
  };

  template <class Type> 
  struct traits_as<Type*, pointer_category> {
    static Type* as(PyObject *obj) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res)) {
	return v;
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	throw std::invalid_argument("bad type");
      }
    }
  };
    
  template <class Type>
  inline Type as(PyObject *obj) {
    return traits_as<Type, typename traits<Type>::category>::as(obj);
  }

  template <class Type> 
  struct traits_check<Type, value_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asval(obj, (Type *)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type> 
  struct traits_check<Type, pointer_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asptr(obj, (Type **)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type>
  inline bool check(PyObject *obj) {
    return traits_check<Type, typename traits<Type>::category>::check(obj);
  }
}


#include <functional>

namespace std {
  template <>
  struct less <PyObject *>
  {
    bool
    operator()(PyObject * v, PyObject *w) const
    { 
      bool res;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      res = PyObject_RichCompareBool(v, w, Py_LT) ? true : false;
      /* This may fall into a case of inconsistent
               eg. ObjA > ObjX > ObjB
               but ObjA < ObjB
      */
      if( PyErr_Occurred() && PyErr_ExceptionMatches(PyExc_TypeError) )
      {
        /* Objects can't be compared, this mostly occurred in Python 3.0 */
        /* Compare their ptr directly for a workaround */
        res = (v < w);
        PyErr_Clear();
      }
      SWIG_PYTHON_THREAD_END_BLOCK;
      return res;
    }
  };

  template <>
  struct less <swig::SwigPtr_PyObject>
  {
    bool
    operator()(const swig::SwigPtr_PyObject& v, const swig::SwigPtr_PyObject& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

  template <>
  struct less <swig::SwigVar_PyObject>
  {
    bool
    operator()(const swig::SwigVar_PyObject& v, const swig::SwigVar_PyObject& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

}

namespace swig {
  template <> struct traits<PyObject *> {
    typedef value_category category;
    static const char* type_name() { return "PyObject *"; }
  };  

  template <>  struct traits_asval<PyObject * > {   
    typedef PyObject * value_type;
    static int asval(PyObject *obj, value_type *val) {
      if (val) *val = obj;
      return SWIG_OK;
    }
  };

  template <> 
  struct traits_check<PyObject *, value_category> {
    static bool check(PyObject *) {
      return true;
    }
  };

  template <>  struct traits_from<PyObject *> {
    typedef PyObject * value_type;
    static PyObject *from(const value_type& val) {
      Py_XINCREF(val);
      return val;
    }
  };
  
}

namespace swig {
  template <class Difference>
  inline size_t
  check_index(Difference i, size_t size, bool insert = false) {
    if ( i < 0 ) {
      if ((size_t) (-i) <= size)
	return (size_t) (i + size);
    } else if ( (size_t) i < size ) {
      return (size_t) i;
    } else if (insert && ((size_t) i == size)) {
      return size;
    }
    throw std::out_of_range("index out of range");
  }

  template <class Difference>
  void
  slice_adjust(Difference i, Difference j, Py_ssize_t step, size_t size, Difference &ii, Difference &jj, bool insert = false) {
    if (step == 0) {
      throw std::invalid_argument("slice step cannot be zero");
    } else if (step > 0) {
      // Required range: 0 <= i < size, 0 <= j < size, i <= j
      if (i < 0) {
        ii = 0;
      } else if (i < (Difference)size) {
        ii = i;
      } else if (insert && (i >= (Difference)size)) {
        ii = (Difference)size;
      }
      if (j < 0) {
        jj = 0;
      } else {
        jj = (j < (Difference)size) ? j : (Difference)size;
      }
      if (jj < ii)
        jj = ii;
    } else {
      // Required range: -1 <= i < size-1, -1 <= j < size-1, i >= j
      if (i < -1) {
        ii = -1;
      } else if (i < (Difference) size) {
        ii = i;
      } else if (i >= (Difference)(size-1)) {
        ii = (Difference)(size-1);
      }
      if (j < -1) {
        jj = -1;
      } else {
        jj = (j < (Difference)size ) ? j : (Difference)(size-1);
      }
      if (ii < jj)
        ii = jj;
    }
  }

  template <class Sequence, class Difference>
  inline typename Sequence::iterator
  getpos(Sequence* self, Difference i)  {
    typename Sequence::iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence, class Difference>
  inline typename Sequence::const_iterator
  cgetpos(const Sequence* self, Difference i)  {
    typename Sequence::const_iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence>
  inline void
  erase(Sequence* seq, const typename Sequence::iterator& position) {
    seq->erase(position);
  }

  template <class Sequence>
  struct traits_reserve {
    static void reserve(Sequence & /*seq*/, typename Sequence::size_type /*n*/) {
      // This should be specialized for types that support reserve
    }
  };

  template <class Sequence, class Difference>
  inline Sequence*
  getslice(const Sequence* self, Difference i, Difference j, Py_ssize_t step) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj);

    if (step > 0) {
      typename Sequence::const_iterator sb = self->begin();
      typename Sequence::const_iterator se = self->begin();
      std::advance(sb,ii);
      std::advance(se,jj);
      if (step == 1) {
        return new Sequence(sb, se);
      } else {
        Sequence *sequence = new Sequence();
        swig::traits_reserve<Sequence>::reserve(*sequence, (jj - ii + step - 1) / step);
        typename Sequence::const_iterator it = sb;
        while (it!=se) {
          sequence->push_back(*it);
          for (Py_ssize_t c=0; c<step && it!=se; ++c)
            it++;
        }
        return sequence;
      } 
    } else {
      Sequence *sequence = new Sequence();
      swig::traits_reserve<Sequence>::reserve(*sequence, (ii - jj - step - 1) / -step);
      typename Sequence::const_reverse_iterator sb = self->rbegin();
      typename Sequence::const_reverse_iterator se = self->rbegin();
      std::advance(sb,size-ii-1);
      std::advance(se,size-jj-1);
      typename Sequence::const_reverse_iterator it = sb;
      while (it!=se) {
        sequence->push_back(*it);
        for (Py_ssize_t c=0; c<-step && it!=se; ++c)
          it++;
      }
      return sequence;
    }
  }

  template <class Sequence, class Difference, class InputSeq>
  inline void
  setslice(Sequence* self, Difference i, Difference j, Py_ssize_t step, const InputSeq& is = InputSeq()) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj, true);
    if (step > 0) {
      if (step == 1) {
        size_t ssize = jj - ii;
        if (ssize <= is.size()) {
          // expanding/staying the same size
          swig::traits_reserve<Sequence>::reserve(*self, self->size() - ssize + is.size());
          typename Sequence::iterator sb = self->begin();
          typename InputSeq::const_iterator isit = is.begin();
          std::advance(sb,ii);
          std::advance(isit, jj - ii);
          self->insert(std::copy(is.begin(), isit, sb), isit, is.end());
        } else {
          // shrinking
          typename Sequence::iterator sb = self->begin();
          typename Sequence::iterator se = self->begin();
          std::advance(sb,ii);
          std::advance(se,jj);
          self->erase(sb,se);
          sb = self->begin();
          std::advance(sb,ii);
          self->insert(sb, is.begin(), is.end());
        }
      } else {
        size_t replacecount = (jj - ii + step - 1) / step;
        if (is.size() != replacecount) {
          char msg[1024];
          sprintf(msg, "attempt to assign sequence of size %lu to extended slice of size %lu", (unsigned long)is.size(), (unsigned long)replacecount);
          throw std::invalid_argument(msg);
        }
        typename Sequence::const_iterator isit = is.begin();
        typename Sequence::iterator it = self->begin();
        std::advance(it,ii);
        for (size_t rc=0; rc<replacecount && it != self->end(); ++rc) {
          *it++ = *isit++;
          for (Py_ssize_t c=0; c<(step-1) && it != self->end(); ++c)
            it++;
        }
      }
    } else {
      size_t replacecount = (ii - jj - step - 1) / -step;
      if (is.size() != replacecount) {
        char msg[1024];
        sprintf(msg, "attempt to assign sequence of size %lu to extended slice of size %lu", (unsigned long)is.size(), (unsigned long)replacecount);
        throw std::invalid_argument(msg);
      }
      typename Sequence::const_iterator isit = is.begin();
      typename Sequence::reverse_iterator it = self->rbegin();
      std::advance(it,size-ii-1);
      for (size_t rc=0; rc<replacecount && it != self->rend(); ++rc) {
        *it++ = *isit++;
        for (Py_ssize_t c=0; c<(-step-1) && it != self->rend(); ++c)
          it++;
      }
    }
  }

  template <class Sequence, class Difference>
  inline void
  delslice(Sequence* self, Difference i, Difference j, Py_ssize_t step) {
    typename Sequence::size_type size = self->size();
    Difference ii = 0;
    Difference jj = 0;
    swig::slice_adjust(i, j, step, size, ii, jj, true);
    if (step > 0) {
      typename Sequence::iterator sb = self->begin();
      std::advance(sb,ii);
      if (step == 1) {
        typename Sequence::iterator se = self->begin();
        std::advance(se,jj);
        self->erase(sb,se);
      } else {
        typename Sequence::iterator it = sb;
        size_t delcount = (jj - ii + step - 1) / step;
        while (delcount) {
          it = self->erase(it);
          for (Py_ssize_t c=0; c<(step-1) && it != self->end(); ++c)
            it++;
          delcount--;
        }
      }
    } else {
      typename Sequence::reverse_iterator sb = self->rbegin();
      std::advance(sb,size-ii-1);
      typename Sequence::reverse_iterator it = sb;
      size_t delcount = (ii - jj - step - 1) / -step;
      while (delcount) {
        it = typename Sequence::reverse_iterator(self->erase((++it).base()));
        for (Py_ssize_t c=0; c<(-step-1) && it != self->rend(); ++c)
          it++;
        delcount--;
      }
    }
  }
}


#if defined(__SUNPRO_CC) && defined(_RWSTD_VER)
#  if !defined(SWIG_NO_STD_NOITERATOR_TRAITS_STL)
#    define SWIG_STD_NOITERATOR_TRAITS_STL
#  endif
#endif

#if !defined(SWIG_STD_NOITERATOR_TRAITS_STL)
#include <iterator>
#else
namespace std {
  template <class Iterator>
  struct iterator_traits {
    typedef ptrdiff_t difference_type;
    typedef typename Iterator::value_type value_type;
  };

  template <class Iterator, class Category,class T, class Reference, class Pointer, class Distance>
  struct iterator_traits<__reverse_bi_iterator<Iterator,Category,T,Reference,Pointer,Distance> > {
    typedef Distance difference_type;
    typedef T value_type;
  };

  template <class T>
  struct iterator_traits<T*> {
    typedef T value_type;
    typedef ptrdiff_t difference_type;
  };

  template<typename _InputIterator>
  inline typename iterator_traits<_InputIterator>::difference_type
  distance(_InputIterator __first, _InputIterator __last)
  {
    typename iterator_traits<_InputIterator>::difference_type __n = 0;
    while (__first != __last) {
      ++__first; ++__n;
    }
    return __n;
  }
}
#endif


namespace swig {
  template<typename OutIterator>
  class SwigPyIterator_T :  public SwigPyIterator
  {
  public:
    typedef OutIterator out_iterator;
    typedef typename std::iterator_traits<out_iterator>::value_type value_type;    
    typedef SwigPyIterator_T<out_iterator> self_type;

    SwigPyIterator_T(out_iterator curr, PyObject *seq)
      : SwigPyIterator(seq), current(curr)
    {
    }

    const out_iterator& get_current() const
    {
      return current;
    }

    
    bool equal (const SwigPyIterator &iter) const
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return (current == iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }
    
    ptrdiff_t distance(const SwigPyIterator &iter) const
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return std::distance(current, iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }    
    
  protected:
    out_iterator current;
  };
  
  template <class ValueType>
  struct from_oper 
  {
    typedef const ValueType& argument_type;
    typedef PyObject *result_type;
    result_type operator()(argument_type v) const
    {
      return swig::from(v);
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyForwardIteratorOpen_T :  public SwigPyIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyForwardIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyForwardIteratorOpen_T(out_iterator curr, PyObject *seq)
      : SwigPyIterator_T<OutIterator>(curr, seq)
    {
    }
    
    PyObject *value() const {
      return from(static_cast<const value_type&>(*(base::current)));
    }
    
    SwigPyIterator *copy() const
    {
      return new self_type(*this);
    }

    SwigPyIterator *incr(size_t n = 1)
    {
      while (n--) {
	++base::current;
      }
      return this;
    }

  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyIteratorOpen_T :  public SwigPyForwardIteratorOpen_T<OutIterator, ValueType, FromOper>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyIteratorOpen_T(out_iterator curr, PyObject *seq)
      : SwigPyForwardIteratorOpen_T<OutIterator>(curr, seq)
    {
    }

    SwigPyIterator *decr(size_t n = 1)
    {
      while (n--) {
	--base::current;
      }
      return this;
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyForwardIteratorClosed_T :  public SwigPyIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;    
    typedef SwigPyForwardIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyForwardIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : SwigPyIterator_T<OutIterator>(curr, seq), begin(first), end(last)
    {
    }
    
    PyObject *value() const {
      if (base::current == end) {
	throw stop_iteration();
      } else {
	return from(static_cast<const value_type&>(*(base::current)));
      }
    }
    
    SwigPyIterator *copy() const
    {
      return new self_type(*this);
    }

    SwigPyIterator *incr(size_t n = 1)
    {
      while (n--) {
	if (base::current == end) {
	  throw stop_iteration();
	} else {
	  ++base::current;
	}
      }
      return this;
    }

  protected:
    out_iterator begin;
    out_iterator end;
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class SwigPyIteratorClosed_T :  public SwigPyForwardIteratorClosed_T<OutIterator,ValueType,FromOper>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef SwigPyIterator_T<out_iterator>  base;
    typedef SwigPyForwardIteratorClosed_T<OutIterator, ValueType, FromOper> base0;
    typedef SwigPyIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    SwigPyIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : SwigPyForwardIteratorClosed_T<OutIterator,ValueType,FromOper>(curr, first, last, seq)
    {
    }

    SwigPyIterator *decr(size_t n = 1)
    {
      while (n--) {
	if (base::current == base0::begin) {
	  throw stop_iteration();
	} else {
	  --base::current;
	}
      }
      return this;
    }
  };


  template<typename OutIter>
  inline SwigPyIterator*
  make_output_forward_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new SwigPyForwardIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new SwigPyIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_forward_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new SwigPyForwardIteratorOpen_T<OutIter>(current, seq);
  }

  template<typename OutIter>
  inline SwigPyIterator*
  make_output_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new SwigPyIteratorOpen_T<OutIter>(current, seq);
  }

}


namespace swig
{
  template <class T>
  struct SwigPySequence_Ref
  {
    SwigPySequence_Ref(PyObject* seq, Py_ssize_t index)
      : _seq(seq), _index(index)
    {
    }
    
    operator T () const
    {
      swig::SwigVar_PyObject item = PySequence_GetItem(_seq, _index);
      try {
	return swig::as<T>(item);
      } catch (const std::invalid_argument& e) {
	char msg[1024];
	sprintf(msg, "in sequence element %d ", (int)_index);
	if (!PyErr_Occurred()) {
	  ::SWIG_Error(SWIG_TypeError,  swig::type_name<T>());
	}
	SWIG_Python_AddErrorMsg(msg);
	SWIG_Python_AddErrorMsg(e.what());
	throw;
      }
    }

    SwigPySequence_Ref& operator=(const T& v)
    {
      PySequence_SetItem(_seq, _index, swig::from<T>(v));
      return *this;
    }

  private:
    PyObject* _seq;
    Py_ssize_t _index;
  };

  template <class T>
  struct SwigPySequence_ArrowProxy
  {
    SwigPySequence_ArrowProxy(const T& x): m_value(x) {}
    const T* operator->() const { return &m_value; }
    operator const T*() const { return &m_value; }
    T m_value;
  };

  template <class T, class Reference >
  struct SwigPySequence_InputIterator
  {
    typedef SwigPySequence_InputIterator<T, Reference > self;

    typedef std::random_access_iterator_tag iterator_category;
    typedef Reference reference;
    typedef T value_type;
    typedef T* pointer;
    typedef Py_ssize_t difference_type;

    SwigPySequence_InputIterator()
    {
    }

    SwigPySequence_InputIterator(PyObject* seq, Py_ssize_t index)
      : _seq(seq), _index(index)
    {
    }

    reference operator*() const
    {
      return reference(_seq, _index);
    }

    SwigPySequence_ArrowProxy<T>
    operator->() const {
      return SwigPySequence_ArrowProxy<T>(operator*());
    }

    bool operator==(const self& ri) const
    {
      return (_index == ri._index) && (_seq == ri._seq);
    }

    bool operator!=(const self& ri) const
    {
      return !(operator==(ri));
    }

    self& operator ++ ()
    {
      ++_index;
      return *this;
    }

    self& operator -- ()
    {
      --_index;
      return *this;
    }

    self& operator += (difference_type n)
    {
      _index += n;
      return *this;
    }

    self operator +(difference_type n) const
    {
      return self(_seq, _index + n);
    }

    self& operator -= (difference_type n)
    {
      _index -= n;
      return *this;
    }

    self operator -(difference_type n) const
    {
      return self(_seq, _index - n);
    }

    difference_type operator - (const self& ri) const
    {
      return _index - ri._index;
    }

    bool operator < (const self& ri) const
    {
      return _index < ri._index;
    }

    reference
    operator[](difference_type n) const
    {
      return reference(_seq, _index + n);
    }

  private:
    PyObject* _seq;
    difference_type _index;
  };

  // STL container wrapper around a Python sequence
  template <class T>
  struct SwigPySequence_Cont
  {
    typedef SwigPySequence_Ref<T> reference;
    typedef const SwigPySequence_Ref<T> const_reference;
    typedef T value_type;
    typedef T* pointer;
    typedef Py_ssize_t difference_type;
    typedef size_t size_type;
    typedef const pointer const_pointer;
    typedef SwigPySequence_InputIterator<T, reference> iterator;
    typedef SwigPySequence_InputIterator<T, const_reference> const_iterator;

    SwigPySequence_Cont(PyObject* seq) : _seq(0)
    {
      if (!PySequence_Check(seq)) {
	throw std::invalid_argument("a sequence is expected");
      }
      _seq = seq;
      Py_INCREF(_seq);
    }

    ~SwigPySequence_Cont()
    {
      Py_XDECREF(_seq);
    }

    size_type size() const
    {
      return static_cast<size_type>(PySequence_Size(_seq));
    }

    bool empty() const
    {
      return size() == 0;
    }

    iterator begin()
    {
      return iterator(_seq, 0);
    }

    const_iterator begin() const
    {
      return const_iterator(_seq, 0);
    }

    iterator end()
    {
      return iterator(_seq, size());
    }

    const_iterator end() const
    {
      return const_iterator(_seq, size());
    }

    reference operator[](difference_type n)
    {
      return reference(_seq, n);
    }

    const_reference operator[](difference_type n)  const
    {
      return const_reference(_seq, n);
    }

    bool check() const
    {
      Py_ssize_t s = size();
      for (Py_ssize_t i = 0; i < s; ++i) {
	swig::SwigVar_PyObject item = PySequence_GetItem(_seq, i);
	if (!swig::check<value_type>(item))
	  return false;
      }
      return true;
    }

  private:
    PyObject* _seq;
  };

}


  #define SWIG_From_double   PyFloat_FromDouble 


namespace swig {
  template <> struct traits< double > {
    typedef value_category category;
    static const char* type_name() { return"double"; }
  };
  template <>  struct traits_asval< double > {
    typedef double value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_double (obj, val);
    }
  };
  template <>  struct traits_from< double > {
    typedef double value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_double  (val);
    }
  };
}


namespace swig {
  template <class SwigPySeq, class Seq>
  inline void
  assign(const SwigPySeq& swigpyseq, Seq* seq) {
    // seq->assign(swigpyseq.begin(), swigpyseq.end()); // not used as not always implemented
    typedef typename SwigPySeq::value_type value_type;
    typename SwigPySeq::const_iterator it = swigpyseq.begin();
    for (;it != swigpyseq.end(); ++it) {
      seq->insert(seq->end(),(value_type)(*it));
    }
  }

  template <class Seq, class T = typename Seq::value_type >
  struct traits_asptr_stdseq {
    typedef Seq sequence;
    typedef T value_type;

    static int asptr(PyObject *obj, sequence **seq) {
      if (obj == Py_None || SWIG_Python_GetSwigThis(obj)) {
	sequence *p;
	swig_type_info *descriptor = swig::type_info<sequence>();
	if (descriptor && SWIG_IsOK(::SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0))) {
	  if (seq) *seq = p;
	  return SWIG_OLDOBJ;
	}
      } else if (PySequence_Check(obj)) {
	try {
	  SwigPySequence_Cont<value_type> swigpyseq(obj);
	  if (seq) {
	    sequence *pseq = new sequence();
	    assign(swigpyseq, pseq);
	    *seq = pseq;
	    return SWIG_NEWOBJ;
	  } else {
	    return swigpyseq.check() ? SWIG_OK : SWIG_ERROR;
	  }
	} catch (std::exception& e) {
	  if (seq) {
	    if (!PyErr_Occurred()) {
	      PyErr_SetString(PyExc_TypeError, e.what());
	    }
	  }
	  return SWIG_ERROR;
	}
      }
      return SWIG_ERROR;
    }
  };

  template <class Seq, class T = typename Seq::value_type >
  struct traits_from_stdseq {
    typedef Seq sequence;
    typedef T value_type;
    typedef typename Seq::size_type size_type;
    typedef typename sequence::const_iterator const_iterator;

    static PyObject *from(const sequence& seq) {
#ifdef SWIG_PYTHON_EXTRA_NATIVE_CONTAINERS
      swig_type_info *desc = swig::type_info<sequence>();
      if (desc && desc->clientdata) {
	return SWIG_InternalNewPointerObj(new sequence(seq), desc, SWIG_POINTER_OWN);
      }
#endif
      size_type size = seq.size();
      if (size <= (size_type)INT_MAX) {
	PyObject *obj = PyTuple_New((Py_ssize_t)size);
	Py_ssize_t i = 0;
	for (const_iterator it = seq.begin(); it != seq.end(); ++it, ++i) {
	  PyTuple_SetItem(obj,i,swig::from<value_type>(*it));
	}
	return obj;
      } else {
	PyErr_SetString(PyExc_OverflowError,"sequence size not valid in python");
	return NULL;
      }
    }
  };
}


  namespace swig {
    template <class T>
    struct traits_reserve<std::vector<T> > {
      static void reserve(std::vector<T> &seq, typename std::vector<T>::size_type n) {
        seq.reserve(n);
      }
    };

    template <class T>
    struct traits_asptr<std::vector<T> >  {
      static int asptr(PyObject *obj, std::vector<T> **vec) {
	return traits_asptr_stdseq<std::vector<T> >::asptr(obj, vec);
      }
    };
    
    template <class T>
    struct traits_from<std::vector<T> > {
      static PyObject *from(const std::vector<T>& vec) {
	return traits_from_stdseq<std::vector<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector< double, std::allocator< double > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "double" "," "std::allocator< double >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_double_Sg__iterator(std::vector< double > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_double_Sg____nonzero__(std::vector< double > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_double_Sg____bool__(std::vector< double > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< double >::size_type std_vector_Sl_double_Sg____len__(std::vector< double > const *self){
      return self->size();
    }

SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong(static_cast< long >(value));
}


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long_SS_long  (unsigned long long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLongLong(value) : PyInt_FromLong(static_cast< long >(value));
}
#endif


SWIGINTERNINLINE PyObject *
SWIG_From_size_t  (size_t value)
{    
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    return SWIG_From_unsigned_SS_long  (static_cast< unsigned long >(value));
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else {
    /* assume sizeof(size_t) <= sizeof(unsigned long long) */
    return SWIG_From_unsigned_SS_long_SS_long  (static_cast< unsigned long long >(value));
  }
#endif
}

SWIGINTERN std::vector< double,std::allocator< double > > *std_vector_Sl_double_Sg____getslice__(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setslice____SWIG_0(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< double,std::allocator< double > >());
    }
SWIGINTERN void std_vector_Sl_double_Sg____setslice____SWIG_1(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j,std::vector< double,std::allocator< double > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delslice__(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delitem____SWIG_0(std::vector< double > *self,std::vector< double >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< double,std::allocator< double > > *std_vector_Sl_double_Sg____getitem____SWIG_0(std::vector< double > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setitem____SWIG_0(std::vector< double > *self,PySliceObject *slice,std::vector< double,std::allocator< double > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setitem____SWIG_1(std::vector< double > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delitem____SWIG_1(std::vector< double > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< double,std::allocator< double > >::difference_type id = i;
      std::vector< double,std::allocator< double > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< double >::value_type const &std_vector_Sl_double_Sg____getitem____SWIG_1(std::vector< double > const *self,std::vector< double >::difference_type i){
      return *(swig::cgetpos(self, i));
    }

namespace swig {
  static PyObject* container_owner_attribute() {
    static PyObject* attr = SWIG_Python_str_FromChar("__swig_container");
    return attr;
  }

  template <typename T>
  struct container_owner {
    // By default, do not add the back-reference (for value types)
    // Specialization below will check the reference for pointer types.
    static bool back_reference(PyObject* /*child*/, PyObject* /*owner*/) {
      return false;
    }
  };

  template <>
  struct container_owner<swig::pointer_category> {  
    /*
     * Call to add a back-reference to the owning object when returning a 
     * reference from a container.  Will only set the reference if child
     * is a SWIG wrapper object that does not own the pointer.
     *
     * returns whether the reference was set or not
     */
    static bool back_reference(PyObject* child, PyObject* owner) {
      SwigPyObject* swigThis = SWIG_Python_GetSwigThis(child);
      if (swigThis && (swigThis->own & SWIG_POINTER_OWN) != SWIG_POINTER_OWN) {
        return PyObject_SetAttr(child, container_owner_attribute(), owner) != -1;
      }
      return false;
    }
  };
}

SWIGINTERN void std_vector_Sl_double_Sg____setitem____SWIG_2(std::vector< double > *self,std::vector< double >::difference_type i,std::vector< double >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< double >::value_type std_vector_Sl_double_Sg__pop(std::vector< double > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< double,std::allocator< double > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_double_Sg__append(std::vector< double > *self,std::vector< double >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< double >::iterator std_vector_Sl_double_Sg__erase__SWIG_0(std::vector< double > *self,std::vector< double >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< double >::iterator std_vector_Sl_double_Sg__erase__SWIG_1(std::vector< double > *self,std::vector< double >::iterator first,std::vector< double >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< double >::iterator std_vector_Sl_double_Sg__insert__SWIG_0(std::vector< double > *self,std::vector< double >::iterator pos,std::vector< double >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_double_Sg__insert__SWIG_1(std::vector< double > *self,std::vector< double >::iterator pos,std::vector< double >::size_type n,std::vector< double >::value_type const &x){ self->insert(pos, n, x); }

SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
  if (PyBytes_Check(obj))
#else
  if (PyUnicode_Check(obj))
#endif
#else  
  if (PyString_Check(obj))
#endif
  {
    char *cstr; Py_ssize_t len;
    int ret = SWIG_OK;
#if PY_VERSION_HEX>=0x03000000
#if !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    if (!alloc && cptr) {
        /* We can't allow converting without allocation, since the internal
           representation of string in Python 3 is UCS-2/UCS-4 but we require
           a UTF-8 representation.
           TODO(bhy) More detailed explanation */
        return SWIG_RuntimeError;
    }
    obj = PyUnicode_AsUTF8String(obj);
    if (!obj)
      return SWIG_TypeError;
    if (alloc)
      *alloc = SWIG_NEWOBJ;
#endif
    if (PyBytes_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#else
    if (PyString_AsStringAndSize(obj, &cstr, &len) == -1)
      return SWIG_TypeError;
#endif
    if (cptr) {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
	  *alloc = SWIG_NEWOBJ;
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
#if PY_VERSION_HEX>=0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
	*cptr = PyBytes_AsString(obj);
#else
	assert(0); /* Should never reach here with Unicode strings in Python 3 */
#endif
#else
	*cptr = SWIG_Python_str_AsChar(obj);
        if (!*cptr)
          ret = SWIG_TypeError;
#endif
      }
    }
    if (psize) *psize = len + 1;
#if PY_VERSION_HEX>=0x03000000 && !defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
    Py_XDECREF(obj);
#endif
    return ret;
  } else {
#if defined(SWIG_PYTHON_2_UNICODE)
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
#error "Cannot use both SWIG_PYTHON_2_UNICODE and SWIG_PYTHON_STRICT_BYTE_CHAR at once"
#endif
#if PY_VERSION_HEX<0x03000000
    if (PyUnicode_Check(obj)) {
      char *cstr; Py_ssize_t len;
      if (!alloc && cptr) {
        return SWIG_RuntimeError;
      }
      obj = PyUnicode_AsUTF8String(obj);
      if (!obj)
        return SWIG_TypeError;
      if (PyString_AsStringAndSize(obj, &cstr, &len) != -1) {
        if (cptr) {
          if (alloc) *alloc = SWIG_NEWOBJ;
          *cptr = reinterpret_cast< char* >(memcpy(new char[len + 1], cstr, sizeof(char)*(len + 1)));
        }
        if (psize) *psize = len + 1;

        Py_XDECREF(obj);
        return SWIG_OK;
      } else {
        Py_XDECREF(obj);
      }
    }
#endif
#endif

    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsPtr_std_string (PyObject * obj, std::string **val) 
{
  char* buf = 0 ; size_t size = 0; int alloc = SWIG_OLDOBJ;
  if (SWIG_IsOK((SWIG_AsCharPtrAndSize(obj, &buf, &size, &alloc)))) {
    if (buf) {
      if (val) *val = new std::string(buf, size - 1);
      if (alloc == SWIG_NEWOBJ) delete[] buf;
      return SWIG_NEWOBJ;
    } else {
      if (val) *val = 0;
      return SWIG_OLDOBJ;
    }
  } else {
    static int init = 0;
    static swig_type_info* descriptor = 0;
    if (!init) {
      descriptor = SWIG_TypeQuery("std::string" " *");
      init = 1;
    }
    if (descriptor) {
      std::string *vptr;
      int res = SWIG_ConvertPtr(obj, (void**)&vptr, descriptor, 0);
      if (SWIG_IsOK(res) && val) *val = vptr;
      return res;
    }
  }
  return SWIG_ERROR;
}


SWIGINTERN int
SWIG_AsVal_std_string (PyObject * obj, std::string *val)
{
  std::string* v = (std::string *) 0;
  int res = SWIG_AsPtr_std_string (obj, &v);
  if (!SWIG_IsOK(res)) return res;
  if (v) {
    if (val) *val = *v;
    if (SWIG_IsNewObj(res)) {
      delete v;
      res = SWIG_DelNewMask(res);
    }
    return res;
  }
  return SWIG_ERROR;
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_InternalNewPointerObj(const_cast< char * >(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
#if PY_VERSION_HEX >= 0x03000000
#if defined(SWIG_PYTHON_STRICT_BYTE_CHAR)
      return PyBytes_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#else
      return PyUnicode_DecodeUTF8(carray, static_cast< Py_ssize_t >(size), "surrogateescape");
#endif
#else
      return PyString_FromStringAndSize(carray, static_cast< Py_ssize_t >(size));
#endif
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject *
SWIG_From_std_string  (const std::string& s)
{
  return SWIG_FromCharPtrAndSize(s.data(), s.size());
}


namespace swig {
  template <> struct traits< std::string > {
    typedef value_category category;
    static const char* type_name() { return"std::string"; }
  };
  template <>  struct traits_asval< std::string > {
    typedef std::string value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_std_string (obj, val);
    }
  };
  template <>  struct traits_from< std::string > {
    typedef std::string value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_std_string  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector< std::string, std::allocator< std::string > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::string" "," "std::allocator< std::string >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_std_string_Sg__iterator(std::vector< std::string > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_string_Sg____nonzero__(std::vector< std::string > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_std_string_Sg____bool__(std::vector< std::string > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< std::string >::size_type std_vector_Sl_std_string_Sg____len__(std::vector< std::string > const *self){
      return self->size();
    }
SWIGINTERN std::vector< std::string,std::allocator< std::string > > *std_vector_Sl_std_string_Sg____getslice__(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setslice____SWIG_0(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< std::string,std::allocator< std::string > >());
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setslice____SWIG_1(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j,std::vector< std::string,std::allocator< std::string > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delslice__(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delitem____SWIG_0(std::vector< std::string > *self,std::vector< std::string >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< std::string,std::allocator< std::string > > *std_vector_Sl_std_string_Sg____getitem____SWIG_0(std::vector< std::string > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem____SWIG_0(std::vector< std::string > *self,PySliceObject *slice,std::vector< std::string,std::allocator< std::string > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem____SWIG_1(std::vector< std::string > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delitem____SWIG_1(std::vector< std::string > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< std::string,std::allocator< std::string > >::difference_type id = i;
      std::vector< std::string,std::allocator< std::string > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< std::string >::value_type const &std_vector_Sl_std_string_Sg____getitem____SWIG_1(std::vector< std::string > const *self,std::vector< std::string >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem____SWIG_2(std::vector< std::string > *self,std::vector< std::string >::difference_type i,std::vector< std::string >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< std::string >::value_type std_vector_Sl_std_string_Sg__pop(std::vector< std::string > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< std::string,std::allocator< std::string > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_std_string_Sg__append(std::vector< std::string > *self,std::vector< std::string >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< std::string >::iterator std_vector_Sl_std_string_Sg__erase__SWIG_0(std::vector< std::string > *self,std::vector< std::string >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< std::string >::iterator std_vector_Sl_std_string_Sg__erase__SWIG_1(std::vector< std::string > *self,std::vector< std::string >::iterator first,std::vector< std::string >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< std::string >::iterator std_vector_Sl_std_string_Sg__insert__SWIG_0(std::vector< std::string > *self,std::vector< std::string >::iterator pos,std::vector< std::string >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_std_string_Sg__insert__SWIG_1(std::vector< std::string > *self,std::vector< std::string >::iterator pos,std::vector< std::string >::size_type n,std::vector< std::string >::value_type const &x){ self->insert(pos, n, x); }

  namespace swig {


















































    template <class T, class U >
    struct traits_asptr<std::pair<T,U> >  {
      typedef std::pair<T,U> value_type;

      static int get_pair(PyObject* first, PyObject* second,
			  std::pair<T,U> **val) 
      {
	if (val) {
	  value_type *vp = (new std::pair<T,U>());
	  T *pfirst = &(vp->first);
	  int res1 = swig::asval((PyObject*)first, pfirst);
	  if (!SWIG_IsOK(res1)) {
	    delete vp;
	    return res1;
	  }
	  U *psecond = &(vp->second);
	  int res2 = swig::asval((PyObject*)second, psecond);
	  if (!SWIG_IsOK(res2)) {
	    delete vp;
	    return res2;
	  }
	  *val = vp;
	  return SWIG_AddNewMask(res1 > res2 ? res1 : res2);
	} else {
	  T *pfirst = 0;
	  int res1 = swig::asval((PyObject*)first, pfirst);
	  if (!SWIG_IsOK(res1)) return res1;
	  U *psecond = 0;
	  int res2 = swig::asval((PyObject*)second, psecond);
	  if (!SWIG_IsOK(res2)) return res2;
	  return res1 > res2 ? res1 : res2;
	}
      }

      static int asptr(PyObject *obj, std::pair<T,U> **val) {
	int res = SWIG_ERROR;
	if (PyTuple_Check(obj)) {
	  if (PyTuple_GET_SIZE(obj) == 2) {
	    res = get_pair(PyTuple_GET_ITEM(obj,0),PyTuple_GET_ITEM(obj,1), val);
	  }
	} else if (PySequence_Check(obj)) {
	  if (PySequence_Size(obj) == 2) {
	    swig::SwigVar_PyObject first = PySequence_GetItem(obj,0);
	    swig::SwigVar_PyObject second = PySequence_GetItem(obj,1);
	    res = get_pair(first, second, val);
	  }
	} else {
	  value_type *p = 0;
	  swig_type_info *descriptor = swig::type_info<value_type>();
	  res = descriptor ? SWIG_ConvertPtr(obj, (void **)&p, descriptor, 0) : SWIG_ERROR;
	  if (SWIG_IsOK(res) && val)  *val = p;
	}
	return res;
      }
    };


    template <class T, class U >
    struct traits_from<std::pair<T,U> >   {
      static PyObject *from(const std::pair<T,U>& val) {
	PyObject* obj = PyTuple_New(2);
	PyTuple_SetItem(obj,0,swig::from(val.first));
	PyTuple_SetItem(obj,1,swig::from(val.second));
	return obj;
      }
    };
  }





































      namespace swig {
	template <>  struct traits<std::pair< std::vector< double,std::allocator< double > >, std::vector< double,std::allocator< double > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::pair<" "std::vector< double,std::allocator< double > >" "," "std::vector< double,std::allocator< double > >" " >";
	  }
	};
      }
    

/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(__clang__) && defined(__has_builtin)
#  if __has_builtin(__builtin_isfinite)
#   define SWIG_isfinite(X) (__builtin_isfinite(X))
#  endif
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN int
SWIG_AsVal_float (PyObject * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double (obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< float >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_float  (float value)
{    
  return SWIG_From_double  (value);
}


namespace swig {
  template <> struct traits< float > {
    typedef value_category category;
    static const char* type_name() { return"float"; }
  };
  template <>  struct traits_asval< float > {
    typedef float value_type;
    static int asval(PyObject *obj, value_type *val) {
      return SWIG_AsVal_float (obj, val);
    }
  };
  template <>  struct traits_from< float > {
    typedef float value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_float  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector< float, std::allocator< float > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "float" "," "std::allocator< float >" " >";
	  }
	};
      }
    
SWIGINTERN swig::SwigPyIterator *std_vector_Sl_float_Sg__iterator(std::vector< float > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_float_Sg____nonzero__(std::vector< float > const *self){
      return !(self->empty());
    }
SWIGINTERN bool std_vector_Sl_float_Sg____bool__(std::vector< float > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector< float >::size_type std_vector_Sl_float_Sg____len__(std::vector< float > const *self){
      return self->size();
    }
SWIGINTERN std::vector< float,std::allocator< float > > *std_vector_Sl_float_Sg____getslice__(std::vector< float > *self,std::vector< float >::difference_type i,std::vector< float >::difference_type j){
      return swig::getslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_float_Sg____setslice____SWIG_0(std::vector< float > *self,std::vector< float >::difference_type i,std::vector< float >::difference_type j){
      swig::setslice(self, i, j, 1, std::vector< float,std::allocator< float > >());
    }
SWIGINTERN void std_vector_Sl_float_Sg____setslice____SWIG_1(std::vector< float > *self,std::vector< float >::difference_type i,std::vector< float >::difference_type j,std::vector< float,std::allocator< float > > const &v){
      swig::setslice(self, i, j, 1, v);
    }
SWIGINTERN void std_vector_Sl_float_Sg____delslice__(std::vector< float > *self,std::vector< float >::difference_type i,std::vector< float >::difference_type j){
      swig::delslice(self, i, j, 1);
    }
SWIGINTERN void std_vector_Sl_float_Sg____delitem____SWIG_0(std::vector< float > *self,std::vector< float >::difference_type i){
      swig::erase(self, swig::getpos(self, i));
    }
SWIGINTERN std::vector< float,std::allocator< float > > *std_vector_Sl_float_Sg____getitem____SWIG_0(std::vector< float > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return NULL;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< float,std::allocator< float > >::difference_type id = i;
      std::vector< float,std::allocator< float > >::difference_type jd = j;
      return swig::getslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_float_Sg____setitem____SWIG_0(std::vector< float > *self,PySliceObject *slice,std::vector< float,std::allocator< float > > const &v){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< float,std::allocator< float > >::difference_type id = i;
      std::vector< float,std::allocator< float > >::difference_type jd = j;
      swig::setslice(self, id, jd, step, v);
    }
SWIGINTERN void std_vector_Sl_float_Sg____setitem____SWIG_1(std::vector< float > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< float,std::allocator< float > >::difference_type id = i;
      std::vector< float,std::allocator< float > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN void std_vector_Sl_float_Sg____delitem____SWIG_1(std::vector< float > *self,PySliceObject *slice){
      Py_ssize_t i, j, step;
      if( !PySlice_Check(slice) ) {
        SWIG_Error(SWIG_TypeError, "Slice object expected.");
        return;
      }
      PySlice_GetIndices(SWIGPY_SLICE_ARG(slice), (Py_ssize_t)self->size(), &i, &j, &step);
      std::vector< float,std::allocator< float > >::difference_type id = i;
      std::vector< float,std::allocator< float > >::difference_type jd = j;
      swig::delslice(self, id, jd, step);
    }
SWIGINTERN std::vector< float >::value_type const &std_vector_Sl_float_Sg____getitem____SWIG_1(std::vector< float > const *self,std::vector< float >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_float_Sg____setitem____SWIG_2(std::vector< float > *self,std::vector< float >::difference_type i,std::vector< float >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN std::vector< float >::value_type std_vector_Sl_float_Sg__pop(std::vector< float > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector< float,std::allocator< float > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN void std_vector_Sl_float_Sg__append(std::vector< float > *self,std::vector< float >::value_type const &x){
      self->push_back(x);
    }
SWIGINTERN std::vector< float >::iterator std_vector_Sl_float_Sg__erase__SWIG_0(std::vector< float > *self,std::vector< float >::iterator pos){ return self->erase(pos); }
SWIGINTERN std::vector< float >::iterator std_vector_Sl_float_Sg__erase__SWIG_1(std::vector< float > *self,std::vector< float >::iterator first,std::vector< float >::iterator last){ return self->erase(first, last); }
SWIGINTERN std::vector< float >::iterator std_vector_Sl_float_Sg__insert__SWIG_0(std::vector< float > *self,std::vector< float >::iterator pos,std::vector< float >::value_type const &x){ return self->insert(pos, x); }
SWIGINTERN void std_vector_Sl_float_Sg__insert__SWIG_1(std::vector< float > *self,std::vector< float >::iterator pos,std::vector< float >::size_type n,std::vector< float >::value_type const &x){ self->insert(pos, n, x); }

      namespace swig {
	template <>  struct traits<std::pair< std::vector< float,std::allocator< float > >, std::vector< float,std::allocator< float > > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::pair<" "std::vector< float,std::allocator< float > >" "," "std::vector< float,std::allocator< float > >" " >";
	  }
	};
      }
    


// include this to avoid needing to define it at compile time 
#ifndef SWIG
# define SWIG
#endif

// needed by numpy.i, harmless otherwise
#define SWIG_FILE_WITH_INIT

// standard library headers we need
#include <cstdlib>
#include <cstring>

// the main library headers
#include "wasserstein/Wasserstein.hh"

using WASSERSTEIN_NAMESPACE::DefaultNetworkSimplex;



#ifndef SWIG_FILE_WITH_INIT
#define NO_IMPORT_ARRAY
#endif
#include "stdio.h"
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/arrayobject.h>


#include <complex> 


SWIGINTERNINLINE PyObject*
  SWIG_From_int  (int value)
{
  return PyInt_FromLong((long) value);
}


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< int >(v);
    }
  }  
  return res;
}


SWIGINTERN int
SWIG_AsVal_bool (PyObject *obj, bool *val)
{
  int r;
  if (!PyBool_Check(obj))
    return SWIG_ERROR;
  r = PyObject_IsTrue(obj);
  if (r == -1)
    return SWIG_ERROR;
  if (val) *val = r ? true : false;
  return SWIG_OK;
}


#if NPY_API_VERSION < 0x00000007
#define NPY_ARRAY_DEFAULT NPY_DEFAULT
#define NPY_ARRAY_FARRAY  NPY_FARRAY
#define NPY_FORTRANORDER  NPY_FORTRAN
#endif


  /* Given a PyObject, return a string describing its type.
   */
  const char* pytype_string(PyObject* py_obj)
  {
    if (py_obj == NULL          ) return "C NULL value";
    if (py_obj == Py_None       ) return "Python None" ;
    if (PyCallable_Check(py_obj)) return "callable"    ;
    if (PyBytes_Check(   py_obj)) return "string"      ;
    if (PyLong_Check(    py_obj)) return "int"         ;
    if (PyFloat_Check(   py_obj)) return "float"       ;
    if (PyDict_Check(    py_obj)) return "dict"        ;
    if (PyList_Check(    py_obj)) return "list"        ;
    if (PyTuple_Check(   py_obj)) return "tuple"       ;

    return "unknown type";
  }

  /* Given a NumPy typecode, return a string describing the type.
   */
  const char* typecode_string(int typecode)
  {
    static const char* type_names[25] = {"bool",
                                         "byte",
                                         "unsigned byte",
                                         "short",
                                         "unsigned short",
                                         "int",
                                         "unsigned int",
                                         "long",
                                         "unsigned long",
                                         "long long",
                                         "unsigned long long",
                                         "float",
                                         "double",
                                         "long double",
                                         "complex float",
                                         "complex double",
                                         "complex long double",
                                         "object",
                                         "string",
                                         "unicode",
                                         "void",
                                         "ntypes",
                                         "notype",
                                         "char",
                                         "unknown"};
    return typecode < 24 ? type_names[typecode] : type_names[24];
  }

  /* Make sure input has correct numpy type.  This now just calls
     PyArray_EquivTypenums().
   */
  int type_match(int actual_type,
                 int desired_type)
  {
    return PyArray_EquivTypenums(actual_type, desired_type);
  }

#ifdef SWIGPY_USE_CAPSULE
  void free_cap(PyObject * cap)
  {
    void* array = (void*) PyCapsule_GetPointer(cap,SWIGPY_CAPSULE_NAME);
    if (array != NULL) free(array);
  }
#endif



SWIGINTERN void wasserstein_EMDBase_Sl_double_Sg__npy_flows(wasserstein::EMDBase< double > *self,double **arr_out,std::ptrdiff_t *n0,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,173,MALLOC_2D_VALUE_ARRAY@*/
  *n0 = self->n0();
  *n1 = self->n1();
  size_t num_elements = size_t(*n0)*size_t(*n1);
  size_t nbytes = num_elements*sizeof(double);
  double * values = (double *) malloc(nbytes);
  if (values == NULL)
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
  *arr_out = values;
/*@SWIG@*/
    memcpy(*arr_out, self->raw_flows().data(), nbytes);
    
    for (size_t i = 0; i < num_elements; i++)
      values[i] *= self->scale();
  }
SWIGINTERN void wasserstein_EMDBase_Sl_double_Sg__npy_dists(wasserstein::EMDBase< double > *self,double **arr_out,std::ptrdiff_t *n0,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,173,MALLOC_2D_VALUE_ARRAY@*/
  *n0 = self->n0();
  *n1 = self->n1();
  size_t num_elements = size_t(*n0)*size_t(*n1);
  size_t nbytes = num_elements*sizeof(double);
  double * values = (double *) malloc(nbytes);
  if (values == NULL)
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
  *arr_out = values;
/*@SWIG@*/
    memcpy(*arr_out, self->ground_dists().data(), nbytes);
  }
SWIGINTERN void wasserstein_EMDBase_Sl_double_Sg__npy_node_potentials(wasserstein::EMDBase< double > *self,double **arr_out0,std::ptrdiff_t *n0,double **arr_out1,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->n0();
  size_t nbytes0 = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = self->n1();
  size_t nbytes1 = size_t(*n1)*sizeof(double);
  *arr_out1 = (double *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<double>, std::vector<double>> vecpair(self->node_potentials());
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN void wasserstein_PairwiseEMDBase_Sl_double_Sg__npy_emds(wasserstein::PairwiseEMDBase< double > *self,double **arr_out,std::ptrdiff_t *n0,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,173,MALLOC_2D_VALUE_ARRAY@*/
  *n0 = self->nevA();
  *n1 = self->nevB();
  size_t num_elements = size_t(*n0)*size_t(*n1);
  size_t nbytes = num_elements*sizeof(double);
  double * values = (double *) malloc(nbytes);
  if (values == NULL)
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
  *arr_out = values;
/*@SWIG@*/
    memcpy(*arr_out, self->emds(false).data(), nbytes);
  }
SWIGINTERN void wasserstein_PairwiseEMDBase_Sl_double_Sg__raw_emds(wasserstein::PairwiseEMDBase< double > *self,double **arr_out0,std::ptrdiff_t *n0){
    if (self->storage() != wasserstein::EMDPairsStorage::FlattenedSymmetric)
      throw std::runtime_error("raw emds only available with raw symmetric storage");

    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->num_emds();
  size_t nbytes = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->emds(true).data(), nbytes);
  }

/* Macros to extract array attributes.
 */
#if NPY_API_VERSION < 0x00000007
#define is_array(a)            ((a) && PyArray_Check((PyArrayObject*)a))
#define array_type(a)          (int)(PyArray_TYPE((PyArrayObject*)a))
#define array_numdims(a)       (((PyArrayObject*)a)->nd)
#define array_dimensions(a)    (((PyArrayObject*)a)->dimensions)
#define array_size(a,i)        (((PyArrayObject*)a)->dimensions[i])
#define array_strides(a)       (((PyArrayObject*)a)->strides)
#define array_stride(a,i)      (((PyArrayObject*)a)->strides[i])
#define array_data(a)          (((PyArrayObject*)a)->data)
#define array_descr(a)         (((PyArrayObject*)a)->descr)
#define array_flags(a)         (((PyArrayObject*)a)->flags)
#define array_clearflags(a,f)  (((PyArrayObject*)a)->flags) &= ~f
#define array_enableflags(a,f) (((PyArrayObject*)a)->flags) = f
#define array_is_fortran(a)    (PyArray_ISFORTRAN((PyArrayObject*)a))
#else
#define is_array(a)            ((a) && PyArray_Check(a))
#define array_type(a)          PyArray_TYPE((PyArrayObject*)a)
#define array_numdims(a)       PyArray_NDIM((PyArrayObject*)a)
#define array_dimensions(a)    PyArray_DIMS((PyArrayObject*)a)
#define array_strides(a)       PyArray_STRIDES((PyArrayObject*)a)
#define array_stride(a,i)      PyArray_STRIDE((PyArrayObject*)a,i)
#define array_size(a,i)        PyArray_DIM((PyArrayObject*)a,i)
#define array_data(a)          PyArray_DATA((PyArrayObject*)a)
#define array_descr(a)         PyArray_DESCR((PyArrayObject*)a)
#define array_flags(a)         PyArray_FLAGS((PyArrayObject*)a)
#define array_enableflags(a,f) PyArray_ENABLEFLAGS((PyArrayObject*)a,f)
#define array_clearflags(a,f)  PyArray_CLEARFLAGS((PyArrayObject*)a,f)
#define array_is_fortran(a)    (PyArray_IS_F_CONTIGUOUS((PyArrayObject*)a))
#endif
#define array_is_contiguous(a) (PyArray_ISCONTIGUOUS((PyArrayObject*)a))
#define array_is_native(a)     (PyArray_ISNOTSWAPPED((PyArrayObject*)a))


  /* Given a PyObject pointer, cast it to a PyArrayObject pointer if
   * legal.  If not, set the python error string appropriately and
   * return NULL.
   */
  PyArrayObject* obj_to_array_no_conversion(PyObject* input,
                                            int        typecode)
  {
    PyArrayObject* ary = NULL;
    if (is_array(input) && (typecode == NPY_NOTYPE ||
                            PyArray_EquivTypenums(array_type(input), typecode)))
    {
      ary = (PyArrayObject*) input;
    }
    else if is_array(input)
    {
      const char* desired_type = typecode_string(typecode);
      const char* actual_type  = typecode_string(array_type(input));
      PyErr_Format(PyExc_TypeError,
                   "Array of type '%s' required.  Array of type '%s' given",
                   desired_type, actual_type);
      ary = NULL;
    }
    else
    {
      const char* desired_type = typecode_string(typecode);
      const char* actual_type  = pytype_string(input);
      PyErr_Format(PyExc_TypeError,
                   "Array of type '%s' required.  A '%s' was given",
                   desired_type,
                   actual_type);
      ary = NULL;
    }
    return ary;
  }

  /* Convert the given PyObject to a NumPy array with the given
   * typecode.  On success, return a valid PyArrayObject* with the
   * correct type.  On failure, the python error string will be set and
   * the routine returns NULL.
   */
  PyArrayObject* obj_to_array_allow_conversion(PyObject* input,
                                               int       typecode,
                                               int*      is_new_object)
  {
    PyArrayObject* ary = NULL;
    PyObject*      py_obj;
    if (is_array(input) && (typecode == NPY_NOTYPE ||
                            PyArray_EquivTypenums(array_type(input),typecode)))
    {
      ary = (PyArrayObject*) input;
      *is_new_object = 0;
    }
    else
    {
      py_obj = PyArray_FROMANY(input, typecode, 0, 0, NPY_ARRAY_DEFAULT);
      /* If NULL, PyArray_FromObject will have set python error value.*/
      ary = (PyArrayObject*) py_obj;
      *is_new_object = 1;
    }
    return ary;
  }

  /* Given a PyArrayObject, check to see if it is contiguous.  If so,
   * return the input pointer and flag it as not a new object.  If it is
   * not contiguous, create a new PyArrayObject using the original data,
   * flag it as a new object and return the pointer.
   */
  PyArrayObject* make_contiguous(PyArrayObject* ary,
                                 int*           is_new_object,
                                 int            min_dims,
                                 int            max_dims)
  {
    PyArrayObject* result;
    if (array_is_contiguous(ary))
    {
      result = ary;
      *is_new_object = 0;
    }
    else
    {
      result = (PyArrayObject*) PyArray_ContiguousFromObject((PyObject*)ary,
                                                              array_type(ary),
                                                              min_dims,
                                                              max_dims);
      *is_new_object = 1;
    }
    return result;
  }

  /* Given a PyArrayObject, check to see if it is Fortran-contiguous.
   * If so, return the input pointer, but do not flag it as not a new
   * object.  If it is not Fortran-contiguous, create a new
   * PyArrayObject using the original data, flag it as a new object
   * and return the pointer.
   */
  PyArrayObject* make_fortran(PyArrayObject* ary,
                              int*           is_new_object)
  {
    PyArrayObject* result;
    if (array_is_fortran(ary))
    {
      result = ary;
      *is_new_object = 0;
    }
    else
    {
      Py_INCREF(array_descr(ary));
      result = (PyArrayObject*) PyArray_FromArray(ary,
                                                  array_descr(ary),
#if NPY_API_VERSION < 0x00000007
                                                  NPY_FORTRANORDER);
#else
                                                  NPY_ARRAY_F_CONTIGUOUS);
#endif
      *is_new_object = 1;
    }
    return result;
  }

  /* Convert a given PyObject to a contiguous PyArrayObject of the
   * specified type.  If the input object is not a contiguous
   * PyArrayObject, a new one will be created and the new object flag
   * will be set.
   */
  PyArrayObject* obj_to_array_contiguous_allow_conversion(PyObject* input,
                                                          int       typecode,
                                                          int*      is_new_object)
  {
    int is_new1 = 0;
    int is_new2 = 0;
    PyArrayObject* ary2;
    PyArrayObject* ary1 = obj_to_array_allow_conversion(input,
                                                        typecode,
                                                        &is_new1);
    if (ary1)
    {
      ary2 = make_contiguous(ary1, &is_new2, 0, 0);
      if ( is_new1 && is_new2)
      {
        Py_DECREF(ary1);
      }
      ary1 = ary2;
    }
    *is_new_object = is_new1 || is_new2;
    return ary1;
  }

  /* Convert a given PyObject to a Fortran-ordered PyArrayObject of the
   * specified type.  If the input object is not a Fortran-ordered
   * PyArrayObject, a new one will be created and the new object flag
   * will be set.
   */
  PyArrayObject* obj_to_array_fortran_allow_conversion(PyObject* input,
                                                       int       typecode,
                                                       int*      is_new_object)
  {
    int is_new1 = 0;
    int is_new2 = 0;
    PyArrayObject* ary2;
    PyArrayObject* ary1 = obj_to_array_allow_conversion(input,
                                                        typecode,
                                                        &is_new1);
    if (ary1)
    {
      ary2 = make_fortran(ary1, &is_new2);
      if (is_new1 && is_new2)
      {
        Py_DECREF(ary1);
      }
      ary1 = ary2;
    }
    *is_new_object = is_new1 || is_new2;
    return ary1;
  }


  /* Test whether a python object is contiguous.  If array is
   * contiguous, return 1.  Otherwise, set the python error string and
   * return 0.
   */
  int require_contiguous(PyArrayObject* ary)
  {
    int contiguous = 1;
    if (!array_is_contiguous(ary))
    {
      PyErr_SetString(PyExc_TypeError,
                      "Array must be contiguous.  A non-contiguous array was given");
      contiguous = 0;
    }
    return contiguous;
  }

  /* Test whether a python object is (C_ or F_) contiguous.  If array is
   * contiguous, return 1.  Otherwise, set the python error string and
   * return 0.
   */
  int require_c_or_f_contiguous(PyArrayObject* ary)
  {
    int contiguous = 1;
    if (!(array_is_contiguous(ary) || array_is_fortran(ary)))
    {
      PyErr_SetString(PyExc_TypeError,
                      "Array must be contiguous (C_ or F_).  A non-contiguous array was given");
      contiguous = 0;
    }
    return contiguous;
  }

  /* Require that a numpy array is not byte-swapped.  If the array is
   * not byte-swapped, return 1.  Otherwise, set the python error string
   * and return 0.
   */
  int require_native(PyArrayObject* ary)
  {
    int native = 1;
    if (!array_is_native(ary))
    {
      PyErr_SetString(PyExc_TypeError,
                      "Array must have native byteorder.  "
                      "A byte-swapped array was given");
      native = 0;
    }
    return native;
  }

  /* Require the given PyArrayObject to have a specified number of
   * dimensions.  If the array has the specified number of dimensions,
   * return 1.  Otherwise, set the python error string and return 0.
   */
  int require_dimensions(PyArrayObject* ary,
                         int            exact_dimensions)
  {
    int success = 1;
    if (array_numdims(ary) != exact_dimensions)
    {
      PyErr_Format(PyExc_TypeError,
                   "Array must have %d dimensions.  Given array has %d dimensions",
                   exact_dimensions,
                   array_numdims(ary));
      success = 0;
    }
    return success;
  }

  /* Require the given PyArrayObject to have one of a list of specified
   * number of dimensions.  If the array has one of the specified number
   * of dimensions, return 1.  Otherwise, set the python error string
   * and return 0.
   */
  int require_dimensions_n(PyArrayObject* ary,
                           int*           exact_dimensions,
                           int            n)
  {
    int success = 0;
    int i;
    char dims_str[255] = "";
    char s[255];
    for (i = 0; i < n && !success; i++)
    {
      if (array_numdims(ary) == exact_dimensions[i])
      {
        success = 1;
      }
    }
    if (!success)
    {
      for (i = 0; i < n-1; i++)
      {
        sprintf(s, "%d, ", exact_dimensions[i]);
        strcat(dims_str,s);
      }
      sprintf(s, " or %d", exact_dimensions[n-1]);
      strcat(dims_str,s);
      PyErr_Format(PyExc_TypeError,
                   "Array must have %s dimensions.  Given array has %d dimensions",
                   dims_str,
                   array_numdims(ary));
    }
    return success;
  }

  /* Require the given PyArrayObject to have a specified shape.  If the
   * array has the specified shape, return 1.  Otherwise, set the python
   * error string and return 0.
   */
  int require_size(PyArrayObject* ary,
                   npy_intp*      size,
                   int            n)
  {
    int i;
    int success = 1;
    size_t len;
    char desired_dims[255] = "[";
    char s[255];
    char actual_dims[255] = "[";
    for(i=0; i < n;i++)
    {
      if (size[i] != -1 &&  size[i] != array_size(ary,i))
      {
        success = 0;
      }
    }
    if (!success)
    {
      for (i = 0; i < n; i++)
      {
        if (size[i] == -1)
        {
          sprintf(s, "*,");
        }
        else
        {
          sprintf(s, "%ld,", (long int)size[i]);
        }
        strcat(desired_dims,s);
      }
      len = strlen(desired_dims);
      desired_dims[len-1] = ']';
      for (i = 0; i < n; i++)
      {
        sprintf(s, "%ld,", (long int)array_size(ary,i));
        strcat(actual_dims,s);
      }
      len = strlen(actual_dims);
      actual_dims[len-1] = ']';
      PyErr_Format(PyExc_TypeError,
                   "Array must have shape of %s.  Given array has shape of %s",
                   desired_dims,
                   actual_dims);
    }
    return success;
  }

  /* Require the given PyArrayObject to to be Fortran ordered.  If the
   * the PyArrayObject is already Fortran ordered, do nothing.  Else,
   * set the Fortran ordering flag and recompute the strides.
   */
  int require_fortran(PyArrayObject* ary)
  {
    int success = 1;
    int nd = array_numdims(ary);
    int i;
    npy_intp * strides = array_strides(ary);
    if (array_is_fortran(ary)) return success;
    int n_non_one = 0;
    /* Set the Fortran ordered flag */
    const npy_intp *dims = array_dimensions(ary);
    for (i=0; i < nd; ++i)
      n_non_one += (dims[i] != 1) ? 1 : 0;
    if (n_non_one > 1)
      array_clearflags(ary,NPY_ARRAY_CARRAY);
    array_enableflags(ary,NPY_ARRAY_FARRAY);
    /* Recompute the strides */
    strides[0] = strides[nd-1];
    for (i=1; i < nd; ++i)
      strides[i] = strides[i-1] * array_size(ary,i-1);
    return success;
  }



SWIGINTERN void wasserstein_ExternalEMDHandler_Sl_double_Sg__npy_evaluate__SWIG_0(wasserstein::ExternalEMDHandler< double > *self,double *emds,std::ptrdiff_t n0){
    self->evaluate(emds, n0);
  }
SWIGINTERN void wasserstein_ExternalEMDHandler_Sl_double_Sg__npy_evaluate__SWIG_1(wasserstein::ExternalEMDHandler< double > *self,double *emds,std::ptrdiff_t n0,double *event_weights,std::ptrdiff_t n1){
    if (n0 != n1)
      throw std::invalid_argument("length of `emds` should match lengh of `event_weights`");
    self->evaluate(emds, n0, event_weights);
  }
SWIGINTERN void wasserstein_ExternalEMDHandler_Sl_double_Sg__npy_evaluate_symmetric(wasserstein::ExternalEMDHandler< double > *self,double *emds,std::ptrdiff_t n0,double *event_weights,std::ptrdiff_t n1){
    if (n0 != n1*(n1 - 1)/2)
      throw std::invalid_argument("length of `emds` should be lengh of `event_weights` choose 2");
    self->evaluate_symmetric(emds, n1, event_weights);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned int >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject*
  SWIG_From_unsigned_SS_int  (unsigned int value)
{
  return PyInt_FromSize_t((size_t) value);
}

SWIGINTERN std::string wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg____repr__(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg__npy_bin_centers(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *self,double **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins();
  size_t nbytes = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_centers().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg__npy_bin_edges(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *self,double **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() + 1;
  size_t nbytes = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_edges().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg__npy_hist_vals_vars(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *self,double **arr_out0,std::ptrdiff_t *n0,double **arr_out1,std::ptrdiff_t *n1,bool overflows=true){
    unsigned nbins = self->nbins() + (overflows ? 2 : 0);
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = nbins;
  size_t nbytes0 = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = nbins;
  size_t nbytes1 = size_t(*n1)*sizeof(double);
  *arr_out1 = (double *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<double>, std::vector<double>> vecpair(self->hist_vals_vars(overflows));
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN std::string wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg____repr__(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg__npy_bin_centers(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *self,double **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins();
  size_t nbytes = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_centers().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg__npy_bin_edges(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *self,double **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() + 1;
  size_t nbytes = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_edges().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg__npy_hist_vals_vars(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *self,double **arr_out0,std::ptrdiff_t *n0,double **arr_out1,std::ptrdiff_t *n1,bool overflows=true){
    unsigned nbins = self->nbins() + (overflows ? 2 : 0);
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = nbins;
  size_t nbytes0 = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = nbins;
  size_t nbytes1 = size_t(*n1)*sizeof(double);
  *arr_out1 = (double *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<double>, std::vector<double>> vecpair(self->hist_vals_vars(overflows));
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN void wasserstein_EMDBase_Sl_float_Sg__npy_flows(wasserstein::EMDBase< float > *self,float **arr_out,std::ptrdiff_t *n0,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,173,MALLOC_2D_VALUE_ARRAY@*/
  *n0 = self->n0();
  *n1 = self->n1();
  size_t num_elements = size_t(*n0)*size_t(*n1);
  size_t nbytes = num_elements*sizeof(float);
  float * values = (float *) malloc(nbytes);
  if (values == NULL)
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
  *arr_out = values;
/*@SWIG@*/
    memcpy(*arr_out, self->raw_flows().data(), nbytes);
    
    for (size_t i = 0; i < num_elements; i++)
      values[i] *= self->scale();
  }
SWIGINTERN void wasserstein_EMDBase_Sl_float_Sg__npy_dists(wasserstein::EMDBase< float > *self,float **arr_out,std::ptrdiff_t *n0,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,173,MALLOC_2D_VALUE_ARRAY@*/
  *n0 = self->n0();
  *n1 = self->n1();
  size_t num_elements = size_t(*n0)*size_t(*n1);
  size_t nbytes = num_elements*sizeof(float);
  float * values = (float *) malloc(nbytes);
  if (values == NULL)
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
  *arr_out = values;
/*@SWIG@*/
    memcpy(*arr_out, self->ground_dists().data(), nbytes);
  }
SWIGINTERN void wasserstein_EMDBase_Sl_float_Sg__npy_node_potentials(wasserstein::EMDBase< float > *self,float **arr_out0,std::ptrdiff_t *n0,float **arr_out1,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->n0();
  size_t nbytes0 = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = self->n1();
  size_t nbytes1 = size_t(*n1)*sizeof(float);
  *arr_out1 = (float *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<float>, std::vector<float>> vecpair(self->node_potentials());
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN void wasserstein_PairwiseEMDBase_Sl_float_Sg__npy_emds(wasserstein::PairwiseEMDBase< float > *self,float **arr_out,std::ptrdiff_t *n0,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,173,MALLOC_2D_VALUE_ARRAY@*/
  *n0 = self->nevA();
  *n1 = self->nevB();
  size_t num_elements = size_t(*n0)*size_t(*n1);
  size_t nbytes = num_elements*sizeof(float);
  float * values = (float *) malloc(nbytes);
  if (values == NULL)
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
  *arr_out = values;
/*@SWIG@*/
    memcpy(*arr_out, self->emds(false).data(), nbytes);
  }
SWIGINTERN void wasserstein_PairwiseEMDBase_Sl_float_Sg__raw_emds(wasserstein::PairwiseEMDBase< float > *self,float **arr_out0,std::ptrdiff_t *n0){
    if (self->storage() != wasserstein::EMDPairsStorage::FlattenedSymmetric)
      throw std::runtime_error("raw emds only available with raw symmetric storage");

    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->num_emds();
  size_t nbytes = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->emds(true).data(), nbytes);
  }
SWIGINTERN void wasserstein_ExternalEMDHandler_Sl_float_Sg__npy_evaluate__SWIG_0(wasserstein::ExternalEMDHandler< float > *self,float *emds,std::ptrdiff_t n0){
    self->evaluate(emds, n0);
  }
SWIGINTERN void wasserstein_ExternalEMDHandler_Sl_float_Sg__npy_evaluate__SWIG_1(wasserstein::ExternalEMDHandler< float > *self,float *emds,std::ptrdiff_t n0,float *event_weights,std::ptrdiff_t n1){
    if (n0 != n1)
      throw std::invalid_argument("length of `emds` should match lengh of `event_weights`");
    self->evaluate(emds, n0, event_weights);
  }
SWIGINTERN void wasserstein_ExternalEMDHandler_Sl_float_Sg__npy_evaluate_symmetric(wasserstein::ExternalEMDHandler< float > *self,float *emds,std::ptrdiff_t n0,float *event_weights,std::ptrdiff_t n1){
    if (n0 != n1*(n1 - 1)/2)
      throw std::invalid_argument("length of `emds` should be lengh of `event_weights` choose 2");
    self->evaluate_symmetric(emds, n1, event_weights);
  }
SWIGINTERN std::string wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg____repr__(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg__npy_bin_centers(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *self,float **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins();
  size_t nbytes = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_centers().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg__npy_bin_edges(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *self,float **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() + 1;
  size_t nbytes = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_edges().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg__npy_hist_vals_vars(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *self,float **arr_out0,std::ptrdiff_t *n0,float **arr_out1,std::ptrdiff_t *n1,bool overflows=true){
    unsigned nbins = self->nbins() + (overflows ? 2 : 0);
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = nbins;
  size_t nbytes0 = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = nbins;
  size_t nbytes1 = size_t(*n1)*sizeof(float);
  *arr_out1 = (float *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<float>, std::vector<float>> vecpair(self->hist_vals_vars(overflows));
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN std::string wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg____repr__(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg__npy_bin_centers(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *self,float **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins();
  size_t nbytes = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_centers().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg__npy_bin_edges(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *self,float **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() + 1;
  size_t nbytes = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->bin_edges().data(), nbytes);
  }
SWIGINTERN void wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg__npy_hist_vals_vars(wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *self,float **arr_out0,std::ptrdiff_t *n0,float **arr_out1,std::ptrdiff_t *n1,bool overflows=true){
    unsigned nbins = self->nbins() + (overflows ? 2 : 0);
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = nbins;
  size_t nbytes0 = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = nbins;
  size_t nbytes1 = size_t(*n1)*sizeof(float);
  *arr_out1 = (float *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<float>, std::vector<float>> vecpair(self->hist_vals_vars(overflows));
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN std::string wasserstein_CorrelationDimension_Sl_float_Sg____repr__(wasserstein::CorrelationDimension< float > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_CorrelationDimension_Sl_float_Sg__npy_corrdim_bins(wasserstein::CorrelationDimension< float > *self,float **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() - 1;
  size_t nbytes = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->corrdim_bins().data(), nbytes);
  }
SWIGINTERN void wasserstein_CorrelationDimension_Sl_float_Sg__npy_corrdims(wasserstein::CorrelationDimension< float > *self,float **arr_out0,std::ptrdiff_t *n0,float **arr_out1,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() - 1;
  size_t nbytes0 = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = self->nbins() - 1;
  size_t nbytes1 = size_t(*n1)*sizeof(float);
  *arr_out1 = (float *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<float>, std::vector<float>> vecpair(self->corrdims());
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN void wasserstein_CorrelationDimension_Sl_float_Sg__npy_cumulative_vals_vars(wasserstein::CorrelationDimension< float > *self,float **arr_out0,std::ptrdiff_t *n0,float **arr_out1,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins();
  size_t nbytes0 = size_t(*n0)*sizeof(float);
  *arr_out0 = (float *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = self->nbins();
  size_t nbytes1 = size_t(*n1)*sizeof(float);
  *arr_out1 = (float *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<float>, std::vector<float>> vecpair(self->cumulative_vals_vars());
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN std::string wasserstein_CorrelationDimension_Sl_double_Sg____repr__(wasserstein::CorrelationDimension< double > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_CorrelationDimension_Sl_double_Sg__npy_corrdim_bins(wasserstein::CorrelationDimension< double > *self,double **arr_out0,std::ptrdiff_t *n0){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() - 1;
  size_t nbytes = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes) + " bytes");
    return;
  }
/*@SWIG@*/
    memcpy(*arr_out0, self->corrdim_bins().data(), nbytes);
  }
SWIGINTERN void wasserstein_CorrelationDimension_Sl_double_Sg__npy_corrdims(wasserstein::CorrelationDimension< double > *self,double **arr_out0,std::ptrdiff_t *n0,double **arr_out1,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins() - 1;
  size_t nbytes0 = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = self->nbins() - 1;
  size_t nbytes1 = size_t(*n1)*sizeof(double);
  *arr_out1 = (double *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<double>, std::vector<double>> vecpair(self->corrdims());
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN void wasserstein_CorrelationDimension_Sl_double_Sg__npy_cumulative_vals_vars(wasserstein::CorrelationDimension< double > *self,double **arr_out0,std::ptrdiff_t *n0,double **arr_out1,std::ptrdiff_t *n1){
    /*@SWIG:wasserstein/swig/wasserstein_common.i,158,PAIRED_1DNUMPY_FROM_VECPAIR@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n0 = self->nbins();
  size_t nbytes0 = size_t(*n0)*sizeof(double);
  *arr_out0 = (double *) malloc(nbytes0);
  if (*arr_out0 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes0) + " bytes");
    return;
  }
/*@SWIG@*/
  /*@SWIG:wasserstein/swig/wasserstein_common.i,141,MALLOC_1D_VALUE_ARRAY@*/
  *n1 = self->nbins();
  size_t nbytes1 = size_t(*n1)*sizeof(double);
  *arr_out1 = (double *) malloc(nbytes1);
  if (*arr_out1 == NULL) {
    throw std::runtime_error("Failed to allocate " + std::to_string(nbytes1) + " bytes");
    return;
  }
/*@SWIG@*/
  std::pair<std::vector<double>, std::vector<double>> vecpair(self->cumulative_vals_vars());
  memcpy(*arr_out0, vecpair.first.data(), nbytes0);
  memcpy(*arr_out1, vecpair.second.data(), nbytes1);
/*@SWIG@*/
  }
SWIGINTERN std::string wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg____repr__(wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__preprocess_CenterWeightedCentroid(wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN double wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_1(wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *self,double *weights0,std::ptrdiff_t n0,double *coords0,std::ptrdiff_t n00,std::ptrdiff_t n01,double *weights1,std::ptrdiff_t n1,double *coords1,std::ptrdiff_t n10,std::ptrdiff_t n11){

    if (n0 != n00 || n1 != n10)
      throw std::invalid_argument("Number of weights does not match number of coordinates");
    if (n01 != n11)
      throw std::invalid_argument("Coordinate dimensions do not match");

    self->set_external_dists(false);

    return (*self)(std::make_tuple(weights0, coords0, n0, n01), std::make_tuple(weights1, coords1, n1, n11));
  }
SWIGINTERN double wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_2(wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *self,double *weights0,std::ptrdiff_t n0,double *weights1,std::ptrdiff_t n1,double *external_dists,std::ptrdiff_t d0,std::ptrdiff_t d1){

    if (n0 != d0 || n1 != d1)
      throw std::invalid_argument("Weights and distance matrix are incompatible");

    // copy distances into vector for network simplex
    std::size_t ndists(std::size_t(d0) * std::size_t(d1));
    self->ground_dists().resize(ndists);
    std::copy(external_dists, external_dists + ndists, self->ground_dists().begin());

    self->set_external_dists(true);

    return (*self)(std::make_tuple(weights0, nullptr, n0, -1), std::make_tuple(weights1, nullptr, n1, -1));
  }
SWIGINTERN std::string wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg____repr__(wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__preprocess_CenterWeightedCentroid(wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN float wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_1(wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *self,float *weights0,std::ptrdiff_t n0,float *coords0,std::ptrdiff_t n00,std::ptrdiff_t n01,float *weights1,std::ptrdiff_t n1,float *coords1,std::ptrdiff_t n10,std::ptrdiff_t n11){

    if (n0 != n00 || n1 != n10)
      throw std::invalid_argument("Number of weights does not match number of coordinates");
    if (n01 != n11)
      throw std::invalid_argument("Coordinate dimensions do not match");

    self->set_external_dists(false);

    return (*self)(std::make_tuple(weights0, coords0, n0, n01), std::make_tuple(weights1, coords1, n1, n11));
  }
SWIGINTERN float wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_2(wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *self,float *weights0,std::ptrdiff_t n0,float *weights1,std::ptrdiff_t n1,float *external_dists,std::ptrdiff_t d0,std::ptrdiff_t d1){

    if (n0 != d0 || n1 != d1)
      throw std::invalid_argument("Weights and distance matrix are incompatible");

    // copy distances into vector for network simplex
    std::size_t ndists(std::size_t(d0) * std::size_t(d1));
    self->ground_dists().resize(ndists);
    std::copy(external_dists, external_dists + ndists, self->ground_dists().begin());

    self->set_external_dists(true);

    return (*self)(std::make_tuple(weights0, nullptr, n0, -1), std::make_tuple(weights1, nullptr, n1, -1));
  }
SWIGINTERN std::string wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg____repr__(wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__preprocess_CenterWeightedCentroid(wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN double wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_1(wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *self,double *weights0,std::ptrdiff_t n0,double *coords0,std::ptrdiff_t n00,std::ptrdiff_t n01,double *weights1,std::ptrdiff_t n1,double *coords1,std::ptrdiff_t n10,std::ptrdiff_t n11){

    if (n0 != n00 || n1 != n10)
      throw std::invalid_argument("Number of weights does not match number of coordinates");
    if (n01 != n11)
      throw std::invalid_argument("Coordinate dimensions do not match");

    self->set_external_dists(false);

    return (*self)(std::make_tuple(weights0, coords0, n0, n01), std::make_tuple(weights1, coords1, n1, n11));
  }
SWIGINTERN double wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_2(wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *self,double *weights0,std::ptrdiff_t n0,double *weights1,std::ptrdiff_t n1,double *external_dists,std::ptrdiff_t d0,std::ptrdiff_t d1){

    if (n0 != d0 || n1 != d1)
      throw std::invalid_argument("Weights and distance matrix are incompatible");

    // copy distances into vector for network simplex
    std::size_t ndists(std::size_t(d0) * std::size_t(d1));
    self->ground_dists().resize(ndists);
    std::copy(external_dists, external_dists + ndists, self->ground_dists().begin());

    self->set_external_dists(true);

    return (*self)(std::make_tuple(weights0, nullptr, n0, -1), std::make_tuple(weights1, nullptr, n1, -1));
  }
SWIGINTERN std::string wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg____repr__(wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__preprocess_CenterWeightedCentroid(wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN float wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_1(wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *self,float *weights0,std::ptrdiff_t n0,float *coords0,std::ptrdiff_t n00,std::ptrdiff_t n01,float *weights1,std::ptrdiff_t n1,float *coords1,std::ptrdiff_t n10,std::ptrdiff_t n11){

    if (n0 != n00 || n1 != n10)
      throw std::invalid_argument("Number of weights does not match number of coordinates");
    if (n01 != n11)
      throw std::invalid_argument("Coordinate dimensions do not match");

    self->set_external_dists(false);

    return (*self)(std::make_tuple(weights0, coords0, n0, n01), std::make_tuple(weights1, coords1, n1, n11));
  }
SWIGINTERN float wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_2(wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *self,float *weights0,std::ptrdiff_t n0,float *weights1,std::ptrdiff_t n1,float *external_dists,std::ptrdiff_t d0,std::ptrdiff_t d1){

    if (n0 != d0 || n1 != d1)
      throw std::invalid_argument("Weights and distance matrix are incompatible");

    // copy distances into vector for network simplex
    std::size_t ndists(std::size_t(d0) * std::size_t(d1));
    self->ground_dists().resize(ndists);
    std::copy(external_dists, external_dists + ndists, self->ground_dists().begin());

    self->set_external_dists(true);

    return (*self)(std::make_tuple(weights0, nullptr, n0, -1), std::make_tuple(weights1, nullptr, n1, -1));
  }
SWIGINTERN std::string wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg____repr__(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg__preprocess_CenterWeightedCentroid(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg___reset_B_events(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *self){
      self->events().resize(self->nevA());
    }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg___add_event(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *self,double *weights,std::ptrdiff_t n,double *coords,std::ptrdiff_t n1,std::ptrdiff_t d,double event_weight=1){

    self->events().emplace_back(weights, coords, n1, d, event_weight);
    self->preprocess_back_event();
  }
SWIGINTERN std::string wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg____repr__(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg__preprocess_CenterWeightedCentroid(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg___reset_B_events(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *self){
      self->events().resize(self->nevA());
    }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg___add_event(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *self,float *weights,std::ptrdiff_t n,float *coords,std::ptrdiff_t n1,std::ptrdiff_t d,float event_weight=1){

    self->events().emplace_back(weights, coords, n1, d, event_weight);
    self->preprocess_back_event();
  }
SWIGINTERN std::string wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg____repr__(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg__preprocess_CenterWeightedCentroid(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg___reset_B_events(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *self){
      self->events().resize(self->nevA());
    }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg___add_event(wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *self,double *weights,std::ptrdiff_t n,double *coords,std::ptrdiff_t n1,std::ptrdiff_t d,double event_weight=1){

    self->events().emplace_back(weights, coords, n1, d, event_weight);
    self->preprocess_back_event();
  }
SWIGINTERN std::string wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg____repr__(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > const *self){
    return self->description();
  }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg__preprocess_CenterWeightedCentroid(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *self){ self->preprocess<wasserstein::CenterWeightedCentroid>(); }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg___reset_B_events(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *self){
      self->events().resize(self->nevA());
    }
SWIGINTERN void wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg___add_event(wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *self,float *weights,std::ptrdiff_t n,float *coords,std::ptrdiff_t n1,std::ptrdiff_t d,float event_weight=1){

    self->events().emplace_back(weights, coords, n1, d, event_weight);
    self->preprocess_back_event();
  }
#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_delete_SwigPyIterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SwigPyIterator" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  delete arg1;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_value" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)((swig::SwigPyIterator const *)arg1)->value();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_incr(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  size_t arg2 = (size_t) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  swig::SwigPyIterator *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:SwigPyIterator_incr", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_incr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_size_t(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_incr" "', argument " "2"" of type '" "size_t""'");
    } 
    arg2 = static_cast< size_t >(val2);
  }
  try {
    result = (swig::SwigPyIterator *)(arg1)->incr(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_decr(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  size_t arg2 = (size_t) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  swig::SwigPyIterator *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:SwigPyIterator_decr", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_decr" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_size_t(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_decr" "', argument " "2"" of type '" "size_t""'");
    } 
    arg2 = static_cast< size_t >(val2);
  }
  try {
    result = (swig::SwigPyIterator *)(arg1)->decr(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_distance(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  ptrdiff_t result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator_distance", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_distance" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator_distance" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator_distance" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  try {
    result = ((swig::SwigPyIterator const *)arg1)->distance((swig::SwigPyIterator const &)*arg2);
  } catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail;
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_equal(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  bool result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator_equal", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_equal" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator_equal" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator_equal" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  try {
    result = (bool)((swig::SwigPyIterator const *)arg1)->equal((swig::SwigPyIterator const &)*arg2);
  } catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail;
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_copy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_copy" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->copy();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_next(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_next" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->next();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___next__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___next__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->__next__();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_previous(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  PyObject *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_previous" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->previous();
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator_advance(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  swig::SwigPyIterator *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator_advance", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator_advance" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator_advance" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *)(arg1)->advance(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___eq__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  bool result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator___eq__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___eq__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___eq__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator___eq__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = (bool)((swig::SwigPyIterator const *)arg1)->operator ==((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___ne__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  bool result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator___ne__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___ne__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___ne__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator___ne__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = (bool)((swig::SwigPyIterator const *)arg1)->operator !=((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___iadd__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  swig::SwigPyIterator *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator___iadd__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___iadd__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___iadd__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *) &(arg1)->operator +=(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___isub__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  swig::SwigPyIterator *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator___isub__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___isub__" "', argument " "1"" of type '" "swig::SwigPyIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___isub__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *) &(arg1)->operator -=(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___add__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  swig::SwigPyIterator *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:SwigPyIterator___add__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___add__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___add__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->operator +(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___sub____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  ptrdiff_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  swig::SwigPyIterator *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___sub__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::SwigPyIterator *)((swig::SwigPyIterator const *)arg1)->operator -(arg2);
  } catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___sub____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  swig::SwigPyIterator *arg1 = (swig::SwigPyIterator *) 0 ;
  swig::SwigPyIterator *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  ptrdiff_t result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_swig__SwigPyIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SwigPyIterator___sub__" "', argument " "1"" of type '" "swig::SwigPyIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::SwigPyIterator * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2, SWIGTYPE_p_swig__SwigPyIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SwigPyIterator___sub__" "', argument " "2"" of type '" "swig::SwigPyIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::SwigPyIterator * >(argp2);
  result = ((swig::SwigPyIterator const *)arg1)->operator -((swig::SwigPyIterator const &)*arg2);
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  PyErr_Clear();
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *_wrap_SwigPyIterator___sub__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "SwigPyIterator___sub__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_NO_NULL | 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_SwigPyIterator___sub____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__SwigPyIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_SwigPyIterator___sub____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *SwigPyIterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_swig__SwigPyIterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vectorDouble_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_iterator" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_double_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___nonzero__" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (bool)std_vector_Sl_double_Sg____nonzero__((std::vector< double > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___bool__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___bool__" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (bool)std_vector_Sl_double_Sg____bool__((std::vector< double > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___len__" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = std_vector_Sl_double_Sg____len__((std::vector< double > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  std::vector< double >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  NULL 
  };
  std::vector< double,std::allocator< double > > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorDouble___getslice__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___getslice__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___getslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___getslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< double >::difference_type >(val3);
  try {
    result = (std::vector< double,std::allocator< double > > *)std_vector_Sl_double_Sg____getslice__(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setslice____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  std::vector< double >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setslice__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___setslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___setslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< double >::difference_type >(val3);
  try {
    std_vector_Sl_double_Sg____setslice____SWIG_0(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setslice____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  std::vector< double >::difference_type arg3 ;
  std::vector< double,std::allocator< double > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setslice__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___setslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___setslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< double >::difference_type >(val3);
  {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorDouble___setslice__" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorDouble___setslice__" "', argument " "4"" of type '" "std::vector< double,std::allocator< double > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_double_Sg____setslice____SWIG_1(arg1,arg2,arg3,(std::vector< double,std::allocator< double > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorDouble___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< double,std::allocator< double > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vectorDouble___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::__setslice__(std::vector< double >::difference_type,std::vector< double >::difference_type)\n"
    "    std::vector< double >::__setslice__(std::vector< double >::difference_type,std::vector< double >::difference_type,std::vector< double,std::allocator< double > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  std::vector< double >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorDouble___delslice__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___delslice__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___delslice__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___delslice__" "', argument " "3"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< double >::difference_type >(val3);
  try {
    std_vector_Sl_double_Sg____delslice__(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___delitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___delitem__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___delitem__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  try {
    std_vector_Sl_double_Sg____delitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___getitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< double,std::allocator< double > > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___getitem__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___getitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    result = (std::vector< double,std::allocator< double > > *)std_vector_Sl_double_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  std::vector< double,std::allocator< double > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setitem__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___setitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorDouble___setitem__" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorDouble___setitem__" "', argument " "3"" of type '" "std::vector< double,std::allocator< double > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_double_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< double,std::allocator< double > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setitem__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___setitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    std_vector_Sl_double_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___delitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___delitem__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble___delitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    std_vector_Sl_double_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorDouble___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorDouble___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::__delitem__(std::vector< double >::difference_type)\n"
    "    std::vector< double >::__delitem__(PySliceObject *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble___getitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< double >::value_type *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___getitem__" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___getitem__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  try {
    result = (std::vector< double >::value_type *) &std_vector_Sl_double_Sg____getitem____SWIG_1((std::vector< double > const *)arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_double(static_cast< double >(*result));
  (void)swig::container_owner<swig::traits<std::vector< double >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorDouble___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorDouble___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::__getitem__(PySliceObject *)\n"
    "    std::vector< double >::__getitem__(std::vector< double >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setitem____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::difference_type arg2 ;
  std::vector< double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble___setitem__" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble___setitem__" "', argument " "2"" of type '" "std::vector< double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble___setitem__" "', argument " "3"" of type '" "std::vector< double >::value_type""'");
  } 
  temp3 = static_cast< std::vector< double >::value_type >(val3);
  arg3 = &temp3;
  try {
    std_vector_Sl_double_Sg____setitem____SWIG_2(arg1,arg2,(double const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorDouble___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< double,std::allocator< double > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vectorDouble___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorDouble___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::__setitem__(PySliceObject *,std::vector< double,std::allocator< double > > const &)\n"
    "    std::vector< double >::__setitem__(PySliceObject *)\n"
    "    std::vector< double >::__setitem__(std::vector< double >::difference_type,std::vector< double >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::value_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_pop" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  try {
    result = (std::vector< double >::value_type)std_vector_Sl_double_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< double >::value_type temp2 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorDouble_append", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_append" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble_append" "', argument " "2"" of type '" "std::vector< double >::value_type""'");
  } 
  temp2 = static_cast< std::vector< double >::value_type >(val2);
  arg2 = &temp2;
  std_vector_Sl_double_Sg__append(arg1,(double const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< double > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< double > *)new std::vector< double >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< double > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< double > *)new std::vector< double >((std::vector< double > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_empty" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (bool)((std::vector< double > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_size" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = ((std::vector< double > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"v",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorDouble_swap", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_swap" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vectorDouble_swap" "', argument " "2"" of type '" "std::vector< double > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorDouble_swap" "', argument " "2"" of type '" "std::vector< double > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< double > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_begin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_begin" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_end(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_end" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_rbegin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::reverse_iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_rbegin" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_rend(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::reverse_iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_rend" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_clear" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_get_allocator" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = ((std::vector< double > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< double >::allocator_type(static_cast< const std::vector< double >::allocator_type& >(result))), SWIGTYPE_p_std__allocatorT_double_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< double > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double >::size_type""'");
  } 
  arg1 = static_cast< std::vector< double >::size_type >(val1);
  result = (std::vector< double > *)new std::vector< double >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_pop_back" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_resize" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble_resize" "', argument " "2"" of type '" "std::vector< double >::size_type""'");
  } 
  arg2 = static_cast< std::vector< double >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< double >::iterator result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_erase" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_erase" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< double >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_erase" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
    }
  }
  result = std_vector_Sl_double_Sg__erase__SWIG_0(arg1,arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::iterator arg2 ;
  std::vector< double >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< double >::iterator result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_erase" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_erase" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< double >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_erase" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_erase" "', argument " "3"" of type '" "std::vector< double >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< double >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_erase" "', argument " "3"" of type '" "std::vector< double >::iterator""'");
    }
  }
  result = std_vector_Sl_double_Sg__erase__SWIG_1(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vectorDouble_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vectorDouble_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::erase(std::vector< double >::iterator)\n"
    "    std::vector< double >::erase(std::vector< double >::iterator,std::vector< double >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_vectorDouble__SWIG_3(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double >::size_type arg1 ;
  std::vector< double >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< double >::value_type temp2 ;
  double val2 ;
  int ecode2 = 0 ;
  std::vector< double > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorDouble" "', argument " "1"" of type '" "std::vector< double >::size_type""'");
  } 
  arg1 = static_cast< std::vector< double >::size_type >(val1);
  ecode2 = SWIG_AsVal_double(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vectorDouble" "', argument " "2"" of type '" "std::vector< double >::value_type""'");
  } 
  temp2 = static_cast< std::vector< double >::value_type >(val2);
  arg2 = &temp2;
  result = (std::vector< double > *)new std::vector< double >(arg1,(std::vector< double >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorDouble(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorDouble", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_vectorDouble__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vectorDouble__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vectorDouble__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_vectorDouble__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorDouble'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::vector()\n"
    "    std::vector< double >::vector(std::vector< double > const &)\n"
    "    std::vector< double >::vector(std::vector< double >::size_type)\n"
    "    std::vector< double >::vector(std::vector< double >::size_type,std::vector< double >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< double >::value_type temp2 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorDouble_push_back", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_push_back" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble_push_back" "', argument " "2"" of type '" "std::vector< double >::value_type""'");
  } 
  temp2 = static_cast< std::vector< double >::value_type >(val2);
  arg2 = &temp2;
  (arg1)->push_back((std::vector< double >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::value_type *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_front" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (std::vector< double >::value_type *) &((std::vector< double > const *)arg1)->front();
  resultobj = SWIG_From_double(static_cast< double >(*result));
  (void)swig::container_owner<swig::traits<std::vector< double >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::value_type *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_back" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = (std::vector< double >::value_type *) &((std::vector< double > const *)arg1)->back();
  resultobj = SWIG_From_double(static_cast< double >(*result));
  (void)swig::container_owner<swig::traits<std::vector< double >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::size_type arg2 ;
  std::vector< double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector< double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorDouble_assign", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_assign" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble_assign" "', argument " "2"" of type '" "std::vector< double >::size_type""'");
  } 
  arg2 = static_cast< std::vector< double >::size_type >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble_assign" "', argument " "3"" of type '" "std::vector< double >::value_type""'");
  } 
  temp3 = static_cast< std::vector< double >::value_type >(val3);
  arg3 = &temp3;
  (arg1)->assign(arg2,(std::vector< double >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::size_type arg2 ;
  std::vector< double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector< double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_resize" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble_resize" "', argument " "2"" of type '" "std::vector< double >::size_type""'");
  } 
  arg2 = static_cast< std::vector< double >::size_type >(val2);
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble_resize" "', argument " "3"" of type '" "std::vector< double >::value_type""'");
  } 
  temp3 = static_cast< std::vector< double >::value_type >(val3);
  arg3 = &temp3;
  (arg1)->resize(arg2,(std::vector< double >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorDouble_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorDouble_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::resize(std::vector< double >::size_type)\n"
    "    std::vector< double >::resize(std::vector< double >::size_type,std::vector< double >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::iterator arg2 ;
  std::vector< double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  std::vector< double >::iterator result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_insert" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_insert" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< double >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_insert" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_double(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble_insert" "', argument " "3"" of type '" "std::vector< double >::value_type""'");
  } 
  temp3 = static_cast< std::vector< double >::value_type >(val3);
  arg3 = &temp3;
  result = std_vector_Sl_double_Sg__insert__SWIG_0(arg1,arg2,(double const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< double >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::iterator arg2 ;
  std::vector< double >::size_type arg3 ;
  std::vector< double >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  std::vector< double >::value_type temp4 ;
  double val4 ;
  int ecode4 = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_insert" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_insert" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< double >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorDouble_insert" "', argument " "2"" of type '" "std::vector< double >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorDouble_insert" "', argument " "3"" of type '" "std::vector< double >::size_type""'");
  } 
  arg3 = static_cast< std::vector< double >::size_type >(val3);
  ecode4 = SWIG_AsVal_double(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vectorDouble_insert" "', argument " "4"" of type '" "std::vector< double >::value_type""'");
  } 
  temp4 = static_cast< std::vector< double >::value_type >(val4);
  arg4 = &temp4;
  std_vector_Sl_double_Sg__insert__SWIG_1(arg1,arg2,arg3,(double const &)*arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorDouble_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorDouble_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< double >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_double(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_vectorDouble_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorDouble_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< double >::insert(std::vector< double >::iterator,std::vector< double >::value_type const &)\n"
    "    std::vector< double >::insert(std::vector< double >::iterator,std::vector< double >::size_type,std::vector< double >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorDouble_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  std::vector< double >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorDouble_reserve", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_reserve" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorDouble_reserve" "', argument " "2"" of type '" "std::vector< double >::size_type""'");
  } 
  arg2 = static_cast< std::vector< double >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorDouble_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorDouble_capacity" "', argument " "1"" of type '" "std::vector< double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  result = ((std::vector< double > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vectorDouble(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< double > *arg1 = (std::vector< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorDouble" "', argument " "1"" of type '" "std::vector< double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vectorDouble_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorDouble_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorString_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_iterator" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_std_string_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___nonzero__" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (bool)std_vector_Sl_std_string_Sg____nonzero__((std::vector< std::string > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___bool__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___bool__" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (bool)std_vector_Sl_std_string_Sg____bool__((std::vector< std::string > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___len__" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = std_vector_Sl_std_string_Sg____len__((std::vector< std::string > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  NULL 
  };
  std::vector< std::string,std::allocator< std::string > > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorString___getslice__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___getslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___getslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___getslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::string >::difference_type >(val3);
  try {
    result = (std::vector< std::string,std::allocator< std::string > > *)std_vector_Sl_std_string_Sg____getslice__(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___setslice____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___setslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___setslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::string >::difference_type >(val3);
  try {
    std_vector_Sl_std_string_Sg____setslice____SWIG_0(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___setslice____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ;
  std::vector< std::string,std::allocator< std::string > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___setslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___setslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::string >::difference_type >(val3);
  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorString___setslice__" "', argument " "4"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString___setslice__" "', argument " "4"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_std_string_Sg____setslice____SWIG_1(arg1,arg2,arg3,(std::vector< std::string,std::allocator< std::string > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorString___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< std::string,std::allocator< std::string > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vectorString___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::__setslice__(std::vector< std::string >::difference_type,std::vector< std::string >::difference_type)\n"
    "    std::vector< std::string >::__setslice__(std::vector< std::string >::difference_type,std::vector< std::string >::difference_type,std::vector< std::string,std::allocator< std::string > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorString___delslice__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___delslice__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___delslice__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString___delslice__" "', argument " "3"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< std::string >::difference_type >(val3);
  try {
    std_vector_Sl_std_string_Sg____delslice__(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___delitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___delitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___delitem__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  try {
    std_vector_Sl_std_string_Sg____delitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___getitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< std::string,std::allocator< std::string > > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___getitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___getitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    result = (std::vector< std::string,std::allocator< std::string > > *)std_vector_Sl_std_string_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___setitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  std::vector< std::string,std::allocator< std::string > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___setitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string,std::allocator< std::string > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_std_string_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< std::string,std::allocator< std::string > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___setitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___setitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    std_vector_Sl_std_string_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___delitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___delitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString___delitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    std_vector_Sl_std_string_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorString___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorString___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::__delitem__(std::vector< std::string >::difference_type)\n"
    "    std::vector< std::string >::__delitem__(PySliceObject *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString___getitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< std::string >::value_type *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___getitem__" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___getitem__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  try {
    result = (std::vector< std::string >::value_type *) &std_vector_Sl_std_string_Sg____getitem____SWIG_1((std::vector< std::string > const *)arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::string >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorString___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorString___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::__getitem__(PySliceObject *)\n"
    "    std::vector< std::string >::__getitem__(std::vector< std::string >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString___setitem____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::difference_type arg2 ;
  std::vector< std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString___setitem__" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString___setitem__" "', argument " "2"" of type '" "std::vector< std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::difference_type >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString___setitem__" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_std_string_Sg____setitem____SWIG_2(arg1,arg2,(std::string const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorString___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< std::string,std::allocator< std::string > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vectorString___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_std_string(argv[2], (std::string**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vectorString___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::__setitem__(PySliceObject *,std::vector< std::string,std::allocator< std::string > > const &)\n"
    "    std::vector< std::string >::__setitem__(PySliceObject *)\n"
    "    std::vector< std::string >::__setitem__(std::vector< std::string >::difference_type,std::vector< std::string >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::value_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_pop" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  try {
    result = std_vector_Sl_std_string_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorString_append", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_append" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vectorString_append" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_append" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_vector_Sl_std_string_Sg__append(arg1,(std::string const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< std::string > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< std::string > *)new std::vector< std::string >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< std::string > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< std::string,std::allocator< std::string > > *ptr = (std::vector< std::string,std::allocator< std::string > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< std::string > *)new std::vector< std::string >((std::vector< std::string > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_empty" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (bool)((std::vector< std::string > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_size" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = ((std::vector< std::string > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"v",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorString_swap", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_swap" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vectorString_swap" "', argument " "2"" of type '" "std::vector< std::string > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_swap" "', argument " "2"" of type '" "std::vector< std::string > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< std::string > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_begin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_begin" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_end(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_end" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_rbegin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::reverse_iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_rbegin" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_rend(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::reverse_iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_rend" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_clear" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< std::string > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_get_allocator" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = ((std::vector< std::string > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< std::string >::allocator_type(static_cast< const std::vector< std::string >::allocator_type& >(result))), SWIGTYPE_p_std__allocatorT_std__string_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< std::string > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::string >::size_type >(val1);
  result = (std::vector< std::string > *)new std::vector< std::string >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_pop_back" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_resize" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString_resize" "', argument " "2"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< std::string >::iterator result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_erase" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_erase" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::string >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_erase" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
    }
  }
  result = std_vector_Sl_std_string_Sg__erase__SWIG_0(arg1,arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::iterator arg2 ;
  std::vector< std::string >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< std::string >::iterator result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_erase" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_erase" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::string >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_erase" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_erase" "', argument " "3"" of type '" "std::vector< std::string >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::string >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_erase" "', argument " "3"" of type '" "std::vector< std::string >::iterator""'");
    }
  }
  result = std_vector_Sl_std_string_Sg__erase__SWIG_1(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vectorString_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vectorString_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::erase(std::vector< std::string >::iterator)\n"
    "    std::vector< std::string >::erase(std::vector< std::string >::iterator,std::vector< std::string >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_vectorString__SWIG_3(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string >::size_type arg1 ;
  std::vector< std::string >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  std::vector< std::string > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorString" "', argument " "1"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg1 = static_cast< std::vector< std::string >::size_type >(val1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vectorString" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vectorString" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::vector< std::string > *)new std::vector< std::string >(arg1,(std::vector< std::string >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorString(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorString", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_vectorString__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vectorString__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vectorString__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vectorString__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorString'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::vector()\n"
    "    std::vector< std::string >::vector(std::vector< std::string > const &)\n"
    "    std::vector< std::string >::vector(std::vector< std::string >::size_type)\n"
    "    std::vector< std::string >::vector(std::vector< std::string >::size_type,std::vector< std::string >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorString_push_back", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_push_back" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vectorString_push_back" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_push_back" "', argument " "2"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_back((std::vector< std::string >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::value_type *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_front" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (std::vector< std::string >::value_type *) &((std::vector< std::string > const *)arg1)->front();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::string >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::value_type *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_back" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = (std::vector< std::string >::value_type *) &((std::vector< std::string > const *)arg1)->back();
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  (void)swig::container_owner<swig::traits<std::vector< std::string >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::size_type arg2 ;
  std::vector< std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorString_assign", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_assign" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString_assign" "', argument " "2"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::size_type >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString_assign" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_assign" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->assign(arg2,(std::vector< std::string >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::size_type arg2 ;
  std::vector< std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_resize" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString_resize" "', argument " "2"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::size_type >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString_resize" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_resize" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->resize(arg2,(std::vector< std::string >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorString_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_std_string(argv[2], (std::string**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vectorString_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::resize(std::vector< std::string >::size_type)\n"
    "    std::vector< std::string >::resize(std::vector< std::string >::size_type,std::vector< std::string >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::iterator arg2 ;
  std::vector< std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  int res3 = SWIG_OLDOBJ ;
  std::vector< std::string >::iterator result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_insert" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_insert" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::string >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_insert" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
    }
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorString_insert" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_insert" "', argument " "3"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  result = std_vector_Sl_std_string_Sg__insert__SWIG_0(arg1,arg2,(std::string const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< std::string >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::iterator arg2 ;
  std::vector< std::string >::size_type arg3 ;
  std::vector< std::string >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_insert" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_insert" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< std::string >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorString_insert" "', argument " "2"" of type '" "std::vector< std::string >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorString_insert" "', argument " "3"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg3 = static_cast< std::vector< std::string >::size_type >(val3);
  {
    std::string *ptr = (std::string *)0;
    res4 = SWIG_AsPtr_std_string(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorString_insert" "', argument " "4"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorString_insert" "', argument " "4"" of type '" "std::vector< std::string >::value_type const &""'"); 
    }
    arg4 = ptr;
  }
  std_vector_Sl_std_string_Sg__insert__SWIG_1(arg1,arg2,arg3,(std::string const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorString_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter) != 0));
      if (_v) {
        int res = SWIG_AsPtr_std_string(argv[2], (std::string**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vectorString_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< std::string,std::allocator< std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< std::string >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_std_string(argv[3], (std::string**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vectorString_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorString_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< std::string >::insert(std::vector< std::string >::iterator,std::vector< std::string >::value_type const &)\n"
    "    std::vector< std::string >::insert(std::vector< std::string >::iterator,std::vector< std::string >::size_type,std::vector< std::string >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorString_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  std::vector< std::string >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorString_reserve", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_reserve" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorString_reserve" "', argument " "2"" of type '" "std::vector< std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector< std::string >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorString_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorString_capacity" "', argument " "1"" of type '" "std::vector< std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  result = ((std::vector< std::string > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vectorString(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< std::string > *arg1 = (std::vector< std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorString" "', argument " "1"" of type '" "std::vector< std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< std::string > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vectorString_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_std__string_std__allocatorT_std__string_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorString_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_pairVectorDouble__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double >,std::vector< double > > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::pair< std::vector< double >,std::vector< double > > *)new std::pair< std::vector< double >,std::vector< double > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_pairVectorDouble__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< double,std::allocator< double > > arg1 ;
  std::vector< double,std::allocator< double > > arg2 ;
  std::pair< std::vector< double >,std::vector< double > > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_pairVectorDouble" "', argument " "1"" of type '" "std::vector< double,std::allocator< double > >""'"); 
    }
    arg1 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    std::vector< double,std::allocator< double > > *ptr = (std::vector< double,std::allocator< double > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_pairVectorDouble" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > >""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (std::pair< std::vector< double >,std::vector< double > > *)new std::pair< std::vector< double >,std::vector< double > >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_pairVectorDouble__SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::pair< std::vector< double >,std::vector< double > > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > *ptr = (std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_pairVectorDouble" "', argument " "1"" of type '" "std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_pairVectorDouble" "', argument " "1"" of type '" "std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::pair< std::vector< double >,std::vector< double > > *)new std::pair< std::vector< double >,std::vector< double > >((std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_pairVectorDouble(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_pairVectorDouble", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_pairVectorDouble__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_pairVectorDouble__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< double,std::allocator< double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector< double,std::allocator< double > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_pairVectorDouble__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_pairVectorDouble'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::pair< std::vector< double >,std::vector< double > >::pair()\n"
    "    std::pair< std::vector< double >,std::vector< double > >::pair(std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > >)\n"
    "    std::pair< std::vector< double >,std::vector< double > >::pair(std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_pairVectorDouble_first_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double >,std::vector< double > > *arg1 = (std::pair< std::vector< double >,std::vector< double > > *) 0 ;
  std::vector< double,std::allocator< double > > *arg2 = (std::vector< double,std::allocator< double > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "pairVectorDouble_first_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorDouble_first_set" "', argument " "1"" of type '" "std::pair< std::vector< double >,std::vector< double > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< double >,std::vector< double > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "pairVectorDouble_first_set" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > *""'"); 
  }
  arg2 = reinterpret_cast< std::vector< double,std::allocator< double > > * >(argp2);
  if (arg1) (arg1)->first = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_pairVectorDouble_first_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double >,std::vector< double > > *arg1 = (std::pair< std::vector< double >,std::vector< double > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorDouble_first_get" "', argument " "1"" of type '" "std::pair< std::vector< double >,std::vector< double > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< double >,std::vector< double > > * >(argp1);
  result = (std::vector< double,std::allocator< double > > *)& ((arg1)->first);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_pairVectorDouble_second_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double >,std::vector< double > > *arg1 = (std::pair< std::vector< double >,std::vector< double > > *) 0 ;
  std::vector< double,std::allocator< double > > *arg2 = (std::vector< double,std::allocator< double > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "pairVectorDouble_second_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorDouble_second_set" "', argument " "1"" of type '" "std::pair< std::vector< double >,std::vector< double > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< double >,std::vector< double > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "pairVectorDouble_second_set" "', argument " "2"" of type '" "std::vector< double,std::allocator< double > > *""'"); 
  }
  arg2 = reinterpret_cast< std::vector< double,std::allocator< double > > * >(argp2);
  if (arg1) (arg1)->second = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_pairVectorDouble_second_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double >,std::vector< double > > *arg1 = (std::pair< std::vector< double >,std::vector< double > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorDouble_second_get" "', argument " "1"" of type '" "std::pair< std::vector< double >,std::vector< double > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< double >,std::vector< double > > * >(argp1);
  result = (std::vector< double,std::allocator< double > > *)& ((arg1)->second);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_double_std__allocatorT_double_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_pairVectorDouble(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< double >,std::vector< double > > *arg1 = (std::pair< std::vector< double >,std::vector< double > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_pairVectorDouble" "', argument " "1"" of type '" "std::pair< std::vector< double >,std::vector< double > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< double >,std::vector< double > > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *pairVectorDouble_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *pairVectorDouble_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_vectorFloat_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  swig::SwigPyIterator *result = 0 ;
  
  arg2 = &swig_obj[0];
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_iterator" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (swig::SwigPyIterator *)std_vector_Sl_float_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__SwigPyIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___nonzero__" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (bool)std_vector_Sl_float_Sg____nonzero__((std::vector< float > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___bool__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___bool__" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (bool)std_vector_Sl_float_Sg____bool__((std::vector< float > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___len__" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = std_vector_Sl_float_Sg____len__((std::vector< float > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  std::vector< float >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  NULL 
  };
  std::vector< float,std::allocator< float > > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorFloat___getslice__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___getslice__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___getslice__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat___getslice__" "', argument " "3"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< float >::difference_type >(val3);
  try {
    result = (std::vector< float,std::allocator< float > > *)std_vector_Sl_float_Sg____getslice__(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setslice____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  std::vector< float >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___setslice__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___setslice__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat___setslice__" "', argument " "3"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< float >::difference_type >(val3);
  try {
    std_vector_Sl_float_Sg____setslice____SWIG_0(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setslice____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  std::vector< float >::difference_type arg3 ;
  std::vector< float,std::allocator< float > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___setslice__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___setslice__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat___setslice__" "', argument " "3"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< float >::difference_type >(val3);
  {
    std::vector< float,std::allocator< float > > *ptr = (std::vector< float,std::allocator< float > > *)0;
    res4 = swig::asptr(swig_obj[3], &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vectorFloat___setslice__" "', argument " "4"" of type '" "std::vector< float,std::allocator< float > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorFloat___setslice__" "', argument " "4"" of type '" "std::vector< float,std::allocator< float > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_float_Sg____setslice____SWIG_1(arg1,arg2,arg3,(std::vector< float,std::allocator< float > > const &)*arg4);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setslice__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat___setslice__", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorFloat___setslice____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector< float,std::allocator< float > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vectorFloat___setslice____SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat___setslice__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::__setslice__(std::vector< float >::difference_type,std::vector< float >::difference_type)\n"
    "    std::vector< float >::__setslice__(std::vector< float >::difference_type,std::vector< float >::difference_type,std::vector< float,std::allocator< float > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  std::vector< float >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorFloat___delslice__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___delslice__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___delslice__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat___delslice__" "', argument " "3"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg3 = static_cast< std::vector< float >::difference_type >(val3);
  try {
    std_vector_Sl_float_Sg____delslice__(arg1,arg2,arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___delitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___delitem__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___delitem__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  try {
    std_vector_Sl_float_Sg____delitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___getitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< float,std::allocator< float > > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___getitem__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat___getitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    result = (std::vector< float,std::allocator< float > > *)std_vector_Sl_float_Sg____getitem____SWIG_0(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setitem____SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  std::vector< float,std::allocator< float > > *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___setitem__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat___setitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  {
    std::vector< float,std::allocator< float > > *ptr = (std::vector< float,std::allocator< float > > *)0;
    res3 = swig::asptr(swig_obj[2], &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vectorFloat___setitem__" "', argument " "3"" of type '" "std::vector< float,std::allocator< float > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorFloat___setitem__" "', argument " "3"" of type '" "std::vector< float,std::allocator< float > > const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_float_Sg____setitem____SWIG_0(arg1,arg2,(std::vector< float,std::allocator< float > > const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___setitem__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat___setitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    std_vector_Sl_float_Sg____setitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___delitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  PySliceObject *arg2 = (PySliceObject *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___delitem__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  {
    if (!PySlice_Check(swig_obj[1])) {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat___delitem__" "', argument " "2"" of type '" "PySliceObject *""'");
    }
    arg2 = (PySliceObject *) swig_obj[1];
  }
  try {
    std_vector_Sl_float_Sg____delitem____SWIG_1(arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  } catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___delitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat___delitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorFloat___delitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorFloat___delitem____SWIG_0(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat___delitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::__delitem__(std::vector< float >::difference_type)\n"
    "    std::vector< float >::__delitem__(PySliceObject *)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat___getitem____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< float >::value_type *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___getitem__" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___getitem__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  try {
    result = (std::vector< float >::value_type *) &std_vector_Sl_float_Sg____getitem____SWIG_1((std::vector< float > const *)arg1,arg2);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_float(static_cast< float >(*result));
  (void)swig::container_owner<swig::traits<std::vector< float >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___getitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat___getitem__", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorFloat___getitem____SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorFloat___getitem____SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat___getitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::__getitem__(PySliceObject *)\n"
    "    std::vector< float >::__getitem__(std::vector< float >::difference_type) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setitem____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::difference_type arg2 ;
  std::vector< float >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector< float >::value_type temp3 ;
  float val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat___setitem__" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat___setitem__" "', argument " "2"" of type '" "std::vector< float >::difference_type""'");
  } 
  arg2 = static_cast< std::vector< float >::difference_type >(val2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat___setitem__" "', argument " "3"" of type '" "std::vector< float >::value_type""'");
  } 
  temp3 = static_cast< std::vector< float >::value_type >(val3);
  arg3 = &temp3;
  try {
    std_vector_Sl_float_Sg____setitem____SWIG_2(arg1,arg2,(float const &)*arg3);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat___setitem__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat___setitem__", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        return _wrap_vectorFloat___setitem____SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = PySlice_Check(argv[1]);
      }
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector< float,std::allocator< float > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vectorFloat___setitem____SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorFloat___setitem____SWIG_2(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat___setitem__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::__setitem__(PySliceObject *,std::vector< float,std::allocator< float > > const &)\n"
    "    std::vector< float >::__setitem__(PySliceObject *)\n"
    "    std::vector< float >::__setitem__(std::vector< float >::difference_type,std::vector< float >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::value_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_pop" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  try {
    result = (std::vector< float >::value_type)std_vector_Sl_float_Sg__pop(arg1);
  } catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< float >::value_type temp2 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorFloat_append", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_append" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat_append" "', argument " "2"" of type '" "std::vector< float >::value_type""'");
  } 
  temp2 = static_cast< std::vector< float >::value_type >(val2);
  arg2 = &temp2;
  std_vector_Sl_float_Sg__append(arg1,(float const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorFloat__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::vector< float > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::vector< float > *)new std::vector< float >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorFloat__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::vector< float > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::vector< float,std::allocator< float > > *ptr = (std::vector< float,std::allocator< float > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vectorFloat" "', argument " "1"" of type '" "std::vector< float > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vectorFloat" "', argument " "1"" of type '" "std::vector< float > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector< float > *)new std::vector< float >((std::vector< float > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_empty" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (bool)((std::vector< float > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_size" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = ((std::vector< float > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"v",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorFloat_swap", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_swap" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vectorFloat_swap" "', argument " "2"" of type '" "std::vector< float > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vectorFloat_swap" "', argument " "2"" of type '" "std::vector< float > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector< float > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_begin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_begin" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_end(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_end" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_rbegin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::reverse_iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_rbegin" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_rend(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::reverse_iterator result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_rend" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::reverse_iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_clear" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  SwigValueWrapper< std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_get_allocator" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = ((std::vector< float > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector< float >::allocator_type(static_cast< const std::vector< float >::allocator_type& >(result))), SWIGTYPE_p_std__allocatorT_float_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorFloat__SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float >::size_type arg1 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< float > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorFloat" "', argument " "1"" of type '" "std::vector< float >::size_type""'");
  } 
  arg1 = static_cast< std::vector< float >::size_type >(val1);
  result = (std::vector< float > *)new std::vector< float >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_pop_back" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_resize" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat_resize" "', argument " "2"" of type '" "std::vector< float >::size_type""'");
  } 
  arg2 = static_cast< std::vector< float >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< float >::iterator result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_erase" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_erase" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< float >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_erase" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
    }
  }
  result = std_vector_Sl_float_Sg__erase__SWIG_0(arg1,arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::iterator arg2 ;
  std::vector< float >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  swig::SwigPyIterator *iter3 = 0 ;
  int res3 ;
  std::vector< float >::iterator result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_erase" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_erase" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< float >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_erase" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(swig_obj[2], SWIG_as_voidptrptr(&iter3), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_erase" "', argument " "3"" of type '" "std::vector< float >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< float >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_erase" "', argument " "3"" of type '" "std::vector< float >::iterator""'");
    }
  }
  result = std_vector_Sl_float_Sg__erase__SWIG_1(arg1,arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_erase(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat_erase", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vectorFloat_erase__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter) != 0));
      if (_v) {
        swig::SwigPyIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vectorFloat_erase__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat_erase'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::erase(std::vector< float >::iterator)\n"
    "    std::vector< float >::erase(std::vector< float >::iterator,std::vector< float >::iterator)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_new_vectorFloat__SWIG_3(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float >::size_type arg1 ;
  std::vector< float >::value_type *arg2 = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector< float >::value_type temp2 ;
  float val2 ;
  int ecode2 = 0 ;
  std::vector< float > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(swig_obj[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vectorFloat" "', argument " "1"" of type '" "std::vector< float >::size_type""'");
  } 
  arg1 = static_cast< std::vector< float >::size_type >(val1);
  ecode2 = SWIG_AsVal_float(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vectorFloat" "', argument " "2"" of type '" "std::vector< float >::value_type""'");
  } 
  temp2 = static_cast< std::vector< float >::value_type >(val2);
  arg2 = &temp2;
  result = (std::vector< float > *)new std::vector< float >(arg1,(std::vector< float >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vectorFloat(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_vectorFloat", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_vectorFloat__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vectorFloat__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vectorFloat__SWIG_1(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_float(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_vectorFloat__SWIG_3(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_vectorFloat'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::vector()\n"
    "    std::vector< float >::vector(std::vector< float > const &)\n"
    "    std::vector< float >::vector(std::vector< float >::size_type)\n"
    "    std::vector< float >::vector(std::vector< float >::size_type,std::vector< float >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector< float >::value_type temp2 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorFloat_push_back", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_push_back" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat_push_back" "', argument " "2"" of type '" "std::vector< float >::value_type""'");
  } 
  temp2 = static_cast< std::vector< float >::value_type >(val2);
  arg2 = &temp2;
  (arg1)->push_back((std::vector< float >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::value_type *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_front" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (std::vector< float >::value_type *) &((std::vector< float > const *)arg1)->front();
  resultobj = SWIG_From_float(static_cast< float >(*result));
  (void)swig::container_owner<swig::traits<std::vector< float >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::value_type *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_back" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = (std::vector< float >::value_type *) &((std::vector< float > const *)arg1)->back();
  resultobj = SWIG_From_float(static_cast< float >(*result));
  (void)swig::container_owner<swig::traits<std::vector< float >::value_type>::category>::back_reference(resultobj, swig_obj[0]);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::size_type arg2 ;
  std::vector< float >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector< float >::value_type temp3 ;
  float val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  (char *)"x",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:vectorFloat_assign", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_assign" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat_assign" "', argument " "2"" of type '" "std::vector< float >::size_type""'");
  } 
  arg2 = static_cast< std::vector< float >::size_type >(val2);
  ecode3 = SWIG_AsVal_float(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat_assign" "', argument " "3"" of type '" "std::vector< float >::value_type""'");
  } 
  temp3 = static_cast< std::vector< float >::value_type >(val3);
  arg3 = &temp3;
  (arg1)->assign(arg2,(std::vector< float >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::size_type arg2 ;
  std::vector< float >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector< float >::value_type temp3 ;
  float val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_resize" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat_resize" "', argument " "2"" of type '" "std::vector< float >::size_type""'");
  } 
  arg2 = static_cast< std::vector< float >::size_type >(val2);
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat_resize" "', argument " "3"" of type '" "std::vector< float >::value_type""'");
  } 
  temp3 = static_cast< std::vector< float >::value_type >(val3);
  arg3 = &temp3;
  (arg1)->resize(arg2,(std::vector< float >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_resize(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat_resize", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vectorFloat_resize__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorFloat_resize__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat_resize'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::resize(std::vector< float >::size_type)\n"
    "    std::vector< float >::resize(std::vector< float >::size_type,std::vector< float >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::iterator arg2 ;
  std::vector< float >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  std::vector< float >::value_type temp3 ;
  float val3 ;
  int ecode3 = 0 ;
  std::vector< float >::iterator result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_insert" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_insert" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< float >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_insert" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_float(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat_insert" "', argument " "3"" of type '" "std::vector< float >::value_type""'");
  } 
  temp3 = static_cast< std::vector< float >::value_type >(val3);
  arg3 = &temp3;
  result = std_vector_Sl_float_Sg__insert__SWIG_0(arg1,arg2,(float const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector< float >::iterator & >(result)),
    swig::SwigPyIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::iterator arg2 ;
  std::vector< float >::size_type arg3 ;
  std::vector< float >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::SwigPyIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  std::vector< float >::value_type temp4 ;
  float val4 ;
  int ecode4 = 0 ;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_insert" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], SWIG_as_voidptrptr(&iter2), swig::SwigPyIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_insert" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
  } else {
    swig::SwigPyIterator_T<std::vector< float >::iterator > *iter_t = dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vectorFloat_insert" "', argument " "2"" of type '" "std::vector< float >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vectorFloat_insert" "', argument " "3"" of type '" "std::vector< float >::size_type""'");
  } 
  arg3 = static_cast< std::vector< float >::size_type >(val3);
  ecode4 = SWIG_AsVal_float(swig_obj[3], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vectorFloat_insert" "', argument " "4"" of type '" "std::vector< float >::value_type""'");
  } 
  temp4 = static_cast< std::vector< float >::value_type >(val4);
  arg4 = &temp4;
  std_vector_Sl_float_Sg__insert__SWIG_1(arg1,arg2,arg3,(float const &)*arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_insert(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[5] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "vectorFloat_insert", 0, 4, argv))) SWIG_fail;
  --argc;
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_float(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vectorFloat_insert__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::SwigPyIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::SwigPyIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::SwigPyIterator_T<std::vector< float >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_float(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_vectorFloat_insert__SWIG_1(self, argc, argv);
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'vectorFloat_insert'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::vector< float >::insert(std::vector< float >::iterator,std::vector< float >::value_type const &)\n"
    "    std::vector< float >::insert(std::vector< float >::iterator,std::vector< float >::size_type,std::vector< float >::value_type const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_vectorFloat_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  std::vector< float >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:vectorFloat_reserve", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_reserve" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vectorFloat_reserve" "', argument " "2"" of type '" "std::vector< float >::size_type""'");
  } 
  arg2 = static_cast< std::vector< float >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vectorFloat_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float >::size_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vectorFloat_capacity" "', argument " "1"" of type '" "std::vector< float > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  result = ((std::vector< float > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vectorFloat(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector< float > *arg1 = (std::vector< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vectorFloat" "', argument " "1"" of type '" "std::vector< float > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector< float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vectorFloat_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *vectorFloat_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_pairVectorFloat__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **SWIGUNUSEDPARM(swig_obj)) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float >,std::vector< float > > *result = 0 ;
  
  if ((nobjs < 0) || (nobjs > 0)) SWIG_fail;
  result = (std::pair< std::vector< float >,std::vector< float > > *)new std::pair< std::vector< float >,std::vector< float > >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_pairVectorFloat__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::vector< float,std::allocator< float > > arg1 ;
  std::vector< float,std::allocator< float > > arg2 ;
  std::pair< std::vector< float >,std::vector< float > > *result = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  {
    std::vector< float,std::allocator< float > > *ptr = (std::vector< float,std::allocator< float > > *)0;
    int res = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_pairVectorFloat" "', argument " "1"" of type '" "std::vector< float,std::allocator< float > >""'"); 
    }
    arg1 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    std::vector< float,std::allocator< float > > *ptr = (std::vector< float,std::allocator< float > > *)0;
    int res = swig::asptr(swig_obj[1], &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_pairVectorFloat" "', argument " "2"" of type '" "std::vector< float,std::allocator< float > >""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  result = (std::pair< std::vector< float >,std::vector< float > > *)new std::pair< std::vector< float >,std::vector< float > >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_pairVectorFloat__SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  std::pair< std::vector< float >,std::vector< float > > *result = 0 ;
  
  if ((nobjs < 1) || (nobjs > 1)) SWIG_fail;
  {
    std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > *ptr = (std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > *)0;
    res1 = swig::asptr(swig_obj[0], &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_pairVectorFloat" "', argument " "1"" of type '" "std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_pairVectorFloat" "', argument " "1"" of type '" "std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::pair< std::vector< float >,std::vector< float > > *)new std::pair< std::vector< float >,std::vector< float > >((std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_pairVectorFloat(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[3] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "new_pairVectorFloat", 0, 2, argv))) SWIG_fail;
  --argc;
  if (argc == 0) {
    return _wrap_new_pairVectorFloat__SWIG_0(self, argc, argv);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_pairVectorFloat__SWIG_2(self, argc, argv);
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector< float,std::allocator< float > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector< float,std::allocator< float > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_pairVectorFloat__SWIG_1(self, argc, argv);
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'new_pairVectorFloat'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    std::pair< std::vector< float >,std::vector< float > >::pair()\n"
    "    std::pair< std::vector< float >,std::vector< float > >::pair(std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > >)\n"
    "    std::pair< std::vector< float >,std::vector< float > >::pair(std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > const &)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_pairVectorFloat_first_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float >,std::vector< float > > *arg1 = (std::pair< std::vector< float >,std::vector< float > > *) 0 ;
  std::vector< float,std::allocator< float > > *arg2 = (std::vector< float,std::allocator< float > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "pairVectorFloat_first_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorFloat_first_set" "', argument " "1"" of type '" "std::pair< std::vector< float >,std::vector< float > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< float >,std::vector< float > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "pairVectorFloat_first_set" "', argument " "2"" of type '" "std::vector< float,std::allocator< float > > *""'"); 
  }
  arg2 = reinterpret_cast< std::vector< float,std::allocator< float > > * >(argp2);
  if (arg1) (arg1)->first = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_pairVectorFloat_first_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float >,std::vector< float > > *arg1 = (std::pair< std::vector< float >,std::vector< float > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorFloat_first_get" "', argument " "1"" of type '" "std::pair< std::vector< float >,std::vector< float > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< float >,std::vector< float > > * >(argp1);
  result = (std::vector< float,std::allocator< float > > *)& ((arg1)->first);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_pairVectorFloat_second_set(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float >,std::vector< float > > *arg1 = (std::pair< std::vector< float >,std::vector< float > > *) 0 ;
  std::vector< float,std::allocator< float > > *arg2 = (std::vector< float,std::allocator< float > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject *swig_obj[2] ;
  
  if (!SWIG_Python_UnpackTuple(args, "pairVectorFloat_second_set", 2, 2, swig_obj)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorFloat_second_set" "', argument " "1"" of type '" "std::pair< std::vector< float >,std::vector< float > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< float >,std::vector< float > > * >(argp1);
  res2 = SWIG_ConvertPtr(swig_obj[1], &argp2,SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "pairVectorFloat_second_set" "', argument " "2"" of type '" "std::vector< float,std::allocator< float > > *""'"); 
  }
  arg2 = reinterpret_cast< std::vector< float,std::allocator< float > > * >(argp2);
  if (arg1) (arg1)->second = *arg2;
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_pairVectorFloat_second_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float >,std::vector< float > > *arg1 = (std::pair< std::vector< float >,std::vector< float > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "pairVectorFloat_second_get" "', argument " "1"" of type '" "std::pair< std::vector< float >,std::vector< float > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< float >,std::vector< float > > * >(argp1);
  result = (std::vector< float,std::allocator< float > > *)& ((arg1)->second);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorT_float_std__allocatorT_float_t_t, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_pairVectorFloat(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::pair< std::vector< float >,std::vector< float > > *arg1 = (std::pair< std::vector< float >,std::vector< float > > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_pairVectorFloat" "', argument " "1"" of type '" "std::pair< std::vector< float >,std::vector< float > > *""'"); 
  }
  arg1 = reinterpret_cast< std::pair< std::vector< float >,std::vector< float > > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *pairVectorFloat_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *pairVectorFloat_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN int Swig_var_PI_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable PI is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_PI_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_double(static_cast< double >(wasserstein::PI));
  return pyobj;
}


SWIGINTERN int Swig_var_TWOPI_set(PyObject *) {
  SWIG_Error(SWIG_AttributeError,"Variable TWOPI is read-only.");
  return 1;
}


SWIGINTERN PyObject *Swig_var_TWOPI_get(void) {
  PyObject *pyobj = 0;
  
  pyobj = SWIG_From_double(static_cast< double >(wasserstein::TWOPI));
  return pyobj;
}


SWIGINTERN PyObject *_wrap_check_emd_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDStatus arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  char * kwnames[] = {
    (char *)"status",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O:check_emd_status", kwnames, &obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "check_emd_status" "', argument " "1"" of type '" "wasserstein::EMDStatus""'");
  } 
  arg1 = static_cast< wasserstein::EMDStatus >(val1);
  {
    try {
      wasserstein::check_emd_status(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_EMDBaseFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EMDBaseFloat64" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_R" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->R(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_beta" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->beta(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_set_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"R",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat64_set_R", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_set_R" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_set_R" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_R(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_set_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"beta",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat64_set_beta", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_set_beta" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_set_beta" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_beta(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_set_network_simplex_params(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  std::size_t arg2 = (std::size_t) 100000 ;
  double arg3 = (double) 1000 ;
  double arg4 = (double) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OOO:EMDBaseFloat64_set_network_simplex_params", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_set_network_simplex_params" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_size_t(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_set_network_simplex_params" "', argument " "2"" of type '" "std::size_t""'");
    } 
    arg2 = static_cast< std::size_t >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_double(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "EMDBaseFloat64_set_network_simplex_params" "', argument " "3"" of type '" "double""'");
    } 
    arg3 = static_cast< double >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_double(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "EMDBaseFloat64_set_network_simplex_params" "', argument " "4"" of type '" "double""'");
    } 
    arg4 = static_cast< double >(val4);
  }
  {
    try {
      (arg1)->set_network_simplex_params(arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_norm" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::EMDBase< double > const *)arg1)->norm(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_set_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"norm",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat64_set_norm", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_set_norm" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_set_norm" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_norm(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_do_timing(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_do_timing" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::EMDBase< double > const *)arg1)->do_timing(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_set_do_timing(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"timing",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat64_set_do_timing", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_set_do_timing" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_set_do_timing" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_do_timing(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_external_dists(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_external_dists" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::EMDBase< double > const *)arg1)->external_dists(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_set_external_dists(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"exdists",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat64_set_external_dists", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_set_external_dists" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_set_external_dists" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_external_dists(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_n0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_n0" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< double > const *)arg1)->n0(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_n1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_n1" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< double > const *)arg1)->n1(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_extra(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::ExtraParticle result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_extra" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (wasserstein::ExtraParticle)((wasserstein::EMDBase< double > const *)arg1)->extra(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_emd(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_emd" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->emd(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::EMDStatus result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_status" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (wasserstein::EMDStatus)((wasserstein::EMDBase< double > const *)arg1)->status(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_weightdiff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_weightdiff" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->weightdiff(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_scale(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_scale" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->scale(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_n_iter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::size_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_n_iter" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< double > const *)arg1)->n_iter(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_dists_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_dists_vec" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< double > const *)arg1)->dists(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_flows_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_flows_vec" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< double > const *)arg1)->flows(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_flow__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  double result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_flow" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_flow" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "EMDBaseFloat64_flow" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->flow(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_flow__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  std::size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  double result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_flow" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat64_flow" "', argument " "2"" of type '" "std::size_t""'");
  } 
  arg2 = static_cast< std::size_t >(val2);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->flow(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_flow(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "EMDBaseFloat64_flow", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_EMDBaseFloat64_flow__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_EMDBaseFloat64_flow__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'EMDBaseFloat64_flow'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::EMDBase< double >::flow(wasserstein::index_type,wasserstein::index_type) const\n"
    "    wasserstein::EMDBase< double >::flow(std::size_t) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_node_potentials(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_node_potentials" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< double > const *)arg1)->node_potentials(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_duration(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_duration" "', argument " "1"" of type '" "wasserstein::EMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< double > const *)arg1)->duration(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_clear" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      (arg1)->clear(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_flows(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  std::ptrdiff_t *arg4 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim1_temp2 ;
  std::ptrdiff_t dim2_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim1_temp2;
    arg4 = &dim2_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_flows" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      wasserstein_EMDBase_Sl_double_Sg__npy_flows(arg1,arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[2] = {
      *arg3, *arg4 
    };
    PyObject* obj = PyArray_SimpleNewFromData(2, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_dists(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  std::ptrdiff_t *arg4 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim1_temp2 ;
  std::ptrdiff_t dim2_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim1_temp2;
    arg4 = &dim2_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_dists" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      wasserstein_EMDBase_Sl_double_Sg__npy_dists(arg1,arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[2] = {
      *arg3, *arg4 
    };
    PyObject* obj = PyArray_SimpleNewFromData(2, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat64_npy_node_potentials(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< double > *arg1 = (wasserstein::EMDBase< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  double **arg4 = (double **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  double *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat64_npy_node_potentials" "', argument " "1"" of type '" "wasserstein::EMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< double > * >(argp1);
  {
    try {
      wasserstein_EMDBase_Sl_double_Sg__npy_node_potentials(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *EMDBaseFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__EMDBaseT_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_delete_PairwiseEMDBaseFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseEMDBaseFloat64" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_R" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::PairwiseEMDBase< double > const *)arg1)->R(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"R",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat64_set_R", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_R" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_set_R" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_R(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_beta" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::PairwiseEMDBase< double > const *)arg1)->beta(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"beta",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat64_set_beta", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_beta" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_set_beta" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->set_beta(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_norm" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< double > const *)arg1)->norm(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"norm",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat64_set_norm", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_norm" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_set_norm" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_norm(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_network_simplex_params(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  std::size_t arg2 = (std::size_t) 100000 ;
  double arg3 = (double) 1000 ;
  double arg4 = (double) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OOO:PairwiseEMDBaseFloat64_set_network_simplex_params", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_network_simplex_params" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_size_t(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_set_network_simplex_params" "', argument " "2"" of type '" "std::size_t""'");
    } 
    arg2 = static_cast< std::size_t >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_double(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDBaseFloat64_set_network_simplex_params" "', argument " "3"" of type '" "double""'");
    } 
    arg3 = static_cast< double >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_double(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "PairwiseEMDBaseFloat64_set_network_simplex_params" "', argument " "4"" of type '" "double""'");
    } 
    arg4 = static_cast< double >(val4);
  }
  {
    try {
      (arg1)->set_network_simplex_params(arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_external_emd_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  wasserstein::ExternalEMDHandler< double > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"handler",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat64_set_external_emd_handler", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_external_emd_handler" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PairwiseEMDBaseFloat64_set_external_emd_handler" "', argument " "2"" of type '" "wasserstein::ExternalEMDHandler< double > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseEMDBaseFloat64_set_external_emd_handler" "', argument " "2"" of type '" "wasserstein::ExternalEMDHandler< double > &""'"); 
  }
  arg2 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp2);
  {
    try {
      (arg1)->set_external_emd_handler(*arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_have_external_emd_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_have_external_emd_handler" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< double > const *)arg1)->have_external_emd_handler(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_num_threads(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_num_threads" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (int)((wasserstein::PairwiseEMDBase< double > const *)arg1)->num_threads(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"chunksize",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      (arg1)->set_omp_dynamic_chunksize(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_omp_dynamic_chunksize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_omp_dynamic_chunksize" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (int)((wasserstein::PairwiseEMDBase< double > const *)arg1)->omp_dynamic_chunksize(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_set_request_mode(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"mode",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat64_set_request_mode", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_set_request_mode" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_set_request_mode" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_request_mode(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_request_mode(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_request_mode" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< double > const *)arg1)->request_mode(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_duration(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_duration" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::PairwiseEMDBase< double > const *)arg1)->duration(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_nevA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_nevA" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMDBase< double > const *)arg1)->nevA(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_nevB(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_nevB" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMDBase< double > const *)arg1)->nevB(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_num_emds(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_num_emds" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMDBase< double > const *)arg1)->num_emds(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_storage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::EMDPairsStorage result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_storage" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (wasserstein::EMDPairsStorage)((wasserstein::PairwiseEMDBase< double > const *)arg1)->storage(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_errored(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_errored" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< double > const *)arg1)->errored(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_error_messages(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string,std::allocator< std::string > > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_error_messages" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      result = (std::vector< std::string,std::allocator< std::string > > *) &((wasserstein::PairwiseEMDBase< double > const *)arg1)->error_messages(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< std::string,std::allocator< std::string > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_emds_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  bool arg2 = (bool) false ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"raw",  NULL 
  };
  std::vector< double,std::allocator< double > > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:PairwiseEMDBaseFloat64_emds_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_emds_vec" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_emds_vec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = (std::vector< double,std::allocator< double > > *) &(arg1)->emds(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_emd(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  int arg4 = (int) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  (char *)"thread",  NULL 
  };
  double result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO|O:PairwiseEMDBaseFloat64_emd", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_emd" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat64_emd" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDBaseFloat64_emd" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  if (obj3) {
    ecode4 = SWIG_AsVal_int(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "PairwiseEMDBaseFloat64_emd" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = static_cast< int >(val4);
  }
  {
    try {
      result = (double)(arg1)->emd(arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_emds(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  std::ptrdiff_t *arg4 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim1_temp2 ;
  std::ptrdiff_t dim2_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim1_temp2;
    arg4 = &dim2_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_emds" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMDBase_Sl_double_Sg__npy_emds(arg1,arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[2] = {
      *arg3, *arg4 
    };
    PyObject* obj = PyArray_SimpleNewFromData(2, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat64_raw_emds(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< double > *arg1 = (wasserstein::PairwiseEMDBase< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat64_raw_emds" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< double > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMDBase_Sl_double_Sg__raw_emds(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PairwiseEMDBaseFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_delete_ExternalEMDHandlerFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ExternalEMDHandlerFloat64" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat64_description" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  {
    try {
      result = ((wasserstein::ExternalEMDHandler< double > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64_num_calls(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::size_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat64_num_calls" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  {
    try {
      result = ((wasserstein::ExternalEMDHandler< double > const *)arg1)->num_calls(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64___call__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  double arg2 ;
  double arg3 = (double) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"emd",  (char *)"weight",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO|O:ExternalEMDHandlerFloat64___call__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat64___call__" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ExternalEMDHandlerFloat64___call__" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  if (obj2) {
    ecode3 = SWIG_AsVal_double(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ExternalEMDHandlerFloat64___call__" "', argument " "3"" of type '" "double""'");
    } 
    arg3 = static_cast< double >(val3);
  }
  {
    try {
      (arg1)->operator ()(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64_evaluate__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat64_evaluate" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    try {
      wasserstein_ExternalEMDHandler_Sl_double_Sg__npy_evaluate__SWIG_0(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64_evaluate__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat64_evaluate" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2],
      NPY_DOUBLE,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    try {
      wasserstein_ExternalEMDHandler_Sl_double_Sg__npy_evaluate__SWIG_1(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64_evaluate(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "ExternalEMDHandlerFloat64_evaluate", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_ExternalEMDHandlerFloat64_evaluate__SWIG_0(self, argc, argv);
        }
        return _wrap_ExternalEMDHandlerFloat64_evaluate__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_ExternalEMDHandlerFloat64_evaluate__SWIG_1(self, argc, argv);
          }
          return _wrap_ExternalEMDHandlerFloat64_evaluate__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'ExternalEMDHandlerFloat64_evaluate'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::ExternalEMDHandler< double >::npy_evaluate(double *,std::ptrdiff_t)\n"
    "    wasserstein::ExternalEMDHandler< double >::npy_evaluate(double *,std::ptrdiff_t,double *,std::ptrdiff_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat64_evaluate_symmetric(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< double > *arg1 = (wasserstein::ExternalEMDHandler< double > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"emds",  (char *)"event_weights",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:ExternalEMDHandlerFloat64_evaluate_symmetric", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat64_evaluate_symmetric" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< double > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(obj1,
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(obj2,
      NPY_DOUBLE,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    try {
      wasserstein_ExternalEMDHandler_Sl_double_Sg__npy_evaluate_symmetric(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *ExternalEMDHandlerFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Histogram1DHandlerLogFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  double arg2 ;
  double arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"nbins",  (char *)"axis_min",  (char *)"axis_max",  NULL 
  };
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:new_Histogram1DHandlerLogFloat64", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Histogram1DHandlerLogFloat64" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Histogram1DHandlerLogFloat64" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Histogram1DHandlerLogFloat64" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      result = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *)new wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Histogram1DHandlerLogFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Histogram1DHandlerLogFloat64" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_nbins(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_nbins" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = (unsigned int)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->nbins(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_axis_min(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_axis_min" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->axis_min(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_axis_max(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_axis_max" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->axis_max(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_description" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_hist_vals_vars_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerLogFloat64_hist_vals_vars_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_hist_vals_vars_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Histogram1DHandlerLogFloat64_hist_vals_vars_vec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->hist_vals_vars(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_bin_centers_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_bin_centers_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->bin_centers(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_bin_edges_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_bin_edges_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1)->bin_edges(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64___iadd__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"other",  NULL 
  };
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:Histogram1DHandlerLogFloat64___iadd__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64___iadd__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Histogram1DHandlerLogFloat64___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Histogram1DHandlerLogFloat64___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > const &""'"); 
  }
  arg2 = reinterpret_cast< wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > * >(argp2);
  {
    try {
      result = (wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > *) &(arg1)->operator +=((wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > const &)*arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64___repr__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      result = wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg____repr__((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_bin_centers(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_bin_centers" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg__npy_bin_centers(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_bin_edges(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_bin_edges" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg__npy_bin_edges(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat64_hist_vals_vars(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  double **arg4 = (double **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  bool arg6 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  double *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  bool val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerLogFloat64_hist_vals_vars", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat64_hist_vals_vars" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > * >(argp1);
  if (obj1) {
    ecode6 = SWIG_AsVal_bool(obj1, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Histogram1DHandlerLogFloat64_hist_vals_vars" "', argument " "6"" of type '" "bool""'");
    } 
    arg6 = static_cast< bool >(val6);
  }
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_double_Sg__npy_hist_vals_vars(arg1,arg2,arg3,arg4,arg5,arg6); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Histogram1DHandlerLogFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Histogram1DHandlerLogFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Histogram1DHandlerFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  double arg2 ;
  double arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"nbins",  (char *)"axis_min",  (char *)"axis_max",  NULL 
  };
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:new_Histogram1DHandlerFloat64", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Histogram1DHandlerFloat64" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Histogram1DHandlerFloat64" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Histogram1DHandlerFloat64" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      result = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *)new wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Histogram1DHandlerFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Histogram1DHandlerFloat64" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_nbins(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_nbins" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = (unsigned int)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->nbins(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_axis_min(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_axis_min" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->axis_min(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_axis_max(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_axis_max" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = (double)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->axis_max(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_description" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_hist_vals_vars_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerFloat64_hist_vals_vars_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_hist_vals_vars_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Histogram1DHandlerFloat64_hist_vals_vars_vec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->hist_vals_vars(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_bin_centers_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_bin_centers_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->bin_centers(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_bin_edges_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_bin_edges_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1)->bin_edges(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64___iadd__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"other",  NULL 
  };
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:Histogram1DHandlerFloat64___iadd__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64___iadd__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Histogram1DHandlerFloat64___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Histogram1DHandlerFloat64___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > const &""'"); 
  }
  arg2 = reinterpret_cast< wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > * >(argp2);
  {
    try {
      result = (wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > *) &(arg1)->operator +=((wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > const &)*arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64___repr__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      result = wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg____repr__((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_bin_centers(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_bin_centers" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg__npy_bin_centers(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_bin_edges(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_bin_edges" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg__npy_bin_edges(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat64_hist_vals_vars(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  double **arg4 = (double **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  bool arg6 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  double *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  bool val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerFloat64_hist_vals_vars", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat64_hist_vals_vars" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > * >(argp1);
  if (obj1) {
    ecode6 = SWIG_AsVal_bool(obj1, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Histogram1DHandlerFloat64_hist_vals_vars" "', argument " "6"" of type '" "bool""'");
    } 
    arg6 = static_cast< bool >(val6);
  }
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_double_Sg__npy_hist_vals_vars(arg1,arg2,arg3,arg4,arg5,arg6); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Histogram1DHandlerFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Histogram1DHandlerFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_delete_EMDBaseFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EMDBaseFloat32" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_R" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->R(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_beta" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->beta(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_set_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"R",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat32_set_R", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_set_R" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_set_R" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  {
    try {
      (arg1)->set_R(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_set_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"beta",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat32_set_beta", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_set_beta" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_set_beta" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  {
    try {
      (arg1)->set_beta(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_set_network_simplex_params(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  std::size_t arg2 = (std::size_t) 100000 ;
  float arg3 = (float) 1000 ;
  float arg4 = (float) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  float val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OOO:EMDBaseFloat32_set_network_simplex_params", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_set_network_simplex_params" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_size_t(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_set_network_simplex_params" "', argument " "2"" of type '" "std::size_t""'");
    } 
    arg2 = static_cast< std::size_t >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_float(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "EMDBaseFloat32_set_network_simplex_params" "', argument " "3"" of type '" "float""'");
    } 
    arg3 = static_cast< float >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_float(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "EMDBaseFloat32_set_network_simplex_params" "', argument " "4"" of type '" "float""'");
    } 
    arg4 = static_cast< float >(val4);
  }
  {
    try {
      (arg1)->set_network_simplex_params(arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_norm" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::EMDBase< float > const *)arg1)->norm(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_set_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"norm",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat32_set_norm", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_set_norm" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_set_norm" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_norm(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_do_timing(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_do_timing" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::EMDBase< float > const *)arg1)->do_timing(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_set_do_timing(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"timing",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat32_set_do_timing", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_set_do_timing" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_set_do_timing" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_do_timing(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_external_dists(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_external_dists" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::EMDBase< float > const *)arg1)->external_dists(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_set_external_dists(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"exdists",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:EMDBaseFloat32_set_external_dists", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_set_external_dists" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_set_external_dists" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_external_dists(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_n0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_n0" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< float > const *)arg1)->n0(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_n1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_n1" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< float > const *)arg1)->n1(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_extra(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::ExtraParticle result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_extra" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (wasserstein::ExtraParticle)((wasserstein::EMDBase< float > const *)arg1)->extra(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_emd(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_emd" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->emd(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_status(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::EMDStatus result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_status" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (wasserstein::EMDStatus)((wasserstein::EMDBase< float > const *)arg1)->status(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_weightdiff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_weightdiff" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->weightdiff(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_scale(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_scale" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->scale(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_n_iter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::size_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_n_iter" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< float > const *)arg1)->n_iter(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_dists_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_dists_vec" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< float > const *)arg1)->dists(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_flows_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_flows_vec" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< float > const *)arg1)->flows(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_flow__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  float result;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_flow" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_flow" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "EMDBaseFloat32_flow" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->flow(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_flow__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  std::size_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  float result;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_flow" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDBaseFloat32_flow" "', argument " "2"" of type '" "std::size_t""'");
  } 
  arg2 = static_cast< std::size_t >(val2);
  {
    try {
      result = (float)((wasserstein::EMDBase< float > const *)arg1)->flow(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_flow(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "EMDBaseFloat32_flow", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_EMDBaseFloat32_flow__SWIG_1(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_EMDBaseFloat32_flow__SWIG_0(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'EMDBaseFloat32_flow'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::EMDBase< float >::flow(wasserstein::index_type,wasserstein::index_type) const\n"
    "    wasserstein::EMDBase< float >::flow(std::size_t) const\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_node_potentials(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_node_potentials" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::EMDBase< float > const *)arg1)->node_potentials(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_duration(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_duration" "', argument " "1"" of type '" "wasserstein::EMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      result = (double)((wasserstein::EMDBase< float > const *)arg1)->duration(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_clear" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      (arg1)->clear(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_flows(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  std::ptrdiff_t *arg4 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim1_temp2 ;
  std::ptrdiff_t dim2_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim1_temp2;
    arg4 = &dim2_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_flows" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      wasserstein_EMDBase_Sl_float_Sg__npy_flows(arg1,arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[2] = {
      *arg3, *arg4 
    };
    PyObject* obj = PyArray_SimpleNewFromData(2, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_dists(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  std::ptrdiff_t *arg4 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim1_temp2 ;
  std::ptrdiff_t dim2_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim1_temp2;
    arg4 = &dim2_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_dists" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      wasserstein_EMDBase_Sl_float_Sg__npy_dists(arg1,arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[2] = {
      *arg3, *arg4 
    };
    PyObject* obj = PyArray_SimpleNewFromData(2, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDBaseFloat32_npy_node_potentials(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMDBase< float > *arg1 = (wasserstein::EMDBase< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  float **arg4 = (float **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  float *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDBaseFloat32_npy_node_potentials" "', argument " "1"" of type '" "wasserstein::EMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMDBase< float > * >(argp1);
  {
    try {
      wasserstein_EMDBase_Sl_float_Sg__npy_node_potentials(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *EMDBaseFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__EMDBaseT_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_delete_PairwiseEMDBaseFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseEMDBaseFloat32" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_R" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::PairwiseEMDBase< float > const *)arg1)->R(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_R(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"R",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat32_set_R", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_R" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_set_R" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  {
    try {
      (arg1)->set_R(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_beta" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::PairwiseEMDBase< float > const *)arg1)->beta(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_beta(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"beta",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat32_set_beta", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_beta" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_set_beta" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  {
    try {
      (arg1)->set_beta(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_norm" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< float > const *)arg1)->norm(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_norm(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"norm",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat32_set_norm", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_norm" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_set_norm" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_norm(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_network_simplex_params(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  std::size_t arg2 = (std::size_t) 100000 ;
  float arg3 = (float) 1000 ;
  float arg4 = (float) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  float val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OOO:PairwiseEMDBaseFloat32_set_network_simplex_params", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_network_simplex_params" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_size_t(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_set_network_simplex_params" "', argument " "2"" of type '" "std::size_t""'");
    } 
    arg2 = static_cast< std::size_t >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_float(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDBaseFloat32_set_network_simplex_params" "', argument " "3"" of type '" "float""'");
    } 
    arg3 = static_cast< float >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_float(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "PairwiseEMDBaseFloat32_set_network_simplex_params" "', argument " "4"" of type '" "float""'");
    } 
    arg4 = static_cast< float >(val4);
  }
  {
    try {
      (arg1)->set_network_simplex_params(arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_external_emd_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  wasserstein::ExternalEMDHandler< float > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"handler",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat32_set_external_emd_handler", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_external_emd_handler" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PairwiseEMDBaseFloat32_set_external_emd_handler" "', argument " "2"" of type '" "wasserstein::ExternalEMDHandler< float > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PairwiseEMDBaseFloat32_set_external_emd_handler" "', argument " "2"" of type '" "wasserstein::ExternalEMDHandler< float > &""'"); 
  }
  arg2 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp2);
  {
    try {
      (arg1)->set_external_emd_handler(*arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_have_external_emd_handler(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_have_external_emd_handler" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< float > const *)arg1)->have_external_emd_handler(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_num_threads(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_num_threads" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (int)((wasserstein::PairwiseEMDBase< float > const *)arg1)->num_threads(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"chunksize",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize" "', argument " "2"" of type '" "int""'");
  } 
  arg2 = static_cast< int >(val2);
  {
    try {
      (arg1)->set_omp_dynamic_chunksize(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_omp_dynamic_chunksize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_omp_dynamic_chunksize" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (int)((wasserstein::PairwiseEMDBase< float > const *)arg1)->omp_dynamic_chunksize(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_set_request_mode(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  bool arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"mode",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:PairwiseEMDBaseFloat32_set_request_mode", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_set_request_mode" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_set_request_mode" "', argument " "2"" of type '" "bool""'");
  } 
  arg2 = static_cast< bool >(val2);
  {
    try {
      (arg1)->set_request_mode(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_request_mode(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_request_mode" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< float > const *)arg1)->request_mode(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_duration(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  double result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_duration" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (double)((wasserstein::PairwiseEMDBase< float > const *)arg1)->duration(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_nevA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_nevA" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMDBase< float > const *)arg1)->nevA(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_nevB(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_nevB" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMDBase< float > const *)arg1)->nevB(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_num_emds(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::index_type result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_num_emds" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMDBase< float > const *)arg1)->num_emds(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_storage(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  wasserstein::EMDPairsStorage result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_storage" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (wasserstein::EMDPairsStorage)((wasserstein::PairwiseEMDBase< float > const *)arg1)->storage(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_errored(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  bool result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_errored" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (bool)((wasserstein::PairwiseEMDBase< float > const *)arg1)->errored(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_error_messages(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< std::string,std::allocator< std::string > > *result = 0 ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_error_messages" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      result = (std::vector< std::string,std::allocator< std::string > > *) &((wasserstein::PairwiseEMDBase< float > const *)arg1)->error_messages(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< std::string,std::allocator< std::string > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_emds_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  bool arg2 = (bool) false ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"raw",  NULL 
  };
  std::vector< float,std::allocator< float > > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:PairwiseEMDBaseFloat32_emds_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_emds_vec" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_emds_vec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = (std::vector< float,std::allocator< float > > *) &(arg1)->emds(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_emd(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  int arg4 = (int) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"i",  (char *)"j",  (char *)"thread",  NULL 
  };
  float result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO|O:PairwiseEMDBaseFloat32_emd", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_emd" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDBaseFloat32_emd" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDBaseFloat32_emd" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  if (obj3) {
    ecode4 = SWIG_AsVal_int(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "PairwiseEMDBaseFloat32_emd" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = static_cast< int >(val4);
  }
  {
    try {
      result = (float)(arg1)->emd(arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_emds(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  std::ptrdiff_t *arg4 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim1_temp2 ;
  std::ptrdiff_t dim2_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim1_temp2;
    arg4 = &dim2_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_emds" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMDBase_Sl_float_Sg__npy_emds(arg1,arg2,arg3,arg4); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[2] = {
      *arg3, *arg4 
    };
    PyObject* obj = PyArray_SimpleNewFromData(2, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDBaseFloat32_raw_emds(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMDBase< float > *arg1 = (wasserstein::PairwiseEMDBase< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDBaseFloat32_raw_emds" "', argument " "1"" of type '" "wasserstein::PairwiseEMDBase< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMDBase< float > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMDBase_Sl_float_Sg__raw_emds(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PairwiseEMDBaseFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__PairwiseEMDBaseT_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_delete_ExternalEMDHandlerFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ExternalEMDHandlerFloat32" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat32_description" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  {
    try {
      result = ((wasserstein::ExternalEMDHandler< float > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32_num_calls(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::size_t result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat32_num_calls" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  {
    try {
      result = ((wasserstein::ExternalEMDHandler< float > const *)arg1)->num_calls(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32___call__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  float arg2 ;
  float arg3 = (float) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"emd",  (char *)"weight",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO|O:ExternalEMDHandlerFloat32___call__", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat32___call__" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ExternalEMDHandlerFloat32___call__" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  if (obj2) {
    ecode3 = SWIG_AsVal_float(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "ExternalEMDHandlerFloat32___call__" "', argument " "3"" of type '" "float""'");
    } 
    arg3 = static_cast< float >(val3);
  }
  {
    try {
      (arg1)->operator ()(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32_evaluate__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat32_evaluate" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    try {
      wasserstein_ExternalEMDHandler_Sl_float_Sg__npy_evaluate__SWIG_0(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32_evaluate__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat32_evaluate" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2],
      NPY_FLOAT,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    try {
      wasserstein_ExternalEMDHandler_Sl_float_Sg__npy_evaluate__SWIG_1(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32_evaluate(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "ExternalEMDHandlerFloat32_evaluate", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_ExternalEMDHandlerFloat32_evaluate__SWIG_0(self, argc, argv);
        }
        return _wrap_ExternalEMDHandlerFloat32_evaluate__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          if (argc <= 3) {
            return _wrap_ExternalEMDHandlerFloat32_evaluate__SWIG_1(self, argc, argv);
          }
          return _wrap_ExternalEMDHandlerFloat32_evaluate__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'ExternalEMDHandlerFloat32_evaluate'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::ExternalEMDHandler< float >::npy_evaluate(float *,std::ptrdiff_t)\n"
    "    wasserstein::ExternalEMDHandler< float >::npy_evaluate(float *,std::ptrdiff_t,float *,std::ptrdiff_t)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_ExternalEMDHandlerFloat32_evaluate_symmetric(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::ExternalEMDHandler< float > *arg1 = (wasserstein::ExternalEMDHandler< float > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"emds",  (char *)"event_weights",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:ExternalEMDHandlerFloat32_evaluate_symmetric", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ExternalEMDHandlerFloat32_evaluate_symmetric" "', argument " "1"" of type '" "wasserstein::ExternalEMDHandler< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::ExternalEMDHandler< float > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(obj1,
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(obj2,
      NPY_FLOAT,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    try {
      wasserstein_ExternalEMDHandler_Sl_float_Sg__npy_evaluate_symmetric(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *ExternalEMDHandlerFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__ExternalEMDHandlerT_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Histogram1DHandlerLogFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  float arg2 ;
  float arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"nbins",  (char *)"axis_min",  (char *)"axis_max",  NULL 
  };
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:new_Histogram1DHandlerLogFloat32", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Histogram1DHandlerLogFloat32" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Histogram1DHandlerLogFloat32" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  ecode3 = SWIG_AsVal_float(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Histogram1DHandlerLogFloat32" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = static_cast< float >(val3);
  {
    try {
      result = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *)new wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Histogram1DHandlerLogFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Histogram1DHandlerLogFloat32" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_nbins(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_nbins" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = (unsigned int)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->nbins(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_axis_min(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_axis_min" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->axis_min(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_axis_max(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_axis_max" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->axis_max(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_description" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_hist_vals_vars_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerLogFloat32_hist_vals_vars_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_hist_vals_vars_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Histogram1DHandlerLogFloat32_hist_vals_vars_vec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->hist_vals_vars(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_bin_centers_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_bin_centers_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->bin_centers(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_bin_edges_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_bin_edges_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1)->bin_edges(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32___iadd__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"other",  NULL 
  };
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:Histogram1DHandlerLogFloat32___iadd__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32___iadd__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Histogram1DHandlerLogFloat32___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Histogram1DHandlerLogFloat32___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > const &""'"); 
  }
  arg2 = reinterpret_cast< wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > * >(argp2);
  {
    try {
      result = (wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > *) &(arg1)->operator +=((wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > const &)*arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32___repr__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      result = wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg____repr__((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_bin_centers(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_bin_centers" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg__npy_bin_centers(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_bin_edges(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_bin_edges" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg__npy_bin_edges(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerLogFloat32_hist_vals_vars(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  float **arg4 = (float **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  bool arg6 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  float *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  bool val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerLogFloat32_hist_vals_vars", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerLogFloat32_hist_vals_vars" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > * >(argp1);
  if (obj1) {
    ecode6 = SWIG_AsVal_bool(obj1, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Histogram1DHandlerLogFloat32_hist_vals_vars" "', argument " "6"" of type '" "bool""'");
    } 
    arg6 = static_cast< bool >(val6);
  }
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_log_Sc_float_Sg__npy_hist_vals_vars(arg1,arg2,arg3,arg4,arg5,arg6); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Histogram1DHandlerLogFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Histogram1DHandlerLogFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_Histogram1DHandlerFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  float arg2 ;
  float arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"nbins",  (char *)"axis_min",  (char *)"axis_max",  NULL 
  };
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:new_Histogram1DHandlerFloat32", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Histogram1DHandlerFloat32" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Histogram1DHandlerFloat32" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  ecode3 = SWIG_AsVal_float(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Histogram1DHandlerFloat32" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = static_cast< float >(val3);
  {
    try {
      result = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *)new wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Histogram1DHandlerFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Histogram1DHandlerFloat32" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_nbins(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  unsigned int result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_nbins" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = (unsigned int)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->nbins(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_axis_min(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_axis_min" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->axis_min(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_axis_max(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  float result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_axis_max" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = (float)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->axis_max(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_description" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_hist_vals_vars_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerFloat32_hist_vals_vars_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_hist_vals_vars_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Histogram1DHandlerFloat32_hist_vals_vars_vec" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->hist_vals_vars(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_bin_centers_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_bin_centers_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->bin_centers(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_bin_edges_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_bin_edges_vec" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1)->bin_edges(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32___iadd__(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"other",  NULL 
  };
  wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OO:Histogram1DHandlerFloat32___iadd__", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32___iadd__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Histogram1DHandlerFloat32___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Histogram1DHandlerFloat32___iadd__" "', argument " "2"" of type '" "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > const &""'"); 
  }
  arg2 = reinterpret_cast< wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > * >(argp2);
  {
    try {
      result = (wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > *) &(arg1)->operator +=((wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > const &)*arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32___repr__" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      result = wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg____repr__((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_bin_centers(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_bin_centers" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg__npy_bin_centers(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_bin_edges(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_bin_edges" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg__npy_bin_edges(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Histogram1DHandlerFloat32_hist_vals_vars(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *arg1 = (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  float **arg4 = (float **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  bool arg6 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  float *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  bool val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"overflows",  NULL 
  };
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:Histogram1DHandlerFloat32_hist_vals_vars", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Histogram1DHandlerFloat32_hist_vals_vars" "', argument " "1"" of type '" "wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > * >(argp1);
  if (obj1) {
    ecode6 = SWIG_AsVal_bool(obj1, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Histogram1DHandlerFloat32_hist_vals_vars" "', argument " "6"" of type '" "bool""'");
    } 
    arg6 = static_cast< bool >(val6);
  }
  {
    try {
      wasserstein_Histogram1DHandler_Sl_boost_histogram_axis_transform_id_Sc_float_Sg__npy_hist_vals_vars(arg1,arg2,arg3,arg4,arg5,arg6); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Histogram1DHandlerFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *Histogram1DHandlerFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CorrelationDimensionFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  float arg2 ;
  float arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  float val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"nbins",  (char *)"axis_min",  (char *)"axis_max",  NULL 
  };
  wasserstein::CorrelationDimension< float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:new_CorrelationDimensionFloat32", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CorrelationDimensionFloat32" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CorrelationDimensionFloat32" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  ecode3 = SWIG_AsVal_float(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CorrelationDimensionFloat32" "', argument " "3"" of type '" "float""'");
  } 
  arg3 = static_cast< float >(val3);
  {
    try {
      result = (wasserstein::CorrelationDimension< float > *)new wasserstein::CorrelationDimension< float >(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32_corrdims_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  float arg2 = (float) std::numeric_limits< float >::epsilon() ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"eps",  NULL 
  };
  std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:CorrelationDimensionFloat32_corrdims_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32_corrdims_vec" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_float(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CorrelationDimensionFloat32_corrdims_vec" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = static_cast< float >(val2);
  }
  {
    try {
      result = ((wasserstein::CorrelationDimension< float > const *)arg1)->corrdims(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32_corrdim_bins_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< float,std::allocator< float > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32_corrdim_bins_vec" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      result = ((wasserstein::CorrelationDimension< float > const *)arg1)->corrdim_bins(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< float,std::allocator< float > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32_cumulative_vals_vars_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32_cumulative_vals_vars_vec" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      result = ((wasserstein::CorrelationDimension< float > const *)arg1)->cumulative_vals_vars(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32___repr__" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      result = wasserstein_CorrelationDimension_Sl_float_Sg____repr__((wasserstein::CorrelationDimension< float > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32_corrdim_bins(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32_corrdim_bins" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      wasserstein_CorrelationDimension_Sl_float_Sg__npy_corrdim_bins(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32_corrdims(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  float **arg4 = (float **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  float *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32_corrdims" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      wasserstein_CorrelationDimension_Sl_float_Sg__npy_corrdims(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat32_cumulative_vals_vars(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  float **arg2 = (float **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  float **arg4 = (float **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  float *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat32_cumulative_vals_vars" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      wasserstein_CorrelationDimension_Sl_float_Sg__npy_cumulative_vals_vars(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_FLOAT, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CorrelationDimensionFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< float > *arg1 = (wasserstein::CorrelationDimension< float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CorrelationDimensionFloat32" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CorrelationDimensionFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__CorrelationDimensionT_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CorrelationDimensionFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_CorrelationDimensionFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  double arg2 ;
  double arg3 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char * kwnames[] = {
    (char *)"nbins",  (char *)"axis_min",  (char *)"axis_max",  NULL 
  };
  wasserstein::CorrelationDimension< double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO:new_CorrelationDimensionFloat64", kwnames, &obj0, &obj1, &obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_CorrelationDimensionFloat64" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_CorrelationDimensionFloat64" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_CorrelationDimensionFloat64" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      result = (wasserstein::CorrelationDimension< double > *)new wasserstein::CorrelationDimension< double >(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64_corrdims_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  double arg2 = (double) std::numeric_limits< double >::epsilon() ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"eps",  NULL 
  };
  std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:CorrelationDimensionFloat64_corrdims_vec", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64_corrdims_vec" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_double(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "CorrelationDimensionFloat64_corrdims_vec" "', argument " "2"" of type '" "double""'");
    } 
    arg2 = static_cast< double >(val2);
  }
  {
    try {
      result = ((wasserstein::CorrelationDimension< double > const *)arg1)->corrdims(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64_corrdim_bins_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::vector< double,std::allocator< double > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64_corrdim_bins_vec" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      result = ((wasserstein::CorrelationDimension< double > const *)arg1)->corrdim_bins(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::vector< double,std::allocator< double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64_cumulative_vals_vars_vec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64_cumulative_vals_vars_vec" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      result = ((wasserstein::CorrelationDimension< double > const *)arg1)->cumulative_vals_vars(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = swig::from(static_cast< std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64___repr__" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      result = wasserstein_CorrelationDimension_Sl_double_Sg____repr__((wasserstein::CorrelationDimension< double > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64_corrdim_bins(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64_corrdim_bins" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      wasserstein_CorrelationDimension_Sl_double_Sg__npy_corrdim_bins(arg1,arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64_corrdims(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  double **arg4 = (double **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  double *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64_corrdims" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      wasserstein_CorrelationDimension_Sl_double_Sg__npy_corrdims(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_CorrelationDimensionFloat64_cumulative_vals_vars(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  double **arg2 = (double **) 0 ;
  std::ptrdiff_t *arg3 = (std::ptrdiff_t *) 0 ;
  double **arg4 = (double **) 0 ;
  std::ptrdiff_t *arg5 = (std::ptrdiff_t *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double *data_temp2 = NULL ;
  std::ptrdiff_t dim_temp2 ;
  double *data_temp4 = NULL ;
  std::ptrdiff_t dim_temp4 ;
  PyObject *swig_obj[1] ;
  
  {
    arg2 = &data_temp2;
    arg3 = &dim_temp2;
  }
  {
    arg4 = &data_temp4;
    arg5 = &dim_temp4;
  }
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "CorrelationDimensionFloat64_cumulative_vals_vars" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      wasserstein_CorrelationDimension_Sl_double_Sg__npy_cumulative_vals_vars(arg1,arg2,arg3,arg4,arg5); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  {
    npy_intp dims[1] = {
      *arg3 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg2));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg2), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg2), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  {
    npy_intp dims[1] = {
      *arg5 
    };
    PyObject* obj = PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)(*arg4));
    PyArrayObject* array = (PyArrayObject*) obj;
    
    if (!array) SWIG_fail;
    
#ifdef SWIGPY_USE_CAPSULE
    PyObject* cap = PyCapsule_New((void*)(*arg4), SWIGPY_CAPSULE_NAME, free_cap);
#else
    PyObject* cap = PyCObject_FromVoidPtr((void*)(*arg4), free);
#endif
    
#if NPY_API_VERSION < 0x00000007
    PyArray_BASE(array) = cap;
#else
    PyArray_SetBaseObject(array,cap);
#endif
    
    resultobj = SWIG_Python_AppendOutput(resultobj,obj);
  }
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_CorrelationDimensionFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::CorrelationDimension< double > *arg1 = (wasserstein::CorrelationDimension< double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_CorrelationDimensionFloat64" "', argument " "1"" of type '" "wasserstein::CorrelationDimension< double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::CorrelationDimension< double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *CorrelationDimensionFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__CorrelationDimensionT_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *CorrelationDimensionFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_EMDFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  double arg1 = (double) 1 ;
  double arg2 = (double) 1 ;
  bool arg3 = (bool) false ;
  bool arg4 = (bool) false ;
  bool arg5 = (bool) false ;
  std::size_t arg6 = (std::size_t) 100000 ;
  double arg7 = (double) 1000 ;
  double arg8 = (double) 1 ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  size_t val6 ;
  int ecode6 = 0 ;
  double val7 ;
  int ecode7 = 0 ;
  double val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"do_timing",  (char *)"external_dists",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOO:new_EMDFloat64", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_double(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_EMDFloat64" "', argument " "1"" of type '" "double""'");
    } 
    arg1 = static_cast< double >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_double(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_EMDFloat64" "', argument " "2"" of type '" "double""'");
    } 
    arg2 = static_cast< double >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_EMDFloat64" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_bool(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_EMDFloat64" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = static_cast< bool >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_bool(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_EMDFloat64" "', argument " "5"" of type '" "bool""'");
    } 
    arg5 = static_cast< bool >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_size_t(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_EMDFloat64" "', argument " "6"" of type '" "std::size_t""'");
    } 
    arg6 = static_cast< std::size_t >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_double(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_EMDFloat64" "', argument " "7"" of type '" "double""'");
    } 
    arg7 = static_cast< double >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_double(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_EMDFloat64" "', argument " "8"" of type '" "double""'");
    } 
    arg8 = static_cast< double >(val8);
  }
  {
    try {
      result = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *)new wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_EMDFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EMDFloat64" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"write_preprocessors",  NULL 
  };
  std::string result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:EMDFloat64_description", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat64_description" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDFloat64_description" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *)arg1)->description(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat64___repr__" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    try {
      result = wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg____repr__((wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat64_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat64_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    try {
      wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat64___call____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  double *arg7 = (double *) 0 ;
  std::ptrdiff_t arg8 ;
  double *arg9 = (double *) 0 ;
  std::ptrdiff_t arg10 ;
  std::ptrdiff_t arg11 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array7 = NULL ;
  int is_new_object7 = 0 ;
  PyArrayObject *array9 = NULL ;
  int is_new_object9 = 0 ;
  double result;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat64___call__" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2], NPY_DOUBLE,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 2) ||
      !require_size(array4, size, 2)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array7 = obj_to_array_contiguous_allow_conversion(swig_obj[3],
      NPY_DOUBLE,
      &is_new_object7);
    if (!array7 || !require_dimensions(array7, 1) ||
      !require_size(array7, size, 1)) SWIG_fail;
    arg7 = (double*) array_data(array7);
    arg8 = (std::ptrdiff_t) array_size(array7,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array9 = obj_to_array_contiguous_allow_conversion(swig_obj[4], NPY_DOUBLE,
      &is_new_object9);
    if (!array9 || !require_dimensions(array9, 2) ||
      !require_size(array9, size, 2)) SWIG_fail;
    arg9 = (double*) array_data(array9);
    arg10 = (std::ptrdiff_t) array_size(array9,0);
    arg11 = (std::ptrdiff_t) array_size(array9,1);
  }
  {
    try {
      result = (double)wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_1(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat64___call____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  double *arg6 = (double *) 0 ;
  std::ptrdiff_t arg7 ;
  std::ptrdiff_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array6 = NULL ;
  int is_new_object6 = 0 ;
  double result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat64___call__" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2],
      NPY_DOUBLE,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array6 = obj_to_array_contiguous_allow_conversion(swig_obj[3], NPY_DOUBLE,
      &is_new_object6);
    if (!array6 || !require_dimensions(array6, 2) ||
      !require_size(array6, size, 2)) SWIG_fail;
    arg6 = (double*) array_data(array6);
    arg7 = (std::ptrdiff_t) array_size(array6,0);
    arg8 = (std::ptrdiff_t) array_size(array6,1);
  }
  {
    try {
      result = (double)wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_2(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat64___call__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "EMDFloat64___call__", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            if (argc <= 4) {
              return _wrap_EMDFloat64___call____SWIG_2(self, argc, argv);
            }
            if (argc <= 5) {
              return _wrap_EMDFloat64___call____SWIG_2(self, argc, argv);
            }
            return _wrap_EMDFloat64___call____SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            {
              _v = is_array(argv[4]) || PySequence_Check(argv[4]);
            }
            if (_v) {
              if (argc <= 5) {
                return _wrap_EMDFloat64___call____SWIG_1(self, argc, argv);
              }
              if (argc <= 6) {
                return _wrap_EMDFloat64___call____SWIG_1(self, argc, argv);
              }
              return _wrap_EMDFloat64___call____SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'EMDFloat64___call__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >::operator ()(double *,std::ptrdiff_t,double *,std::ptrdiff_t,std::ptrdiff_t,double *,std::ptrdiff_t,double *,std::ptrdiff_t,std::ptrdiff_t)\n"
    "    wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >::operator ()(double *,std::ptrdiff_t,double *,std::ptrdiff_t,double *,std::ptrdiff_t,std::ptrdiff_t)\n");
  return 0;
}


SWIGINTERN PyObject *EMDFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *EMDFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_EMDFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  float arg1 = (float) 1 ;
  float arg2 = (float) 1 ;
  bool arg3 = (bool) false ;
  bool arg4 = (bool) false ;
  bool arg5 = (bool) false ;
  std::size_t arg6 = (std::size_t) 100000 ;
  float arg7 = (float) 1000 ;
  float arg8 = (float) 1 ;
  float val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  size_t val6 ;
  int ecode6 = 0 ;
  float val7 ;
  int ecode7 = 0 ;
  float val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"do_timing",  (char *)"external_dists",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOO:new_EMDFloat32", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_float(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_EMDFloat32" "', argument " "1"" of type '" "float""'");
    } 
    arg1 = static_cast< float >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_float(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_EMDFloat32" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = static_cast< float >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_EMDFloat32" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_bool(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_EMDFloat32" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = static_cast< bool >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_bool(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_EMDFloat32" "', argument " "5"" of type '" "bool""'");
    } 
    arg5 = static_cast< bool >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_size_t(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_EMDFloat32" "', argument " "6"" of type '" "std::size_t""'");
    } 
    arg6 = static_cast< std::size_t >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_float(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_EMDFloat32" "', argument " "7"" of type '" "float""'");
    } 
    arg7 = static_cast< float >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_float(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_EMDFloat32" "', argument " "8"" of type '" "float""'");
    } 
    arg8 = static_cast< float >(val8);
  }
  {
    try {
      result = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *)new wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_EMDFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EMDFloat32" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"write_preprocessors",  NULL 
  };
  std::string result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:EMDFloat32_description", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat32_description" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDFloat32_description" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *)arg1)->description(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat32___repr__" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    try {
      result = wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg____repr__((wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat32_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat32_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    try {
      wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat32___call____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  float *arg7 = (float *) 0 ;
  std::ptrdiff_t arg8 ;
  float *arg9 = (float *) 0 ;
  std::ptrdiff_t arg10 ;
  std::ptrdiff_t arg11 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array7 = NULL ;
  int is_new_object7 = 0 ;
  PyArrayObject *array9 = NULL ;
  int is_new_object9 = 0 ;
  float result;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat32___call__" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2], NPY_FLOAT,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 2) ||
      !require_size(array4, size, 2)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array7 = obj_to_array_contiguous_allow_conversion(swig_obj[3],
      NPY_FLOAT,
      &is_new_object7);
    if (!array7 || !require_dimensions(array7, 1) ||
      !require_size(array7, size, 1)) SWIG_fail;
    arg7 = (float*) array_data(array7);
    arg8 = (std::ptrdiff_t) array_size(array7,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array9 = obj_to_array_contiguous_allow_conversion(swig_obj[4], NPY_FLOAT,
      &is_new_object9);
    if (!array9 || !require_dimensions(array9, 2) ||
      !require_size(array9, size, 2)) SWIG_fail;
    arg9 = (float*) array_data(array9);
    arg10 = (std::ptrdiff_t) array_size(array9,0);
    arg11 = (std::ptrdiff_t) array_size(array9,1);
  }
  {
    try {
      result = (float)wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_1(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat32___call____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  float *arg6 = (float *) 0 ;
  std::ptrdiff_t arg7 ;
  std::ptrdiff_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array6 = NULL ;
  int is_new_object6 = 0 ;
  float result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDFloat32___call__" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2],
      NPY_FLOAT,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array6 = obj_to_array_contiguous_allow_conversion(swig_obj[3], NPY_FLOAT,
      &is_new_object6);
    if (!array6 || !require_dimensions(array6, 2) ||
      !require_size(array6, size, 2)) SWIG_fail;
    arg6 = (float*) array_data(array6);
    arg7 = (std::ptrdiff_t) array_size(array6,0);
    arg8 = (std::ptrdiff_t) array_size(array6,1);
  }
  {
    try {
      result = (float)wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__operator_Sp__SP___SWIG_2(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDFloat32___call__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "EMDFloat32___call__", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            if (argc <= 4) {
              return _wrap_EMDFloat32___call____SWIG_2(self, argc, argv);
            }
            if (argc <= 5) {
              return _wrap_EMDFloat32___call____SWIG_2(self, argc, argv);
            }
            return _wrap_EMDFloat32___call____SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            {
              _v = is_array(argv[4]) || PySequence_Check(argv[4]);
            }
            if (_v) {
              if (argc <= 5) {
                return _wrap_EMDFloat32___call____SWIG_1(self, argc, argv);
              }
              if (argc <= 6) {
                return _wrap_EMDFloat32___call____SWIG_1(self, argc, argv);
              }
              return _wrap_EMDFloat32___call____SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'EMDFloat32___call__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >::operator ()(float *,std::ptrdiff_t,float *,std::ptrdiff_t,std::ptrdiff_t,float *,std::ptrdiff_t,float *,std::ptrdiff_t,std::ptrdiff_t)\n"
    "    wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >::operator ()(float *,std::ptrdiff_t,float *,std::ptrdiff_t,float *,std::ptrdiff_t,std::ptrdiff_t)\n");
  return 0;
}


SWIGINTERN PyObject *EMDFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *EMDFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_EMDYPhiFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  double arg1 = (double) 1 ;
  double arg2 = (double) 1 ;
  bool arg3 = (bool) false ;
  bool arg4 = (bool) false ;
  bool arg5 = (bool) false ;
  std::size_t arg6 = (std::size_t) 100000 ;
  double arg7 = (double) 1000 ;
  double arg8 = (double) 1 ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  size_t val6 ;
  int ecode6 = 0 ;
  double val7 ;
  int ecode7 = 0 ;
  double val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"do_timing",  (char *)"external_dists",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOO:new_EMDYPhiFloat64", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_double(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_EMDYPhiFloat64" "', argument " "1"" of type '" "double""'");
    } 
    arg1 = static_cast< double >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_double(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_EMDYPhiFloat64" "', argument " "2"" of type '" "double""'");
    } 
    arg2 = static_cast< double >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_EMDYPhiFloat64" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_bool(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_EMDYPhiFloat64" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = static_cast< bool >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_bool(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_EMDYPhiFloat64" "', argument " "5"" of type '" "bool""'");
    } 
    arg5 = static_cast< bool >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_size_t(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_EMDYPhiFloat64" "', argument " "6"" of type '" "std::size_t""'");
    } 
    arg6 = static_cast< std::size_t >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_double(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_EMDYPhiFloat64" "', argument " "7"" of type '" "double""'");
    } 
    arg7 = static_cast< double >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_double(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_EMDYPhiFloat64" "', argument " "8"" of type '" "double""'");
    } 
    arg8 = static_cast< double >(val8);
  }
  {
    try {
      result = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *)new wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_EMDYPhiFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EMDYPhiFloat64" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"write_preprocessors",  NULL 
  };
  std::string result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:EMDYPhiFloat64_description", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat64_description" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDYPhiFloat64_description" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *)arg1)->description(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat64___repr__" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    try {
      result = wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg____repr__((wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat64_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat64_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    try {
      wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat64___call____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  double *arg7 = (double *) 0 ;
  std::ptrdiff_t arg8 ;
  double *arg9 = (double *) 0 ;
  std::ptrdiff_t arg10 ;
  std::ptrdiff_t arg11 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array7 = NULL ;
  int is_new_object7 = 0 ;
  PyArrayObject *array9 = NULL ;
  int is_new_object9 = 0 ;
  double result;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat64___call__" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2], NPY_DOUBLE,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 2) ||
      !require_size(array4, size, 2)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array7 = obj_to_array_contiguous_allow_conversion(swig_obj[3],
      NPY_DOUBLE,
      &is_new_object7);
    if (!array7 || !require_dimensions(array7, 1) ||
      !require_size(array7, size, 1)) SWIG_fail;
    arg7 = (double*) array_data(array7);
    arg8 = (std::ptrdiff_t) array_size(array7,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array9 = obj_to_array_contiguous_allow_conversion(swig_obj[4], NPY_DOUBLE,
      &is_new_object9);
    if (!array9 || !require_dimensions(array9, 2) ||
      !require_size(array9, size, 2)) SWIG_fail;
    arg9 = (double*) array_data(array9);
    arg10 = (std::ptrdiff_t) array_size(array9,0);
    arg11 = (std::ptrdiff_t) array_size(array9,1);
  }
  {
    try {
      result = (double)wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_1(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat64___call____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  double *arg6 = (double *) 0 ;
  std::ptrdiff_t arg7 ;
  std::ptrdiff_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array6 = NULL ;
  int is_new_object6 = 0 ;
  double result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat64___call__" "', argument " "1"" of type '" "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_DOUBLE,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2],
      NPY_DOUBLE,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array6 = obj_to_array_contiguous_allow_conversion(swig_obj[3], NPY_DOUBLE,
      &is_new_object6);
    if (!array6 || !require_dimensions(array6, 2) ||
      !require_size(array6, size, 2)) SWIG_fail;
    arg6 = (double*) array_data(array6);
    arg7 = (std::ptrdiff_t) array_size(array6,0);
    arg8 = (std::ptrdiff_t) array_size(array6,1);
  }
  {
    try {
      result = (double)wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_2(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat64___call__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "EMDYPhiFloat64___call__", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            if (argc <= 4) {
              return _wrap_EMDYPhiFloat64___call____SWIG_2(self, argc, argv);
            }
            if (argc <= 5) {
              return _wrap_EMDYPhiFloat64___call____SWIG_2(self, argc, argv);
            }
            return _wrap_EMDYPhiFloat64___call____SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            {
              _v = is_array(argv[4]) || PySequence_Check(argv[4]);
            }
            if (_v) {
              if (argc <= 5) {
                return _wrap_EMDYPhiFloat64___call____SWIG_1(self, argc, argv);
              }
              if (argc <= 6) {
                return _wrap_EMDYPhiFloat64___call____SWIG_1(self, argc, argv);
              }
              return _wrap_EMDYPhiFloat64___call____SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'EMDYPhiFloat64___call__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >::operator ()(double *,std::ptrdiff_t,double *,std::ptrdiff_t,std::ptrdiff_t,double *,std::ptrdiff_t,double *,std::ptrdiff_t,std::ptrdiff_t)\n"
    "    wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >::operator ()(double *,std::ptrdiff_t,double *,std::ptrdiff_t,double *,std::ptrdiff_t,std::ptrdiff_t)\n");
  return 0;
}


SWIGINTERN PyObject *EMDYPhiFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *EMDYPhiFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_EMDYPhiFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  float arg1 = (float) 1 ;
  float arg2 = (float) 1 ;
  bool arg3 = (bool) false ;
  bool arg4 = (bool) false ;
  bool arg5 = (bool) false ;
  std::size_t arg6 = (std::size_t) 100000 ;
  float arg7 = (float) 1000 ;
  float arg8 = (float) 1 ;
  float val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  bool val5 ;
  int ecode5 = 0 ;
  size_t val6 ;
  int ecode6 = 0 ;
  float val7 ;
  int ecode7 = 0 ;
  float val8 ;
  int ecode8 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"do_timing",  (char *)"external_dists",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  NULL 
  };
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOO:new_EMDYPhiFloat32", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_float(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_EMDYPhiFloat32" "', argument " "1"" of type '" "float""'");
    } 
    arg1 = static_cast< float >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_float(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_EMDYPhiFloat32" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = static_cast< float >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_EMDYPhiFloat32" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_bool(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_EMDYPhiFloat32" "', argument " "4"" of type '" "bool""'");
    } 
    arg4 = static_cast< bool >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_bool(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_EMDYPhiFloat32" "', argument " "5"" of type '" "bool""'");
    } 
    arg5 = static_cast< bool >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_size_t(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_EMDYPhiFloat32" "', argument " "6"" of type '" "std::size_t""'");
    } 
    arg6 = static_cast< std::size_t >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_float(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_EMDYPhiFloat32" "', argument " "7"" of type '" "float""'");
    } 
    arg7 = static_cast< float >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_float(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_EMDYPhiFloat32" "', argument " "8"" of type '" "float""'");
    } 
    arg8 = static_cast< float >(val8);
  }
  {
    try {
      result = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *)new wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_EMDYPhiFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EMDYPhiFloat32" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"write_preprocessors",  NULL 
  };
  std::string result;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:EMDYPhiFloat32_description", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat32_description" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EMDYPhiFloat32_description" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      result = ((wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *)arg1)->description(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat32___repr__" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    try {
      result = wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg____repr__((wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat32_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat32_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    try {
      wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat32___call____SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  float *arg7 = (float *) 0 ;
  std::ptrdiff_t arg8 ;
  float *arg9 = (float *) 0 ;
  std::ptrdiff_t arg10 ;
  std::ptrdiff_t arg11 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array7 = NULL ;
  int is_new_object7 = 0 ;
  PyArrayObject *array9 = NULL ;
  int is_new_object9 = 0 ;
  float result;
  
  if ((nobjs < 5) || (nobjs > 5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat32___call__" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2], NPY_FLOAT,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 2) ||
      !require_size(array4, size, 2)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array7 = obj_to_array_contiguous_allow_conversion(swig_obj[3],
      NPY_FLOAT,
      &is_new_object7);
    if (!array7 || !require_dimensions(array7, 1) ||
      !require_size(array7, size, 1)) SWIG_fail;
    arg7 = (float*) array_data(array7);
    arg8 = (std::ptrdiff_t) array_size(array7,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array9 = obj_to_array_contiguous_allow_conversion(swig_obj[4], NPY_FLOAT,
      &is_new_object9);
    if (!array9 || !require_dimensions(array9, 2) ||
      !require_size(array9, size, 2)) SWIG_fail;
    arg9 = (float*) array_data(array9);
    arg10 = (std::ptrdiff_t) array_size(array9,0);
    arg11 = (std::ptrdiff_t) array_size(array9,1);
  }
  {
    try {
      result = (float)wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_1(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object7 && array7)
    {
      Py_DECREF(array7); 
    }
  }
  {
    if (is_new_object9 && array9)
    {
      Py_DECREF(array9); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat32___call____SWIG_2(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *arg1 = (wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  float *arg6 = (float *) 0 ;
  std::ptrdiff_t arg7 ;
  std::ptrdiff_t arg8 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int is_new_object2 = 0 ;
  PyArrayObject *array4 = NULL ;
  int is_new_object4 = 0 ;
  PyArrayObject *array6 = NULL ;
  int is_new_object6 = 0 ;
  float result;
  
  if ((nobjs < 4) || (nobjs > 4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EMDYPhiFloat32___call__" "', argument " "1"" of type '" "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > * >(argp1);
  {
    npy_intp size[1] = {
      -1 
    };
    array2 = obj_to_array_contiguous_allow_conversion(swig_obj[1],
      NPY_FLOAT,
      &is_new_object2);
    if (!array2 || !require_dimensions(array2, 1) ||
      !require_size(array2, size, 1)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = (std::ptrdiff_t) array_size(array2,0);
  }
  {
    npy_intp size[1] = {
      -1 
    };
    array4 = obj_to_array_contiguous_allow_conversion(swig_obj[2],
      NPY_FLOAT,
      &is_new_object4);
    if (!array4 || !require_dimensions(array4, 1) ||
      !require_size(array4, size, 1)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
  }
  {
    npy_intp size[2] = {
      -1, -1 
    };
    array6 = obj_to_array_contiguous_allow_conversion(swig_obj[3], NPY_FLOAT,
      &is_new_object6);
    if (!array6 || !require_dimensions(array6, 2) ||
      !require_size(array6, size, 2)) SWIG_fail;
    arg6 = (float*) array_data(array6);
    arg7 = (std::ptrdiff_t) array_size(array6,0);
    arg8 = (std::ptrdiff_t) array_size(array6,1);
  }
  {
    try {
      result = (float)wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__operator_Sp__SP___SWIG_2(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return resultobj;
fail:
  {
    if (is_new_object2 && array2)
    {
      Py_DECREF(array2); 
    }
  }
  {
    if (is_new_object4 && array4)
    {
      Py_DECREF(array4); 
    }
  }
  {
    if (is_new_object6 && array6)
    {
      Py_DECREF(array6); 
    }
  }
  return NULL;
}


SWIGINTERN PyObject *_wrap_EMDYPhiFloat32___call__(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[6] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "EMDYPhiFloat32___call__", 0, 5, argv))) SWIG_fail;
  --argc;
  if (argc == 4) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            if (argc <= 4) {
              return _wrap_EMDYPhiFloat32___call____SWIG_2(self, argc, argv);
            }
            if (argc <= 5) {
              return _wrap_EMDYPhiFloat32___call____SWIG_2(self, argc, argv);
            }
            return _wrap_EMDYPhiFloat32___call____SWIG_2(self, argc, argv);
          }
        }
      }
    }
  }
  if (argc == 5) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        _v = is_array(argv[1]) || PySequence_Check(argv[1]);
      }
      if (_v) {
        {
          _v = is_array(argv[2]) || PySequence_Check(argv[2]);
        }
        if (_v) {
          {
            _v = is_array(argv[3]) || PySequence_Check(argv[3]);
          }
          if (_v) {
            {
              _v = is_array(argv[4]) || PySequence_Check(argv[4]);
            }
            if (_v) {
              if (argc <= 5) {
                return _wrap_EMDYPhiFloat32___call____SWIG_1(self, argc, argv);
              }
              if (argc <= 6) {
                return _wrap_EMDYPhiFloat32___call____SWIG_1(self, argc, argv);
              }
              return _wrap_EMDYPhiFloat32___call____SWIG_1(self, argc, argv);
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'EMDYPhiFloat32___call__'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >::operator ()(float *,std::ptrdiff_t,float *,std::ptrdiff_t,std::ptrdiff_t,float *,std::ptrdiff_t,float *,std::ptrdiff_t,std::ptrdiff_t)\n"
    "    wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >::operator ()(float *,std::ptrdiff_t,float *,std::ptrdiff_t,float *,std::ptrdiff_t,std::ptrdiff_t)\n");
  return 0;
}


SWIGINTERN PyObject *EMDYPhiFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *EMDYPhiFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_PairwiseEMDFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  double arg1 = (double) 1 ;
  double arg2 = (double) 1 ;
  bool arg3 = (bool) false ;
  int arg4 = (int) -1 ;
  wasserstein::index_type arg5 = (wasserstein::index_type) -10 ;
  unsigned int arg6 = (unsigned int) 1 ;
  bool arg7 = (bool) false ;
  bool arg8 = (bool) true ;
  bool arg9 = (bool) false ;
  unsigned int arg10 = (unsigned int) 10 ;
  std::size_t arg11 = (std::size_t) 100000 ;
  double arg12 = (double) 1000 ;
  double arg13 = (double) 1 ;
  std::ostream &arg14_defvalue = std::cout ;
  std::ostream *arg14 = (std::ostream *) &arg14_defvalue ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  ptrdiff_t val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  bool val8 ;
  int ecode8 = 0 ;
  bool val9 ;
  int ecode9 = 0 ;
  unsigned int val10 ;
  int ecode10 = 0 ;
  size_t val11 ;
  int ecode11 = 0 ;
  double val12 ;
  int ecode12 = 0 ;
  double val13 ;
  int ecode13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"num_threads",  (char *)"print_every",  (char *)"verbose",  (char *)"request_mode",  (char *)"store_sym_emds_raw",  (char *)"throw_on_error",  (char *)"omp_dynamic_chunksize",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  (char *)"os",  NULL 
  };
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOOOOOOOO:new_PairwiseEMDFloat64", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9, &obj10, &obj11, &obj12, &obj13)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_double(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_PairwiseEMDFloat64" "', argument " "1"" of type '" "double""'");
    } 
    arg1 = static_cast< double >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_double(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_PairwiseEMDFloat64" "', argument " "2"" of type '" "double""'");
    } 
    arg2 = static_cast< double >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_PairwiseEMDFloat64" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_int(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_PairwiseEMDFloat64" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = static_cast< int >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_ptrdiff_t(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_PairwiseEMDFloat64" "', argument " "5"" of type '" "wasserstein::index_type""'");
    } 
    arg5 = static_cast< wasserstein::index_type >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_PairwiseEMDFloat64" "', argument " "6"" of type '" "unsigned int""'");
    } 
    arg6 = static_cast< unsigned int >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_bool(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_PairwiseEMDFloat64" "', argument " "7"" of type '" "bool""'");
    } 
    arg7 = static_cast< bool >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_bool(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_PairwiseEMDFloat64" "', argument " "8"" of type '" "bool""'");
    } 
    arg8 = static_cast< bool >(val8);
  }
  if (obj8) {
    ecode9 = SWIG_AsVal_bool(obj8, &val9);
    if (!SWIG_IsOK(ecode9)) {
      SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "new_PairwiseEMDFloat64" "', argument " "9"" of type '" "bool""'");
    } 
    arg9 = static_cast< bool >(val9);
  }
  if (obj9) {
    ecode10 = SWIG_AsVal_unsigned_SS_int(obj9, &val10);
    if (!SWIG_IsOK(ecode10)) {
      SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "new_PairwiseEMDFloat64" "', argument " "10"" of type '" "unsigned int""'");
    } 
    arg10 = static_cast< unsigned int >(val10);
  }
  if (obj10) {
    ecode11 = SWIG_AsVal_size_t(obj10, &val11);
    if (!SWIG_IsOK(ecode11)) {
      SWIG_exception_fail(SWIG_ArgError(ecode11), "in method '" "new_PairwiseEMDFloat64" "', argument " "11"" of type '" "std::size_t""'");
    } 
    arg11 = static_cast< std::size_t >(val11);
  }
  if (obj11) {
    ecode12 = SWIG_AsVal_double(obj11, &val12);
    if (!SWIG_IsOK(ecode12)) {
      SWIG_exception_fail(SWIG_ArgError(ecode12), "in method '" "new_PairwiseEMDFloat64" "', argument " "12"" of type '" "double""'");
    } 
    arg12 = static_cast< double >(val12);
  }
  if (obj12) {
    ecode13 = SWIG_AsVal_double(obj12, &val13);
    if (!SWIG_IsOK(ecode13)) {
      SWIG_exception_fail(SWIG_ArgError(ecode13), "in method '" "new_PairwiseEMDFloat64" "', argument " "13"" of type '" "double""'");
    } 
    arg13 = static_cast< double >(val13);
  }
  if (obj13) {
    res14 = SWIG_ConvertPtr(obj13, &argp14, SWIGTYPE_p_std__ostream,  0 );
    if (!SWIG_IsOK(res14)) {
      SWIG_exception_fail(SWIG_ArgError(res14), "in method '" "new_PairwiseEMDFloat64" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    if (!argp14) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PairwiseEMDFloat64" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    arg14 = reinterpret_cast< std::ostream * >(argp14);
  }
  {
    try {
      result = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *)new wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,*arg14); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PairwiseEMDFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseEMDFloat64" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64_description" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"free_memory",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:PairwiseEMDFloat64_clear", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64_clear" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDFloat64_clear" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      (arg1)->clear(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_init__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  wasserstein::index_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDFloat64_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  {
    try {
      (arg1)->init(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_init__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDFloat64_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDFloat64_init" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  {
    try {
      (arg1)->init(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_init(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PairwiseEMDFloat64_init", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PairwiseEMDFloat64_init__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_PairwiseEMDFloat64_init__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PairwiseEMDFloat64_init'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double >::init(wasserstein::index_type)\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double >::init(wasserstein::index_type,wasserstein::index_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_compute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64_compute" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    try {
      {
        SWIG_PYTHON_THREAD_BEGIN_ALLOW;
        (arg1)->compute();
        SWIG_PYTHON_THREAD_END_ALLOW;
      } 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64___repr__" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    try {
      result = wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg____repr__((wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64__reset_B_events(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64__reset_B_events" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg___reset_B_events(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat64__add_event(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  double arg7 = (double) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int i2 = 1 ;
  PyArrayObject *array4 = NULL ;
  double val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"weights",  (char *)"coords",  (char *)"event_weight",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO|O:PairwiseEMDFloat64__add_event", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat64__add_event" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > * >(argp1);
  {
    array2 = obj_to_array_no_conversion(obj1, NPY_DOUBLE);
    if (!array2 || !require_dimensions(array2,1) || !require_contiguous(array2)
      || !require_native(array2)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = 1;
    for (i2=0; i2 < array_numdims(array2); ++i2) arg3 *= array_size(array2,i2);
  }
  {
    array4 = obj_to_array_no_conversion(obj2, NPY_DOUBLE);
    if (!array4 || !require_dimensions(array4,2) || !require_contiguous(array4)
      || !require_native(array4)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  if (obj3) {
    ecode7 = SWIG_AsVal_double(obj3, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "PairwiseEMDFloat64__add_event" "', argument " "7"" of type '" "double""'");
    } 
    arg7 = static_cast< double >(val7);
  }
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_double_Sg___add_event(arg1,arg2,arg3,arg4,arg5,arg6,arg7); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PairwiseEMDFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PairwiseEMDFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_PairwiseEMDFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  float arg1 = (float) 1 ;
  float arg2 = (float) 1 ;
  bool arg3 = (bool) false ;
  int arg4 = (int) -1 ;
  wasserstein::index_type arg5 = (wasserstein::index_type) -10 ;
  unsigned int arg6 = (unsigned int) 1 ;
  bool arg7 = (bool) false ;
  bool arg8 = (bool) true ;
  bool arg9 = (bool) false ;
  unsigned int arg10 = (unsigned int) 10 ;
  std::size_t arg11 = (std::size_t) 100000 ;
  float arg12 = (float) 1000 ;
  float arg13 = (float) 1 ;
  std::ostream &arg14_defvalue = std::cout ;
  std::ostream *arg14 = (std::ostream *) &arg14_defvalue ;
  float val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  ptrdiff_t val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  bool val8 ;
  int ecode8 = 0 ;
  bool val9 ;
  int ecode9 = 0 ;
  unsigned int val10 ;
  int ecode10 = 0 ;
  size_t val11 ;
  int ecode11 = 0 ;
  float val12 ;
  int ecode12 = 0 ;
  float val13 ;
  int ecode13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"num_threads",  (char *)"print_every",  (char *)"verbose",  (char *)"request_mode",  (char *)"store_sym_emds_raw",  (char *)"throw_on_error",  (char *)"omp_dynamic_chunksize",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  (char *)"os",  NULL 
  };
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOOOOOOOO:new_PairwiseEMDFloat32", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9, &obj10, &obj11, &obj12, &obj13)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_float(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_PairwiseEMDFloat32" "', argument " "1"" of type '" "float""'");
    } 
    arg1 = static_cast< float >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_float(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_PairwiseEMDFloat32" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = static_cast< float >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_PairwiseEMDFloat32" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_int(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_PairwiseEMDFloat32" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = static_cast< int >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_ptrdiff_t(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_PairwiseEMDFloat32" "', argument " "5"" of type '" "wasserstein::index_type""'");
    } 
    arg5 = static_cast< wasserstein::index_type >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_PairwiseEMDFloat32" "', argument " "6"" of type '" "unsigned int""'");
    } 
    arg6 = static_cast< unsigned int >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_bool(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_PairwiseEMDFloat32" "', argument " "7"" of type '" "bool""'");
    } 
    arg7 = static_cast< bool >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_bool(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_PairwiseEMDFloat32" "', argument " "8"" of type '" "bool""'");
    } 
    arg8 = static_cast< bool >(val8);
  }
  if (obj8) {
    ecode9 = SWIG_AsVal_bool(obj8, &val9);
    if (!SWIG_IsOK(ecode9)) {
      SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "new_PairwiseEMDFloat32" "', argument " "9"" of type '" "bool""'");
    } 
    arg9 = static_cast< bool >(val9);
  }
  if (obj9) {
    ecode10 = SWIG_AsVal_unsigned_SS_int(obj9, &val10);
    if (!SWIG_IsOK(ecode10)) {
      SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "new_PairwiseEMDFloat32" "', argument " "10"" of type '" "unsigned int""'");
    } 
    arg10 = static_cast< unsigned int >(val10);
  }
  if (obj10) {
    ecode11 = SWIG_AsVal_size_t(obj10, &val11);
    if (!SWIG_IsOK(ecode11)) {
      SWIG_exception_fail(SWIG_ArgError(ecode11), "in method '" "new_PairwiseEMDFloat32" "', argument " "11"" of type '" "std::size_t""'");
    } 
    arg11 = static_cast< std::size_t >(val11);
  }
  if (obj11) {
    ecode12 = SWIG_AsVal_float(obj11, &val12);
    if (!SWIG_IsOK(ecode12)) {
      SWIG_exception_fail(SWIG_ArgError(ecode12), "in method '" "new_PairwiseEMDFloat32" "', argument " "12"" of type '" "float""'");
    } 
    arg12 = static_cast< float >(val12);
  }
  if (obj12) {
    ecode13 = SWIG_AsVal_float(obj12, &val13);
    if (!SWIG_IsOK(ecode13)) {
      SWIG_exception_fail(SWIG_ArgError(ecode13), "in method '" "new_PairwiseEMDFloat32" "', argument " "13"" of type '" "float""'");
    } 
    arg13 = static_cast< float >(val13);
  }
  if (obj13) {
    res14 = SWIG_ConvertPtr(obj13, &argp14, SWIGTYPE_p_std__ostream,  0 );
    if (!SWIG_IsOK(res14)) {
      SWIG_exception_fail(SWIG_ArgError(res14), "in method '" "new_PairwiseEMDFloat32" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    if (!argp14) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PairwiseEMDFloat32" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    arg14 = reinterpret_cast< std::ostream * >(argp14);
  }
  {
    try {
      result = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *)new wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,*arg14); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PairwiseEMDFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseEMDFloat32" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32_description" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"free_memory",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:PairwiseEMDFloat32_clear", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32_clear" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDFloat32_clear" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      (arg1)->clear(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_init__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  wasserstein::index_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDFloat32_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  {
    try {
      (arg1)->init(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_init__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDFloat32_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDFloat32_init" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  {
    try {
      (arg1)->init(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_init(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PairwiseEMDFloat32_init", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PairwiseEMDFloat32_init__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_PairwiseEMDFloat32_init__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PairwiseEMDFloat32_init'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float >::init(wasserstein::index_type)\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float >::init(wasserstein::index_type,wasserstein::index_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_compute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32_compute" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    try {
      {
        SWIG_PYTHON_THREAD_BEGIN_ALLOW;
        (arg1)->compute();
        SWIG_PYTHON_THREAD_END_ALLOW;
      } 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32___repr__" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    try {
      result = wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg____repr__((wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32__reset_B_events(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32__reset_B_events" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg___reset_B_events(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDFloat32__add_event(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  float arg7 = (float) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int i2 = 1 ;
  PyArrayObject *array4 = NULL ;
  float val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"weights",  (char *)"coords",  (char *)"event_weight",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO|O:PairwiseEMDFloat32__add_event", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDFloat32__add_event" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > * >(argp1);
  {
    array2 = obj_to_array_no_conversion(obj1, NPY_FLOAT);
    if (!array2 || !require_dimensions(array2,1) || !require_contiguous(array2)
      || !require_native(array2)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = 1;
    for (i2=0; i2 < array_numdims(array2); ++i2) arg3 *= array_size(array2,i2);
  }
  {
    array4 = obj_to_array_no_conversion(obj2, NPY_FLOAT);
    if (!array4 || !require_dimensions(array4,2) || !require_contiguous(array4)
      || !require_native(array4)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  if (obj3) {
    ecode7 = SWIG_AsVal_float(obj3, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "PairwiseEMDFloat32__add_event" "', argument " "7"" of type '" "float""'");
    } 
    arg7 = static_cast< float >(val7);
  }
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArrayEvent_Sc_wasserstein_EuclideanArrayDistance_Sg__Sc_float_Sg___add_event(arg1,arg2,arg3,arg4,arg5,arg6,arg7); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PairwiseEMDFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PairwiseEMDFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_PairwiseEMDYPhiFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  double arg1 = (double) 1 ;
  double arg2 = (double) 1 ;
  bool arg3 = (bool) false ;
  int arg4 = (int) -1 ;
  wasserstein::index_type arg5 = (wasserstein::index_type) -10 ;
  unsigned int arg6 = (unsigned int) 1 ;
  bool arg7 = (bool) false ;
  bool arg8 = (bool) true ;
  bool arg9 = (bool) false ;
  unsigned int arg10 = (unsigned int) 10 ;
  std::size_t arg11 = (std::size_t) 100000 ;
  double arg12 = (double) 1000 ;
  double arg13 = (double) 1 ;
  std::ostream &arg14_defvalue = std::cout ;
  std::ostream *arg14 = (std::ostream *) &arg14_defvalue ;
  double val1 ;
  int ecode1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  ptrdiff_t val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  bool val8 ;
  int ecode8 = 0 ;
  bool val9 ;
  int ecode9 = 0 ;
  unsigned int val10 ;
  int ecode10 = 0 ;
  size_t val11 ;
  int ecode11 = 0 ;
  double val12 ;
  int ecode12 = 0 ;
  double val13 ;
  int ecode13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"num_threads",  (char *)"print_every",  (char *)"verbose",  (char *)"request_mode",  (char *)"store_sym_emds_raw",  (char *)"throw_on_error",  (char *)"omp_dynamic_chunksize",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  (char *)"os",  NULL 
  };
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOOOOOOOO:new_PairwiseEMDYPhiFloat64", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9, &obj10, &obj11, &obj12, &obj13)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_double(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "1"" of type '" "double""'");
    } 
    arg1 = static_cast< double >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_double(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "2"" of type '" "double""'");
    } 
    arg2 = static_cast< double >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_int(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = static_cast< int >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_ptrdiff_t(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "5"" of type '" "wasserstein::index_type""'");
    } 
    arg5 = static_cast< wasserstein::index_type >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "6"" of type '" "unsigned int""'");
    } 
    arg6 = static_cast< unsigned int >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_bool(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "7"" of type '" "bool""'");
    } 
    arg7 = static_cast< bool >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_bool(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "8"" of type '" "bool""'");
    } 
    arg8 = static_cast< bool >(val8);
  }
  if (obj8) {
    ecode9 = SWIG_AsVal_bool(obj8, &val9);
    if (!SWIG_IsOK(ecode9)) {
      SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "9"" of type '" "bool""'");
    } 
    arg9 = static_cast< bool >(val9);
  }
  if (obj9) {
    ecode10 = SWIG_AsVal_unsigned_SS_int(obj9, &val10);
    if (!SWIG_IsOK(ecode10)) {
      SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "10"" of type '" "unsigned int""'");
    } 
    arg10 = static_cast< unsigned int >(val10);
  }
  if (obj10) {
    ecode11 = SWIG_AsVal_size_t(obj10, &val11);
    if (!SWIG_IsOK(ecode11)) {
      SWIG_exception_fail(SWIG_ArgError(ecode11), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "11"" of type '" "std::size_t""'");
    } 
    arg11 = static_cast< std::size_t >(val11);
  }
  if (obj11) {
    ecode12 = SWIG_AsVal_double(obj11, &val12);
    if (!SWIG_IsOK(ecode12)) {
      SWIG_exception_fail(SWIG_ArgError(ecode12), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "12"" of type '" "double""'");
    } 
    arg12 = static_cast< double >(val12);
  }
  if (obj12) {
    ecode13 = SWIG_AsVal_double(obj12, &val13);
    if (!SWIG_IsOK(ecode13)) {
      SWIG_exception_fail(SWIG_ArgError(ecode13), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "13"" of type '" "double""'");
    } 
    arg13 = static_cast< double >(val13);
  }
  if (obj13) {
    res14 = SWIG_ConvertPtr(obj13, &argp14, SWIGTYPE_p_std__ostream,  0 );
    if (!SWIG_IsOK(res14)) {
      SWIG_exception_fail(SWIG_ArgError(res14), "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    if (!argp14) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PairwiseEMDYPhiFloat64" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    arg14 = reinterpret_cast< std::ostream * >(argp14);
  }
  {
    try {
      result = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *)new wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,*arg14); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PairwiseEMDYPhiFloat64(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseEMDYPhiFloat64" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64_description" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"free_memory",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:PairwiseEMDYPhiFloat64_clear", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64_clear" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDYPhiFloat64_clear" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      (arg1)->clear(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_init__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  wasserstein::index_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDYPhiFloat64_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  {
    try {
      (arg1)->init(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_init__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDYPhiFloat64_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDYPhiFloat64_init" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  {
    try {
      (arg1)->init(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_init(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PairwiseEMDYPhiFloat64_init", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PairwiseEMDYPhiFloat64_init__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_PairwiseEMDYPhiFloat64_init__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PairwiseEMDYPhiFloat64_init'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double >::init(wasserstein::index_type)\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double >::init(wasserstein::index_type,wasserstein::index_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_compute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64_compute" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    try {
      {
        SWIG_PYTHON_THREAD_BEGIN_ALLOW;
        (arg1)->compute();
        SWIG_PYTHON_THREAD_END_ALLOW;
      } 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64___repr__" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    try {
      result = wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg____repr__((wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64__reset_B_events(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64__reset_B_events" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg___reset_B_events(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat64__add_event(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *) 0 ;
  double *arg2 = (double *) 0 ;
  std::ptrdiff_t arg3 ;
  double *arg4 = (double *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  double arg7 = (double) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int i2 = 1 ;
  PyArrayObject *array4 = NULL ;
  double val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"weights",  (char *)"coords",  (char *)"event_weight",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO|O:PairwiseEMDYPhiFloat64__add_event", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat64__add_event" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > * >(argp1);
  {
    array2 = obj_to_array_no_conversion(obj1, NPY_DOUBLE);
    if (!array2 || !require_dimensions(array2,1) || !require_contiguous(array2)
      || !require_native(array2)) SWIG_fail;
    arg2 = (double*) array_data(array2);
    arg3 = 1;
    for (i2=0; i2 < array_numdims(array2); ++i2) arg3 *= array_size(array2,i2);
  }
  {
    array4 = obj_to_array_no_conversion(obj2, NPY_DOUBLE);
    if (!array4 || !require_dimensions(array4,2) || !require_contiguous(array4)
      || !require_native(array4)) SWIG_fail;
    arg4 = (double*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  if (obj3) {
    ecode7 = SWIG_AsVal_double(obj3, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "PairwiseEMDYPhiFloat64__add_event" "', argument " "7"" of type '" "double""'");
    } 
    arg7 = static_cast< double >(val7);
  }
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_double_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_double_Sg___add_event(arg1,arg2,arg3,arg4,arg5,arg6,arg7); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PairwiseEMDYPhiFloat64_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PairwiseEMDYPhiFloat64_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

SWIGINTERN PyObject *_wrap_new_PairwiseEMDYPhiFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  float arg1 = (float) 1 ;
  float arg2 = (float) 1 ;
  bool arg3 = (bool) false ;
  int arg4 = (int) -1 ;
  wasserstein::index_type arg5 = (wasserstein::index_type) -10 ;
  unsigned int arg6 = (unsigned int) 1 ;
  bool arg7 = (bool) false ;
  bool arg8 = (bool) true ;
  bool arg9 = (bool) false ;
  unsigned int arg10 = (unsigned int) 10 ;
  std::size_t arg11 = (std::size_t) 100000 ;
  float arg12 = (float) 1000 ;
  float arg13 = (float) 1 ;
  std::ostream &arg14_defvalue = std::cout ;
  std::ostream *arg14 = (std::ostream *) &arg14_defvalue ;
  float val1 ;
  int ecode1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  ptrdiff_t val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  bool val7 ;
  int ecode7 = 0 ;
  bool val8 ;
  int ecode8 = 0 ;
  bool val9 ;
  int ecode9 = 0 ;
  unsigned int val10 ;
  int ecode10 = 0 ;
  size_t val11 ;
  int ecode11 = 0 ;
  float val12 ;
  int ecode12 = 0 ;
  float val13 ;
  int ecode13 = 0 ;
  void *argp14 = 0 ;
  int res14 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  PyObject * obj10 = 0 ;
  PyObject * obj11 = 0 ;
  PyObject * obj12 = 0 ;
  PyObject * obj13 = 0 ;
  char * kwnames[] = {
    (char *)"R",  (char *)"beta",  (char *)"norm",  (char *)"num_threads",  (char *)"print_every",  (char *)"verbose",  (char *)"request_mode",  (char *)"store_sym_emds_raw",  (char *)"throw_on_error",  (char *)"omp_dynamic_chunksize",  (char *)"n_iter_max",  (char *)"epsilon_large_factor",  (char *)"epsilon_small_factor",  (char *)"os",  NULL 
  };
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *result = 0 ;
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|OOOOOOOOOOOOOO:new_PairwiseEMDYPhiFloat32", kwnames, &obj0, &obj1, &obj2, &obj3, &obj4, &obj5, &obj6, &obj7, &obj8, &obj9, &obj10, &obj11, &obj12, &obj13)) SWIG_fail;
  if (obj0) {
    ecode1 = SWIG_AsVal_float(obj0, &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "1"" of type '" "float""'");
    } 
    arg1 = static_cast< float >(val1);
  }
  if (obj1) {
    ecode2 = SWIG_AsVal_float(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = static_cast< float >(val2);
  }
  if (obj2) {
    ecode3 = SWIG_AsVal_bool(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "3"" of type '" "bool""'");
    } 
    arg3 = static_cast< bool >(val3);
  }
  if (obj3) {
    ecode4 = SWIG_AsVal_int(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = static_cast< int >(val4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_ptrdiff_t(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "5"" of type '" "wasserstein::index_type""'");
    } 
    arg5 = static_cast< wasserstein::index_type >(val5);
  }
  if (obj5) {
    ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
    if (!SWIG_IsOK(ecode6)) {
      SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "6"" of type '" "unsigned int""'");
    } 
    arg6 = static_cast< unsigned int >(val6);
  }
  if (obj6) {
    ecode7 = SWIG_AsVal_bool(obj6, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "7"" of type '" "bool""'");
    } 
    arg7 = static_cast< bool >(val7);
  }
  if (obj7) {
    ecode8 = SWIG_AsVal_bool(obj7, &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "8"" of type '" "bool""'");
    } 
    arg8 = static_cast< bool >(val8);
  }
  if (obj8) {
    ecode9 = SWIG_AsVal_bool(obj8, &val9);
    if (!SWIG_IsOK(ecode9)) {
      SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "9"" of type '" "bool""'");
    } 
    arg9 = static_cast< bool >(val9);
  }
  if (obj9) {
    ecode10 = SWIG_AsVal_unsigned_SS_int(obj9, &val10);
    if (!SWIG_IsOK(ecode10)) {
      SWIG_exception_fail(SWIG_ArgError(ecode10), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "10"" of type '" "unsigned int""'");
    } 
    arg10 = static_cast< unsigned int >(val10);
  }
  if (obj10) {
    ecode11 = SWIG_AsVal_size_t(obj10, &val11);
    if (!SWIG_IsOK(ecode11)) {
      SWIG_exception_fail(SWIG_ArgError(ecode11), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "11"" of type '" "std::size_t""'");
    } 
    arg11 = static_cast< std::size_t >(val11);
  }
  if (obj11) {
    ecode12 = SWIG_AsVal_float(obj11, &val12);
    if (!SWIG_IsOK(ecode12)) {
      SWIG_exception_fail(SWIG_ArgError(ecode12), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "12"" of type '" "float""'");
    } 
    arg12 = static_cast< float >(val12);
  }
  if (obj12) {
    ecode13 = SWIG_AsVal_float(obj12, &val13);
    if (!SWIG_IsOK(ecode13)) {
      SWIG_exception_fail(SWIG_ArgError(ecode13), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "13"" of type '" "float""'");
    } 
    arg13 = static_cast< float >(val13);
  }
  if (obj13) {
    res14 = SWIG_ConvertPtr(obj13, &argp14, SWIGTYPE_p_std__ostream,  0 );
    if (!SWIG_IsOK(res14)) {
      SWIG_exception_fail(SWIG_ArgError(res14), "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    if (!argp14) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_PairwiseEMDYPhiFloat32" "', argument " "14"" of type '" "std::ostream &""'"); 
    }
    arg14 = reinterpret_cast< std::ostream * >(argp14);
  }
  {
    try {
      result = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *)new wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float >(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,*arg14); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_PairwiseEMDYPhiFloat32(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PairwiseEMDYPhiFloat32" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    try {
      delete arg1; 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_description(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32_description" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    try {
      result = ((wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > const *)arg1)->description(); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  bool arg2 = (bool) true ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"free_memory",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|O:PairwiseEMDYPhiFloat32_clear", kwnames, &obj0, &obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32_clear" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  if (obj1) {
    ecode2 = SWIG_AsVal_bool(obj1, &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDYPhiFloat32_clear" "', argument " "2"" of type '" "bool""'");
    } 
    arg2 = static_cast< bool >(val2);
  }
  {
    try {
      (arg1)->clear(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_init__SWIG_0(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  wasserstein::index_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  
  if ((nobjs < 2) || (nobjs > 2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDYPhiFloat32_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  {
    try {
      (arg1)->init(arg2); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_init__SWIG_1(PyObject *SWIGUNUSEDPARM(self), Py_ssize_t nobjs, PyObject **swig_obj) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  wasserstein::index_type arg2 ;
  wasserstein::index_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  
  if ((nobjs < 3) || (nobjs > 3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32_init" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(swig_obj[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PairwiseEMDYPhiFloat32_init" "', argument " "2"" of type '" "wasserstein::index_type""'");
  } 
  arg2 = static_cast< wasserstein::index_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(swig_obj[2], &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "PairwiseEMDYPhiFloat32_init" "', argument " "3"" of type '" "wasserstein::index_type""'");
  } 
  arg3 = static_cast< wasserstein::index_type >(val3);
  {
    try {
      (arg1)->init(arg2,arg3); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_init(PyObject *self, PyObject *args) {
  Py_ssize_t argc;
  PyObject *argv[4] = {
    0
  };
  
  if (!(argc = SWIG_Python_UnpackTuple(args, "PairwiseEMDYPhiFloat32_init", 0, 3, argv))) SWIG_fail;
  --argc;
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PairwiseEMDYPhiFloat32_init__SWIG_0(self, argc, argv);
      }
    }
  }
  if (argc == 3) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_ptrdiff_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_PairwiseEMDYPhiFloat32_init__SWIG_1(self, argc, argv);
        }
      }
    }
  }
  
fail:
  SWIG_Python_RaiseOrModifyTypeError("Wrong number or type of arguments for overloaded function 'PairwiseEMDYPhiFloat32_init'.\n"
    "  Possible C/C++ prototypes are:\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float >::init(wasserstein::index_type)\n"
    "    wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float >::init(wasserstein::index_type,wasserstein::index_type)\n");
  return 0;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_compute(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32_compute" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    try {
      {
        SWIG_PYTHON_THREAD_BEGIN_ALLOW;
        (arg1)->compute();
        SWIG_PYTHON_THREAD_END_ALLOW;
      } 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32___repr__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  std::string result;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32___repr__" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > const *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    try {
      result = wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg____repr__((wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > const *)arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg__preprocess_CenterWeightedCentroid(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32__reset_B_events(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject *swig_obj[1] ;
  
  if (!args) SWIG_fail;
  swig_obj[0] = args;
  res1 = SWIG_ConvertPtr(swig_obj[0], &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32__reset_B_events" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg___reset_B_events(arg1); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PairwiseEMDYPhiFloat32__add_event(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *arg1 = (wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *) 0 ;
  float *arg2 = (float *) 0 ;
  std::ptrdiff_t arg3 ;
  float *arg4 = (float *) 0 ;
  std::ptrdiff_t arg5 ;
  std::ptrdiff_t arg6 ;
  float arg7 = (float) 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyArrayObject *array2 = NULL ;
  int i2 = 1 ;
  PyArrayObject *array4 = NULL ;
  float val7 ;
  int ecode7 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char * kwnames[] = {
    (char *)"self",  (char *)"weights",  (char *)"coords",  (char *)"event_weight",  NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOO|O:PairwiseEMDYPhiFloat32__add_event", kwnames, &obj0, &obj1, &obj2, &obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PairwiseEMDYPhiFloat32__add_event" "', argument " "1"" of type '" "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *""'"); 
  }
  arg1 = reinterpret_cast< wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > * >(argp1);
  {
    array2 = obj_to_array_no_conversion(obj1, NPY_FLOAT);
    if (!array2 || !require_dimensions(array2,1) || !require_contiguous(array2)
      || !require_native(array2)) SWIG_fail;
    arg2 = (float*) array_data(array2);
    arg3 = 1;
    for (i2=0; i2 < array_numdims(array2); ++i2) arg3 *= array_size(array2,i2);
  }
  {
    array4 = obj_to_array_no_conversion(obj2, NPY_FLOAT);
    if (!array4 || !require_dimensions(array4,2) || !require_contiguous(array4)
      || !require_native(array4)) SWIG_fail;
    arg4 = (float*) array_data(array4);
    arg5 = (std::ptrdiff_t) array_size(array4,0);
    arg6 = (std::ptrdiff_t) array_size(array4,1);
  }
  if (obj3) {
    ecode7 = SWIG_AsVal_float(obj3, &val7);
    if (!SWIG_IsOK(ecode7)) {
      SWIG_exception_fail(SWIG_ArgError(ecode7), "in method '" "PairwiseEMDYPhiFloat32__add_event" "', argument " "7"" of type '" "float""'");
    } 
    arg7 = static_cast< float >(val7);
  }
  {
    try {
      wasserstein_PairwiseEMD_Sl_wasserstein_EMD_Sl_float_Sc_wasserstein_DefaultArray2Event_Sc_wasserstein_YPhiArrayDistance_Sg__Sc_float_Sg___add_event(arg1,arg2,arg3,arg4,arg5,arg6,arg7); 
    }
    /*@SWIG:/usr/local/Cellar/swig/4.0.2/share/swig/4.0.2/typemaps/exception.swg,58,SWIG_CATCH_STDEXCEPT@*/  /* catching std::exception  */
    catch (std::invalid_argument& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::domain_error& e) {
      SWIG_exception_fail(SWIG_ValueError, e.what() );
    } catch (std::overflow_error& e) {
      SWIG_exception_fail(SWIG_OverflowError, e.what() );
    } catch (std::out_of_range& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::length_error& e) {
      SWIG_exception_fail(SWIG_IndexError, e.what() );
    } catch (std::runtime_error& e) {
      SWIG_exception_fail(SWIG_RuntimeError, e.what() );
    } catch (std::exception& e) {
      SWIG_exception_fail(SWIG_SystemError, e.what() );
    }
    /*@SWIG@*/
    catch (...) {
      SWIG_exception_fail(SWIG_UnknownError, "unknown exception");
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *PairwiseEMDYPhiFloat32_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!SWIG_Python_UnpackTuple(args, "swigregister", 1, 1, &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *PairwiseEMDYPhiFloat32_swiginit(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  return SWIG_Python_InitShadowInstance(args);
}

static PyMethodDef SwigMethods[] = {
	 { "SWIG_PyInstanceMethod_New", SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { "SWIG_PyStaticMethod_New", SWIG_PyStaticMethod_New, METH_O, NULL},
	 { "delete_SwigPyIterator", _wrap_delete_SwigPyIterator, METH_O, "delete_SwigPyIterator(SwigPyIterator self)"},
	 { "SwigPyIterator_value", _wrap_SwigPyIterator_value, METH_O, "SwigPyIterator_value(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator_incr", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_incr, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator_incr(SwigPyIterator self, size_t n=1) -> SwigPyIterator"},
	 { "SwigPyIterator_decr", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_decr, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator_decr(SwigPyIterator self, size_t n=1) -> SwigPyIterator"},
	 { "SwigPyIterator_distance", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_distance, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator_distance(SwigPyIterator self, SwigPyIterator x) -> ptrdiff_t"},
	 { "SwigPyIterator_equal", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_equal, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator_equal(SwigPyIterator self, SwigPyIterator x) -> bool"},
	 { "SwigPyIterator_copy", _wrap_SwigPyIterator_copy, METH_O, "SwigPyIterator_copy(SwigPyIterator self) -> SwigPyIterator"},
	 { "SwigPyIterator_next", _wrap_SwigPyIterator_next, METH_O, "SwigPyIterator_next(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator___next__", _wrap_SwigPyIterator___next__, METH_O, "SwigPyIterator___next__(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator_previous", _wrap_SwigPyIterator_previous, METH_O, "SwigPyIterator_previous(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator_advance", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_advance, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator_advance(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___eq__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___eq__, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator___eq__(SwigPyIterator self, SwigPyIterator x) -> bool"},
	 { "SwigPyIterator___ne__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___ne__, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator___ne__(SwigPyIterator self, SwigPyIterator x) -> bool"},
	 { "SwigPyIterator___iadd__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___iadd__, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator___iadd__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___isub__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___isub__, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator___isub__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___add__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___add__, METH_VARARGS|METH_KEYWORDS, "SwigPyIterator___add__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___sub__", _wrap_SwigPyIterator___sub__, METH_VARARGS, "\n"
		"SwigPyIterator___sub__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator\n"
		"SwigPyIterator___sub__(SwigPyIterator self, SwigPyIterator x) -> ptrdiff_t\n"
		""},
	 { "SwigPyIterator_swigregister", SwigPyIterator_swigregister, METH_O, NULL},
	 { "vectorDouble_iterator", _wrap_vectorDouble_iterator, METH_O, "vectorDouble_iterator(vectorDouble self) -> SwigPyIterator"},
	 { "vectorDouble___nonzero__", _wrap_vectorDouble___nonzero__, METH_O, "vectorDouble___nonzero__(vectorDouble self) -> bool"},
	 { "vectorDouble___bool__", _wrap_vectorDouble___bool__, METH_O, "vectorDouble___bool__(vectorDouble self) -> bool"},
	 { "vectorDouble___len__", _wrap_vectorDouble___len__, METH_O, "vectorDouble___len__(vectorDouble self) -> std::vector< double >::size_type"},
	 { "vectorDouble___getslice__", (PyCFunction)(void(*)(void))_wrap_vectorDouble___getslice__, METH_VARARGS|METH_KEYWORDS, "vectorDouble___getslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j) -> vectorDouble"},
	 { "vectorDouble___setslice__", _wrap_vectorDouble___setslice__, METH_VARARGS, "\n"
		"vectorDouble___setslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j)\n"
		"vectorDouble___setslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j, vectorDouble v)\n"
		""},
	 { "vectorDouble___delslice__", (PyCFunction)(void(*)(void))_wrap_vectorDouble___delslice__, METH_VARARGS|METH_KEYWORDS, "vectorDouble___delslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j)"},
	 { "vectorDouble___delitem__", _wrap_vectorDouble___delitem__, METH_VARARGS, "\n"
		"vectorDouble___delitem__(vectorDouble self, std::vector< double >::difference_type i)\n"
		"vectorDouble___delitem__(vectorDouble self, PySliceObject * slice)\n"
		""},
	 { "vectorDouble___getitem__", _wrap_vectorDouble___getitem__, METH_VARARGS, "\n"
		"vectorDouble___getitem__(vectorDouble self, PySliceObject * slice) -> vectorDouble\n"
		"vectorDouble___getitem__(vectorDouble self, std::vector< double >::difference_type i) -> std::vector< double >::value_type const &\n"
		""},
	 { "vectorDouble___setitem__", _wrap_vectorDouble___setitem__, METH_VARARGS, "\n"
		"vectorDouble___setitem__(vectorDouble self, PySliceObject * slice, vectorDouble v)\n"
		"vectorDouble___setitem__(vectorDouble self, PySliceObject * slice)\n"
		"vectorDouble___setitem__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::value_type const & x)\n"
		""},
	 { "vectorDouble_pop", _wrap_vectorDouble_pop, METH_O, "vectorDouble_pop(vectorDouble self) -> std::vector< double >::value_type"},
	 { "vectorDouble_append", (PyCFunction)(void(*)(void))_wrap_vectorDouble_append, METH_VARARGS|METH_KEYWORDS, "vectorDouble_append(vectorDouble self, std::vector< double >::value_type const & x)"},
	 { "vectorDouble_empty", _wrap_vectorDouble_empty, METH_O, "vectorDouble_empty(vectorDouble self) -> bool"},
	 { "vectorDouble_size", _wrap_vectorDouble_size, METH_O, "vectorDouble_size(vectorDouble self) -> std::vector< double >::size_type"},
	 { "vectorDouble_swap", (PyCFunction)(void(*)(void))_wrap_vectorDouble_swap, METH_VARARGS|METH_KEYWORDS, "vectorDouble_swap(vectorDouble self, vectorDouble v)"},
	 { "vectorDouble_begin", _wrap_vectorDouble_begin, METH_O, "vectorDouble_begin(vectorDouble self) -> std::vector< double >::iterator"},
	 { "vectorDouble_end", _wrap_vectorDouble_end, METH_O, "vectorDouble_end(vectorDouble self) -> std::vector< double >::iterator"},
	 { "vectorDouble_rbegin", _wrap_vectorDouble_rbegin, METH_O, "vectorDouble_rbegin(vectorDouble self) -> std::vector< double >::reverse_iterator"},
	 { "vectorDouble_rend", _wrap_vectorDouble_rend, METH_O, "vectorDouble_rend(vectorDouble self) -> std::vector< double >::reverse_iterator"},
	 { "vectorDouble_clear", _wrap_vectorDouble_clear, METH_O, "vectorDouble_clear(vectorDouble self)"},
	 { "vectorDouble_get_allocator", _wrap_vectorDouble_get_allocator, METH_O, "vectorDouble_get_allocator(vectorDouble self) -> std::vector< double >::allocator_type"},
	 { "vectorDouble_pop_back", _wrap_vectorDouble_pop_back, METH_O, "vectorDouble_pop_back(vectorDouble self)"},
	 { "vectorDouble_erase", _wrap_vectorDouble_erase, METH_VARARGS, "\n"
		"vectorDouble_erase(vectorDouble self, std::vector< double >::iterator pos) -> std::vector< double >::iterator\n"
		"vectorDouble_erase(vectorDouble self, std::vector< double >::iterator first, std::vector< double >::iterator last) -> std::vector< double >::iterator\n"
		""},
	 { "new_vectorDouble", _wrap_new_vectorDouble, METH_VARARGS, "\n"
		"vectorDouble()\n"
		"vectorDouble(vectorDouble other)\n"
		"vectorDouble(std::vector< double >::size_type size)\n"
		"new_vectorDouble(std::vector< double >::size_type size, std::vector< double >::value_type const & value) -> vectorDouble\n"
		""},
	 { "vectorDouble_push_back", (PyCFunction)(void(*)(void))_wrap_vectorDouble_push_back, METH_VARARGS|METH_KEYWORDS, "vectorDouble_push_back(vectorDouble self, std::vector< double >::value_type const & x)"},
	 { "vectorDouble_front", _wrap_vectorDouble_front, METH_O, "vectorDouble_front(vectorDouble self) -> std::vector< double >::value_type const &"},
	 { "vectorDouble_back", _wrap_vectorDouble_back, METH_O, "vectorDouble_back(vectorDouble self) -> std::vector< double >::value_type const &"},
	 { "vectorDouble_assign", (PyCFunction)(void(*)(void))_wrap_vectorDouble_assign, METH_VARARGS|METH_KEYWORDS, "vectorDouble_assign(vectorDouble self, std::vector< double >::size_type n, std::vector< double >::value_type const & x)"},
	 { "vectorDouble_resize", _wrap_vectorDouble_resize, METH_VARARGS, "\n"
		"vectorDouble_resize(vectorDouble self, std::vector< double >::size_type new_size)\n"
		"vectorDouble_resize(vectorDouble self, std::vector< double >::size_type new_size, std::vector< double >::value_type const & x)\n"
		""},
	 { "vectorDouble_insert", _wrap_vectorDouble_insert, METH_VARARGS, "\n"
		"vectorDouble_insert(vectorDouble self, std::vector< double >::iterator pos, std::vector< double >::value_type const & x) -> std::vector< double >::iterator\n"
		"vectorDouble_insert(vectorDouble self, std::vector< double >::iterator pos, std::vector< double >::size_type n, std::vector< double >::value_type const & x)\n"
		""},
	 { "vectorDouble_reserve", (PyCFunction)(void(*)(void))_wrap_vectorDouble_reserve, METH_VARARGS|METH_KEYWORDS, "vectorDouble_reserve(vectorDouble self, std::vector< double >::size_type n)"},
	 { "vectorDouble_capacity", _wrap_vectorDouble_capacity, METH_O, "vectorDouble_capacity(vectorDouble self) -> std::vector< double >::size_type"},
	 { "delete_vectorDouble", _wrap_delete_vectorDouble, METH_O, "delete_vectorDouble(vectorDouble self)"},
	 { "vectorDouble_swigregister", vectorDouble_swigregister, METH_O, NULL},
	 { "vectorDouble_swiginit", vectorDouble_swiginit, METH_VARARGS, NULL},
	 { "vectorString_iterator", _wrap_vectorString_iterator, METH_O, "vectorString_iterator(vectorString self) -> SwigPyIterator"},
	 { "vectorString___nonzero__", _wrap_vectorString___nonzero__, METH_O, "vectorString___nonzero__(vectorString self) -> bool"},
	 { "vectorString___bool__", _wrap_vectorString___bool__, METH_O, "vectorString___bool__(vectorString self) -> bool"},
	 { "vectorString___len__", _wrap_vectorString___len__, METH_O, "vectorString___len__(vectorString self) -> std::vector< std::string >::size_type"},
	 { "vectorString___getslice__", (PyCFunction)(void(*)(void))_wrap_vectorString___getslice__, METH_VARARGS|METH_KEYWORDS, "vectorString___getslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j) -> vectorString"},
	 { "vectorString___setslice__", _wrap_vectorString___setslice__, METH_VARARGS, "\n"
		"vectorString___setslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j)\n"
		"vectorString___setslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j, vectorString v)\n"
		""},
	 { "vectorString___delslice__", (PyCFunction)(void(*)(void))_wrap_vectorString___delslice__, METH_VARARGS|METH_KEYWORDS, "vectorString___delslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j)"},
	 { "vectorString___delitem__", _wrap_vectorString___delitem__, METH_VARARGS, "\n"
		"vectorString___delitem__(vectorString self, std::vector< std::string >::difference_type i)\n"
		"vectorString___delitem__(vectorString self, PySliceObject * slice)\n"
		""},
	 { "vectorString___getitem__", _wrap_vectorString___getitem__, METH_VARARGS, "\n"
		"vectorString___getitem__(vectorString self, PySliceObject * slice) -> vectorString\n"
		"vectorString___getitem__(vectorString self, std::vector< std::string >::difference_type i) -> std::vector< std::string >::value_type const &\n"
		""},
	 { "vectorString___setitem__", _wrap_vectorString___setitem__, METH_VARARGS, "\n"
		"vectorString___setitem__(vectorString self, PySliceObject * slice, vectorString v)\n"
		"vectorString___setitem__(vectorString self, PySliceObject * slice)\n"
		"vectorString___setitem__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::value_type const & x)\n"
		""},
	 { "vectorString_pop", _wrap_vectorString_pop, METH_O, "vectorString_pop(vectorString self) -> std::vector< std::string >::value_type"},
	 { "vectorString_append", (PyCFunction)(void(*)(void))_wrap_vectorString_append, METH_VARARGS|METH_KEYWORDS, "vectorString_append(vectorString self, std::vector< std::string >::value_type const & x)"},
	 { "vectorString_empty", _wrap_vectorString_empty, METH_O, "vectorString_empty(vectorString self) -> bool"},
	 { "vectorString_size", _wrap_vectorString_size, METH_O, "vectorString_size(vectorString self) -> std::vector< std::string >::size_type"},
	 { "vectorString_swap", (PyCFunction)(void(*)(void))_wrap_vectorString_swap, METH_VARARGS|METH_KEYWORDS, "vectorString_swap(vectorString self, vectorString v)"},
	 { "vectorString_begin", _wrap_vectorString_begin, METH_O, "vectorString_begin(vectorString self) -> std::vector< std::string >::iterator"},
	 { "vectorString_end", _wrap_vectorString_end, METH_O, "vectorString_end(vectorString self) -> std::vector< std::string >::iterator"},
	 { "vectorString_rbegin", _wrap_vectorString_rbegin, METH_O, "vectorString_rbegin(vectorString self) -> std::vector< std::string >::reverse_iterator"},
	 { "vectorString_rend", _wrap_vectorString_rend, METH_O, "vectorString_rend(vectorString self) -> std::vector< std::string >::reverse_iterator"},
	 { "vectorString_clear", _wrap_vectorString_clear, METH_O, "vectorString_clear(vectorString self)"},
	 { "vectorString_get_allocator", _wrap_vectorString_get_allocator, METH_O, "vectorString_get_allocator(vectorString self) -> std::vector< std::string >::allocator_type"},
	 { "vectorString_pop_back", _wrap_vectorString_pop_back, METH_O, "vectorString_pop_back(vectorString self)"},
	 { "vectorString_erase", _wrap_vectorString_erase, METH_VARARGS, "\n"
		"vectorString_erase(vectorString self, std::vector< std::string >::iterator pos) -> std::vector< std::string >::iterator\n"
		"vectorString_erase(vectorString self, std::vector< std::string >::iterator first, std::vector< std::string >::iterator last) -> std::vector< std::string >::iterator\n"
		""},
	 { "new_vectorString", _wrap_new_vectorString, METH_VARARGS, "\n"
		"vectorString()\n"
		"vectorString(vectorString other)\n"
		"vectorString(std::vector< std::string >::size_type size)\n"
		"new_vectorString(std::vector< std::string >::size_type size, std::vector< std::string >::value_type const & value) -> vectorString\n"
		""},
	 { "vectorString_push_back", (PyCFunction)(void(*)(void))_wrap_vectorString_push_back, METH_VARARGS|METH_KEYWORDS, "vectorString_push_back(vectorString self, std::vector< std::string >::value_type const & x)"},
	 { "vectorString_front", _wrap_vectorString_front, METH_O, "vectorString_front(vectorString self) -> std::vector< std::string >::value_type const &"},
	 { "vectorString_back", _wrap_vectorString_back, METH_O, "vectorString_back(vectorString self) -> std::vector< std::string >::value_type const &"},
	 { "vectorString_assign", (PyCFunction)(void(*)(void))_wrap_vectorString_assign, METH_VARARGS|METH_KEYWORDS, "vectorString_assign(vectorString self, std::vector< std::string >::size_type n, std::vector< std::string >::value_type const & x)"},
	 { "vectorString_resize", _wrap_vectorString_resize, METH_VARARGS, "\n"
		"vectorString_resize(vectorString self, std::vector< std::string >::size_type new_size)\n"
		"vectorString_resize(vectorString self, std::vector< std::string >::size_type new_size, std::vector< std::string >::value_type const & x)\n"
		""},
	 { "vectorString_insert", _wrap_vectorString_insert, METH_VARARGS, "\n"
		"vectorString_insert(vectorString self, std::vector< std::string >::iterator pos, std::vector< std::string >::value_type const & x) -> std::vector< std::string >::iterator\n"
		"vectorString_insert(vectorString self, std::vector< std::string >::iterator pos, std::vector< std::string >::size_type n, std::vector< std::string >::value_type const & x)\n"
		""},
	 { "vectorString_reserve", (PyCFunction)(void(*)(void))_wrap_vectorString_reserve, METH_VARARGS|METH_KEYWORDS, "vectorString_reserve(vectorString self, std::vector< std::string >::size_type n)"},
	 { "vectorString_capacity", _wrap_vectorString_capacity, METH_O, "vectorString_capacity(vectorString self) -> std::vector< std::string >::size_type"},
	 { "delete_vectorString", _wrap_delete_vectorString, METH_O, "delete_vectorString(vectorString self)"},
	 { "vectorString_swigregister", vectorString_swigregister, METH_O, NULL},
	 { "vectorString_swiginit", vectorString_swiginit, METH_VARARGS, NULL},
	 { "new_pairVectorDouble", _wrap_new_pairVectorDouble, METH_VARARGS, "\n"
		"pairVectorDouble()\n"
		"pairVectorDouble(vectorDouble first, vectorDouble second)\n"
		"new_pairVectorDouble(pairVectorDouble other) -> pairVectorDouble\n"
		""},
	 { "pairVectorDouble_first_set", _wrap_pairVectorDouble_first_set, METH_VARARGS, "pairVectorDouble_first_set(pairVectorDouble self, vectorDouble first)"},
	 { "pairVectorDouble_first_get", _wrap_pairVectorDouble_first_get, METH_O, "pairVectorDouble_first_get(pairVectorDouble self) -> vectorDouble"},
	 { "pairVectorDouble_second_set", _wrap_pairVectorDouble_second_set, METH_VARARGS, "pairVectorDouble_second_set(pairVectorDouble self, vectorDouble second)"},
	 { "pairVectorDouble_second_get", _wrap_pairVectorDouble_second_get, METH_O, "pairVectorDouble_second_get(pairVectorDouble self) -> vectorDouble"},
	 { "delete_pairVectorDouble", _wrap_delete_pairVectorDouble, METH_O, "delete_pairVectorDouble(pairVectorDouble self)"},
	 { "pairVectorDouble_swigregister", pairVectorDouble_swigregister, METH_O, NULL},
	 { "pairVectorDouble_swiginit", pairVectorDouble_swiginit, METH_VARARGS, NULL},
	 { "vectorFloat_iterator", _wrap_vectorFloat_iterator, METH_O, "vectorFloat_iterator(vectorFloat self) -> SwigPyIterator"},
	 { "vectorFloat___nonzero__", _wrap_vectorFloat___nonzero__, METH_O, "vectorFloat___nonzero__(vectorFloat self) -> bool"},
	 { "vectorFloat___bool__", _wrap_vectorFloat___bool__, METH_O, "vectorFloat___bool__(vectorFloat self) -> bool"},
	 { "vectorFloat___len__", _wrap_vectorFloat___len__, METH_O, "vectorFloat___len__(vectorFloat self) -> std::vector< float >::size_type"},
	 { "vectorFloat___getslice__", (PyCFunction)(void(*)(void))_wrap_vectorFloat___getslice__, METH_VARARGS|METH_KEYWORDS, "vectorFloat___getslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j) -> vectorFloat"},
	 { "vectorFloat___setslice__", _wrap_vectorFloat___setslice__, METH_VARARGS, "\n"
		"vectorFloat___setslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j)\n"
		"vectorFloat___setslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j, vectorFloat v)\n"
		""},
	 { "vectorFloat___delslice__", (PyCFunction)(void(*)(void))_wrap_vectorFloat___delslice__, METH_VARARGS|METH_KEYWORDS, "vectorFloat___delslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j)"},
	 { "vectorFloat___delitem__", _wrap_vectorFloat___delitem__, METH_VARARGS, "\n"
		"vectorFloat___delitem__(vectorFloat self, std::vector< float >::difference_type i)\n"
		"vectorFloat___delitem__(vectorFloat self, PySliceObject * slice)\n"
		""},
	 { "vectorFloat___getitem__", _wrap_vectorFloat___getitem__, METH_VARARGS, "\n"
		"vectorFloat___getitem__(vectorFloat self, PySliceObject * slice) -> vectorFloat\n"
		"vectorFloat___getitem__(vectorFloat self, std::vector< float >::difference_type i) -> std::vector< float >::value_type const &\n"
		""},
	 { "vectorFloat___setitem__", _wrap_vectorFloat___setitem__, METH_VARARGS, "\n"
		"vectorFloat___setitem__(vectorFloat self, PySliceObject * slice, vectorFloat v)\n"
		"vectorFloat___setitem__(vectorFloat self, PySliceObject * slice)\n"
		"vectorFloat___setitem__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::value_type const & x)\n"
		""},
	 { "vectorFloat_pop", _wrap_vectorFloat_pop, METH_O, "vectorFloat_pop(vectorFloat self) -> std::vector< float >::value_type"},
	 { "vectorFloat_append", (PyCFunction)(void(*)(void))_wrap_vectorFloat_append, METH_VARARGS|METH_KEYWORDS, "vectorFloat_append(vectorFloat self, std::vector< float >::value_type const & x)"},
	 { "vectorFloat_empty", _wrap_vectorFloat_empty, METH_O, "vectorFloat_empty(vectorFloat self) -> bool"},
	 { "vectorFloat_size", _wrap_vectorFloat_size, METH_O, "vectorFloat_size(vectorFloat self) -> std::vector< float >::size_type"},
	 { "vectorFloat_swap", (PyCFunction)(void(*)(void))_wrap_vectorFloat_swap, METH_VARARGS|METH_KEYWORDS, "vectorFloat_swap(vectorFloat self, vectorFloat v)"},
	 { "vectorFloat_begin", _wrap_vectorFloat_begin, METH_O, "vectorFloat_begin(vectorFloat self) -> std::vector< float >::iterator"},
	 { "vectorFloat_end", _wrap_vectorFloat_end, METH_O, "vectorFloat_end(vectorFloat self) -> std::vector< float >::iterator"},
	 { "vectorFloat_rbegin", _wrap_vectorFloat_rbegin, METH_O, "vectorFloat_rbegin(vectorFloat self) -> std::vector< float >::reverse_iterator"},
	 { "vectorFloat_rend", _wrap_vectorFloat_rend, METH_O, "vectorFloat_rend(vectorFloat self) -> std::vector< float >::reverse_iterator"},
	 { "vectorFloat_clear", _wrap_vectorFloat_clear, METH_O, "vectorFloat_clear(vectorFloat self)"},
	 { "vectorFloat_get_allocator", _wrap_vectorFloat_get_allocator, METH_O, "vectorFloat_get_allocator(vectorFloat self) -> std::vector< float >::allocator_type"},
	 { "vectorFloat_pop_back", _wrap_vectorFloat_pop_back, METH_O, "vectorFloat_pop_back(vectorFloat self)"},
	 { "vectorFloat_erase", _wrap_vectorFloat_erase, METH_VARARGS, "\n"
		"vectorFloat_erase(vectorFloat self, std::vector< float >::iterator pos) -> std::vector< float >::iterator\n"
		"vectorFloat_erase(vectorFloat self, std::vector< float >::iterator first, std::vector< float >::iterator last) -> std::vector< float >::iterator\n"
		""},
	 { "new_vectorFloat", _wrap_new_vectorFloat, METH_VARARGS, "\n"
		"vectorFloat()\n"
		"vectorFloat(vectorFloat other)\n"
		"vectorFloat(std::vector< float >::size_type size)\n"
		"new_vectorFloat(std::vector< float >::size_type size, std::vector< float >::value_type const & value) -> vectorFloat\n"
		""},
	 { "vectorFloat_push_back", (PyCFunction)(void(*)(void))_wrap_vectorFloat_push_back, METH_VARARGS|METH_KEYWORDS, "vectorFloat_push_back(vectorFloat self, std::vector< float >::value_type const & x)"},
	 { "vectorFloat_front", _wrap_vectorFloat_front, METH_O, "vectorFloat_front(vectorFloat self) -> std::vector< float >::value_type const &"},
	 { "vectorFloat_back", _wrap_vectorFloat_back, METH_O, "vectorFloat_back(vectorFloat self) -> std::vector< float >::value_type const &"},
	 { "vectorFloat_assign", (PyCFunction)(void(*)(void))_wrap_vectorFloat_assign, METH_VARARGS|METH_KEYWORDS, "vectorFloat_assign(vectorFloat self, std::vector< float >::size_type n, std::vector< float >::value_type const & x)"},
	 { "vectorFloat_resize", _wrap_vectorFloat_resize, METH_VARARGS, "\n"
		"vectorFloat_resize(vectorFloat self, std::vector< float >::size_type new_size)\n"
		"vectorFloat_resize(vectorFloat self, std::vector< float >::size_type new_size, std::vector< float >::value_type const & x)\n"
		""},
	 { "vectorFloat_insert", _wrap_vectorFloat_insert, METH_VARARGS, "\n"
		"vectorFloat_insert(vectorFloat self, std::vector< float >::iterator pos, std::vector< float >::value_type const & x) -> std::vector< float >::iterator\n"
		"vectorFloat_insert(vectorFloat self, std::vector< float >::iterator pos, std::vector< float >::size_type n, std::vector< float >::value_type const & x)\n"
		""},
	 { "vectorFloat_reserve", (PyCFunction)(void(*)(void))_wrap_vectorFloat_reserve, METH_VARARGS|METH_KEYWORDS, "vectorFloat_reserve(vectorFloat self, std::vector< float >::size_type n)"},
	 { "vectorFloat_capacity", _wrap_vectorFloat_capacity, METH_O, "vectorFloat_capacity(vectorFloat self) -> std::vector< float >::size_type"},
	 { "delete_vectorFloat", _wrap_delete_vectorFloat, METH_O, "delete_vectorFloat(vectorFloat self)"},
	 { "vectorFloat_swigregister", vectorFloat_swigregister, METH_O, NULL},
	 { "vectorFloat_swiginit", vectorFloat_swiginit, METH_VARARGS, NULL},
	 { "new_pairVectorFloat", _wrap_new_pairVectorFloat, METH_VARARGS, "\n"
		"pairVectorFloat()\n"
		"pairVectorFloat(vectorFloat first, vectorFloat second)\n"
		"new_pairVectorFloat(pairVectorFloat other) -> pairVectorFloat\n"
		""},
	 { "pairVectorFloat_first_set", _wrap_pairVectorFloat_first_set, METH_VARARGS, "pairVectorFloat_first_set(pairVectorFloat self, vectorFloat first)"},
	 { "pairVectorFloat_first_get", _wrap_pairVectorFloat_first_get, METH_O, "pairVectorFloat_first_get(pairVectorFloat self) -> vectorFloat"},
	 { "pairVectorFloat_second_set", _wrap_pairVectorFloat_second_set, METH_VARARGS, "pairVectorFloat_second_set(pairVectorFloat self, vectorFloat second)"},
	 { "pairVectorFloat_second_get", _wrap_pairVectorFloat_second_get, METH_O, "pairVectorFloat_second_get(pairVectorFloat self) -> vectorFloat"},
	 { "delete_pairVectorFloat", _wrap_delete_pairVectorFloat, METH_O, "delete_pairVectorFloat(pairVectorFloat self)"},
	 { "pairVectorFloat_swigregister", pairVectorFloat_swigregister, METH_O, NULL},
	 { "pairVectorFloat_swiginit", pairVectorFloat_swiginit, METH_VARARGS, NULL},
	 { "check_emd_status", (PyCFunction)(void(*)(void))_wrap_check_emd_status, METH_VARARGS|METH_KEYWORDS, "check_emd_status(wasserstein::EMDStatus status)"},
	 { "delete_EMDBaseFloat64", _wrap_delete_EMDBaseFloat64, METH_O, "delete_EMDBaseFloat64(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_R", _wrap_EMDBaseFloat64_R, METH_O, "EMDBaseFloat64_R(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_beta", _wrap_EMDBaseFloat64_beta, METH_O, "EMDBaseFloat64_beta(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_set_R", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_R, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat64_set_R(EMDBaseFloat64 self, double R)"},
	 { "EMDBaseFloat64_set_beta", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_beta, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat64_set_beta(EMDBaseFloat64 self, double beta)"},
	 { "EMDBaseFloat64_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat64_set_network_simplex_params(EMDBaseFloat64 self, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1)"},
	 { "EMDBaseFloat64_norm", _wrap_EMDBaseFloat64_norm, METH_O, "EMDBaseFloat64_norm(EMDBaseFloat64 self) -> bool"},
	 { "EMDBaseFloat64_set_norm", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_norm, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat64_set_norm(EMDBaseFloat64 self, bool norm)"},
	 { "EMDBaseFloat64_do_timing", _wrap_EMDBaseFloat64_do_timing, METH_O, "EMDBaseFloat64_do_timing(EMDBaseFloat64 self) -> bool"},
	 { "EMDBaseFloat64_set_do_timing", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_do_timing, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat64_set_do_timing(EMDBaseFloat64 self, bool timing)"},
	 { "EMDBaseFloat64_external_dists", _wrap_EMDBaseFloat64_external_dists, METH_O, "EMDBaseFloat64_external_dists(EMDBaseFloat64 self) -> bool"},
	 { "EMDBaseFloat64_set_external_dists", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_external_dists, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat64_set_external_dists(EMDBaseFloat64 self, bool exdists)"},
	 { "EMDBaseFloat64_n0", _wrap_EMDBaseFloat64_n0, METH_O, "EMDBaseFloat64_n0(EMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat64_n1", _wrap_EMDBaseFloat64_n1, METH_O, "EMDBaseFloat64_n1(EMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat64_extra", _wrap_EMDBaseFloat64_extra, METH_O, "EMDBaseFloat64_extra(EMDBaseFloat64 self) -> wasserstein::ExtraParticle"},
	 { "EMDBaseFloat64_emd", _wrap_EMDBaseFloat64_emd, METH_O, "EMDBaseFloat64_emd(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_status", _wrap_EMDBaseFloat64_status, METH_O, "EMDBaseFloat64_status(EMDBaseFloat64 self) -> wasserstein::EMDStatus"},
	 { "EMDBaseFloat64_weightdiff", _wrap_EMDBaseFloat64_weightdiff, METH_O, "EMDBaseFloat64_weightdiff(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_scale", _wrap_EMDBaseFloat64_scale, METH_O, "EMDBaseFloat64_scale(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_n_iter", _wrap_EMDBaseFloat64_n_iter, METH_O, "EMDBaseFloat64_n_iter(EMDBaseFloat64 self) -> std::size_t"},
	 { "EMDBaseFloat64_dists_vec", _wrap_EMDBaseFloat64_dists_vec, METH_O, "EMDBaseFloat64_dists_vec(EMDBaseFloat64 self) -> vectorDouble"},
	 { "EMDBaseFloat64_flows_vec", _wrap_EMDBaseFloat64_flows_vec, METH_O, "EMDBaseFloat64_flows_vec(EMDBaseFloat64 self) -> vectorDouble"},
	 { "EMDBaseFloat64_flow", _wrap_EMDBaseFloat64_flow, METH_VARARGS, "\n"
		"EMDBaseFloat64_flow(EMDBaseFloat64 self, wasserstein::index_type i, wasserstein::index_type j) -> double\n"
		"EMDBaseFloat64_flow(EMDBaseFloat64 self, std::size_t ind) -> double\n"
		""},
	 { "EMDBaseFloat64_node_potentials", _wrap_EMDBaseFloat64_node_potentials, METH_O, "EMDBaseFloat64_node_potentials(EMDBaseFloat64 self) -> pairVectorDouble"},
	 { "EMDBaseFloat64_duration", _wrap_EMDBaseFloat64_duration, METH_O, "EMDBaseFloat64_duration(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_clear", _wrap_EMDBaseFloat64_clear, METH_O, "EMDBaseFloat64_clear(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_flows", _wrap_EMDBaseFloat64_flows, METH_O, "EMDBaseFloat64_flows(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_dists", _wrap_EMDBaseFloat64_dists, METH_O, "EMDBaseFloat64_dists(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_npy_node_potentials", _wrap_EMDBaseFloat64_npy_node_potentials, METH_O, "EMDBaseFloat64_npy_node_potentials(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_swigregister", EMDBaseFloat64_swigregister, METH_O, NULL},
	 { "delete_PairwiseEMDBaseFloat64", _wrap_delete_PairwiseEMDBaseFloat64, METH_O, "delete_PairwiseEMDBaseFloat64(PairwiseEMDBaseFloat64 self)"},
	 { "PairwiseEMDBaseFloat64_R", _wrap_PairwiseEMDBaseFloat64_R, METH_O, "PairwiseEMDBaseFloat64_R(PairwiseEMDBaseFloat64 self) -> double"},
	 { "PairwiseEMDBaseFloat64_set_R", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_R, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_R(PairwiseEMDBaseFloat64 self, double R)"},
	 { "PairwiseEMDBaseFloat64_beta", _wrap_PairwiseEMDBaseFloat64_beta, METH_O, "PairwiseEMDBaseFloat64_beta(PairwiseEMDBaseFloat64 self) -> double"},
	 { "PairwiseEMDBaseFloat64_set_beta", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_beta, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_beta(PairwiseEMDBaseFloat64 self, double beta)"},
	 { "PairwiseEMDBaseFloat64_norm", _wrap_PairwiseEMDBaseFloat64_norm, METH_O, "PairwiseEMDBaseFloat64_norm(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_set_norm", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_norm, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_norm(PairwiseEMDBaseFloat64 self, bool norm)"},
	 { "PairwiseEMDBaseFloat64_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_network_simplex_params(PairwiseEMDBaseFloat64 self, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1)"},
	 { "PairwiseEMDBaseFloat64_set_external_emd_handler", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_external_emd_handler, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_external_emd_handler(PairwiseEMDBaseFloat64 self, ExternalEMDHandlerFloat64 handler)"},
	 { "PairwiseEMDBaseFloat64_have_external_emd_handler", _wrap_PairwiseEMDBaseFloat64_have_external_emd_handler, METH_O, "PairwiseEMDBaseFloat64_have_external_emd_handler(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_num_threads", _wrap_PairwiseEMDBaseFloat64_num_threads, METH_O, "PairwiseEMDBaseFloat64_num_threads(PairwiseEMDBaseFloat64 self) -> int"},
	 { "PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize(PairwiseEMDBaseFloat64 self, int chunksize)"},
	 { "PairwiseEMDBaseFloat64_omp_dynamic_chunksize", _wrap_PairwiseEMDBaseFloat64_omp_dynamic_chunksize, METH_O, "PairwiseEMDBaseFloat64_omp_dynamic_chunksize(PairwiseEMDBaseFloat64 self) -> int"},
	 { "PairwiseEMDBaseFloat64_set_request_mode", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_request_mode, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_set_request_mode(PairwiseEMDBaseFloat64 self, bool mode)"},
	 { "PairwiseEMDBaseFloat64_request_mode", _wrap_PairwiseEMDBaseFloat64_request_mode, METH_O, "PairwiseEMDBaseFloat64_request_mode(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_duration", _wrap_PairwiseEMDBaseFloat64_duration, METH_O, "PairwiseEMDBaseFloat64_duration(PairwiseEMDBaseFloat64 self) -> double"},
	 { "PairwiseEMDBaseFloat64_nevA", _wrap_PairwiseEMDBaseFloat64_nevA, METH_O, "PairwiseEMDBaseFloat64_nevA(PairwiseEMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat64_nevB", _wrap_PairwiseEMDBaseFloat64_nevB, METH_O, "PairwiseEMDBaseFloat64_nevB(PairwiseEMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat64_num_emds", _wrap_PairwiseEMDBaseFloat64_num_emds, METH_O, "PairwiseEMDBaseFloat64_num_emds(PairwiseEMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat64_storage", _wrap_PairwiseEMDBaseFloat64_storage, METH_O, "PairwiseEMDBaseFloat64_storage(PairwiseEMDBaseFloat64 self) -> wasserstein::EMDPairsStorage"},
	 { "PairwiseEMDBaseFloat64_errored", _wrap_PairwiseEMDBaseFloat64_errored, METH_O, "PairwiseEMDBaseFloat64_errored(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_error_messages", _wrap_PairwiseEMDBaseFloat64_error_messages, METH_O, "PairwiseEMDBaseFloat64_error_messages(PairwiseEMDBaseFloat64 self) -> vectorString"},
	 { "PairwiseEMDBaseFloat64_emds_vec", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_emds_vec, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_emds_vec(PairwiseEMDBaseFloat64 self, bool raw=False) -> vectorDouble"},
	 { "PairwiseEMDBaseFloat64_emd", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_emd, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat64_emd(PairwiseEMDBaseFloat64 self, wasserstein::index_type i, wasserstein::index_type j, int thread=0) -> double"},
	 { "PairwiseEMDBaseFloat64_emds", _wrap_PairwiseEMDBaseFloat64_emds, METH_O, "PairwiseEMDBaseFloat64_emds(PairwiseEMDBaseFloat64 self)"},
	 { "PairwiseEMDBaseFloat64_raw_emds", _wrap_PairwiseEMDBaseFloat64_raw_emds, METH_O, "PairwiseEMDBaseFloat64_raw_emds(PairwiseEMDBaseFloat64 self)"},
	 { "PairwiseEMDBaseFloat64_swigregister", PairwiseEMDBaseFloat64_swigregister, METH_O, NULL},
	 { "delete_ExternalEMDHandlerFloat64", _wrap_delete_ExternalEMDHandlerFloat64, METH_O, "delete_ExternalEMDHandlerFloat64(ExternalEMDHandlerFloat64 self)"},
	 { "ExternalEMDHandlerFloat64_description", _wrap_ExternalEMDHandlerFloat64_description, METH_O, "ExternalEMDHandlerFloat64_description(ExternalEMDHandlerFloat64 self) -> std::string"},
	 { "ExternalEMDHandlerFloat64_num_calls", _wrap_ExternalEMDHandlerFloat64_num_calls, METH_O, "ExternalEMDHandlerFloat64_num_calls(ExternalEMDHandlerFloat64 self) -> std::size_t"},
	 { "ExternalEMDHandlerFloat64___call__", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat64___call__, METH_VARARGS|METH_KEYWORDS, "ExternalEMDHandlerFloat64___call__(ExternalEMDHandlerFloat64 self, double emd, double weight=1)"},
	 { "ExternalEMDHandlerFloat64_evaluate", _wrap_ExternalEMDHandlerFloat64_evaluate, METH_VARARGS, "\n"
		"ExternalEMDHandlerFloat64_evaluate(ExternalEMDHandlerFloat64 self, double * emds)\n"
		"ExternalEMDHandlerFloat64_evaluate(ExternalEMDHandlerFloat64 self, double * emds, double * event_weights)\n"
		""},
	 { "ExternalEMDHandlerFloat64_evaluate_symmetric", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat64_evaluate_symmetric, METH_VARARGS|METH_KEYWORDS, "ExternalEMDHandlerFloat64_evaluate_symmetric(ExternalEMDHandlerFloat64 self, double * emds, double * event_weights)"},
	 { "ExternalEMDHandlerFloat64_swigregister", ExternalEMDHandlerFloat64_swigregister, METH_O, NULL},
	 { "new_Histogram1DHandlerLogFloat64", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerLogFloat64, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerLogFloat64(unsigned int nbins, double axis_min, double axis_max) -> Histogram1DHandlerLogFloat64"},
	 { "delete_Histogram1DHandlerLogFloat64", _wrap_delete_Histogram1DHandlerLogFloat64, METH_O, "delete_Histogram1DHandlerLogFloat64(Histogram1DHandlerLogFloat64 self)"},
	 { "Histogram1DHandlerLogFloat64_nbins", _wrap_Histogram1DHandlerLogFloat64_nbins, METH_O, "Histogram1DHandlerLogFloat64_nbins(Histogram1DHandlerLogFloat64 self) -> unsigned int"},
	 { "Histogram1DHandlerLogFloat64_axis_min", _wrap_Histogram1DHandlerLogFloat64_axis_min, METH_O, "Histogram1DHandlerLogFloat64_axis_min(Histogram1DHandlerLogFloat64 self) -> double"},
	 { "Histogram1DHandlerLogFloat64_axis_max", _wrap_Histogram1DHandlerLogFloat64_axis_max, METH_O, "Histogram1DHandlerLogFloat64_axis_max(Histogram1DHandlerLogFloat64 self) -> double"},
	 { "Histogram1DHandlerLogFloat64_description", _wrap_Histogram1DHandlerLogFloat64_description, METH_O, "Histogram1DHandlerLogFloat64_description(Histogram1DHandlerLogFloat64 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat64_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat64_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerLogFloat64_hist_vals_vars_vec(Histogram1DHandlerLogFloat64 self, bool overflows=True) -> pairVectorDouble"},
	 { "Histogram1DHandlerLogFloat64_bin_centers_vec", _wrap_Histogram1DHandlerLogFloat64_bin_centers_vec, METH_O, "Histogram1DHandlerLogFloat64_bin_centers_vec(Histogram1DHandlerLogFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerLogFloat64_bin_edges_vec", _wrap_Histogram1DHandlerLogFloat64_bin_edges_vec, METH_O, "Histogram1DHandlerLogFloat64_bin_edges_vec(Histogram1DHandlerLogFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerLogFloat64___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat64___iadd__, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerLogFloat64___iadd__(Histogram1DHandlerLogFloat64 self, Histogram1DHandlerLogFloat64 other) -> Histogram1DHandlerLogFloat64"},
	 { "Histogram1DHandlerLogFloat64___repr__", _wrap_Histogram1DHandlerLogFloat64___repr__, METH_O, "Histogram1DHandlerLogFloat64___repr__(Histogram1DHandlerLogFloat64 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat64_bin_centers", _wrap_Histogram1DHandlerLogFloat64_bin_centers, METH_O, "Histogram1DHandlerLogFloat64_bin_centers(Histogram1DHandlerLogFloat64 self)"},
	 { "Histogram1DHandlerLogFloat64_bin_edges", _wrap_Histogram1DHandlerLogFloat64_bin_edges, METH_O, "Histogram1DHandlerLogFloat64_bin_edges(Histogram1DHandlerLogFloat64 self)"},
	 { "Histogram1DHandlerLogFloat64_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat64_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerLogFloat64_hist_vals_vars(Histogram1DHandlerLogFloat64 self, bool overflows=True)"},
	 { "Histogram1DHandlerLogFloat64_swigregister", Histogram1DHandlerLogFloat64_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerLogFloat64_swiginit", Histogram1DHandlerLogFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_Histogram1DHandlerFloat64", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerFloat64, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerFloat64(unsigned int nbins, double axis_min, double axis_max) -> Histogram1DHandlerFloat64"},
	 { "delete_Histogram1DHandlerFloat64", _wrap_delete_Histogram1DHandlerFloat64, METH_O, "delete_Histogram1DHandlerFloat64(Histogram1DHandlerFloat64 self)"},
	 { "Histogram1DHandlerFloat64_nbins", _wrap_Histogram1DHandlerFloat64_nbins, METH_O, "Histogram1DHandlerFloat64_nbins(Histogram1DHandlerFloat64 self) -> unsigned int"},
	 { "Histogram1DHandlerFloat64_axis_min", _wrap_Histogram1DHandlerFloat64_axis_min, METH_O, "Histogram1DHandlerFloat64_axis_min(Histogram1DHandlerFloat64 self) -> double"},
	 { "Histogram1DHandlerFloat64_axis_max", _wrap_Histogram1DHandlerFloat64_axis_max, METH_O, "Histogram1DHandlerFloat64_axis_max(Histogram1DHandlerFloat64 self) -> double"},
	 { "Histogram1DHandlerFloat64_description", _wrap_Histogram1DHandlerFloat64_description, METH_O, "Histogram1DHandlerFloat64_description(Histogram1DHandlerFloat64 self) -> std::string"},
	 { "Histogram1DHandlerFloat64_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat64_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerFloat64_hist_vals_vars_vec(Histogram1DHandlerFloat64 self, bool overflows=True) -> pairVectorDouble"},
	 { "Histogram1DHandlerFloat64_bin_centers_vec", _wrap_Histogram1DHandlerFloat64_bin_centers_vec, METH_O, "Histogram1DHandlerFloat64_bin_centers_vec(Histogram1DHandlerFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerFloat64_bin_edges_vec", _wrap_Histogram1DHandlerFloat64_bin_edges_vec, METH_O, "Histogram1DHandlerFloat64_bin_edges_vec(Histogram1DHandlerFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerFloat64___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat64___iadd__, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerFloat64___iadd__(Histogram1DHandlerFloat64 self, Histogram1DHandlerFloat64 other) -> Histogram1DHandlerFloat64"},
	 { "Histogram1DHandlerFloat64___repr__", _wrap_Histogram1DHandlerFloat64___repr__, METH_O, "Histogram1DHandlerFloat64___repr__(Histogram1DHandlerFloat64 self) -> std::string"},
	 { "Histogram1DHandlerFloat64_bin_centers", _wrap_Histogram1DHandlerFloat64_bin_centers, METH_O, "Histogram1DHandlerFloat64_bin_centers(Histogram1DHandlerFloat64 self)"},
	 { "Histogram1DHandlerFloat64_bin_edges", _wrap_Histogram1DHandlerFloat64_bin_edges, METH_O, "Histogram1DHandlerFloat64_bin_edges(Histogram1DHandlerFloat64 self)"},
	 { "Histogram1DHandlerFloat64_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat64_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerFloat64_hist_vals_vars(Histogram1DHandlerFloat64 self, bool overflows=True)"},
	 { "Histogram1DHandlerFloat64_swigregister", Histogram1DHandlerFloat64_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerFloat64_swiginit", Histogram1DHandlerFloat64_swiginit, METH_VARARGS, NULL},
	 { "delete_EMDBaseFloat32", _wrap_delete_EMDBaseFloat32, METH_O, "delete_EMDBaseFloat32(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_R", _wrap_EMDBaseFloat32_R, METH_O, "EMDBaseFloat32_R(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_beta", _wrap_EMDBaseFloat32_beta, METH_O, "EMDBaseFloat32_beta(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_set_R", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_R, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat32_set_R(EMDBaseFloat32 self, float R)"},
	 { "EMDBaseFloat32_set_beta", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_beta, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat32_set_beta(EMDBaseFloat32 self, float beta)"},
	 { "EMDBaseFloat32_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat32_set_network_simplex_params(EMDBaseFloat32 self, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1)"},
	 { "EMDBaseFloat32_norm", _wrap_EMDBaseFloat32_norm, METH_O, "EMDBaseFloat32_norm(EMDBaseFloat32 self) -> bool"},
	 { "EMDBaseFloat32_set_norm", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_norm, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat32_set_norm(EMDBaseFloat32 self, bool norm)"},
	 { "EMDBaseFloat32_do_timing", _wrap_EMDBaseFloat32_do_timing, METH_O, "EMDBaseFloat32_do_timing(EMDBaseFloat32 self) -> bool"},
	 { "EMDBaseFloat32_set_do_timing", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_do_timing, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat32_set_do_timing(EMDBaseFloat32 self, bool timing)"},
	 { "EMDBaseFloat32_external_dists", _wrap_EMDBaseFloat32_external_dists, METH_O, "EMDBaseFloat32_external_dists(EMDBaseFloat32 self) -> bool"},
	 { "EMDBaseFloat32_set_external_dists", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_external_dists, METH_VARARGS|METH_KEYWORDS, "EMDBaseFloat32_set_external_dists(EMDBaseFloat32 self, bool exdists)"},
	 { "EMDBaseFloat32_n0", _wrap_EMDBaseFloat32_n0, METH_O, "EMDBaseFloat32_n0(EMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat32_n1", _wrap_EMDBaseFloat32_n1, METH_O, "EMDBaseFloat32_n1(EMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat32_extra", _wrap_EMDBaseFloat32_extra, METH_O, "EMDBaseFloat32_extra(EMDBaseFloat32 self) -> wasserstein::ExtraParticle"},
	 { "EMDBaseFloat32_emd", _wrap_EMDBaseFloat32_emd, METH_O, "EMDBaseFloat32_emd(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_status", _wrap_EMDBaseFloat32_status, METH_O, "EMDBaseFloat32_status(EMDBaseFloat32 self) -> wasserstein::EMDStatus"},
	 { "EMDBaseFloat32_weightdiff", _wrap_EMDBaseFloat32_weightdiff, METH_O, "EMDBaseFloat32_weightdiff(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_scale", _wrap_EMDBaseFloat32_scale, METH_O, "EMDBaseFloat32_scale(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_n_iter", _wrap_EMDBaseFloat32_n_iter, METH_O, "EMDBaseFloat32_n_iter(EMDBaseFloat32 self) -> std::size_t"},
	 { "EMDBaseFloat32_dists_vec", _wrap_EMDBaseFloat32_dists_vec, METH_O, "EMDBaseFloat32_dists_vec(EMDBaseFloat32 self) -> vectorFloat"},
	 { "EMDBaseFloat32_flows_vec", _wrap_EMDBaseFloat32_flows_vec, METH_O, "EMDBaseFloat32_flows_vec(EMDBaseFloat32 self) -> vectorFloat"},
	 { "EMDBaseFloat32_flow", _wrap_EMDBaseFloat32_flow, METH_VARARGS, "\n"
		"EMDBaseFloat32_flow(EMDBaseFloat32 self, wasserstein::index_type i, wasserstein::index_type j) -> float\n"
		"EMDBaseFloat32_flow(EMDBaseFloat32 self, std::size_t ind) -> float\n"
		""},
	 { "EMDBaseFloat32_node_potentials", _wrap_EMDBaseFloat32_node_potentials, METH_O, "EMDBaseFloat32_node_potentials(EMDBaseFloat32 self) -> pairVectorFloat"},
	 { "EMDBaseFloat32_duration", _wrap_EMDBaseFloat32_duration, METH_O, "EMDBaseFloat32_duration(EMDBaseFloat32 self) -> double"},
	 { "EMDBaseFloat32_clear", _wrap_EMDBaseFloat32_clear, METH_O, "EMDBaseFloat32_clear(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_flows", _wrap_EMDBaseFloat32_flows, METH_O, "EMDBaseFloat32_flows(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_dists", _wrap_EMDBaseFloat32_dists, METH_O, "EMDBaseFloat32_dists(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_npy_node_potentials", _wrap_EMDBaseFloat32_npy_node_potentials, METH_O, "EMDBaseFloat32_npy_node_potentials(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_swigregister", EMDBaseFloat32_swigregister, METH_O, NULL},
	 { "delete_PairwiseEMDBaseFloat32", _wrap_delete_PairwiseEMDBaseFloat32, METH_O, "delete_PairwiseEMDBaseFloat32(PairwiseEMDBaseFloat32 self)"},
	 { "PairwiseEMDBaseFloat32_R", _wrap_PairwiseEMDBaseFloat32_R, METH_O, "PairwiseEMDBaseFloat32_R(PairwiseEMDBaseFloat32 self) -> float"},
	 { "PairwiseEMDBaseFloat32_set_R", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_R, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_R(PairwiseEMDBaseFloat32 self, float R)"},
	 { "PairwiseEMDBaseFloat32_beta", _wrap_PairwiseEMDBaseFloat32_beta, METH_O, "PairwiseEMDBaseFloat32_beta(PairwiseEMDBaseFloat32 self) -> float"},
	 { "PairwiseEMDBaseFloat32_set_beta", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_beta, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_beta(PairwiseEMDBaseFloat32 self, float beta)"},
	 { "PairwiseEMDBaseFloat32_norm", _wrap_PairwiseEMDBaseFloat32_norm, METH_O, "PairwiseEMDBaseFloat32_norm(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_set_norm", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_norm, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_norm(PairwiseEMDBaseFloat32 self, bool norm)"},
	 { "PairwiseEMDBaseFloat32_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_network_simplex_params(PairwiseEMDBaseFloat32 self, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1)"},
	 { "PairwiseEMDBaseFloat32_set_external_emd_handler", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_external_emd_handler, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_external_emd_handler(PairwiseEMDBaseFloat32 self, ExternalEMDHandlerFloat32 handler)"},
	 { "PairwiseEMDBaseFloat32_have_external_emd_handler", _wrap_PairwiseEMDBaseFloat32_have_external_emd_handler, METH_O, "PairwiseEMDBaseFloat32_have_external_emd_handler(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_num_threads", _wrap_PairwiseEMDBaseFloat32_num_threads, METH_O, "PairwiseEMDBaseFloat32_num_threads(PairwiseEMDBaseFloat32 self) -> int"},
	 { "PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize(PairwiseEMDBaseFloat32 self, int chunksize)"},
	 { "PairwiseEMDBaseFloat32_omp_dynamic_chunksize", _wrap_PairwiseEMDBaseFloat32_omp_dynamic_chunksize, METH_O, "PairwiseEMDBaseFloat32_omp_dynamic_chunksize(PairwiseEMDBaseFloat32 self) -> int"},
	 { "PairwiseEMDBaseFloat32_set_request_mode", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_request_mode, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_set_request_mode(PairwiseEMDBaseFloat32 self, bool mode)"},
	 { "PairwiseEMDBaseFloat32_request_mode", _wrap_PairwiseEMDBaseFloat32_request_mode, METH_O, "PairwiseEMDBaseFloat32_request_mode(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_duration", _wrap_PairwiseEMDBaseFloat32_duration, METH_O, "PairwiseEMDBaseFloat32_duration(PairwiseEMDBaseFloat32 self) -> double"},
	 { "PairwiseEMDBaseFloat32_nevA", _wrap_PairwiseEMDBaseFloat32_nevA, METH_O, "PairwiseEMDBaseFloat32_nevA(PairwiseEMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat32_nevB", _wrap_PairwiseEMDBaseFloat32_nevB, METH_O, "PairwiseEMDBaseFloat32_nevB(PairwiseEMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat32_num_emds", _wrap_PairwiseEMDBaseFloat32_num_emds, METH_O, "PairwiseEMDBaseFloat32_num_emds(PairwiseEMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat32_storage", _wrap_PairwiseEMDBaseFloat32_storage, METH_O, "PairwiseEMDBaseFloat32_storage(PairwiseEMDBaseFloat32 self) -> wasserstein::EMDPairsStorage"},
	 { "PairwiseEMDBaseFloat32_errored", _wrap_PairwiseEMDBaseFloat32_errored, METH_O, "PairwiseEMDBaseFloat32_errored(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_error_messages", _wrap_PairwiseEMDBaseFloat32_error_messages, METH_O, "PairwiseEMDBaseFloat32_error_messages(PairwiseEMDBaseFloat32 self) -> vectorString"},
	 { "PairwiseEMDBaseFloat32_emds_vec", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_emds_vec, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_emds_vec(PairwiseEMDBaseFloat32 self, bool raw=False) -> vectorFloat"},
	 { "PairwiseEMDBaseFloat32_emd", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_emd, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDBaseFloat32_emd(PairwiseEMDBaseFloat32 self, wasserstein::index_type i, wasserstein::index_type j, int thread=0) -> float"},
	 { "PairwiseEMDBaseFloat32_emds", _wrap_PairwiseEMDBaseFloat32_emds, METH_O, "PairwiseEMDBaseFloat32_emds(PairwiseEMDBaseFloat32 self)"},
	 { "PairwiseEMDBaseFloat32_raw_emds", _wrap_PairwiseEMDBaseFloat32_raw_emds, METH_O, "PairwiseEMDBaseFloat32_raw_emds(PairwiseEMDBaseFloat32 self)"},
	 { "PairwiseEMDBaseFloat32_swigregister", PairwiseEMDBaseFloat32_swigregister, METH_O, NULL},
	 { "delete_ExternalEMDHandlerFloat32", _wrap_delete_ExternalEMDHandlerFloat32, METH_O, "delete_ExternalEMDHandlerFloat32(ExternalEMDHandlerFloat32 self)"},
	 { "ExternalEMDHandlerFloat32_description", _wrap_ExternalEMDHandlerFloat32_description, METH_O, "ExternalEMDHandlerFloat32_description(ExternalEMDHandlerFloat32 self) -> std::string"},
	 { "ExternalEMDHandlerFloat32_num_calls", _wrap_ExternalEMDHandlerFloat32_num_calls, METH_O, "ExternalEMDHandlerFloat32_num_calls(ExternalEMDHandlerFloat32 self) -> std::size_t"},
	 { "ExternalEMDHandlerFloat32___call__", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat32___call__, METH_VARARGS|METH_KEYWORDS, "ExternalEMDHandlerFloat32___call__(ExternalEMDHandlerFloat32 self, float emd, float weight=1)"},
	 { "ExternalEMDHandlerFloat32_evaluate", _wrap_ExternalEMDHandlerFloat32_evaluate, METH_VARARGS, "\n"
		"ExternalEMDHandlerFloat32_evaluate(ExternalEMDHandlerFloat32 self, float * emds)\n"
		"ExternalEMDHandlerFloat32_evaluate(ExternalEMDHandlerFloat32 self, float * emds, float * event_weights)\n"
		""},
	 { "ExternalEMDHandlerFloat32_evaluate_symmetric", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat32_evaluate_symmetric, METH_VARARGS|METH_KEYWORDS, "ExternalEMDHandlerFloat32_evaluate_symmetric(ExternalEMDHandlerFloat32 self, float * emds, float * event_weights)"},
	 { "ExternalEMDHandlerFloat32_swigregister", ExternalEMDHandlerFloat32_swigregister, METH_O, NULL},
	 { "new_Histogram1DHandlerLogFloat32", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerLogFloat32, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerLogFloat32(unsigned int nbins, float axis_min, float axis_max) -> Histogram1DHandlerLogFloat32"},
	 { "delete_Histogram1DHandlerLogFloat32", _wrap_delete_Histogram1DHandlerLogFloat32, METH_O, "delete_Histogram1DHandlerLogFloat32(Histogram1DHandlerLogFloat32 self)"},
	 { "Histogram1DHandlerLogFloat32_nbins", _wrap_Histogram1DHandlerLogFloat32_nbins, METH_O, "Histogram1DHandlerLogFloat32_nbins(Histogram1DHandlerLogFloat32 self) -> unsigned int"},
	 { "Histogram1DHandlerLogFloat32_axis_min", _wrap_Histogram1DHandlerLogFloat32_axis_min, METH_O, "Histogram1DHandlerLogFloat32_axis_min(Histogram1DHandlerLogFloat32 self) -> float"},
	 { "Histogram1DHandlerLogFloat32_axis_max", _wrap_Histogram1DHandlerLogFloat32_axis_max, METH_O, "Histogram1DHandlerLogFloat32_axis_max(Histogram1DHandlerLogFloat32 self) -> float"},
	 { "Histogram1DHandlerLogFloat32_description", _wrap_Histogram1DHandlerLogFloat32_description, METH_O, "Histogram1DHandlerLogFloat32_description(Histogram1DHandlerLogFloat32 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat32_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat32_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerLogFloat32_hist_vals_vars_vec(Histogram1DHandlerLogFloat32 self, bool overflows=True) -> pairVectorFloat"},
	 { "Histogram1DHandlerLogFloat32_bin_centers_vec", _wrap_Histogram1DHandlerLogFloat32_bin_centers_vec, METH_O, "Histogram1DHandlerLogFloat32_bin_centers_vec(Histogram1DHandlerLogFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerLogFloat32_bin_edges_vec", _wrap_Histogram1DHandlerLogFloat32_bin_edges_vec, METH_O, "Histogram1DHandlerLogFloat32_bin_edges_vec(Histogram1DHandlerLogFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerLogFloat32___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat32___iadd__, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerLogFloat32___iadd__(Histogram1DHandlerLogFloat32 self, Histogram1DHandlerLogFloat32 other) -> Histogram1DHandlerLogFloat32"},
	 { "Histogram1DHandlerLogFloat32___repr__", _wrap_Histogram1DHandlerLogFloat32___repr__, METH_O, "Histogram1DHandlerLogFloat32___repr__(Histogram1DHandlerLogFloat32 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat32_bin_centers", _wrap_Histogram1DHandlerLogFloat32_bin_centers, METH_O, "Histogram1DHandlerLogFloat32_bin_centers(Histogram1DHandlerLogFloat32 self)"},
	 { "Histogram1DHandlerLogFloat32_bin_edges", _wrap_Histogram1DHandlerLogFloat32_bin_edges, METH_O, "Histogram1DHandlerLogFloat32_bin_edges(Histogram1DHandlerLogFloat32 self)"},
	 { "Histogram1DHandlerLogFloat32_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat32_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerLogFloat32_hist_vals_vars(Histogram1DHandlerLogFloat32 self, bool overflows=True)"},
	 { "Histogram1DHandlerLogFloat32_swigregister", Histogram1DHandlerLogFloat32_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerLogFloat32_swiginit", Histogram1DHandlerLogFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_Histogram1DHandlerFloat32", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerFloat32, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerFloat32(unsigned int nbins, float axis_min, float axis_max) -> Histogram1DHandlerFloat32"},
	 { "delete_Histogram1DHandlerFloat32", _wrap_delete_Histogram1DHandlerFloat32, METH_O, "delete_Histogram1DHandlerFloat32(Histogram1DHandlerFloat32 self)"},
	 { "Histogram1DHandlerFloat32_nbins", _wrap_Histogram1DHandlerFloat32_nbins, METH_O, "Histogram1DHandlerFloat32_nbins(Histogram1DHandlerFloat32 self) -> unsigned int"},
	 { "Histogram1DHandlerFloat32_axis_min", _wrap_Histogram1DHandlerFloat32_axis_min, METH_O, "Histogram1DHandlerFloat32_axis_min(Histogram1DHandlerFloat32 self) -> float"},
	 { "Histogram1DHandlerFloat32_axis_max", _wrap_Histogram1DHandlerFloat32_axis_max, METH_O, "Histogram1DHandlerFloat32_axis_max(Histogram1DHandlerFloat32 self) -> float"},
	 { "Histogram1DHandlerFloat32_description", _wrap_Histogram1DHandlerFloat32_description, METH_O, "Histogram1DHandlerFloat32_description(Histogram1DHandlerFloat32 self) -> std::string"},
	 { "Histogram1DHandlerFloat32_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat32_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerFloat32_hist_vals_vars_vec(Histogram1DHandlerFloat32 self, bool overflows=True) -> pairVectorFloat"},
	 { "Histogram1DHandlerFloat32_bin_centers_vec", _wrap_Histogram1DHandlerFloat32_bin_centers_vec, METH_O, "Histogram1DHandlerFloat32_bin_centers_vec(Histogram1DHandlerFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerFloat32_bin_edges_vec", _wrap_Histogram1DHandlerFloat32_bin_edges_vec, METH_O, "Histogram1DHandlerFloat32_bin_edges_vec(Histogram1DHandlerFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerFloat32___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat32___iadd__, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerFloat32___iadd__(Histogram1DHandlerFloat32 self, Histogram1DHandlerFloat32 other) -> Histogram1DHandlerFloat32"},
	 { "Histogram1DHandlerFloat32___repr__", _wrap_Histogram1DHandlerFloat32___repr__, METH_O, "Histogram1DHandlerFloat32___repr__(Histogram1DHandlerFloat32 self) -> std::string"},
	 { "Histogram1DHandlerFloat32_bin_centers", _wrap_Histogram1DHandlerFloat32_bin_centers, METH_O, "Histogram1DHandlerFloat32_bin_centers(Histogram1DHandlerFloat32 self)"},
	 { "Histogram1DHandlerFloat32_bin_edges", _wrap_Histogram1DHandlerFloat32_bin_edges, METH_O, "Histogram1DHandlerFloat32_bin_edges(Histogram1DHandlerFloat32 self)"},
	 { "Histogram1DHandlerFloat32_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat32_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "Histogram1DHandlerFloat32_hist_vals_vars(Histogram1DHandlerFloat32 self, bool overflows=True)"},
	 { "Histogram1DHandlerFloat32_swigregister", Histogram1DHandlerFloat32_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerFloat32_swiginit", Histogram1DHandlerFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_CorrelationDimensionFloat32", (PyCFunction)(void(*)(void))_wrap_new_CorrelationDimensionFloat32, METH_VARARGS|METH_KEYWORDS, "new_CorrelationDimensionFloat32(unsigned int nbins, float axis_min, float axis_max) -> CorrelationDimensionFloat32"},
	 { "CorrelationDimensionFloat32_corrdims_vec", (PyCFunction)(void(*)(void))_wrap_CorrelationDimensionFloat32_corrdims_vec, METH_VARARGS|METH_KEYWORDS, "CorrelationDimensionFloat32_corrdims_vec(CorrelationDimensionFloat32 self, float eps=std::numeric_limits< float >::epsilon()) -> pairVectorFloat"},
	 { "CorrelationDimensionFloat32_corrdim_bins_vec", _wrap_CorrelationDimensionFloat32_corrdim_bins_vec, METH_O, "CorrelationDimensionFloat32_corrdim_bins_vec(CorrelationDimensionFloat32 self) -> vectorFloat"},
	 { "CorrelationDimensionFloat32_cumulative_vals_vars_vec", _wrap_CorrelationDimensionFloat32_cumulative_vals_vars_vec, METH_O, "CorrelationDimensionFloat32_cumulative_vals_vars_vec(CorrelationDimensionFloat32 self) -> pairVectorFloat"},
	 { "CorrelationDimensionFloat32___repr__", _wrap_CorrelationDimensionFloat32___repr__, METH_O, "CorrelationDimensionFloat32___repr__(CorrelationDimensionFloat32 self) -> std::string"},
	 { "CorrelationDimensionFloat32_corrdim_bins", _wrap_CorrelationDimensionFloat32_corrdim_bins, METH_O, "CorrelationDimensionFloat32_corrdim_bins(CorrelationDimensionFloat32 self)"},
	 { "CorrelationDimensionFloat32_corrdims", _wrap_CorrelationDimensionFloat32_corrdims, METH_O, "CorrelationDimensionFloat32_corrdims(CorrelationDimensionFloat32 self)"},
	 { "CorrelationDimensionFloat32_cumulative_vals_vars", _wrap_CorrelationDimensionFloat32_cumulative_vals_vars, METH_O, "CorrelationDimensionFloat32_cumulative_vals_vars(CorrelationDimensionFloat32 self)"},
	 { "delete_CorrelationDimensionFloat32", _wrap_delete_CorrelationDimensionFloat32, METH_O, "delete_CorrelationDimensionFloat32(CorrelationDimensionFloat32 self)"},
	 { "CorrelationDimensionFloat32_swigregister", CorrelationDimensionFloat32_swigregister, METH_O, NULL},
	 { "CorrelationDimensionFloat32_swiginit", CorrelationDimensionFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_CorrelationDimensionFloat64", (PyCFunction)(void(*)(void))_wrap_new_CorrelationDimensionFloat64, METH_VARARGS|METH_KEYWORDS, "new_CorrelationDimensionFloat64(unsigned int nbins, double axis_min, double axis_max) -> CorrelationDimensionFloat64"},
	 { "CorrelationDimensionFloat64_corrdims_vec", (PyCFunction)(void(*)(void))_wrap_CorrelationDimensionFloat64_corrdims_vec, METH_VARARGS|METH_KEYWORDS, "CorrelationDimensionFloat64_corrdims_vec(CorrelationDimensionFloat64 self, double eps=std::numeric_limits< double >::epsilon()) -> pairVectorDouble"},
	 { "CorrelationDimensionFloat64_corrdim_bins_vec", _wrap_CorrelationDimensionFloat64_corrdim_bins_vec, METH_O, "CorrelationDimensionFloat64_corrdim_bins_vec(CorrelationDimensionFloat64 self) -> vectorDouble"},
	 { "CorrelationDimensionFloat64_cumulative_vals_vars_vec", _wrap_CorrelationDimensionFloat64_cumulative_vals_vars_vec, METH_O, "CorrelationDimensionFloat64_cumulative_vals_vars_vec(CorrelationDimensionFloat64 self) -> pairVectorDouble"},
	 { "CorrelationDimensionFloat64___repr__", _wrap_CorrelationDimensionFloat64___repr__, METH_O, "CorrelationDimensionFloat64___repr__(CorrelationDimensionFloat64 self) -> std::string"},
	 { "CorrelationDimensionFloat64_corrdim_bins", _wrap_CorrelationDimensionFloat64_corrdim_bins, METH_O, "CorrelationDimensionFloat64_corrdim_bins(CorrelationDimensionFloat64 self)"},
	 { "CorrelationDimensionFloat64_corrdims", _wrap_CorrelationDimensionFloat64_corrdims, METH_O, "CorrelationDimensionFloat64_corrdims(CorrelationDimensionFloat64 self)"},
	 { "CorrelationDimensionFloat64_cumulative_vals_vars", _wrap_CorrelationDimensionFloat64_cumulative_vals_vars, METH_O, "CorrelationDimensionFloat64_cumulative_vals_vars(CorrelationDimensionFloat64 self)"},
	 { "delete_CorrelationDimensionFloat64", _wrap_delete_CorrelationDimensionFloat64, METH_O, "delete_CorrelationDimensionFloat64(CorrelationDimensionFloat64 self)"},
	 { "CorrelationDimensionFloat64_swigregister", CorrelationDimensionFloat64_swigregister, METH_O, NULL},
	 { "CorrelationDimensionFloat64_swiginit", CorrelationDimensionFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_EMDFloat64", (PyCFunction)(void(*)(void))_wrap_new_EMDFloat64, METH_VARARGS|METH_KEYWORDS, "new_EMDFloat64(double R=1, double beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1) -> EMDFloat64"},
	 { "delete_EMDFloat64", _wrap_delete_EMDFloat64, METH_O, "delete_EMDFloat64(EMDFloat64 self)"},
	 { "EMDFloat64_description", (PyCFunction)(void(*)(void))_wrap_EMDFloat64_description, METH_VARARGS|METH_KEYWORDS, "EMDFloat64_description(EMDFloat64 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDFloat64___repr__", _wrap_EMDFloat64___repr__, METH_O, "EMDFloat64___repr__(EMDFloat64 self) -> std::string"},
	 { "EMDFloat64_preprocess_CenterWeightedCentroid", _wrap_EMDFloat64_preprocess_CenterWeightedCentroid, METH_O, "EMDFloat64_preprocess_CenterWeightedCentroid(EMDFloat64 self)"},
	 { "EMDFloat64___call__", _wrap_EMDFloat64___call__, METH_VARARGS, "\n"
		"EMDFloat64___call__(EMDFloat64 self, double * weights0, double * coords0, double * weights1, double * coords1) -> double\n"
		"EMDFloat64___call__(EMDFloat64 self, double * weights0, double * weights1, double * external_dists) -> double\n"
		""},
	 { "EMDFloat64_swigregister", EMDFloat64_swigregister, METH_O, NULL},
	 { "EMDFloat64_swiginit", EMDFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_EMDFloat32", (PyCFunction)(void(*)(void))_wrap_new_EMDFloat32, METH_VARARGS|METH_KEYWORDS, "new_EMDFloat32(float R=1, float beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1) -> EMDFloat32"},
	 { "delete_EMDFloat32", _wrap_delete_EMDFloat32, METH_O, "delete_EMDFloat32(EMDFloat32 self)"},
	 { "EMDFloat32_description", (PyCFunction)(void(*)(void))_wrap_EMDFloat32_description, METH_VARARGS|METH_KEYWORDS, "EMDFloat32_description(EMDFloat32 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDFloat32___repr__", _wrap_EMDFloat32___repr__, METH_O, "EMDFloat32___repr__(EMDFloat32 self) -> std::string"},
	 { "EMDFloat32_preprocess_CenterWeightedCentroid", _wrap_EMDFloat32_preprocess_CenterWeightedCentroid, METH_O, "EMDFloat32_preprocess_CenterWeightedCentroid(EMDFloat32 self)"},
	 { "EMDFloat32___call__", _wrap_EMDFloat32___call__, METH_VARARGS, "\n"
		"EMDFloat32___call__(EMDFloat32 self, float * weights0, float * coords0, float * weights1, float * coords1) -> float\n"
		"EMDFloat32___call__(EMDFloat32 self, float * weights0, float * weights1, float * external_dists) -> float\n"
		""},
	 { "EMDFloat32_swigregister", EMDFloat32_swigregister, METH_O, NULL},
	 { "EMDFloat32_swiginit", EMDFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_EMDYPhiFloat64", (PyCFunction)(void(*)(void))_wrap_new_EMDYPhiFloat64, METH_VARARGS|METH_KEYWORDS, "new_EMDYPhiFloat64(double R=1, double beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1) -> EMDYPhiFloat64"},
	 { "delete_EMDYPhiFloat64", _wrap_delete_EMDYPhiFloat64, METH_O, "delete_EMDYPhiFloat64(EMDYPhiFloat64 self)"},
	 { "EMDYPhiFloat64_description", (PyCFunction)(void(*)(void))_wrap_EMDYPhiFloat64_description, METH_VARARGS|METH_KEYWORDS, "EMDYPhiFloat64_description(EMDYPhiFloat64 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDYPhiFloat64___repr__", _wrap_EMDYPhiFloat64___repr__, METH_O, "EMDYPhiFloat64___repr__(EMDYPhiFloat64 self) -> std::string"},
	 { "EMDYPhiFloat64_preprocess_CenterWeightedCentroid", _wrap_EMDYPhiFloat64_preprocess_CenterWeightedCentroid, METH_O, "EMDYPhiFloat64_preprocess_CenterWeightedCentroid(EMDYPhiFloat64 self)"},
	 { "EMDYPhiFloat64___call__", _wrap_EMDYPhiFloat64___call__, METH_VARARGS, "\n"
		"EMDYPhiFloat64___call__(EMDYPhiFloat64 self, double * weights0, double * coords0, double * weights1, double * coords1) -> double\n"
		"EMDYPhiFloat64___call__(EMDYPhiFloat64 self, double * weights0, double * weights1, double * external_dists) -> double\n"
		""},
	 { "EMDYPhiFloat64_swigregister", EMDYPhiFloat64_swigregister, METH_O, NULL},
	 { "EMDYPhiFloat64_swiginit", EMDYPhiFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_EMDYPhiFloat32", (PyCFunction)(void(*)(void))_wrap_new_EMDYPhiFloat32, METH_VARARGS|METH_KEYWORDS, "new_EMDYPhiFloat32(float R=1, float beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1) -> EMDYPhiFloat32"},
	 { "delete_EMDYPhiFloat32", _wrap_delete_EMDYPhiFloat32, METH_O, "delete_EMDYPhiFloat32(EMDYPhiFloat32 self)"},
	 { "EMDYPhiFloat32_description", (PyCFunction)(void(*)(void))_wrap_EMDYPhiFloat32_description, METH_VARARGS|METH_KEYWORDS, "EMDYPhiFloat32_description(EMDYPhiFloat32 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDYPhiFloat32___repr__", _wrap_EMDYPhiFloat32___repr__, METH_O, "EMDYPhiFloat32___repr__(EMDYPhiFloat32 self) -> std::string"},
	 { "EMDYPhiFloat32_preprocess_CenterWeightedCentroid", _wrap_EMDYPhiFloat32_preprocess_CenterWeightedCentroid, METH_O, "EMDYPhiFloat32_preprocess_CenterWeightedCentroid(EMDYPhiFloat32 self)"},
	 { "EMDYPhiFloat32___call__", _wrap_EMDYPhiFloat32___call__, METH_VARARGS, "\n"
		"EMDYPhiFloat32___call__(EMDYPhiFloat32 self, float * weights0, float * coords0, float * weights1, float * coords1) -> float\n"
		"EMDYPhiFloat32___call__(EMDYPhiFloat32 self, float * weights0, float * weights1, float * external_dists) -> float\n"
		""},
	 { "EMDYPhiFloat32_swigregister", EMDYPhiFloat32_swigregister, METH_O, NULL},
	 { "EMDYPhiFloat32_swiginit", EMDYPhiFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDFloat64", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDFloat64, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDFloat64(double R=1, double beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDFloat64"},
	 { "delete_PairwiseEMDFloat64", _wrap_delete_PairwiseEMDFloat64, METH_O, "delete_PairwiseEMDFloat64(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64_description", _wrap_PairwiseEMDFloat64_description, METH_O, "PairwiseEMDFloat64_description(PairwiseEMDFloat64 self) -> std::string"},
	 { "PairwiseEMDFloat64_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat64_clear, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDFloat64_clear(PairwiseEMDFloat64 self, bool free_memory=True)"},
	 { "PairwiseEMDFloat64_init", _wrap_PairwiseEMDFloat64_init, METH_VARARGS, "\n"
		"PairwiseEMDFloat64_init(PairwiseEMDFloat64 self, wasserstein::index_type nev)\n"
		"PairwiseEMDFloat64_init(PairwiseEMDFloat64 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDFloat64_compute", _wrap_PairwiseEMDFloat64_compute, METH_O, "PairwiseEMDFloat64_compute(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64___repr__", _wrap_PairwiseEMDFloat64___repr__, METH_O, "PairwiseEMDFloat64___repr__(PairwiseEMDFloat64 self) -> std::string"},
	 { "PairwiseEMDFloat64_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDFloat64_preprocess_CenterWeightedCentroid, METH_O, "PairwiseEMDFloat64_preprocess_CenterWeightedCentroid(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64__reset_B_events", _wrap_PairwiseEMDFloat64__reset_B_events, METH_O, "PairwiseEMDFloat64__reset_B_events(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat64__add_event, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDFloat64__add_event(PairwiseEMDFloat64 self, double * weights, double * coords, double event_weight=1)"},
	 { "PairwiseEMDFloat64_swigregister", PairwiseEMDFloat64_swigregister, METH_O, NULL},
	 { "PairwiseEMDFloat64_swiginit", PairwiseEMDFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDFloat32", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDFloat32, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDFloat32(float R=1, float beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDFloat32"},
	 { "delete_PairwiseEMDFloat32", _wrap_delete_PairwiseEMDFloat32, METH_O, "delete_PairwiseEMDFloat32(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32_description", _wrap_PairwiseEMDFloat32_description, METH_O, "PairwiseEMDFloat32_description(PairwiseEMDFloat32 self) -> std::string"},
	 { "PairwiseEMDFloat32_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat32_clear, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDFloat32_clear(PairwiseEMDFloat32 self, bool free_memory=True)"},
	 { "PairwiseEMDFloat32_init", _wrap_PairwiseEMDFloat32_init, METH_VARARGS, "\n"
		"PairwiseEMDFloat32_init(PairwiseEMDFloat32 self, wasserstein::index_type nev)\n"
		"PairwiseEMDFloat32_init(PairwiseEMDFloat32 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDFloat32_compute", _wrap_PairwiseEMDFloat32_compute, METH_O, "PairwiseEMDFloat32_compute(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32___repr__", _wrap_PairwiseEMDFloat32___repr__, METH_O, "PairwiseEMDFloat32___repr__(PairwiseEMDFloat32 self) -> std::string"},
	 { "PairwiseEMDFloat32_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDFloat32_preprocess_CenterWeightedCentroid, METH_O, "PairwiseEMDFloat32_preprocess_CenterWeightedCentroid(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32__reset_B_events", _wrap_PairwiseEMDFloat32__reset_B_events, METH_O, "PairwiseEMDFloat32__reset_B_events(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat32__add_event, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDFloat32__add_event(PairwiseEMDFloat32 self, float * weights, float * coords, float event_weight=1)"},
	 { "PairwiseEMDFloat32_swigregister", PairwiseEMDFloat32_swigregister, METH_O, NULL},
	 { "PairwiseEMDFloat32_swiginit", PairwiseEMDFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDYPhiFloat64", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDYPhiFloat64, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDYPhiFloat64(double R=1, double beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDYPhiFloat64"},
	 { "delete_PairwiseEMDYPhiFloat64", _wrap_delete_PairwiseEMDYPhiFloat64, METH_O, "delete_PairwiseEMDYPhiFloat64(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64_description", _wrap_PairwiseEMDYPhiFloat64_description, METH_O, "PairwiseEMDYPhiFloat64_description(PairwiseEMDYPhiFloat64 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat64_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat64_clear, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDYPhiFloat64_clear(PairwiseEMDYPhiFloat64 self, bool free_memory=True)"},
	 { "PairwiseEMDYPhiFloat64_init", _wrap_PairwiseEMDYPhiFloat64_init, METH_VARARGS, "\n"
		"PairwiseEMDYPhiFloat64_init(PairwiseEMDYPhiFloat64 self, wasserstein::index_type nev)\n"
		"PairwiseEMDYPhiFloat64_init(PairwiseEMDYPhiFloat64 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDYPhiFloat64_compute", _wrap_PairwiseEMDYPhiFloat64_compute, METH_O, "PairwiseEMDYPhiFloat64_compute(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64___repr__", _wrap_PairwiseEMDYPhiFloat64___repr__, METH_O, "PairwiseEMDYPhiFloat64___repr__(PairwiseEMDYPhiFloat64 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid, METH_O, "PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64__reset_B_events", _wrap_PairwiseEMDYPhiFloat64__reset_B_events, METH_O, "PairwiseEMDYPhiFloat64__reset_B_events(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat64__add_event, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDYPhiFloat64__add_event(PairwiseEMDYPhiFloat64 self, double * weights, double * coords, double event_weight=1)"},
	 { "PairwiseEMDYPhiFloat64_swigregister", PairwiseEMDYPhiFloat64_swigregister, METH_O, NULL},
	 { "PairwiseEMDYPhiFloat64_swiginit", PairwiseEMDYPhiFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDYPhiFloat32", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDYPhiFloat32, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDYPhiFloat32(float R=1, float beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDYPhiFloat32"},
	 { "delete_PairwiseEMDYPhiFloat32", _wrap_delete_PairwiseEMDYPhiFloat32, METH_O, "delete_PairwiseEMDYPhiFloat32(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32_description", _wrap_PairwiseEMDYPhiFloat32_description, METH_O, "PairwiseEMDYPhiFloat32_description(PairwiseEMDYPhiFloat32 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat32_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat32_clear, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDYPhiFloat32_clear(PairwiseEMDYPhiFloat32 self, bool free_memory=True)"},
	 { "PairwiseEMDYPhiFloat32_init", _wrap_PairwiseEMDYPhiFloat32_init, METH_VARARGS, "\n"
		"PairwiseEMDYPhiFloat32_init(PairwiseEMDYPhiFloat32 self, wasserstein::index_type nev)\n"
		"PairwiseEMDYPhiFloat32_init(PairwiseEMDYPhiFloat32 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDYPhiFloat32_compute", _wrap_PairwiseEMDYPhiFloat32_compute, METH_O, "PairwiseEMDYPhiFloat32_compute(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32___repr__", _wrap_PairwiseEMDYPhiFloat32___repr__, METH_O, "PairwiseEMDYPhiFloat32___repr__(PairwiseEMDYPhiFloat32 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid, METH_O, "PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32__reset_B_events", _wrap_PairwiseEMDYPhiFloat32__reset_B_events, METH_O, "PairwiseEMDYPhiFloat32__reset_B_events(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat32__add_event, METH_VARARGS|METH_KEYWORDS, "PairwiseEMDYPhiFloat32__add_event(PairwiseEMDYPhiFloat32 self, float * weights, float * coords, float event_weight=1)"},
	 { "PairwiseEMDYPhiFloat32_swigregister", PairwiseEMDYPhiFloat32_swigregister, METH_O, NULL},
	 { "PairwiseEMDYPhiFloat32_swiginit", PairwiseEMDYPhiFloat32_swiginit, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};

static PyMethodDef SwigMethods_proxydocs[] = {
	 { "SWIG_PyInstanceMethod_New", SWIG_PyInstanceMethod_New, METH_O, NULL},
	 { "SWIG_PyStaticMethod_New", SWIG_PyStaticMethod_New, METH_O, NULL},
	 { "delete_SwigPyIterator", _wrap_delete_SwigPyIterator, METH_O, "delete_SwigPyIterator(SwigPyIterator self)"},
	 { "SwigPyIterator_value", _wrap_SwigPyIterator_value, METH_O, "value(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator_incr", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_incr, METH_VARARGS|METH_KEYWORDS, "incr(SwigPyIterator self, size_t n=1) -> SwigPyIterator"},
	 { "SwigPyIterator_decr", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_decr, METH_VARARGS|METH_KEYWORDS, "decr(SwigPyIterator self, size_t n=1) -> SwigPyIterator"},
	 { "SwigPyIterator_distance", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_distance, METH_VARARGS|METH_KEYWORDS, "distance(SwigPyIterator self, SwigPyIterator x) -> ptrdiff_t"},
	 { "SwigPyIterator_equal", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_equal, METH_VARARGS|METH_KEYWORDS, "equal(SwigPyIterator self, SwigPyIterator x) -> bool"},
	 { "SwigPyIterator_copy", _wrap_SwigPyIterator_copy, METH_O, "copy(SwigPyIterator self) -> SwigPyIterator"},
	 { "SwigPyIterator_next", _wrap_SwigPyIterator_next, METH_O, "next(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator___next__", _wrap_SwigPyIterator___next__, METH_O, "__next__(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator_previous", _wrap_SwigPyIterator_previous, METH_O, "previous(SwigPyIterator self) -> PyObject *"},
	 { "SwigPyIterator_advance", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator_advance, METH_VARARGS|METH_KEYWORDS, "advance(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___eq__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___eq__, METH_VARARGS|METH_KEYWORDS, "__eq__(SwigPyIterator self, SwigPyIterator x) -> bool"},
	 { "SwigPyIterator___ne__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___ne__, METH_VARARGS|METH_KEYWORDS, "__ne__(SwigPyIterator self, SwigPyIterator x) -> bool"},
	 { "SwigPyIterator___iadd__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___iadd__, METH_VARARGS|METH_KEYWORDS, "__iadd__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___isub__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___isub__, METH_VARARGS|METH_KEYWORDS, "__isub__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___add__", (PyCFunction)(void(*)(void))_wrap_SwigPyIterator___add__, METH_VARARGS|METH_KEYWORDS, "__add__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator"},
	 { "SwigPyIterator___sub__", _wrap_SwigPyIterator___sub__, METH_VARARGS, "\n"
		"__sub__(SwigPyIterator self, ptrdiff_t n) -> SwigPyIterator\n"
		"__sub__(SwigPyIterator self, SwigPyIterator x) -> ptrdiff_t\n"
		""},
	 { "SwigPyIterator_swigregister", SwigPyIterator_swigregister, METH_O, NULL},
	 { "vectorDouble_iterator", _wrap_vectorDouble_iterator, METH_O, "iterator(vectorDouble self) -> SwigPyIterator"},
	 { "vectorDouble___nonzero__", _wrap_vectorDouble___nonzero__, METH_O, "__nonzero__(vectorDouble self) -> bool"},
	 { "vectorDouble___bool__", _wrap_vectorDouble___bool__, METH_O, "__bool__(vectorDouble self) -> bool"},
	 { "vectorDouble___len__", _wrap_vectorDouble___len__, METH_O, "__len__(vectorDouble self) -> std::vector< double >::size_type"},
	 { "vectorDouble___getslice__", (PyCFunction)(void(*)(void))_wrap_vectorDouble___getslice__, METH_VARARGS|METH_KEYWORDS, "__getslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j) -> vectorDouble"},
	 { "vectorDouble___setslice__", _wrap_vectorDouble___setslice__, METH_VARARGS, "\n"
		"__setslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j)\n"
		"__setslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j, vectorDouble v)\n"
		""},
	 { "vectorDouble___delslice__", (PyCFunction)(void(*)(void))_wrap_vectorDouble___delslice__, METH_VARARGS|METH_KEYWORDS, "__delslice__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::difference_type j)"},
	 { "vectorDouble___delitem__", _wrap_vectorDouble___delitem__, METH_VARARGS, "\n"
		"__delitem__(vectorDouble self, std::vector< double >::difference_type i)\n"
		"__delitem__(vectorDouble self, PySliceObject * slice)\n"
		""},
	 { "vectorDouble___getitem__", _wrap_vectorDouble___getitem__, METH_VARARGS, "\n"
		"__getitem__(vectorDouble self, PySliceObject * slice) -> vectorDouble\n"
		"__getitem__(vectorDouble self, std::vector< double >::difference_type i) -> std::vector< double >::value_type const &\n"
		""},
	 { "vectorDouble___setitem__", _wrap_vectorDouble___setitem__, METH_VARARGS, "\n"
		"__setitem__(vectorDouble self, PySliceObject * slice, vectorDouble v)\n"
		"__setitem__(vectorDouble self, PySliceObject * slice)\n"
		"__setitem__(vectorDouble self, std::vector< double >::difference_type i, std::vector< double >::value_type const & x)\n"
		""},
	 { "vectorDouble_pop", _wrap_vectorDouble_pop, METH_O, "pop(vectorDouble self) -> std::vector< double >::value_type"},
	 { "vectorDouble_append", (PyCFunction)(void(*)(void))_wrap_vectorDouble_append, METH_VARARGS|METH_KEYWORDS, "append(vectorDouble self, std::vector< double >::value_type const & x)"},
	 { "vectorDouble_empty", _wrap_vectorDouble_empty, METH_O, "empty(vectorDouble self) -> bool"},
	 { "vectorDouble_size", _wrap_vectorDouble_size, METH_O, "size(vectorDouble self) -> std::vector< double >::size_type"},
	 { "vectorDouble_swap", (PyCFunction)(void(*)(void))_wrap_vectorDouble_swap, METH_VARARGS|METH_KEYWORDS, "swap(vectorDouble self, vectorDouble v)"},
	 { "vectorDouble_begin", _wrap_vectorDouble_begin, METH_O, "begin(vectorDouble self) -> std::vector< double >::iterator"},
	 { "vectorDouble_end", _wrap_vectorDouble_end, METH_O, "end(vectorDouble self) -> std::vector< double >::iterator"},
	 { "vectorDouble_rbegin", _wrap_vectorDouble_rbegin, METH_O, "rbegin(vectorDouble self) -> std::vector< double >::reverse_iterator"},
	 { "vectorDouble_rend", _wrap_vectorDouble_rend, METH_O, "rend(vectorDouble self) -> std::vector< double >::reverse_iterator"},
	 { "vectorDouble_clear", _wrap_vectorDouble_clear, METH_O, "clear(vectorDouble self)"},
	 { "vectorDouble_get_allocator", _wrap_vectorDouble_get_allocator, METH_O, "get_allocator(vectorDouble self) -> std::vector< double >::allocator_type"},
	 { "vectorDouble_pop_back", _wrap_vectorDouble_pop_back, METH_O, "pop_back(vectorDouble self)"},
	 { "vectorDouble_erase", _wrap_vectorDouble_erase, METH_VARARGS, "\n"
		"erase(vectorDouble self, std::vector< double >::iterator pos) -> std::vector< double >::iterator\n"
		"erase(vectorDouble self, std::vector< double >::iterator first, std::vector< double >::iterator last) -> std::vector< double >::iterator\n"
		""},
	 { "new_vectorDouble", _wrap_new_vectorDouble, METH_VARARGS, "\n"
		"vectorDouble()\n"
		"vectorDouble(vectorDouble other)\n"
		"vectorDouble(std::vector< double >::size_type size)\n"
		"new_vectorDouble(std::vector< double >::size_type size, std::vector< double >::value_type const & value) -> vectorDouble\n"
		""},
	 { "vectorDouble_push_back", (PyCFunction)(void(*)(void))_wrap_vectorDouble_push_back, METH_VARARGS|METH_KEYWORDS, "push_back(vectorDouble self, std::vector< double >::value_type const & x)"},
	 { "vectorDouble_front", _wrap_vectorDouble_front, METH_O, "front(vectorDouble self) -> std::vector< double >::value_type const &"},
	 { "vectorDouble_back", _wrap_vectorDouble_back, METH_O, "back(vectorDouble self) -> std::vector< double >::value_type const &"},
	 { "vectorDouble_assign", (PyCFunction)(void(*)(void))_wrap_vectorDouble_assign, METH_VARARGS|METH_KEYWORDS, "assign(vectorDouble self, std::vector< double >::size_type n, std::vector< double >::value_type const & x)"},
	 { "vectorDouble_resize", _wrap_vectorDouble_resize, METH_VARARGS, "\n"
		"resize(vectorDouble self, std::vector< double >::size_type new_size)\n"
		"resize(vectorDouble self, std::vector< double >::size_type new_size, std::vector< double >::value_type const & x)\n"
		""},
	 { "vectorDouble_insert", _wrap_vectorDouble_insert, METH_VARARGS, "\n"
		"insert(vectorDouble self, std::vector< double >::iterator pos, std::vector< double >::value_type const & x) -> std::vector< double >::iterator\n"
		"insert(vectorDouble self, std::vector< double >::iterator pos, std::vector< double >::size_type n, std::vector< double >::value_type const & x)\n"
		""},
	 { "vectorDouble_reserve", (PyCFunction)(void(*)(void))_wrap_vectorDouble_reserve, METH_VARARGS|METH_KEYWORDS, "reserve(vectorDouble self, std::vector< double >::size_type n)"},
	 { "vectorDouble_capacity", _wrap_vectorDouble_capacity, METH_O, "capacity(vectorDouble self) -> std::vector< double >::size_type"},
	 { "delete_vectorDouble", _wrap_delete_vectorDouble, METH_O, "delete_vectorDouble(vectorDouble self)"},
	 { "vectorDouble_swigregister", vectorDouble_swigregister, METH_O, NULL},
	 { "vectorDouble_swiginit", vectorDouble_swiginit, METH_VARARGS, NULL},
	 { "vectorString_iterator", _wrap_vectorString_iterator, METH_O, "iterator(vectorString self) -> SwigPyIterator"},
	 { "vectorString___nonzero__", _wrap_vectorString___nonzero__, METH_O, "__nonzero__(vectorString self) -> bool"},
	 { "vectorString___bool__", _wrap_vectorString___bool__, METH_O, "__bool__(vectorString self) -> bool"},
	 { "vectorString___len__", _wrap_vectorString___len__, METH_O, "__len__(vectorString self) -> std::vector< std::string >::size_type"},
	 { "vectorString___getslice__", (PyCFunction)(void(*)(void))_wrap_vectorString___getslice__, METH_VARARGS|METH_KEYWORDS, "__getslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j) -> vectorString"},
	 { "vectorString___setslice__", _wrap_vectorString___setslice__, METH_VARARGS, "\n"
		"__setslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j)\n"
		"__setslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j, vectorString v)\n"
		""},
	 { "vectorString___delslice__", (PyCFunction)(void(*)(void))_wrap_vectorString___delslice__, METH_VARARGS|METH_KEYWORDS, "__delslice__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::difference_type j)"},
	 { "vectorString___delitem__", _wrap_vectorString___delitem__, METH_VARARGS, "\n"
		"__delitem__(vectorString self, std::vector< std::string >::difference_type i)\n"
		"__delitem__(vectorString self, PySliceObject * slice)\n"
		""},
	 { "vectorString___getitem__", _wrap_vectorString___getitem__, METH_VARARGS, "\n"
		"__getitem__(vectorString self, PySliceObject * slice) -> vectorString\n"
		"__getitem__(vectorString self, std::vector< std::string >::difference_type i) -> std::vector< std::string >::value_type const &\n"
		""},
	 { "vectorString___setitem__", _wrap_vectorString___setitem__, METH_VARARGS, "\n"
		"__setitem__(vectorString self, PySliceObject * slice, vectorString v)\n"
		"__setitem__(vectorString self, PySliceObject * slice)\n"
		"__setitem__(vectorString self, std::vector< std::string >::difference_type i, std::vector< std::string >::value_type const & x)\n"
		""},
	 { "vectorString_pop", _wrap_vectorString_pop, METH_O, "pop(vectorString self) -> std::vector< std::string >::value_type"},
	 { "vectorString_append", (PyCFunction)(void(*)(void))_wrap_vectorString_append, METH_VARARGS|METH_KEYWORDS, "append(vectorString self, std::vector< std::string >::value_type const & x)"},
	 { "vectorString_empty", _wrap_vectorString_empty, METH_O, "empty(vectorString self) -> bool"},
	 { "vectorString_size", _wrap_vectorString_size, METH_O, "size(vectorString self) -> std::vector< std::string >::size_type"},
	 { "vectorString_swap", (PyCFunction)(void(*)(void))_wrap_vectorString_swap, METH_VARARGS|METH_KEYWORDS, "swap(vectorString self, vectorString v)"},
	 { "vectorString_begin", _wrap_vectorString_begin, METH_O, "begin(vectorString self) -> std::vector< std::string >::iterator"},
	 { "vectorString_end", _wrap_vectorString_end, METH_O, "end(vectorString self) -> std::vector< std::string >::iterator"},
	 { "vectorString_rbegin", _wrap_vectorString_rbegin, METH_O, "rbegin(vectorString self) -> std::vector< std::string >::reverse_iterator"},
	 { "vectorString_rend", _wrap_vectorString_rend, METH_O, "rend(vectorString self) -> std::vector< std::string >::reverse_iterator"},
	 { "vectorString_clear", _wrap_vectorString_clear, METH_O, "clear(vectorString self)"},
	 { "vectorString_get_allocator", _wrap_vectorString_get_allocator, METH_O, "get_allocator(vectorString self) -> std::vector< std::string >::allocator_type"},
	 { "vectorString_pop_back", _wrap_vectorString_pop_back, METH_O, "pop_back(vectorString self)"},
	 { "vectorString_erase", _wrap_vectorString_erase, METH_VARARGS, "\n"
		"erase(vectorString self, std::vector< std::string >::iterator pos) -> std::vector< std::string >::iterator\n"
		"erase(vectorString self, std::vector< std::string >::iterator first, std::vector< std::string >::iterator last) -> std::vector< std::string >::iterator\n"
		""},
	 { "new_vectorString", _wrap_new_vectorString, METH_VARARGS, "\n"
		"vectorString()\n"
		"vectorString(vectorString other)\n"
		"vectorString(std::vector< std::string >::size_type size)\n"
		"new_vectorString(std::vector< std::string >::size_type size, std::vector< std::string >::value_type const & value) -> vectorString\n"
		""},
	 { "vectorString_push_back", (PyCFunction)(void(*)(void))_wrap_vectorString_push_back, METH_VARARGS|METH_KEYWORDS, "push_back(vectorString self, std::vector< std::string >::value_type const & x)"},
	 { "vectorString_front", _wrap_vectorString_front, METH_O, "front(vectorString self) -> std::vector< std::string >::value_type const &"},
	 { "vectorString_back", _wrap_vectorString_back, METH_O, "back(vectorString self) -> std::vector< std::string >::value_type const &"},
	 { "vectorString_assign", (PyCFunction)(void(*)(void))_wrap_vectorString_assign, METH_VARARGS|METH_KEYWORDS, "assign(vectorString self, std::vector< std::string >::size_type n, std::vector< std::string >::value_type const & x)"},
	 { "vectorString_resize", _wrap_vectorString_resize, METH_VARARGS, "\n"
		"resize(vectorString self, std::vector< std::string >::size_type new_size)\n"
		"resize(vectorString self, std::vector< std::string >::size_type new_size, std::vector< std::string >::value_type const & x)\n"
		""},
	 { "vectorString_insert", _wrap_vectorString_insert, METH_VARARGS, "\n"
		"insert(vectorString self, std::vector< std::string >::iterator pos, std::vector< std::string >::value_type const & x) -> std::vector< std::string >::iterator\n"
		"insert(vectorString self, std::vector< std::string >::iterator pos, std::vector< std::string >::size_type n, std::vector< std::string >::value_type const & x)\n"
		""},
	 { "vectorString_reserve", (PyCFunction)(void(*)(void))_wrap_vectorString_reserve, METH_VARARGS|METH_KEYWORDS, "reserve(vectorString self, std::vector< std::string >::size_type n)"},
	 { "vectorString_capacity", _wrap_vectorString_capacity, METH_O, "capacity(vectorString self) -> std::vector< std::string >::size_type"},
	 { "delete_vectorString", _wrap_delete_vectorString, METH_O, "delete_vectorString(vectorString self)"},
	 { "vectorString_swigregister", vectorString_swigregister, METH_O, NULL},
	 { "vectorString_swiginit", vectorString_swiginit, METH_VARARGS, NULL},
	 { "new_pairVectorDouble", _wrap_new_pairVectorDouble, METH_VARARGS, "\n"
		"pairVectorDouble()\n"
		"pairVectorDouble(vectorDouble first, vectorDouble second)\n"
		"new_pairVectorDouble(pairVectorDouble other) -> pairVectorDouble\n"
		""},
	 { "pairVectorDouble_first_set", _wrap_pairVectorDouble_first_set, METH_VARARGS, "pairVectorDouble_first_set(pairVectorDouble self, vectorDouble first)"},
	 { "pairVectorDouble_first_get", _wrap_pairVectorDouble_first_get, METH_O, "pairVectorDouble_first_get(pairVectorDouble self) -> vectorDouble"},
	 { "pairVectorDouble_second_set", _wrap_pairVectorDouble_second_set, METH_VARARGS, "pairVectorDouble_second_set(pairVectorDouble self, vectorDouble second)"},
	 { "pairVectorDouble_second_get", _wrap_pairVectorDouble_second_get, METH_O, "pairVectorDouble_second_get(pairVectorDouble self) -> vectorDouble"},
	 { "delete_pairVectorDouble", _wrap_delete_pairVectorDouble, METH_O, "delete_pairVectorDouble(pairVectorDouble self)"},
	 { "pairVectorDouble_swigregister", pairVectorDouble_swigregister, METH_O, NULL},
	 { "pairVectorDouble_swiginit", pairVectorDouble_swiginit, METH_VARARGS, NULL},
	 { "vectorFloat_iterator", _wrap_vectorFloat_iterator, METH_O, "iterator(vectorFloat self) -> SwigPyIterator"},
	 { "vectorFloat___nonzero__", _wrap_vectorFloat___nonzero__, METH_O, "__nonzero__(vectorFloat self) -> bool"},
	 { "vectorFloat___bool__", _wrap_vectorFloat___bool__, METH_O, "__bool__(vectorFloat self) -> bool"},
	 { "vectorFloat___len__", _wrap_vectorFloat___len__, METH_O, "__len__(vectorFloat self) -> std::vector< float >::size_type"},
	 { "vectorFloat___getslice__", (PyCFunction)(void(*)(void))_wrap_vectorFloat___getslice__, METH_VARARGS|METH_KEYWORDS, "__getslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j) -> vectorFloat"},
	 { "vectorFloat___setslice__", _wrap_vectorFloat___setslice__, METH_VARARGS, "\n"
		"__setslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j)\n"
		"__setslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j, vectorFloat v)\n"
		""},
	 { "vectorFloat___delslice__", (PyCFunction)(void(*)(void))_wrap_vectorFloat___delslice__, METH_VARARGS|METH_KEYWORDS, "__delslice__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::difference_type j)"},
	 { "vectorFloat___delitem__", _wrap_vectorFloat___delitem__, METH_VARARGS, "\n"
		"__delitem__(vectorFloat self, std::vector< float >::difference_type i)\n"
		"__delitem__(vectorFloat self, PySliceObject * slice)\n"
		""},
	 { "vectorFloat___getitem__", _wrap_vectorFloat___getitem__, METH_VARARGS, "\n"
		"__getitem__(vectorFloat self, PySliceObject * slice) -> vectorFloat\n"
		"__getitem__(vectorFloat self, std::vector< float >::difference_type i) -> std::vector< float >::value_type const &\n"
		""},
	 { "vectorFloat___setitem__", _wrap_vectorFloat___setitem__, METH_VARARGS, "\n"
		"__setitem__(vectorFloat self, PySliceObject * slice, vectorFloat v)\n"
		"__setitem__(vectorFloat self, PySliceObject * slice)\n"
		"__setitem__(vectorFloat self, std::vector< float >::difference_type i, std::vector< float >::value_type const & x)\n"
		""},
	 { "vectorFloat_pop", _wrap_vectorFloat_pop, METH_O, "pop(vectorFloat self) -> std::vector< float >::value_type"},
	 { "vectorFloat_append", (PyCFunction)(void(*)(void))_wrap_vectorFloat_append, METH_VARARGS|METH_KEYWORDS, "append(vectorFloat self, std::vector< float >::value_type const & x)"},
	 { "vectorFloat_empty", _wrap_vectorFloat_empty, METH_O, "empty(vectorFloat self) -> bool"},
	 { "vectorFloat_size", _wrap_vectorFloat_size, METH_O, "size(vectorFloat self) -> std::vector< float >::size_type"},
	 { "vectorFloat_swap", (PyCFunction)(void(*)(void))_wrap_vectorFloat_swap, METH_VARARGS|METH_KEYWORDS, "swap(vectorFloat self, vectorFloat v)"},
	 { "vectorFloat_begin", _wrap_vectorFloat_begin, METH_O, "begin(vectorFloat self) -> std::vector< float >::iterator"},
	 { "vectorFloat_end", _wrap_vectorFloat_end, METH_O, "end(vectorFloat self) -> std::vector< float >::iterator"},
	 { "vectorFloat_rbegin", _wrap_vectorFloat_rbegin, METH_O, "rbegin(vectorFloat self) -> std::vector< float >::reverse_iterator"},
	 { "vectorFloat_rend", _wrap_vectorFloat_rend, METH_O, "rend(vectorFloat self) -> std::vector< float >::reverse_iterator"},
	 { "vectorFloat_clear", _wrap_vectorFloat_clear, METH_O, "clear(vectorFloat self)"},
	 { "vectorFloat_get_allocator", _wrap_vectorFloat_get_allocator, METH_O, "get_allocator(vectorFloat self) -> std::vector< float >::allocator_type"},
	 { "vectorFloat_pop_back", _wrap_vectorFloat_pop_back, METH_O, "pop_back(vectorFloat self)"},
	 { "vectorFloat_erase", _wrap_vectorFloat_erase, METH_VARARGS, "\n"
		"erase(vectorFloat self, std::vector< float >::iterator pos) -> std::vector< float >::iterator\n"
		"erase(vectorFloat self, std::vector< float >::iterator first, std::vector< float >::iterator last) -> std::vector< float >::iterator\n"
		""},
	 { "new_vectorFloat", _wrap_new_vectorFloat, METH_VARARGS, "\n"
		"vectorFloat()\n"
		"vectorFloat(vectorFloat other)\n"
		"vectorFloat(std::vector< float >::size_type size)\n"
		"new_vectorFloat(std::vector< float >::size_type size, std::vector< float >::value_type const & value) -> vectorFloat\n"
		""},
	 { "vectorFloat_push_back", (PyCFunction)(void(*)(void))_wrap_vectorFloat_push_back, METH_VARARGS|METH_KEYWORDS, "push_back(vectorFloat self, std::vector< float >::value_type const & x)"},
	 { "vectorFloat_front", _wrap_vectorFloat_front, METH_O, "front(vectorFloat self) -> std::vector< float >::value_type const &"},
	 { "vectorFloat_back", _wrap_vectorFloat_back, METH_O, "back(vectorFloat self) -> std::vector< float >::value_type const &"},
	 { "vectorFloat_assign", (PyCFunction)(void(*)(void))_wrap_vectorFloat_assign, METH_VARARGS|METH_KEYWORDS, "assign(vectorFloat self, std::vector< float >::size_type n, std::vector< float >::value_type const & x)"},
	 { "vectorFloat_resize", _wrap_vectorFloat_resize, METH_VARARGS, "\n"
		"resize(vectorFloat self, std::vector< float >::size_type new_size)\n"
		"resize(vectorFloat self, std::vector< float >::size_type new_size, std::vector< float >::value_type const & x)\n"
		""},
	 { "vectorFloat_insert", _wrap_vectorFloat_insert, METH_VARARGS, "\n"
		"insert(vectorFloat self, std::vector< float >::iterator pos, std::vector< float >::value_type const & x) -> std::vector< float >::iterator\n"
		"insert(vectorFloat self, std::vector< float >::iterator pos, std::vector< float >::size_type n, std::vector< float >::value_type const & x)\n"
		""},
	 { "vectorFloat_reserve", (PyCFunction)(void(*)(void))_wrap_vectorFloat_reserve, METH_VARARGS|METH_KEYWORDS, "reserve(vectorFloat self, std::vector< float >::size_type n)"},
	 { "vectorFloat_capacity", _wrap_vectorFloat_capacity, METH_O, "capacity(vectorFloat self) -> std::vector< float >::size_type"},
	 { "delete_vectorFloat", _wrap_delete_vectorFloat, METH_O, "delete_vectorFloat(vectorFloat self)"},
	 { "vectorFloat_swigregister", vectorFloat_swigregister, METH_O, NULL},
	 { "vectorFloat_swiginit", vectorFloat_swiginit, METH_VARARGS, NULL},
	 { "new_pairVectorFloat", _wrap_new_pairVectorFloat, METH_VARARGS, "\n"
		"pairVectorFloat()\n"
		"pairVectorFloat(vectorFloat first, vectorFloat second)\n"
		"new_pairVectorFloat(pairVectorFloat other) -> pairVectorFloat\n"
		""},
	 { "pairVectorFloat_first_set", _wrap_pairVectorFloat_first_set, METH_VARARGS, "pairVectorFloat_first_set(pairVectorFloat self, vectorFloat first)"},
	 { "pairVectorFloat_first_get", _wrap_pairVectorFloat_first_get, METH_O, "pairVectorFloat_first_get(pairVectorFloat self) -> vectorFloat"},
	 { "pairVectorFloat_second_set", _wrap_pairVectorFloat_second_set, METH_VARARGS, "pairVectorFloat_second_set(pairVectorFloat self, vectorFloat second)"},
	 { "pairVectorFloat_second_get", _wrap_pairVectorFloat_second_get, METH_O, "pairVectorFloat_second_get(pairVectorFloat self) -> vectorFloat"},
	 { "delete_pairVectorFloat", _wrap_delete_pairVectorFloat, METH_O, "delete_pairVectorFloat(pairVectorFloat self)"},
	 { "pairVectorFloat_swigregister", pairVectorFloat_swigregister, METH_O, NULL},
	 { "pairVectorFloat_swiginit", pairVectorFloat_swiginit, METH_VARARGS, NULL},
	 { "check_emd_status", (PyCFunction)(void(*)(void))_wrap_check_emd_status, METH_VARARGS|METH_KEYWORDS, "check_emd_status(wasserstein::EMDStatus status)"},
	 { "delete_EMDBaseFloat64", _wrap_delete_EMDBaseFloat64, METH_O, "delete_EMDBaseFloat64(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_R", _wrap_EMDBaseFloat64_R, METH_O, "R(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_beta", _wrap_EMDBaseFloat64_beta, METH_O, "beta(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_set_R", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_R, METH_VARARGS|METH_KEYWORDS, "set_R(EMDBaseFloat64 self, double R)"},
	 { "EMDBaseFloat64_set_beta", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_beta, METH_VARARGS|METH_KEYWORDS, "set_beta(EMDBaseFloat64 self, double beta)"},
	 { "EMDBaseFloat64_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "set_network_simplex_params(EMDBaseFloat64 self, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1)"},
	 { "EMDBaseFloat64_norm", _wrap_EMDBaseFloat64_norm, METH_O, "norm(EMDBaseFloat64 self) -> bool"},
	 { "EMDBaseFloat64_set_norm", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_norm, METH_VARARGS|METH_KEYWORDS, "set_norm(EMDBaseFloat64 self, bool norm)"},
	 { "EMDBaseFloat64_do_timing", _wrap_EMDBaseFloat64_do_timing, METH_O, "do_timing(EMDBaseFloat64 self) -> bool"},
	 { "EMDBaseFloat64_set_do_timing", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_do_timing, METH_VARARGS|METH_KEYWORDS, "set_do_timing(EMDBaseFloat64 self, bool timing)"},
	 { "EMDBaseFloat64_external_dists", _wrap_EMDBaseFloat64_external_dists, METH_O, "external_dists(EMDBaseFloat64 self) -> bool"},
	 { "EMDBaseFloat64_set_external_dists", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat64_set_external_dists, METH_VARARGS|METH_KEYWORDS, "set_external_dists(EMDBaseFloat64 self, bool exdists)"},
	 { "EMDBaseFloat64_n0", _wrap_EMDBaseFloat64_n0, METH_O, "n0(EMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat64_n1", _wrap_EMDBaseFloat64_n1, METH_O, "n1(EMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat64_extra", _wrap_EMDBaseFloat64_extra, METH_O, "extra(EMDBaseFloat64 self) -> wasserstein::ExtraParticle"},
	 { "EMDBaseFloat64_emd", _wrap_EMDBaseFloat64_emd, METH_O, "emd(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_status", _wrap_EMDBaseFloat64_status, METH_O, "status(EMDBaseFloat64 self) -> wasserstein::EMDStatus"},
	 { "EMDBaseFloat64_weightdiff", _wrap_EMDBaseFloat64_weightdiff, METH_O, "weightdiff(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_scale", _wrap_EMDBaseFloat64_scale, METH_O, "scale(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_n_iter", _wrap_EMDBaseFloat64_n_iter, METH_O, "n_iter(EMDBaseFloat64 self) -> std::size_t"},
	 { "EMDBaseFloat64_dists_vec", _wrap_EMDBaseFloat64_dists_vec, METH_O, "dists_vec(EMDBaseFloat64 self) -> vectorDouble"},
	 { "EMDBaseFloat64_flows_vec", _wrap_EMDBaseFloat64_flows_vec, METH_O, "flows_vec(EMDBaseFloat64 self) -> vectorDouble"},
	 { "EMDBaseFloat64_flow", _wrap_EMDBaseFloat64_flow, METH_VARARGS, "\n"
		"flow(EMDBaseFloat64 self, wasserstein::index_type i, wasserstein::index_type j) -> double\n"
		"flow(EMDBaseFloat64 self, std::size_t ind) -> double\n"
		""},
	 { "EMDBaseFloat64_node_potentials", _wrap_EMDBaseFloat64_node_potentials, METH_O, "node_potentials(EMDBaseFloat64 self) -> pairVectorDouble"},
	 { "EMDBaseFloat64_duration", _wrap_EMDBaseFloat64_duration, METH_O, "duration(EMDBaseFloat64 self) -> double"},
	 { "EMDBaseFloat64_clear", _wrap_EMDBaseFloat64_clear, METH_O, "clear(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_flows", _wrap_EMDBaseFloat64_flows, METH_O, "flows(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_dists", _wrap_EMDBaseFloat64_dists, METH_O, "dists(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_npy_node_potentials", _wrap_EMDBaseFloat64_npy_node_potentials, METH_O, "npy_node_potentials(EMDBaseFloat64 self)"},
	 { "EMDBaseFloat64_swigregister", EMDBaseFloat64_swigregister, METH_O, NULL},
	 { "delete_PairwiseEMDBaseFloat64", _wrap_delete_PairwiseEMDBaseFloat64, METH_O, "delete_PairwiseEMDBaseFloat64(PairwiseEMDBaseFloat64 self)"},
	 { "PairwiseEMDBaseFloat64_R", _wrap_PairwiseEMDBaseFloat64_R, METH_O, "R(PairwiseEMDBaseFloat64 self) -> double"},
	 { "PairwiseEMDBaseFloat64_set_R", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_R, METH_VARARGS|METH_KEYWORDS, "set_R(PairwiseEMDBaseFloat64 self, double R)"},
	 { "PairwiseEMDBaseFloat64_beta", _wrap_PairwiseEMDBaseFloat64_beta, METH_O, "beta(PairwiseEMDBaseFloat64 self) -> double"},
	 { "PairwiseEMDBaseFloat64_set_beta", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_beta, METH_VARARGS|METH_KEYWORDS, "set_beta(PairwiseEMDBaseFloat64 self, double beta)"},
	 { "PairwiseEMDBaseFloat64_norm", _wrap_PairwiseEMDBaseFloat64_norm, METH_O, "norm(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_set_norm", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_norm, METH_VARARGS|METH_KEYWORDS, "set_norm(PairwiseEMDBaseFloat64 self, bool norm)"},
	 { "PairwiseEMDBaseFloat64_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "set_network_simplex_params(PairwiseEMDBaseFloat64 self, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1)"},
	 { "PairwiseEMDBaseFloat64_set_external_emd_handler", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_external_emd_handler, METH_VARARGS|METH_KEYWORDS, "set_external_emd_handler(PairwiseEMDBaseFloat64 self, ExternalEMDHandlerFloat64 handler)"},
	 { "PairwiseEMDBaseFloat64_have_external_emd_handler", _wrap_PairwiseEMDBaseFloat64_have_external_emd_handler, METH_O, "have_external_emd_handler(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_num_threads", _wrap_PairwiseEMDBaseFloat64_num_threads, METH_O, "num_threads(PairwiseEMDBaseFloat64 self) -> int"},
	 { "PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_omp_dynamic_chunksize, METH_VARARGS|METH_KEYWORDS, "set_omp_dynamic_chunksize(PairwiseEMDBaseFloat64 self, int chunksize)"},
	 { "PairwiseEMDBaseFloat64_omp_dynamic_chunksize", _wrap_PairwiseEMDBaseFloat64_omp_dynamic_chunksize, METH_O, "omp_dynamic_chunksize(PairwiseEMDBaseFloat64 self) -> int"},
	 { "PairwiseEMDBaseFloat64_set_request_mode", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_set_request_mode, METH_VARARGS|METH_KEYWORDS, "set_request_mode(PairwiseEMDBaseFloat64 self, bool mode)"},
	 { "PairwiseEMDBaseFloat64_request_mode", _wrap_PairwiseEMDBaseFloat64_request_mode, METH_O, "request_mode(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_duration", _wrap_PairwiseEMDBaseFloat64_duration, METH_O, "duration(PairwiseEMDBaseFloat64 self) -> double"},
	 { "PairwiseEMDBaseFloat64_nevA", _wrap_PairwiseEMDBaseFloat64_nevA, METH_O, "nevA(PairwiseEMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat64_nevB", _wrap_PairwiseEMDBaseFloat64_nevB, METH_O, "nevB(PairwiseEMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat64_num_emds", _wrap_PairwiseEMDBaseFloat64_num_emds, METH_O, "num_emds(PairwiseEMDBaseFloat64 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat64_storage", _wrap_PairwiseEMDBaseFloat64_storage, METH_O, "storage(PairwiseEMDBaseFloat64 self) -> wasserstein::EMDPairsStorage"},
	 { "PairwiseEMDBaseFloat64_errored", _wrap_PairwiseEMDBaseFloat64_errored, METH_O, "errored(PairwiseEMDBaseFloat64 self) -> bool"},
	 { "PairwiseEMDBaseFloat64_error_messages", _wrap_PairwiseEMDBaseFloat64_error_messages, METH_O, "error_messages(PairwiseEMDBaseFloat64 self) -> vectorString"},
	 { "PairwiseEMDBaseFloat64_emds_vec", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_emds_vec, METH_VARARGS|METH_KEYWORDS, "emds_vec(PairwiseEMDBaseFloat64 self, bool raw=False) -> vectorDouble"},
	 { "PairwiseEMDBaseFloat64_emd", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat64_emd, METH_VARARGS|METH_KEYWORDS, "emd(PairwiseEMDBaseFloat64 self, wasserstein::index_type i, wasserstein::index_type j, int thread=0) -> double"},
	 { "PairwiseEMDBaseFloat64_emds", _wrap_PairwiseEMDBaseFloat64_emds, METH_O, "emds(PairwiseEMDBaseFloat64 self)"},
	 { "PairwiseEMDBaseFloat64_raw_emds", _wrap_PairwiseEMDBaseFloat64_raw_emds, METH_O, "raw_emds(PairwiseEMDBaseFloat64 self)"},
	 { "PairwiseEMDBaseFloat64_swigregister", PairwiseEMDBaseFloat64_swigregister, METH_O, NULL},
	 { "delete_ExternalEMDHandlerFloat64", _wrap_delete_ExternalEMDHandlerFloat64, METH_O, "delete_ExternalEMDHandlerFloat64(ExternalEMDHandlerFloat64 self)"},
	 { "ExternalEMDHandlerFloat64_description", _wrap_ExternalEMDHandlerFloat64_description, METH_O, "description(ExternalEMDHandlerFloat64 self) -> std::string"},
	 { "ExternalEMDHandlerFloat64_num_calls", _wrap_ExternalEMDHandlerFloat64_num_calls, METH_O, "num_calls(ExternalEMDHandlerFloat64 self) -> std::size_t"},
	 { "ExternalEMDHandlerFloat64___call__", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat64___call__, METH_VARARGS|METH_KEYWORDS, "__call__(ExternalEMDHandlerFloat64 self, double emd, double weight=1)"},
	 { "ExternalEMDHandlerFloat64_evaluate", _wrap_ExternalEMDHandlerFloat64_evaluate, METH_VARARGS, "\n"
		"evaluate(ExternalEMDHandlerFloat64 self, double * emds)\n"
		"evaluate(ExternalEMDHandlerFloat64 self, double * emds, double * event_weights)\n"
		""},
	 { "ExternalEMDHandlerFloat64_evaluate_symmetric", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat64_evaluate_symmetric, METH_VARARGS|METH_KEYWORDS, "evaluate_symmetric(ExternalEMDHandlerFloat64 self, double * emds, double * event_weights)"},
	 { "ExternalEMDHandlerFloat64_swigregister", ExternalEMDHandlerFloat64_swigregister, METH_O, NULL},
	 { "new_Histogram1DHandlerLogFloat64", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerLogFloat64, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerLogFloat64(unsigned int nbins, double axis_min, double axis_max) -> Histogram1DHandlerLogFloat64"},
	 { "delete_Histogram1DHandlerLogFloat64", _wrap_delete_Histogram1DHandlerLogFloat64, METH_O, "delete_Histogram1DHandlerLogFloat64(Histogram1DHandlerLogFloat64 self)"},
	 { "Histogram1DHandlerLogFloat64_nbins", _wrap_Histogram1DHandlerLogFloat64_nbins, METH_O, "nbins(Histogram1DHandlerLogFloat64 self) -> unsigned int"},
	 { "Histogram1DHandlerLogFloat64_axis_min", _wrap_Histogram1DHandlerLogFloat64_axis_min, METH_O, "axis_min(Histogram1DHandlerLogFloat64 self) -> double"},
	 { "Histogram1DHandlerLogFloat64_axis_max", _wrap_Histogram1DHandlerLogFloat64_axis_max, METH_O, "axis_max(Histogram1DHandlerLogFloat64 self) -> double"},
	 { "Histogram1DHandlerLogFloat64_description", _wrap_Histogram1DHandlerLogFloat64_description, METH_O, "description(Histogram1DHandlerLogFloat64 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat64_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat64_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars_vec(Histogram1DHandlerLogFloat64 self, bool overflows=True) -> pairVectorDouble"},
	 { "Histogram1DHandlerLogFloat64_bin_centers_vec", _wrap_Histogram1DHandlerLogFloat64_bin_centers_vec, METH_O, "bin_centers_vec(Histogram1DHandlerLogFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerLogFloat64_bin_edges_vec", _wrap_Histogram1DHandlerLogFloat64_bin_edges_vec, METH_O, "bin_edges_vec(Histogram1DHandlerLogFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerLogFloat64___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat64___iadd__, METH_VARARGS|METH_KEYWORDS, "__iadd__(Histogram1DHandlerLogFloat64 self, Histogram1DHandlerLogFloat64 other) -> Histogram1DHandlerLogFloat64"},
	 { "Histogram1DHandlerLogFloat64___repr__", _wrap_Histogram1DHandlerLogFloat64___repr__, METH_O, "__repr__(Histogram1DHandlerLogFloat64 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat64_bin_centers", _wrap_Histogram1DHandlerLogFloat64_bin_centers, METH_O, "bin_centers(Histogram1DHandlerLogFloat64 self)"},
	 { "Histogram1DHandlerLogFloat64_bin_edges", _wrap_Histogram1DHandlerLogFloat64_bin_edges, METH_O, "bin_edges(Histogram1DHandlerLogFloat64 self)"},
	 { "Histogram1DHandlerLogFloat64_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat64_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars(Histogram1DHandlerLogFloat64 self, bool overflows=True)"},
	 { "Histogram1DHandlerLogFloat64_swigregister", Histogram1DHandlerLogFloat64_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerLogFloat64_swiginit", Histogram1DHandlerLogFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_Histogram1DHandlerFloat64", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerFloat64, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerFloat64(unsigned int nbins, double axis_min, double axis_max) -> Histogram1DHandlerFloat64"},
	 { "delete_Histogram1DHandlerFloat64", _wrap_delete_Histogram1DHandlerFloat64, METH_O, "delete_Histogram1DHandlerFloat64(Histogram1DHandlerFloat64 self)"},
	 { "Histogram1DHandlerFloat64_nbins", _wrap_Histogram1DHandlerFloat64_nbins, METH_O, "nbins(Histogram1DHandlerFloat64 self) -> unsigned int"},
	 { "Histogram1DHandlerFloat64_axis_min", _wrap_Histogram1DHandlerFloat64_axis_min, METH_O, "axis_min(Histogram1DHandlerFloat64 self) -> double"},
	 { "Histogram1DHandlerFloat64_axis_max", _wrap_Histogram1DHandlerFloat64_axis_max, METH_O, "axis_max(Histogram1DHandlerFloat64 self) -> double"},
	 { "Histogram1DHandlerFloat64_description", _wrap_Histogram1DHandlerFloat64_description, METH_O, "description(Histogram1DHandlerFloat64 self) -> std::string"},
	 { "Histogram1DHandlerFloat64_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat64_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars_vec(Histogram1DHandlerFloat64 self, bool overflows=True) -> pairVectorDouble"},
	 { "Histogram1DHandlerFloat64_bin_centers_vec", _wrap_Histogram1DHandlerFloat64_bin_centers_vec, METH_O, "bin_centers_vec(Histogram1DHandlerFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerFloat64_bin_edges_vec", _wrap_Histogram1DHandlerFloat64_bin_edges_vec, METH_O, "bin_edges_vec(Histogram1DHandlerFloat64 self) -> vectorDouble"},
	 { "Histogram1DHandlerFloat64___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat64___iadd__, METH_VARARGS|METH_KEYWORDS, "__iadd__(Histogram1DHandlerFloat64 self, Histogram1DHandlerFloat64 other) -> Histogram1DHandlerFloat64"},
	 { "Histogram1DHandlerFloat64___repr__", _wrap_Histogram1DHandlerFloat64___repr__, METH_O, "__repr__(Histogram1DHandlerFloat64 self) -> std::string"},
	 { "Histogram1DHandlerFloat64_bin_centers", _wrap_Histogram1DHandlerFloat64_bin_centers, METH_O, "bin_centers(Histogram1DHandlerFloat64 self)"},
	 { "Histogram1DHandlerFloat64_bin_edges", _wrap_Histogram1DHandlerFloat64_bin_edges, METH_O, "bin_edges(Histogram1DHandlerFloat64 self)"},
	 { "Histogram1DHandlerFloat64_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat64_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars(Histogram1DHandlerFloat64 self, bool overflows=True)"},
	 { "Histogram1DHandlerFloat64_swigregister", Histogram1DHandlerFloat64_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerFloat64_swiginit", Histogram1DHandlerFloat64_swiginit, METH_VARARGS, NULL},
	 { "delete_EMDBaseFloat32", _wrap_delete_EMDBaseFloat32, METH_O, "delete_EMDBaseFloat32(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_R", _wrap_EMDBaseFloat32_R, METH_O, "R(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_beta", _wrap_EMDBaseFloat32_beta, METH_O, "beta(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_set_R", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_R, METH_VARARGS|METH_KEYWORDS, "set_R(EMDBaseFloat32 self, float R)"},
	 { "EMDBaseFloat32_set_beta", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_beta, METH_VARARGS|METH_KEYWORDS, "set_beta(EMDBaseFloat32 self, float beta)"},
	 { "EMDBaseFloat32_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "set_network_simplex_params(EMDBaseFloat32 self, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1)"},
	 { "EMDBaseFloat32_norm", _wrap_EMDBaseFloat32_norm, METH_O, "norm(EMDBaseFloat32 self) -> bool"},
	 { "EMDBaseFloat32_set_norm", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_norm, METH_VARARGS|METH_KEYWORDS, "set_norm(EMDBaseFloat32 self, bool norm)"},
	 { "EMDBaseFloat32_do_timing", _wrap_EMDBaseFloat32_do_timing, METH_O, "do_timing(EMDBaseFloat32 self) -> bool"},
	 { "EMDBaseFloat32_set_do_timing", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_do_timing, METH_VARARGS|METH_KEYWORDS, "set_do_timing(EMDBaseFloat32 self, bool timing)"},
	 { "EMDBaseFloat32_external_dists", _wrap_EMDBaseFloat32_external_dists, METH_O, "external_dists(EMDBaseFloat32 self) -> bool"},
	 { "EMDBaseFloat32_set_external_dists", (PyCFunction)(void(*)(void))_wrap_EMDBaseFloat32_set_external_dists, METH_VARARGS|METH_KEYWORDS, "set_external_dists(EMDBaseFloat32 self, bool exdists)"},
	 { "EMDBaseFloat32_n0", _wrap_EMDBaseFloat32_n0, METH_O, "n0(EMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat32_n1", _wrap_EMDBaseFloat32_n1, METH_O, "n1(EMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "EMDBaseFloat32_extra", _wrap_EMDBaseFloat32_extra, METH_O, "extra(EMDBaseFloat32 self) -> wasserstein::ExtraParticle"},
	 { "EMDBaseFloat32_emd", _wrap_EMDBaseFloat32_emd, METH_O, "emd(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_status", _wrap_EMDBaseFloat32_status, METH_O, "status(EMDBaseFloat32 self) -> wasserstein::EMDStatus"},
	 { "EMDBaseFloat32_weightdiff", _wrap_EMDBaseFloat32_weightdiff, METH_O, "weightdiff(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_scale", _wrap_EMDBaseFloat32_scale, METH_O, "scale(EMDBaseFloat32 self) -> float"},
	 { "EMDBaseFloat32_n_iter", _wrap_EMDBaseFloat32_n_iter, METH_O, "n_iter(EMDBaseFloat32 self) -> std::size_t"},
	 { "EMDBaseFloat32_dists_vec", _wrap_EMDBaseFloat32_dists_vec, METH_O, "dists_vec(EMDBaseFloat32 self) -> vectorFloat"},
	 { "EMDBaseFloat32_flows_vec", _wrap_EMDBaseFloat32_flows_vec, METH_O, "flows_vec(EMDBaseFloat32 self) -> vectorFloat"},
	 { "EMDBaseFloat32_flow", _wrap_EMDBaseFloat32_flow, METH_VARARGS, "\n"
		"flow(EMDBaseFloat32 self, wasserstein::index_type i, wasserstein::index_type j) -> float\n"
		"flow(EMDBaseFloat32 self, std::size_t ind) -> float\n"
		""},
	 { "EMDBaseFloat32_node_potentials", _wrap_EMDBaseFloat32_node_potentials, METH_O, "node_potentials(EMDBaseFloat32 self) -> pairVectorFloat"},
	 { "EMDBaseFloat32_duration", _wrap_EMDBaseFloat32_duration, METH_O, "duration(EMDBaseFloat32 self) -> double"},
	 { "EMDBaseFloat32_clear", _wrap_EMDBaseFloat32_clear, METH_O, "clear(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_flows", _wrap_EMDBaseFloat32_flows, METH_O, "flows(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_dists", _wrap_EMDBaseFloat32_dists, METH_O, "dists(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_npy_node_potentials", _wrap_EMDBaseFloat32_npy_node_potentials, METH_O, "npy_node_potentials(EMDBaseFloat32 self)"},
	 { "EMDBaseFloat32_swigregister", EMDBaseFloat32_swigregister, METH_O, NULL},
	 { "delete_PairwiseEMDBaseFloat32", _wrap_delete_PairwiseEMDBaseFloat32, METH_O, "delete_PairwiseEMDBaseFloat32(PairwiseEMDBaseFloat32 self)"},
	 { "PairwiseEMDBaseFloat32_R", _wrap_PairwiseEMDBaseFloat32_R, METH_O, "R(PairwiseEMDBaseFloat32 self) -> float"},
	 { "PairwiseEMDBaseFloat32_set_R", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_R, METH_VARARGS|METH_KEYWORDS, "set_R(PairwiseEMDBaseFloat32 self, float R)"},
	 { "PairwiseEMDBaseFloat32_beta", _wrap_PairwiseEMDBaseFloat32_beta, METH_O, "beta(PairwiseEMDBaseFloat32 self) -> float"},
	 { "PairwiseEMDBaseFloat32_set_beta", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_beta, METH_VARARGS|METH_KEYWORDS, "set_beta(PairwiseEMDBaseFloat32 self, float beta)"},
	 { "PairwiseEMDBaseFloat32_norm", _wrap_PairwiseEMDBaseFloat32_norm, METH_O, "norm(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_set_norm", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_norm, METH_VARARGS|METH_KEYWORDS, "set_norm(PairwiseEMDBaseFloat32 self, bool norm)"},
	 { "PairwiseEMDBaseFloat32_set_network_simplex_params", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_network_simplex_params, METH_VARARGS|METH_KEYWORDS, "set_network_simplex_params(PairwiseEMDBaseFloat32 self, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1)"},
	 { "PairwiseEMDBaseFloat32_set_external_emd_handler", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_external_emd_handler, METH_VARARGS|METH_KEYWORDS, "set_external_emd_handler(PairwiseEMDBaseFloat32 self, ExternalEMDHandlerFloat32 handler)"},
	 { "PairwiseEMDBaseFloat32_have_external_emd_handler", _wrap_PairwiseEMDBaseFloat32_have_external_emd_handler, METH_O, "have_external_emd_handler(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_num_threads", _wrap_PairwiseEMDBaseFloat32_num_threads, METH_O, "num_threads(PairwiseEMDBaseFloat32 self) -> int"},
	 { "PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_omp_dynamic_chunksize, METH_VARARGS|METH_KEYWORDS, "set_omp_dynamic_chunksize(PairwiseEMDBaseFloat32 self, int chunksize)"},
	 { "PairwiseEMDBaseFloat32_omp_dynamic_chunksize", _wrap_PairwiseEMDBaseFloat32_omp_dynamic_chunksize, METH_O, "omp_dynamic_chunksize(PairwiseEMDBaseFloat32 self) -> int"},
	 { "PairwiseEMDBaseFloat32_set_request_mode", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_set_request_mode, METH_VARARGS|METH_KEYWORDS, "set_request_mode(PairwiseEMDBaseFloat32 self, bool mode)"},
	 { "PairwiseEMDBaseFloat32_request_mode", _wrap_PairwiseEMDBaseFloat32_request_mode, METH_O, "request_mode(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_duration", _wrap_PairwiseEMDBaseFloat32_duration, METH_O, "duration(PairwiseEMDBaseFloat32 self) -> double"},
	 { "PairwiseEMDBaseFloat32_nevA", _wrap_PairwiseEMDBaseFloat32_nevA, METH_O, "nevA(PairwiseEMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat32_nevB", _wrap_PairwiseEMDBaseFloat32_nevB, METH_O, "nevB(PairwiseEMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat32_num_emds", _wrap_PairwiseEMDBaseFloat32_num_emds, METH_O, "num_emds(PairwiseEMDBaseFloat32 self) -> wasserstein::index_type"},
	 { "PairwiseEMDBaseFloat32_storage", _wrap_PairwiseEMDBaseFloat32_storage, METH_O, "storage(PairwiseEMDBaseFloat32 self) -> wasserstein::EMDPairsStorage"},
	 { "PairwiseEMDBaseFloat32_errored", _wrap_PairwiseEMDBaseFloat32_errored, METH_O, "errored(PairwiseEMDBaseFloat32 self) -> bool"},
	 { "PairwiseEMDBaseFloat32_error_messages", _wrap_PairwiseEMDBaseFloat32_error_messages, METH_O, "error_messages(PairwiseEMDBaseFloat32 self) -> vectorString"},
	 { "PairwiseEMDBaseFloat32_emds_vec", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_emds_vec, METH_VARARGS|METH_KEYWORDS, "emds_vec(PairwiseEMDBaseFloat32 self, bool raw=False) -> vectorFloat"},
	 { "PairwiseEMDBaseFloat32_emd", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDBaseFloat32_emd, METH_VARARGS|METH_KEYWORDS, "emd(PairwiseEMDBaseFloat32 self, wasserstein::index_type i, wasserstein::index_type j, int thread=0) -> float"},
	 { "PairwiseEMDBaseFloat32_emds", _wrap_PairwiseEMDBaseFloat32_emds, METH_O, "emds(PairwiseEMDBaseFloat32 self)"},
	 { "PairwiseEMDBaseFloat32_raw_emds", _wrap_PairwiseEMDBaseFloat32_raw_emds, METH_O, "raw_emds(PairwiseEMDBaseFloat32 self)"},
	 { "PairwiseEMDBaseFloat32_swigregister", PairwiseEMDBaseFloat32_swigregister, METH_O, NULL},
	 { "delete_ExternalEMDHandlerFloat32", _wrap_delete_ExternalEMDHandlerFloat32, METH_O, "delete_ExternalEMDHandlerFloat32(ExternalEMDHandlerFloat32 self)"},
	 { "ExternalEMDHandlerFloat32_description", _wrap_ExternalEMDHandlerFloat32_description, METH_O, "description(ExternalEMDHandlerFloat32 self) -> std::string"},
	 { "ExternalEMDHandlerFloat32_num_calls", _wrap_ExternalEMDHandlerFloat32_num_calls, METH_O, "num_calls(ExternalEMDHandlerFloat32 self) -> std::size_t"},
	 { "ExternalEMDHandlerFloat32___call__", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat32___call__, METH_VARARGS|METH_KEYWORDS, "__call__(ExternalEMDHandlerFloat32 self, float emd, float weight=1)"},
	 { "ExternalEMDHandlerFloat32_evaluate", _wrap_ExternalEMDHandlerFloat32_evaluate, METH_VARARGS, "\n"
		"evaluate(ExternalEMDHandlerFloat32 self, float * emds)\n"
		"evaluate(ExternalEMDHandlerFloat32 self, float * emds, float * event_weights)\n"
		""},
	 { "ExternalEMDHandlerFloat32_evaluate_symmetric", (PyCFunction)(void(*)(void))_wrap_ExternalEMDHandlerFloat32_evaluate_symmetric, METH_VARARGS|METH_KEYWORDS, "evaluate_symmetric(ExternalEMDHandlerFloat32 self, float * emds, float * event_weights)"},
	 { "ExternalEMDHandlerFloat32_swigregister", ExternalEMDHandlerFloat32_swigregister, METH_O, NULL},
	 { "new_Histogram1DHandlerLogFloat32", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerLogFloat32, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerLogFloat32(unsigned int nbins, float axis_min, float axis_max) -> Histogram1DHandlerLogFloat32"},
	 { "delete_Histogram1DHandlerLogFloat32", _wrap_delete_Histogram1DHandlerLogFloat32, METH_O, "delete_Histogram1DHandlerLogFloat32(Histogram1DHandlerLogFloat32 self)"},
	 { "Histogram1DHandlerLogFloat32_nbins", _wrap_Histogram1DHandlerLogFloat32_nbins, METH_O, "nbins(Histogram1DHandlerLogFloat32 self) -> unsigned int"},
	 { "Histogram1DHandlerLogFloat32_axis_min", _wrap_Histogram1DHandlerLogFloat32_axis_min, METH_O, "axis_min(Histogram1DHandlerLogFloat32 self) -> float"},
	 { "Histogram1DHandlerLogFloat32_axis_max", _wrap_Histogram1DHandlerLogFloat32_axis_max, METH_O, "axis_max(Histogram1DHandlerLogFloat32 self) -> float"},
	 { "Histogram1DHandlerLogFloat32_description", _wrap_Histogram1DHandlerLogFloat32_description, METH_O, "description(Histogram1DHandlerLogFloat32 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat32_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat32_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars_vec(Histogram1DHandlerLogFloat32 self, bool overflows=True) -> pairVectorFloat"},
	 { "Histogram1DHandlerLogFloat32_bin_centers_vec", _wrap_Histogram1DHandlerLogFloat32_bin_centers_vec, METH_O, "bin_centers_vec(Histogram1DHandlerLogFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerLogFloat32_bin_edges_vec", _wrap_Histogram1DHandlerLogFloat32_bin_edges_vec, METH_O, "bin_edges_vec(Histogram1DHandlerLogFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerLogFloat32___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat32___iadd__, METH_VARARGS|METH_KEYWORDS, "__iadd__(Histogram1DHandlerLogFloat32 self, Histogram1DHandlerLogFloat32 other) -> Histogram1DHandlerLogFloat32"},
	 { "Histogram1DHandlerLogFloat32___repr__", _wrap_Histogram1DHandlerLogFloat32___repr__, METH_O, "__repr__(Histogram1DHandlerLogFloat32 self) -> std::string"},
	 { "Histogram1DHandlerLogFloat32_bin_centers", _wrap_Histogram1DHandlerLogFloat32_bin_centers, METH_O, "bin_centers(Histogram1DHandlerLogFloat32 self)"},
	 { "Histogram1DHandlerLogFloat32_bin_edges", _wrap_Histogram1DHandlerLogFloat32_bin_edges, METH_O, "bin_edges(Histogram1DHandlerLogFloat32 self)"},
	 { "Histogram1DHandlerLogFloat32_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerLogFloat32_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars(Histogram1DHandlerLogFloat32 self, bool overflows=True)"},
	 { "Histogram1DHandlerLogFloat32_swigregister", Histogram1DHandlerLogFloat32_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerLogFloat32_swiginit", Histogram1DHandlerLogFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_Histogram1DHandlerFloat32", (PyCFunction)(void(*)(void))_wrap_new_Histogram1DHandlerFloat32, METH_VARARGS|METH_KEYWORDS, "new_Histogram1DHandlerFloat32(unsigned int nbins, float axis_min, float axis_max) -> Histogram1DHandlerFloat32"},
	 { "delete_Histogram1DHandlerFloat32", _wrap_delete_Histogram1DHandlerFloat32, METH_O, "delete_Histogram1DHandlerFloat32(Histogram1DHandlerFloat32 self)"},
	 { "Histogram1DHandlerFloat32_nbins", _wrap_Histogram1DHandlerFloat32_nbins, METH_O, "nbins(Histogram1DHandlerFloat32 self) -> unsigned int"},
	 { "Histogram1DHandlerFloat32_axis_min", _wrap_Histogram1DHandlerFloat32_axis_min, METH_O, "axis_min(Histogram1DHandlerFloat32 self) -> float"},
	 { "Histogram1DHandlerFloat32_axis_max", _wrap_Histogram1DHandlerFloat32_axis_max, METH_O, "axis_max(Histogram1DHandlerFloat32 self) -> float"},
	 { "Histogram1DHandlerFloat32_description", _wrap_Histogram1DHandlerFloat32_description, METH_O, "description(Histogram1DHandlerFloat32 self) -> std::string"},
	 { "Histogram1DHandlerFloat32_hist_vals_vars_vec", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat32_hist_vals_vars_vec, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars_vec(Histogram1DHandlerFloat32 self, bool overflows=True) -> pairVectorFloat"},
	 { "Histogram1DHandlerFloat32_bin_centers_vec", _wrap_Histogram1DHandlerFloat32_bin_centers_vec, METH_O, "bin_centers_vec(Histogram1DHandlerFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerFloat32_bin_edges_vec", _wrap_Histogram1DHandlerFloat32_bin_edges_vec, METH_O, "bin_edges_vec(Histogram1DHandlerFloat32 self) -> vectorFloat"},
	 { "Histogram1DHandlerFloat32___iadd__", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat32___iadd__, METH_VARARGS|METH_KEYWORDS, "__iadd__(Histogram1DHandlerFloat32 self, Histogram1DHandlerFloat32 other) -> Histogram1DHandlerFloat32"},
	 { "Histogram1DHandlerFloat32___repr__", _wrap_Histogram1DHandlerFloat32___repr__, METH_O, "__repr__(Histogram1DHandlerFloat32 self) -> std::string"},
	 { "Histogram1DHandlerFloat32_bin_centers", _wrap_Histogram1DHandlerFloat32_bin_centers, METH_O, "bin_centers(Histogram1DHandlerFloat32 self)"},
	 { "Histogram1DHandlerFloat32_bin_edges", _wrap_Histogram1DHandlerFloat32_bin_edges, METH_O, "bin_edges(Histogram1DHandlerFloat32 self)"},
	 { "Histogram1DHandlerFloat32_hist_vals_vars", (PyCFunction)(void(*)(void))_wrap_Histogram1DHandlerFloat32_hist_vals_vars, METH_VARARGS|METH_KEYWORDS, "hist_vals_vars(Histogram1DHandlerFloat32 self, bool overflows=True)"},
	 { "Histogram1DHandlerFloat32_swigregister", Histogram1DHandlerFloat32_swigregister, METH_O, NULL},
	 { "Histogram1DHandlerFloat32_swiginit", Histogram1DHandlerFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_CorrelationDimensionFloat32", (PyCFunction)(void(*)(void))_wrap_new_CorrelationDimensionFloat32, METH_VARARGS|METH_KEYWORDS, "new_CorrelationDimensionFloat32(unsigned int nbins, float axis_min, float axis_max) -> CorrelationDimensionFloat32"},
	 { "CorrelationDimensionFloat32_corrdims_vec", (PyCFunction)(void(*)(void))_wrap_CorrelationDimensionFloat32_corrdims_vec, METH_VARARGS|METH_KEYWORDS, "corrdims_vec(CorrelationDimensionFloat32 self, float eps=std::numeric_limits< float >::epsilon()) -> pairVectorFloat"},
	 { "CorrelationDimensionFloat32_corrdim_bins_vec", _wrap_CorrelationDimensionFloat32_corrdim_bins_vec, METH_O, "corrdim_bins_vec(CorrelationDimensionFloat32 self) -> vectorFloat"},
	 { "CorrelationDimensionFloat32_cumulative_vals_vars_vec", _wrap_CorrelationDimensionFloat32_cumulative_vals_vars_vec, METH_O, "cumulative_vals_vars_vec(CorrelationDimensionFloat32 self) -> pairVectorFloat"},
	 { "CorrelationDimensionFloat32___repr__", _wrap_CorrelationDimensionFloat32___repr__, METH_O, "__repr__(CorrelationDimensionFloat32 self) -> std::string"},
	 { "CorrelationDimensionFloat32_corrdim_bins", _wrap_CorrelationDimensionFloat32_corrdim_bins, METH_O, "corrdim_bins(CorrelationDimensionFloat32 self)"},
	 { "CorrelationDimensionFloat32_corrdims", _wrap_CorrelationDimensionFloat32_corrdims, METH_O, "corrdims(CorrelationDimensionFloat32 self)"},
	 { "CorrelationDimensionFloat32_cumulative_vals_vars", _wrap_CorrelationDimensionFloat32_cumulative_vals_vars, METH_O, "cumulative_vals_vars(CorrelationDimensionFloat32 self)"},
	 { "delete_CorrelationDimensionFloat32", _wrap_delete_CorrelationDimensionFloat32, METH_O, "delete_CorrelationDimensionFloat32(CorrelationDimensionFloat32 self)"},
	 { "CorrelationDimensionFloat32_swigregister", CorrelationDimensionFloat32_swigregister, METH_O, NULL},
	 { "CorrelationDimensionFloat32_swiginit", CorrelationDimensionFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_CorrelationDimensionFloat64", (PyCFunction)(void(*)(void))_wrap_new_CorrelationDimensionFloat64, METH_VARARGS|METH_KEYWORDS, "new_CorrelationDimensionFloat64(unsigned int nbins, double axis_min, double axis_max) -> CorrelationDimensionFloat64"},
	 { "CorrelationDimensionFloat64_corrdims_vec", (PyCFunction)(void(*)(void))_wrap_CorrelationDimensionFloat64_corrdims_vec, METH_VARARGS|METH_KEYWORDS, "corrdims_vec(CorrelationDimensionFloat64 self, double eps=std::numeric_limits< double >::epsilon()) -> pairVectorDouble"},
	 { "CorrelationDimensionFloat64_corrdim_bins_vec", _wrap_CorrelationDimensionFloat64_corrdim_bins_vec, METH_O, "corrdim_bins_vec(CorrelationDimensionFloat64 self) -> vectorDouble"},
	 { "CorrelationDimensionFloat64_cumulative_vals_vars_vec", _wrap_CorrelationDimensionFloat64_cumulative_vals_vars_vec, METH_O, "cumulative_vals_vars_vec(CorrelationDimensionFloat64 self) -> pairVectorDouble"},
	 { "CorrelationDimensionFloat64___repr__", _wrap_CorrelationDimensionFloat64___repr__, METH_O, "__repr__(CorrelationDimensionFloat64 self) -> std::string"},
	 { "CorrelationDimensionFloat64_corrdim_bins", _wrap_CorrelationDimensionFloat64_corrdim_bins, METH_O, "corrdim_bins(CorrelationDimensionFloat64 self)"},
	 { "CorrelationDimensionFloat64_corrdims", _wrap_CorrelationDimensionFloat64_corrdims, METH_O, "corrdims(CorrelationDimensionFloat64 self)"},
	 { "CorrelationDimensionFloat64_cumulative_vals_vars", _wrap_CorrelationDimensionFloat64_cumulative_vals_vars, METH_O, "cumulative_vals_vars(CorrelationDimensionFloat64 self)"},
	 { "delete_CorrelationDimensionFloat64", _wrap_delete_CorrelationDimensionFloat64, METH_O, "delete_CorrelationDimensionFloat64(CorrelationDimensionFloat64 self)"},
	 { "CorrelationDimensionFloat64_swigregister", CorrelationDimensionFloat64_swigregister, METH_O, NULL},
	 { "CorrelationDimensionFloat64_swiginit", CorrelationDimensionFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_EMDFloat64", (PyCFunction)(void(*)(void))_wrap_new_EMDFloat64, METH_VARARGS|METH_KEYWORDS, "new_EMDFloat64(double R=1, double beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1) -> EMDFloat64"},
	 { "delete_EMDFloat64", _wrap_delete_EMDFloat64, METH_O, "delete_EMDFloat64(EMDFloat64 self)"},
	 { "EMDFloat64_description", (PyCFunction)(void(*)(void))_wrap_EMDFloat64_description, METH_VARARGS|METH_KEYWORDS, "description(EMDFloat64 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDFloat64___repr__", _wrap_EMDFloat64___repr__, METH_O, "__repr__(EMDFloat64 self) -> std::string"},
	 { "EMDFloat64_preprocess_CenterWeightedCentroid", _wrap_EMDFloat64_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(EMDFloat64 self)"},
	 { "EMDFloat64___call__", _wrap_EMDFloat64___call__, METH_VARARGS, "\n"
		"__call__(EMDFloat64 self, double * weights0, double * coords0, double * weights1, double * coords1) -> double\n"
		"__call__(EMDFloat64 self, double * weights0, double * weights1, double * external_dists) -> double\n"
		""},
	 { "EMDFloat64_swigregister", EMDFloat64_swigregister, METH_O, NULL},
	 { "EMDFloat64_swiginit", EMDFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_EMDFloat32", (PyCFunction)(void(*)(void))_wrap_new_EMDFloat32, METH_VARARGS|METH_KEYWORDS, "new_EMDFloat32(float R=1, float beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1) -> EMDFloat32"},
	 { "delete_EMDFloat32", _wrap_delete_EMDFloat32, METH_O, "delete_EMDFloat32(EMDFloat32 self)"},
	 { "EMDFloat32_description", (PyCFunction)(void(*)(void))_wrap_EMDFloat32_description, METH_VARARGS|METH_KEYWORDS, "description(EMDFloat32 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDFloat32___repr__", _wrap_EMDFloat32___repr__, METH_O, "__repr__(EMDFloat32 self) -> std::string"},
	 { "EMDFloat32_preprocess_CenterWeightedCentroid", _wrap_EMDFloat32_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(EMDFloat32 self)"},
	 { "EMDFloat32___call__", _wrap_EMDFloat32___call__, METH_VARARGS, "\n"
		"__call__(EMDFloat32 self, float * weights0, float * coords0, float * weights1, float * coords1) -> float\n"
		"__call__(EMDFloat32 self, float * weights0, float * weights1, float * external_dists) -> float\n"
		""},
	 { "EMDFloat32_swigregister", EMDFloat32_swigregister, METH_O, NULL},
	 { "EMDFloat32_swiginit", EMDFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_EMDYPhiFloat64", (PyCFunction)(void(*)(void))_wrap_new_EMDYPhiFloat64, METH_VARARGS|METH_KEYWORDS, "new_EMDYPhiFloat64(double R=1, double beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1) -> EMDYPhiFloat64"},
	 { "delete_EMDYPhiFloat64", _wrap_delete_EMDYPhiFloat64, METH_O, "delete_EMDYPhiFloat64(EMDYPhiFloat64 self)"},
	 { "EMDYPhiFloat64_description", (PyCFunction)(void(*)(void))_wrap_EMDYPhiFloat64_description, METH_VARARGS|METH_KEYWORDS, "description(EMDYPhiFloat64 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDYPhiFloat64___repr__", _wrap_EMDYPhiFloat64___repr__, METH_O, "__repr__(EMDYPhiFloat64 self) -> std::string"},
	 { "EMDYPhiFloat64_preprocess_CenterWeightedCentroid", _wrap_EMDYPhiFloat64_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(EMDYPhiFloat64 self)"},
	 { "EMDYPhiFloat64___call__", _wrap_EMDYPhiFloat64___call__, METH_VARARGS, "\n"
		"__call__(EMDYPhiFloat64 self, double * weights0, double * coords0, double * weights1, double * coords1) -> double\n"
		"__call__(EMDYPhiFloat64 self, double * weights0, double * weights1, double * external_dists) -> double\n"
		""},
	 { "EMDYPhiFloat64_swigregister", EMDYPhiFloat64_swigregister, METH_O, NULL},
	 { "EMDYPhiFloat64_swiginit", EMDYPhiFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_EMDYPhiFloat32", (PyCFunction)(void(*)(void))_wrap_new_EMDYPhiFloat32, METH_VARARGS|METH_KEYWORDS, "new_EMDYPhiFloat32(float R=1, float beta=1, bool norm=False, bool do_timing=False, bool external_dists=False, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1) -> EMDYPhiFloat32"},
	 { "delete_EMDYPhiFloat32", _wrap_delete_EMDYPhiFloat32, METH_O, "delete_EMDYPhiFloat32(EMDYPhiFloat32 self)"},
	 { "EMDYPhiFloat32_description", (PyCFunction)(void(*)(void))_wrap_EMDYPhiFloat32_description, METH_VARARGS|METH_KEYWORDS, "description(EMDYPhiFloat32 self, bool write_preprocessors=True) -> std::string"},
	 { "EMDYPhiFloat32___repr__", _wrap_EMDYPhiFloat32___repr__, METH_O, "__repr__(EMDYPhiFloat32 self) -> std::string"},
	 { "EMDYPhiFloat32_preprocess_CenterWeightedCentroid", _wrap_EMDYPhiFloat32_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(EMDYPhiFloat32 self)"},
	 { "EMDYPhiFloat32___call__", _wrap_EMDYPhiFloat32___call__, METH_VARARGS, "\n"
		"__call__(EMDYPhiFloat32 self, float * weights0, float * coords0, float * weights1, float * coords1) -> float\n"
		"__call__(EMDYPhiFloat32 self, float * weights0, float * weights1, float * external_dists) -> float\n"
		""},
	 { "EMDYPhiFloat32_swigregister", EMDYPhiFloat32_swigregister, METH_O, NULL},
	 { "EMDYPhiFloat32_swiginit", EMDYPhiFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDFloat64", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDFloat64, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDFloat64(double R=1, double beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDFloat64"},
	 { "delete_PairwiseEMDFloat64", _wrap_delete_PairwiseEMDFloat64, METH_O, "delete_PairwiseEMDFloat64(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64_description", _wrap_PairwiseEMDFloat64_description, METH_O, "description(PairwiseEMDFloat64 self) -> std::string"},
	 { "PairwiseEMDFloat64_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat64_clear, METH_VARARGS|METH_KEYWORDS, "clear(PairwiseEMDFloat64 self, bool free_memory=True)"},
	 { "PairwiseEMDFloat64_init", _wrap_PairwiseEMDFloat64_init, METH_VARARGS, "\n"
		"init(PairwiseEMDFloat64 self, wasserstein::index_type nev)\n"
		"init(PairwiseEMDFloat64 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDFloat64_compute", _wrap_PairwiseEMDFloat64_compute, METH_O, "compute(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64___repr__", _wrap_PairwiseEMDFloat64___repr__, METH_O, "__repr__(PairwiseEMDFloat64 self) -> std::string"},
	 { "PairwiseEMDFloat64_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDFloat64_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64__reset_B_events", _wrap_PairwiseEMDFloat64__reset_B_events, METH_O, "_reset_B_events(PairwiseEMDFloat64 self)"},
	 { "PairwiseEMDFloat64__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat64__add_event, METH_VARARGS|METH_KEYWORDS, "_add_event(PairwiseEMDFloat64 self, double * weights, double * coords, double event_weight=1)"},
	 { "PairwiseEMDFloat64_swigregister", PairwiseEMDFloat64_swigregister, METH_O, NULL},
	 { "PairwiseEMDFloat64_swiginit", PairwiseEMDFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDFloat32", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDFloat32, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDFloat32(float R=1, float beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDFloat32"},
	 { "delete_PairwiseEMDFloat32", _wrap_delete_PairwiseEMDFloat32, METH_O, "delete_PairwiseEMDFloat32(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32_description", _wrap_PairwiseEMDFloat32_description, METH_O, "description(PairwiseEMDFloat32 self) -> std::string"},
	 { "PairwiseEMDFloat32_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat32_clear, METH_VARARGS|METH_KEYWORDS, "clear(PairwiseEMDFloat32 self, bool free_memory=True)"},
	 { "PairwiseEMDFloat32_init", _wrap_PairwiseEMDFloat32_init, METH_VARARGS, "\n"
		"init(PairwiseEMDFloat32 self, wasserstein::index_type nev)\n"
		"init(PairwiseEMDFloat32 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDFloat32_compute", _wrap_PairwiseEMDFloat32_compute, METH_O, "compute(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32___repr__", _wrap_PairwiseEMDFloat32___repr__, METH_O, "__repr__(PairwiseEMDFloat32 self) -> std::string"},
	 { "PairwiseEMDFloat32_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDFloat32_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32__reset_B_events", _wrap_PairwiseEMDFloat32__reset_B_events, METH_O, "_reset_B_events(PairwiseEMDFloat32 self)"},
	 { "PairwiseEMDFloat32__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDFloat32__add_event, METH_VARARGS|METH_KEYWORDS, "_add_event(PairwiseEMDFloat32 self, float * weights, float * coords, float event_weight=1)"},
	 { "PairwiseEMDFloat32_swigregister", PairwiseEMDFloat32_swigregister, METH_O, NULL},
	 { "PairwiseEMDFloat32_swiginit", PairwiseEMDFloat32_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDYPhiFloat64", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDYPhiFloat64, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDYPhiFloat64(double R=1, double beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, double epsilon_large_factor=1000, double epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDYPhiFloat64"},
	 { "delete_PairwiseEMDYPhiFloat64", _wrap_delete_PairwiseEMDYPhiFloat64, METH_O, "delete_PairwiseEMDYPhiFloat64(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64_description", _wrap_PairwiseEMDYPhiFloat64_description, METH_O, "description(PairwiseEMDYPhiFloat64 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat64_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat64_clear, METH_VARARGS|METH_KEYWORDS, "clear(PairwiseEMDYPhiFloat64 self, bool free_memory=True)"},
	 { "PairwiseEMDYPhiFloat64_init", _wrap_PairwiseEMDYPhiFloat64_init, METH_VARARGS, "\n"
		"init(PairwiseEMDYPhiFloat64 self, wasserstein::index_type nev)\n"
		"init(PairwiseEMDYPhiFloat64 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDYPhiFloat64_compute", _wrap_PairwiseEMDYPhiFloat64_compute, METH_O, "compute(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64___repr__", _wrap_PairwiseEMDYPhiFloat64___repr__, METH_O, "__repr__(PairwiseEMDYPhiFloat64 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDYPhiFloat64_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64__reset_B_events", _wrap_PairwiseEMDYPhiFloat64__reset_B_events, METH_O, "_reset_B_events(PairwiseEMDYPhiFloat64 self)"},
	 { "PairwiseEMDYPhiFloat64__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat64__add_event, METH_VARARGS|METH_KEYWORDS, "_add_event(PairwiseEMDYPhiFloat64 self, double * weights, double * coords, double event_weight=1)"},
	 { "PairwiseEMDYPhiFloat64_swigregister", PairwiseEMDYPhiFloat64_swigregister, METH_O, NULL},
	 { "PairwiseEMDYPhiFloat64_swiginit", PairwiseEMDYPhiFloat64_swiginit, METH_VARARGS, NULL},
	 { "new_PairwiseEMDYPhiFloat32", (PyCFunction)(void(*)(void))_wrap_new_PairwiseEMDYPhiFloat32, METH_VARARGS|METH_KEYWORDS, "new_PairwiseEMDYPhiFloat32(float R=1, float beta=1, bool norm=False, int num_threads=-1, wasserstein::index_type print_every=-10, unsigned int verbose=1, bool request_mode=False, bool store_sym_emds_raw=True, bool throw_on_error=False, unsigned int omp_dynamic_chunksize=10, std::size_t n_iter_max=100000, float epsilon_large_factor=1000, float epsilon_small_factor=1, std::ostream & os=std::cout) -> PairwiseEMDYPhiFloat32"},
	 { "delete_PairwiseEMDYPhiFloat32", _wrap_delete_PairwiseEMDYPhiFloat32, METH_O, "delete_PairwiseEMDYPhiFloat32(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32_description", _wrap_PairwiseEMDYPhiFloat32_description, METH_O, "description(PairwiseEMDYPhiFloat32 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat32_clear", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat32_clear, METH_VARARGS|METH_KEYWORDS, "clear(PairwiseEMDYPhiFloat32 self, bool free_memory=True)"},
	 { "PairwiseEMDYPhiFloat32_init", _wrap_PairwiseEMDYPhiFloat32_init, METH_VARARGS, "\n"
		"init(PairwiseEMDYPhiFloat32 self, wasserstein::index_type nev)\n"
		"init(PairwiseEMDYPhiFloat32 self, wasserstein::index_type nevA, wasserstein::index_type nevB)\n"
		""},
	 { "PairwiseEMDYPhiFloat32_compute", _wrap_PairwiseEMDYPhiFloat32_compute, METH_O, "compute(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32___repr__", _wrap_PairwiseEMDYPhiFloat32___repr__, METH_O, "__repr__(PairwiseEMDYPhiFloat32 self) -> std::string"},
	 { "PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid", _wrap_PairwiseEMDYPhiFloat32_preprocess_CenterWeightedCentroid, METH_O, "preprocess_CenterWeightedCentroid(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32__reset_B_events", _wrap_PairwiseEMDYPhiFloat32__reset_B_events, METH_O, "_reset_B_events(PairwiseEMDYPhiFloat32 self)"},
	 { "PairwiseEMDYPhiFloat32__add_event", (PyCFunction)(void(*)(void))_wrap_PairwiseEMDYPhiFloat32__add_event, METH_VARARGS|METH_KEYWORDS, "_add_event(PairwiseEMDYPhiFloat32 self, float * weights, float * coords, float event_weight=1)"},
	 { "PairwiseEMDYPhiFloat32_swigregister", PairwiseEMDYPhiFloat32_swigregister, METH_O, NULL},
	 { "PairwiseEMDYPhiFloat32_swiginit", PairwiseEMDYPhiFloat32_swiginit, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_wasserstein__CorrelationDimensionT_double_tTo_p_wasserstein__ExternalEMDHandlerT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::ExternalEMDHandler< double > *) (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) ((wasserstein::CorrelationDimension< double > *) x));
}
static void *_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_tTo_p_wasserstein__ExternalEMDHandlerT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::ExternalEMDHandler< double > *)  ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *) x));
}
static void *_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_tTo_p_wasserstein__ExternalEMDHandlerT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::ExternalEMDHandler< double > *)  ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *) x));
}
static void *_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::EMDBase< double > *)  ((wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex > *) x));
}
static void *_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::EMDBase< double > *)  ((wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex > *) x));
}
static void *_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_tTo_p_wasserstein__PairwiseEMDBaseT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::PairwiseEMDBase< double > *)  ((wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex >,double > *) x));
}
static void *_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_tTo_p_wasserstein__PairwiseEMDBaseT_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::PairwiseEMDBase< double > *)  ((wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex >,double > *) x));
}
static void *_p_wasserstein__CorrelationDimensionT_float_tTo_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *)  ((wasserstein::CorrelationDimension< float > *) x));
}
static void *_p_wasserstein__CorrelationDimensionT_double_tTo_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *)  ((wasserstein::CorrelationDimension< double > *) x));
}
static void *_p_wasserstein__CorrelationDimensionT_float_tTo_p_wasserstein__ExternalEMDHandlerT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::ExternalEMDHandler< float > *) (wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) ((wasserstein::CorrelationDimension< float > *) x));
}
static void *_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_tTo_p_wasserstein__ExternalEMDHandlerT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::ExternalEMDHandler< float > *)  ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *) x));
}
static void *_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_tTo_p_wasserstein__ExternalEMDHandlerT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::ExternalEMDHandler< float > *)  ((wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *) x));
}
static void *_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::EMDBase< float > *)  ((wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex > *) x));
}
static void *_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::EMDBase< float > *)  ((wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex > *) x));
}
static void *_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_tTo_p_wasserstein__PairwiseEMDBaseT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::PairwiseEMDBase< float > *)  ((wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex >,float > *) x));
}
static void *_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_tTo_p_wasserstein__PairwiseEMDBaseT_float_t(void *x, int *SWIGUNUSEDPARM(newmemory)) {
    return (void *)((wasserstein::PairwiseEMDBase< float > *)  ((wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex >,float > *) x));
}
static swig_type_info _swigt__p_Axis = {"_p_Axis", "Axis *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Base = {"_p_Base", "Base *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Event = {"_p_Event", "Event *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Hist1DBase = {"_p_Hist1DBase", "Hist1DBase *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_NetworkSimplex = {"_p_NetworkSimplex", "NetworkSimplex *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_PairwiseDistance = {"_p_PairwiseDistance", "PairwiseDistance *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_ParticleCollection = {"_p_ParticleCollection", "ParticleCollection *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Self = {"_p_Self", "Self *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_Transform = {"_p_Transform", "Transform *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WeightCollection = {"_p_WeightCollection", "WeightCollection *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_allocator_type = {"_p_allocator_type", "allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_difference_type = {"_p_difference_type", "difference_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_double = {"_p_double", "wasserstein::default_value_type *|double *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_first_type = {"_p_first_type", "first_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_PyObject = {"_p_p_PyObject", "PyObject **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_double = {"_p_p_double", "double **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_float = {"_p_p_float", "float **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_second_type = {"_p_second_type", "second_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_size_type = {"_p_size_type", "size_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_double_t = {"_p_std__allocatorT_double_t", "std::vector< double >::allocator_type *|std::allocator< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_float_t = {"_p_std__allocatorT_float_t", "std::vector< float >::allocator_type *|std::allocator< float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__allocatorT_std__string_t = {"_p_std__allocatorT_std__string_t", "std::vector< std::string >::allocator_type *|std::allocator< std::string > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__invalid_argument = {"_p_std__invalid_argument", "std::invalid_argument *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__ostream = {"_p_std__ostream", "std::ostream *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t = {"_p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t", "std::pair< std::vector< double >,std::vector< double > > *|std::pair< std::vector< double,std::allocator< double > >,std::vector< double,std::allocator< double > > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t = {"_p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t", "std::pair< std::vector< float,std::allocator< float > >,std::vector< float,std::allocator< float > > > *|std::pair< std::vector< float >,std::vector< float > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__ptrdiff_t = {"_p_std__ptrdiff_t", "std::ptrdiff_t *|wasserstein::index_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_double_std__allocatorT_double_t_t = {"_p_std__vectorT_double_std__allocatorT_double_t_t", "std::vector< double,std::allocator< double > > *|std::vector< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_float_std__allocatorT_float_t_t = {"_p_std__vectorT_float_std__allocatorT_float_t_t", "std::vector< float > *|std::vector< float,std::allocator< float > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorT_std__string_std__allocatorT_std__string_t_t = {"_p_std__vectorT_std__string_std__allocatorT_std__string_t_t", "std::vector< std::string,std::allocator< std::string > > *|std::vector< std::string > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_swig__SwigPyIterator = {"_p_swig__SwigPyIterator", "swig::SwigPyIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_type = {"_p_value_type", "value_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__CorrelationDimensionT_double_t = {"_p_wasserstein__CorrelationDimensionT_double_t", "wasserstein::CorrelationDimension< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__CorrelationDimensionT_float_t = {"_p_wasserstein__CorrelationDimensionT_float_t", "wasserstein::CorrelationDimension< float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__EMDBaseT_double_t = {"_p_wasserstein__EMDBaseT_double_t", "wasserstein::EMDBase< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__EMDBaseT_float_t = {"_p_wasserstein__EMDBaseT_float_t", "wasserstein::EMDBase< float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t = {"_p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t", "wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *|wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t = {"_p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t", "wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *|wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t = {"_p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t", "wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance > *|wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t = {"_p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t", "wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance > *|wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__ExternalEMDHandlerT_double_t = {"_p_wasserstein__ExternalEMDHandlerT_double_t", "wasserstein::ExternalEMDHandler< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__ExternalEMDHandlerT_float_t = {"_p_wasserstein__ExternalEMDHandlerT_float_t", "wasserstein::ExternalEMDHandler< float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t = {"_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t", "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double >::Transform,double > *|wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t = {"_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t", "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float >::Transform,float > *|wasserstein::Histogram1DHandler< boost::histogram::axis::transform::id,float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t = {"_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t", "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double >::Transform,double > *|wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t = {"_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t", "wasserstein::Histogram1DHandler< wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float >::Transform,float > *|wasserstein::Histogram1DHandler< boost::histogram::axis::transform::log,float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__PairwiseEMDBaseT_double_t = {"_p_wasserstein__PairwiseEMDBaseT_double_t", "wasserstein::PairwiseEMDBase< double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__PairwiseEMDBaseT_float_t = {"_p_wasserstein__PairwiseEMDBaseT_float_t", "wasserstein::PairwiseEMDBase< float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t = {"_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t", "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,double > *|wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex >,double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t = {"_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t", "wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,double > *|wasserstein::PairwiseEMD< wasserstein::EMD< double,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex >,double > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t = {"_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t", "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance >,float > *|wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArray2Event,wasserstein::YPhiArrayDistance,wasserstein::DefaultNetworkSimplex >,float > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t = {"_p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t", "wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance >,float > *|wasserstein::PairwiseEMD< wasserstein::EMD< float,wasserstein::DefaultArrayEvent,wasserstein::EuclideanArrayDistance,wasserstein::DefaultNetworkSimplex >,float > *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_Axis,
  &_swigt__p_Base,
  &_swigt__p_Event,
  &_swigt__p_Hist1DBase,
  &_swigt__p_NetworkSimplex,
  &_swigt__p_PairwiseDistance,
  &_swigt__p_ParticleCollection,
  &_swigt__p_Self,
  &_swigt__p_Transform,
  &_swigt__p_WeightCollection,
  &_swigt__p_allocator_type,
  &_swigt__p_char,
  &_swigt__p_difference_type,
  &_swigt__p_double,
  &_swigt__p_first_type,
  &_swigt__p_p_PyObject,
  &_swigt__p_p_double,
  &_swigt__p_p_float,
  &_swigt__p_second_type,
  &_swigt__p_size_type,
  &_swigt__p_std__allocatorT_double_t,
  &_swigt__p_std__allocatorT_float_t,
  &_swigt__p_std__allocatorT_std__string_t,
  &_swigt__p_std__invalid_argument,
  &_swigt__p_std__ostream,
  &_swigt__p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t,
  &_swigt__p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t,
  &_swigt__p_std__ptrdiff_t,
  &_swigt__p_std__vectorT_double_std__allocatorT_double_t_t,
  &_swigt__p_std__vectorT_float_std__allocatorT_float_t_t,
  &_swigt__p_std__vectorT_std__string_std__allocatorT_std__string_t_t,
  &_swigt__p_swig__SwigPyIterator,
  &_swigt__p_value_type,
  &_swigt__p_wasserstein__CorrelationDimensionT_double_t,
  &_swigt__p_wasserstein__CorrelationDimensionT_float_t,
  &_swigt__p_wasserstein__EMDBaseT_double_t,
  &_swigt__p_wasserstein__EMDBaseT_float_t,
  &_swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  &_swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  &_swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  &_swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  &_swigt__p_wasserstein__ExternalEMDHandlerT_double_t,
  &_swigt__p_wasserstein__ExternalEMDHandlerT_float_t,
  &_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t,
  &_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t,
  &_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t,
  &_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t,
  &_swigt__p_wasserstein__PairwiseEMDBaseT_double_t,
  &_swigt__p_wasserstein__PairwiseEMDBaseT_float_t,
  &_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t,
  &_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t,
  &_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t,
  &_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t,
};

static swig_cast_info _swigc__p_Axis[] = {  {&_swigt__p_Axis, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Base[] = {  {&_swigt__p_Base, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Event[] = {  {&_swigt__p_Event, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Hist1DBase[] = {  {&_swigt__p_Hist1DBase, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_NetworkSimplex[] = {  {&_swigt__p_NetworkSimplex, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_PairwiseDistance[] = {  {&_swigt__p_PairwiseDistance, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_ParticleCollection[] = {  {&_swigt__p_ParticleCollection, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Self[] = {  {&_swigt__p_Self, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_Transform[] = {  {&_swigt__p_Transform, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WeightCollection[] = {  {&_swigt__p_WeightCollection, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_allocator_type[] = {  {&_swigt__p_allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_difference_type[] = {  {&_swigt__p_difference_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_double[] = {  {&_swigt__p_double, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_first_type[] = {  {&_swigt__p_first_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_PyObject[] = {  {&_swigt__p_p_PyObject, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_double[] = {  {&_swigt__p_p_double, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_float[] = {  {&_swigt__p_p_float, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_second_type[] = {  {&_swigt__p_second_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_size_type[] = {  {&_swigt__p_size_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_double_t[] = {  {&_swigt__p_std__allocatorT_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_float_t[] = {  {&_swigt__p_std__allocatorT_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__allocatorT_std__string_t[] = {  {&_swigt__p_std__allocatorT_std__string_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__invalid_argument[] = {  {&_swigt__p_std__invalid_argument, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__ostream[] = {  {&_swigt__p_std__ostream, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t[] = {  {&_swigt__p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t[] = {  {&_swigt__p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__ptrdiff_t[] = {  {&_swigt__p_std__ptrdiff_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_double_std__allocatorT_double_t_t[] = {  {&_swigt__p_std__vectorT_double_std__allocatorT_double_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_float_std__allocatorT_float_t_t[] = {  {&_swigt__p_std__vectorT_float_std__allocatorT_float_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorT_std__string_std__allocatorT_std__string_t_t[] = {  {&_swigt__p_std__vectorT_std__string_std__allocatorT_std__string_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_swig__SwigPyIterator[] = {  {&_swigt__p_swig__SwigPyIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_type[] = {  {&_swigt__p_value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__CorrelationDimensionT_double_t[] = {  {&_swigt__p_wasserstein__CorrelationDimensionT_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__CorrelationDimensionT_float_t[] = {  {&_swigt__p_wasserstein__CorrelationDimensionT_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__EMDBaseT_double_t[] = {  {&_swigt__p_wasserstein__EMDBaseT_double_t, 0, 0, 0},  {&_swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, _p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_double_t, 0, 0},  {&_swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, _p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_double_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__EMDBaseT_float_t[] = {  {&_swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, _p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_float_t, 0, 0},  {&_swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, _p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_tTo_p_wasserstein__EMDBaseT_float_t, 0, 0},  {&_swigt__p_wasserstein__EMDBaseT_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t[] = {  {&_swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t[] = {  {&_swigt__p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t[] = {  {&_swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t[] = {  {&_swigt__p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__ExternalEMDHandlerT_double_t[] = {  {&_swigt__p_wasserstein__ExternalEMDHandlerT_double_t, 0, 0, 0},  {&_swigt__p_wasserstein__CorrelationDimensionT_double_t, _p_wasserstein__CorrelationDimensionT_double_tTo_p_wasserstein__ExternalEMDHandlerT_double_t, 0, 0},  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, _p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_tTo_p_wasserstein__ExternalEMDHandlerT_double_t, 0, 0},  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, _p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_tTo_p_wasserstein__ExternalEMDHandlerT_double_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__ExternalEMDHandlerT_float_t[] = {  {&_swigt__p_wasserstein__ExternalEMDHandlerT_float_t, 0, 0, 0},  {&_swigt__p_wasserstein__CorrelationDimensionT_float_t, _p_wasserstein__CorrelationDimensionT_float_tTo_p_wasserstein__ExternalEMDHandlerT_float_t, 0, 0},  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, _p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_tTo_p_wasserstein__ExternalEMDHandlerT_float_t, 0, 0},  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, _p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_tTo_p_wasserstein__ExternalEMDHandlerT_float_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t[] = {  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t[] = {  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t[] = {  {&_swigt__p_wasserstein__CorrelationDimensionT_double_t, _p_wasserstein__CorrelationDimensionT_double_tTo_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0, 0},  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t[] = {  {&_swigt__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0, 0, 0},  {&_swigt__p_wasserstein__CorrelationDimensionT_float_t, _p_wasserstein__CorrelationDimensionT_float_tTo_p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__PairwiseEMDBaseT_double_t[] = {  {&_swigt__p_wasserstein__PairwiseEMDBaseT_double_t, 0, 0, 0},  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, _p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_tTo_p_wasserstein__PairwiseEMDBaseT_double_t, 0, 0},  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, _p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_tTo_p_wasserstein__PairwiseEMDBaseT_double_t, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__PairwiseEMDBaseT_float_t[] = {  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, _p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_tTo_p_wasserstein__PairwiseEMDBaseT_float_t, 0, 0},  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, _p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_tTo_p_wasserstein__PairwiseEMDBaseT_float_t, 0, 0},  {&_swigt__p_wasserstein__PairwiseEMDBaseT_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t[] = {  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t[] = {  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t[] = {  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t[] = {  {&_swigt__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_Axis,
  _swigc__p_Base,
  _swigc__p_Event,
  _swigc__p_Hist1DBase,
  _swigc__p_NetworkSimplex,
  _swigc__p_PairwiseDistance,
  _swigc__p_ParticleCollection,
  _swigc__p_Self,
  _swigc__p_Transform,
  _swigc__p_WeightCollection,
  _swigc__p_allocator_type,
  _swigc__p_char,
  _swigc__p_difference_type,
  _swigc__p_double,
  _swigc__p_first_type,
  _swigc__p_p_PyObject,
  _swigc__p_p_double,
  _swigc__p_p_float,
  _swigc__p_second_type,
  _swigc__p_size_type,
  _swigc__p_std__allocatorT_double_t,
  _swigc__p_std__allocatorT_float_t,
  _swigc__p_std__allocatorT_std__string_t,
  _swigc__p_std__invalid_argument,
  _swigc__p_std__ostream,
  _swigc__p_std__pairT_std__vectorT_double_std__allocatorT_double_t_t_std__vectorT_double_std__allocatorT_double_t_t_t,
  _swigc__p_std__pairT_std__vectorT_float_std__allocatorT_float_t_t_std__vectorT_float_std__allocatorT_float_t_t_t,
  _swigc__p_std__ptrdiff_t,
  _swigc__p_std__vectorT_double_std__allocatorT_double_t_t,
  _swigc__p_std__vectorT_float_std__allocatorT_float_t_t,
  _swigc__p_std__vectorT_std__string_std__allocatorT_std__string_t_t,
  _swigc__p_swig__SwigPyIterator,
  _swigc__p_value_type,
  _swigc__p_wasserstein__CorrelationDimensionT_double_t,
  _swigc__p_wasserstein__CorrelationDimensionT_float_t,
  _swigc__p_wasserstein__EMDBaseT_double_t,
  _swigc__p_wasserstein__EMDBaseT_float_t,
  _swigc__p_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  _swigc__p_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  _swigc__p_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  _swigc__p_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t,
  _swigc__p_wasserstein__ExternalEMDHandlerT_double_t,
  _swigc__p_wasserstein__ExternalEMDHandlerT_float_t,
  _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_double_t,
  _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__id_float_t,
  _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_double_t,
  _swigc__p_wasserstein__Histogram1DHandlerT_boost__histogram__axis__transform__log_float_t,
  _swigc__p_wasserstein__PairwiseEMDBaseT_double_t,
  _swigc__p_wasserstein__PairwiseEMDBaseT_float_t,
  _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t,
  _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_double_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_double_t,
  _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArray2Event_wasserstein__YPhiArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t,
  _swigc__p_wasserstein__PairwiseEMDT_wasserstein__EMDT_float_wasserstein__DefaultArrayEvent_wasserstein__EuclideanArrayDistance_wasserstein__DefaultNetworkSimplex_t_float_t,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
#if PY_VERSION_HEX >= 0x03000000
    return PyUnicode_InternFromString("<Swig global variables>");
#else
    return PyString_FromString("<Swig global variables>");
#endif
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
#if PY_VERSION_HEX >= 0x03000000
    PyObject *str = PyUnicode_InternFromString("(");
    PyObject *tail;
    PyObject *joined;
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      tail = PyUnicode_FromString(var->name);
      joined = PyUnicode_Concat(str, tail);
      Py_DecRef(str);
      Py_DecRef(tail);
      str = joined;
      if (var->next) {
        tail = PyUnicode_InternFromString(", ");
        joined = PyUnicode_Concat(str, tail);
        Py_DecRef(str);
        Py_DecRef(tail);
        str = joined;
      }
    }
    tail = PyUnicode_InternFromString(")");
    joined = PyUnicode_Concat(str, tail);
    Py_DecRef(str);
    Py_DecRef(tail);
    str = joined;
#else
    PyObject *str = PyString_FromString("(");
    swig_globalvar *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
#endif
    return str;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_Format(PyExc_AttributeError, "Unknown C global variable '%s'", n);
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;
    if (!type_init) {
      const PyTypeObject tmp = {
#if PY_VERSION_HEX >= 0x03000000
        PyVarObject_HEAD_INIT(NULL, 0)
#else
        PyObject_HEAD_INIT(NULL)
        0,                                  /* ob_size */
#endif
        "swigvarlink",                      /* tp_name */
        sizeof(swig_varlinkobject),         /* tp_basicsize */
        0,                                  /* tp_itemsize */
        (destructor) swig_varlink_dealloc,  /* tp_dealloc */
        0,                                  /* tp_print */
        (getattrfunc) swig_varlink_getattr, /* tp_getattr */
        (setattrfunc) swig_varlink_setattr, /* tp_setattr */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc) swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
        0,                                  /* tp_del */
        0,                                  /* tp_version_tag */
#if PY_VERSION_HEX >= 0x03040000
        0,                                  /* tp_finalize */
#endif
#if PY_VERSION_HEX >= 0x03080000
        0,                                  /* tp_vectorcall */
#endif
#if (PY_VERSION_HEX >= 0x03080000) && (PY_VERSION_HEX < 0x03090000)
        0,                                  /* tp_print */
#endif
#ifdef COUNT_ALLOCS
        0,                                  /* tp_allocs */
        0,                                  /* tp_frees */
        0,                                  /* tp_maxalloc */
        0,                                  /* tp_prev */
        0                                   /* tp_next */
#endif
      };
      varlink_type = tmp;
      type_init = 1;
      if (PyType_Ready(&varlink_type) < 0)
      return NULL;
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, const char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        memcpy(gv->name, name, size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *globals = 0;
    if (!globals) {
      globals = SWIG_newvarlink();
    }
    return globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_InternalNewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (!c) continue;
      c = strstr(c, "swig_ptr: ");
      if (c) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
          if (ptr) {
            size_t shift = (ci->ptype) - types;
            swig_type_info *ty = types_initial[shift];
            size_t ldoc = (c - methods[i].ml_doc);
            size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
            char *ndoc = (char*)malloc(ldoc + lptr + 10);
            if (ndoc) {
              char *buff = ndoc;
              memcpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              memcpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
  /* -----------------------------------------------------------------------------
   * Method creation and docstring support functions
   * ----------------------------------------------------------------------------- */
  
  /* -----------------------------------------------------------------------------
   * Function to find the method definition with the correct docstring for the
   * proxy module as opposed to the low-level API
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyMethodDef *SWIG_PythonGetProxyDoc(const char *name) {
    /* Find the function in the modified method table */
    size_t offset = 0;
    int found = 0;
    while (SwigMethods_proxydocs[offset].ml_meth != NULL) {
      if (strcmp(SwigMethods_proxydocs[offset].ml_name, name) == 0) {
        found = 1;
        break;
      }
      offset++;
    }
    /* Use the copy with the modified docstring if available */
    return found ? &SwigMethods_proxydocs[offset] : NULL;
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyInstanceMethod_New() used in Python 3
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyInstanceMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
#if PY_VERSION_HEX >= 0x03000000
    return PyInstanceMethod_New(func);
#else
    return PyMethod_New(func, NULL, NULL);
#endif
  }
  
  /* -----------------------------------------------------------------------------
   * Wrapper of PyStaticMethod_New()
   * It is exported to the generated module, used for -fastproxy
   * ----------------------------------------------------------------------------- */
  
  SWIGINTERN PyObject *SWIG_PyStaticMethod_New(PyObject *SWIGUNUSEDPARM(self), PyObject *func) {
    if (PyCFunction_Check(func)) {
      PyCFunctionObject *funcobj = (PyCFunctionObject *)func;
      PyMethodDef *ml = SWIG_PythonGetProxyDoc(funcobj->m_ml->ml_name);
      if (ml)
      func = PyCFunction_NewEx(ml, funcobj->m_self, funcobj->m_module);
    }
    return PyStaticMethod_New(func);
  }
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif

SWIGEXPORT 
#if PY_VERSION_HEX >= 0x03000000
PyObject*
#else
void
#endif
SWIG_init(void) {
  PyObject *m, *d, *md, *globals;
  
#if PY_VERSION_HEX >= 0x03000000
  static struct PyModuleDef SWIG_module = {
    PyModuleDef_HEAD_INIT,
    SWIG_name,
    NULL,
    -1,
    SwigMethods,
    NULL,
    NULL,
    NULL,
    NULL
  };
#endif
  
#if defined(SWIGPYTHON_BUILTIN)
  static SwigPyClientData SwigPyObject_clientdata = {
    0, 0, 0, 0, 0, 0, 0
  };
  static PyGetSetDef this_getset_def = {
    (char *)"this", &SwigPyBuiltin_ThisClosure, NULL, NULL, NULL
  };
  static SwigPyGetSet thisown_getset_closure = {
    SwigPyObject_own,
    SwigPyObject_own
  };
  static PyGetSetDef thisown_getset_def = {
    (char *)"thisown", SwigPyBuiltin_GetterClosure, SwigPyBuiltin_SetterClosure, NULL, &thisown_getset_closure
  };
  PyTypeObject *builtin_pytype;
  int builtin_base_count;
  swig_type_info *builtin_basetype;
  PyObject *tuple;
  PyGetSetDescrObject *static_getset;
  PyTypeObject *metatype;
  PyTypeObject *swigpyobject;
  SwigPyClientData *cd;
  PyObject *public_interface, *public_symbol;
  PyObject *this_descr;
  PyObject *thisown_descr;
  PyObject *self = 0;
  int i;
  
  (void)builtin_pytype;
  (void)builtin_base_count;
  (void)builtin_basetype;
  (void)tuple;
  (void)static_getset;
  (void)self;
  
  /* Metaclass is used to implement static member variables */
  metatype = SwigPyObjectType();
  assert(metatype);
#endif
  
  (void)globals;
  
  /* Create singletons now to avoid potential deadlocks with multi-threaded usage after module initialization */
  SWIG_This();
  SWIG_Python_TypeCache();
  SwigPyPacked_type();
#ifndef SWIGPYTHON_BUILTIN
  SwigPyObject_type();
#endif
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
#if PY_VERSION_HEX >= 0x03000000
  m = PyModule_Create(&SWIG_module);
#else
  m = Py_InitModule(SWIG_name, SwigMethods);
#endif
  
  md = d = PyModule_GetDict(m);
  (void)md;
  
  SWIG_InitializeModule(0);
  
#ifdef SWIGPYTHON_BUILTIN
  swigpyobject = SwigPyObject_TypeOnce();
  
  SwigPyObject_stype = SWIG_MangledTypeQuery("_p_SwigPyObject");
  assert(SwigPyObject_stype);
  cd = (SwigPyClientData*) SwigPyObject_stype->clientdata;
  if (!cd) {
    SwigPyObject_stype->clientdata = &SwigPyObject_clientdata;
    SwigPyObject_clientdata.pytype = swigpyobject;
  } else if (swigpyobject->tp_basicsize != cd->pytype->tp_basicsize) {
    PyErr_SetString(PyExc_RuntimeError, "Import error: attempted to load two incompatible swig-generated modules.");
# if PY_VERSION_HEX >= 0x03000000
    return NULL;
# else
    return;
# endif
  }
  
  /* All objects have a 'this' attribute */
  this_descr = PyDescr_NewGetSet(SwigPyObject_type(), &this_getset_def);
  (void)this_descr;
  
  /* All objects have a 'thisown' attribute */
  thisown_descr = PyDescr_NewGetSet(SwigPyObject_type(), &thisown_getset_def);
  (void)thisown_descr;
  
  public_interface = PyList_New(0);
  public_symbol = 0;
  (void)public_symbol;
  
  PyDict_SetItemString(md, "__all__", public_interface);
  Py_DECREF(public_interface);
  for (i = 0; SwigMethods[i].ml_name != NULL; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, SwigMethods[i].ml_name);
  for (i = 0; swig_const_table[i].name != 0; ++i)
  SwigPyBuiltin_AddPublicSymbol(public_interface, swig_const_table[i].name);
#endif
  
  SWIG_InstallConstants(d,swig_const_table);
  
  
  // thread safe initialization
  swig::container_owner_attribute();
  
  
  import_array();
  
  globals = SWIG_globals();
  if (!globals) {
    PyErr_SetString(PyExc_TypeError, "Failure to create SWIG globals.");
#if PY_VERSION_HEX >= 0x03000000
    return NULL;
#else
    return;
#endif
  }
  PyDict_SetItemString(md, "cvar", globals);
  Py_DECREF(globals);
  SWIG_addvarlink(globals, "PI", Swig_var_PI_get, Swig_var_PI_set);
  SWIG_addvarlink(globals, "TWOPI", Swig_var_TWOPI_get, Swig_var_TWOPI_set);
  SWIG_Python_SetConstant(d, "EMDStatus_Success",SWIG_From_int(static_cast< int >(wasserstein::EMDStatus::Success)));
  SWIG_Python_SetConstant(d, "EMDStatus_Empty",SWIG_From_int(static_cast< int >(wasserstein::EMDStatus::Empty)));
  SWIG_Python_SetConstant(d, "EMDStatus_SupplyMismatch",SWIG_From_int(static_cast< int >(wasserstein::EMDStatus::SupplyMismatch)));
  SWIG_Python_SetConstant(d, "EMDStatus_Unbounded",SWIG_From_int(static_cast< int >(wasserstein::EMDStatus::Unbounded)));
  SWIG_Python_SetConstant(d, "EMDStatus_MaxIterReached",SWIG_From_int(static_cast< int >(wasserstein::EMDStatus::MaxIterReached)));
  SWIG_Python_SetConstant(d, "EMDStatus_Infeasible",SWIG_From_int(static_cast< int >(wasserstein::EMDStatus::Infeasible)));
  SWIG_Python_SetConstant(d, "ExtraParticle_Neither",SWIG_From_int(static_cast< int >(wasserstein::ExtraParticle::Neither)));
  SWIG_Python_SetConstant(d, "ExtraParticle_Zero",SWIG_From_int(static_cast< int >(wasserstein::ExtraParticle::Zero)));
  SWIG_Python_SetConstant(d, "ExtraParticle_One",SWIG_From_int(static_cast< int >(wasserstein::ExtraParticle::One)));
  SWIG_Python_SetConstant(d, "EMDPairsStorage_Full",SWIG_From_int(static_cast< int >(wasserstein::EMDPairsStorage::Full)));
  SWIG_Python_SetConstant(d, "EMDPairsStorage_FullSymmetric",SWIG_From_int(static_cast< int >(wasserstein::EMDPairsStorage::FullSymmetric)));
  SWIG_Python_SetConstant(d, "EMDPairsStorage_FlattenedSymmetric",SWIG_From_int(static_cast< int >(wasserstein::EMDPairsStorage::FlattenedSymmetric)));
  SWIG_Python_SetConstant(d, "EMDPairsStorage_External",SWIG_From_int(static_cast< int >(wasserstein::EMDPairsStorage::External)));
  
  /* Initialize threading */
  SWIG_PYTHON_INITIALIZE_THREADS;
#if PY_VERSION_HEX >= 0x03000000
  return m;
#else
  return;
#endif
}

